Package: olsrr
Type: Package
Title: Tools for Building OLS Regression Models
Version: 0.7.0
Authors@R: person("Aravind", "Hebbali", email = "hebbali.aravind@gmail.com", role = c("aut", "cre"))
Description: Tools designed to make it easier for users, particularly beginner/intermediate R users 
    to build ordinary least squares regression models. Includes comprehensive regression output, 
    heteroskedasticity tests, collinearity diagnostics, residual diagnostics, measures of influence, 
    model fit assessment and variable selection procedures.
Depends: R(>= 3.3)
Imports: car, ggplot2, goftest, graphics, gridExtra, nortest, stats,
        utils, xplorerr
Suggests: covr, descriptr, knitr, rmarkdown, testthat, vdiffr
License: MIT + file LICENSE
URL: https://olsrr.rsquaredacademy.com/,
        https://github.com/rsquaredacademy/olsrr
BugReports: https://github.com/rsquaredacademy/olsrr/issues
Encoding: UTF-8
LazyData: true
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-21 10:53:56 UTC; HP
Author: Aravind Hebbali [aut, cre]
Maintainer: Aravind Hebbali <hebbali.aravind@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-22 09:00:08 UTC
Built: R 4.6.0; ; 2026-04-25 14:06:40 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(ols_anova,default)
S3method(ols_coll_diag,default)
S3method(ols_compare_model_performance,lm)
S3method(ols_correlations,default)
S3method(ols_model_performance,default)
S3method(ols_pure_error_anova,default)
S3method(ols_regress,default)
S3method(ols_regress,lm)
S3method(ols_step_all_possible,default)
S3method(ols_step_backward_adj_r2,default)
S3method(ols_step_backward_aic,default)
S3method(ols_step_backward_p,default)
S3method(ols_step_backward_r2,default)
S3method(ols_step_backward_sbc,default)
S3method(ols_step_backward_sbic,default)
S3method(ols_step_best_subset,default)
S3method(ols_step_both_adj_r2,default)
S3method(ols_step_both_aic,default)
S3method(ols_step_both_p,default)
S3method(ols_step_both_r2,default)
S3method(ols_step_both_sbc,default)
S3method(ols_step_both_sbic,default)
S3method(ols_step_forward_adj_r2,default)
S3method(ols_step_forward_aic,default)
S3method(ols_step_forward_p,default)
S3method(ols_step_forward_r2,default)
S3method(ols_step_forward_sbc,default)
S3method(ols_step_forward_sbic,default)
S3method(ols_test_bartlett,default)
S3method(ols_test_breusch_pagan,default)
S3method(ols_test_f,default)
S3method(ols_test_normality,default)
S3method(ols_test_normality,lm)
S3method(ols_test_score,default)
S3method(plot,ols_step_all_possible)
S3method(plot,ols_step_backward_adj_r2)
S3method(plot,ols_step_backward_aic)
S3method(plot,ols_step_backward_p)
S3method(plot,ols_step_backward_r2)
S3method(plot,ols_step_backward_sbc)
S3method(plot,ols_step_backward_sbic)
S3method(plot,ols_step_best_subset)
S3method(plot,ols_step_both_adj_r2)
S3method(plot,ols_step_both_aic)
S3method(plot,ols_step_both_p)
S3method(plot,ols_step_both_r2)
S3method(plot,ols_step_both_sbc)
S3method(plot,ols_step_both_sbic)
S3method(plot,ols_step_forward_adj_r2)
S3method(plot,ols_step_forward_aic)
S3method(plot,ols_step_forward_p)
S3method(plot,ols_step_forward_r2)
S3method(plot,ols_step_forward_sbc)
S3method(plot,ols_step_forward_sbic)
S3method(print,ols_anova)
S3method(print,ols_coll_diag)
S3method(print,ols_correlations)
S3method(print,ols_model_performance)
S3method(print,ols_pure_error_anova)
S3method(print,ols_regress)
S3method(print,ols_step_all_possible)
S3method(print,ols_step_backward_adj_r2)
S3method(print,ols_step_backward_aic)
S3method(print,ols_step_backward_p)
S3method(print,ols_step_backward_r2)
S3method(print,ols_step_backward_sbc)
S3method(print,ols_step_backward_sbic)
S3method(print,ols_step_best_subset)
S3method(print,ols_step_both_adj_r2)
S3method(print,ols_step_both_aic)
S3method(print,ols_step_both_p)
S3method(print,ols_step_both_r2)
S3method(print,ols_step_both_sbc)
S3method(print,ols_step_both_sbic)
S3method(print,ols_step_forward_adj_r2)
S3method(print,ols_step_forward_aic)
S3method(print,ols_step_forward_p)
S3method(print,ols_step_forward_r2)
S3method(print,ols_step_forward_sbc)
S3method(print,ols_step_forward_sbic)
S3method(print,ols_test_bartlett)
S3method(print,ols_test_breusch_pagan)
S3method(print,ols_test_f)
S3method(print,ols_test_normality)
S3method(print,ols_test_score)
export(ols_aic)
export(ols_anova)
export(ols_apc)
export(ols_coll_diag)
export(ols_compare_model_performance)
export(ols_correlations)
export(ols_count_obs)
export(ols_count_parameters)
export(ols_eigen_cindex)
export(ols_fpe)
export(ols_get_call)
export(ols_get_data)
export(ols_get_deviance)
export(ols_get_df)
export(ols_get_formula)
export(ols_get_interaction_terms)
export(ols_get_intercept)
export(ols_get_model_matrix)
export(ols_get_parameters)
export(ols_get_predicted)
export(ols_get_predictors)
export(ols_get_residuals)
export(ols_get_response)
export(ols_get_sigma)
export(ols_get_variables)
export(ols_get_vcov)
export(ols_hadi)
export(ols_has_intercept)
export(ols_hsp)
export(ols_launch_app)
export(ols_leverage)
export(ols_mallows_cp)
export(ols_model_info)
export(ols_model_performance)
export(ols_msep)
export(ols_plot_added_variable)
export(ols_plot_comp_plus_resid)
export(ols_plot_cooksd_bar)
export(ols_plot_cooksd_chart)
export(ols_plot_dfbetas)
export(ols_plot_dffits)
export(ols_plot_diagnostics)
export(ols_plot_fm)
export(ols_plot_hadi)
export(ols_plot_obs_fit)
export(ols_plot_reg_line)
export(ols_plot_resid_box)
export(ols_plot_resid_fit)
export(ols_plot_resid_fit_spread)
export(ols_plot_resid_hist)
export(ols_plot_resid_lev)
export(ols_plot_resid_pot)
export(ols_plot_resid_qq)
export(ols_plot_resid_regressor)
export(ols_plot_resid_spread)
export(ols_plot_resid_stand)
export(ols_plot_resid_stud)
export(ols_plot_resid_stud_fit)
export(ols_plot_response)
export(ols_pred_rsq)
export(ols_prep_avplot_data)
export(ols_prep_cdplot_data)
export(ols_prep_cdplot_outliers)
export(ols_prep_dfbeta_data)
export(ols_prep_dfbeta_outliers)
export(ols_prep_dsrvf_data)
export(ols_prep_outlier_obs)
export(ols_prep_regress_x)
export(ols_prep_regress_y)
export(ols_prep_rfsplot_fmdata)
export(ols_prep_rfsplot_rsdata)
export(ols_prep_rstudlev_data)
export(ols_prep_rvsrplot_data)
export(ols_prep_srchart_data)
export(ols_prep_srplot_data)
export(ols_press)
export(ols_pure_error_anova)
export(ols_regress)
export(ols_sbc)
export(ols_sbic)
export(ols_step_all_possible)
export(ols_step_all_possible_betas)
export(ols_step_backward_adj_r2)
export(ols_step_backward_aic)
export(ols_step_backward_p)
export(ols_step_backward_r2)
export(ols_step_backward_sbc)
export(ols_step_backward_sbic)
export(ols_step_best_subset)
export(ols_step_both_adj_r2)
export(ols_step_both_aic)
export(ols_step_both_p)
export(ols_step_both_r2)
export(ols_step_both_sbc)
export(ols_step_both_sbic)
export(ols_step_forward_adj_r2)
export(ols_step_forward_aic)
export(ols_step_forward_p)
export(ols_step_forward_r2)
export(ols_step_forward_sbc)
export(ols_step_forward_sbic)
export(ols_test_bartlett)
export(ols_test_breusch_pagan)
export(ols_test_correlation)
export(ols_test_f)
export(ols_test_normality)
export(ols_test_outlier)
export(ols_test_score)
export(ols_vif_tol)
export(rvsr_plot_shiny)
importFrom(car,Anova)
importFrom(ggplot2,aes)
importFrom(ggplot2,aes_string)
importFrom(ggplot2,annotate)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_boxplot)
importFrom(ggplot2,geom_dotplot)
importFrom(ggplot2,geom_histogram)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_label)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_linerange)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_smooth)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_fill_manual)
importFrom(ggplot2,scale_shape_manual)
importFrom(ggplot2,scale_size_manual)
importFrom(ggplot2,stat_qq)
importFrom(ggplot2,stat_smooth)
importFrom(ggplot2,theme)
importFrom(ggplot2,xlab)
importFrom(ggplot2,xlim)
importFrom(ggplot2,ylab)
importFrom(ggplot2,ylim)
importFrom(goftest,cvm.test)
importFrom(graphics,hist)
importFrom(gridExtra,grid.arrange)
importFrom(gridExtra,marrangeGrob)
importFrom(nortest,ad.test)
importFrom(stats,anova)
importFrom(stats,as.formula)
importFrom(stats,coefficients)
importFrom(stats,complete.cases)
importFrom(stats,confint.lm)
importFrom(stats,cooks.distance)
importFrom(stats,cor)
importFrom(stats,deviance)
importFrom(stats,df.residual)
importFrom(stats,dfbetas)
importFrom(stats,dffits)
importFrom(stats,dnorm)
importFrom(stats,ecdf)
importFrom(stats,fitted)
importFrom(stats,fitted.values)
importFrom(stats,formula)
importFrom(stats,hatvalues)
importFrom(stats,ks.test)
importFrom(stats,lm)
importFrom(stats,logLik)
importFrom(stats,lsfit)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,na.omit)
importFrom(stats,pchisq)
importFrom(stats,pf)
importFrom(stats,pt)
importFrom(stats,qnorm)
importFrom(stats,qqline)
importFrom(stats,qqnorm)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,residuals)
importFrom(stats,rstandard)
importFrom(stats,rstudent)
importFrom(stats,sd)
importFrom(stats,shapiro.test)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(utils,combn)
importFrom(utils,tail)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# olsrr

<!-- badges: start -->

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/olsrr)](https://cran.r-project.org/package=olsrr)
[![R build
status](https://github.com/rsquaredacademy/olsrr/workflows/R-CMD-check/badge.svg)](https://github.com/rsquaredacademy/olsrr/actions)
[![Coverage
status](https://codecov.io/gh/rsquaredacademy/olsrr/branch/master/graph/badge.svg)](https://app.codecov.io/github/rsquaredacademy/olsrr?branch=master)
<!-- badges: end -->

## Overview

The olsrr package provides following tools for building OLS regression
models using R:

- Comprehensive Regression Output
- Variable Selection Procedures
- Heteroskedasticity Tests
- Collinearity Diagnostics
- Model Fit Assessment
- Measures of Influence
- Residual Diagnostics
- Variable Contribution Assessment

## Installation

``` r
# Install release version from CRAN
install.packages("olsrr")

# Install development version from GitHub
# install.packages("pak")
pak::pak("rsquaredacademy/olsrr")
```

## Articles

- [Quick
  Overview](https://olsrr.rsquaredacademy.com/articles/intro.html)
- [Variable Selection
  Methods](https://olsrr.rsquaredacademy.com/articles/variable_selection.html)
- [Residual
  Diagnostics](https://olsrr.rsquaredacademy.com/articles/residual_diagnostics.html)
- [Heteroskedasticity](https://olsrr.rsquaredacademy.com/articles/heteroskedasticity.html)
- [Measures of
  Influence](https://olsrr.rsquaredacademy.com/articles/influence_measures.html)
- [Collinearity
  Diagnostics](https://olsrr.rsquaredacademy.com/articles/regression_diagnostics.html)

## Usage

    #> Warning: package 'dplyr' was built under R version 4.5.3
    #> Warning: package 'ggplot2' was built under R version 4.5.3
    #> Warning: package 'gridExtra' was built under R version 4.5.3

olsrr uses consistent prefix `ols_` for easy tab completion. If you know
how to write a `formula` or build models using `lm`, you will find olsrr
very useful. Most of the functions use an object of class `lm` as input.
So you just need to build a model using `lm` and then pass it onto the
functions in olsrr. Below is a quick demo:

#### Regression

``` r
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_regress(model)
#>                          Model Summary                          
#> ---------------------------------------------------------------
#> R                       0.914       RMSE                 2.409 
#> R-Squared               0.835       MSE                  5.801 
#> Adj. R-Squared          0.811       Coef. Var           13.051 
#> Pred R-Squared          0.771       AIC                159.070 
#> MAE                     1.858       SBC                167.864 
#> ---------------------------------------------------------------
#>  RMSE: Root Mean Square Error 
#>  MSE: Mean Square Error 
#>  MAE: Mean Absolute Error 
#>  AIC: Akaike Information Criteria 
#>  SBC: Schwarz Bayesian Criteria 
#> 
#>                                ANOVA                                 
#> --------------------------------------------------------------------
#>                 Sum of                                              
#>                Squares        DF    Mean Square      F         Sig. 
#> --------------------------------------------------------------------
#> Regression     940.412         4        235.103    34.195    0.0000 
#> Residual       185.635        27          6.875                     
#> Total         1126.047        31                                    
#> --------------------------------------------------------------------
#> 
#>                                   Parameter Estimates                                    
#> ----------------------------------------------------------------------------------------
#>       model      Beta    Std. Error    Std. Beta      t        Sig      lower     upper 
#> ----------------------------------------------------------------------------------------
#> (Intercept)    27.330         8.639                  3.164    0.004     9.604    45.055 
#>        disp     0.003         0.011        0.055     0.248    0.806    -0.019     0.025 
#>          hp    -0.019         0.016       -0.212    -1.196    0.242    -0.051     0.013 
#>          wt    -4.609         1.266       -0.748    -3.641    0.001    -7.206    -2.012 
#>        qsec     0.544         0.466        0.161     1.166    0.254    -0.413     1.501 
#> ----------------------------------------------------------------------------------------
```

## Getting Help

If you encounter a bug, please file a minimal reproducible example using
[reprex](https://reprex.tidyverse.org/index.html) on github. For
questions and clarifications, use
[StackOverflow](https://stackoverflow.com/).

## Code of Conduct

Please note that the olsrr project is released with a [Contributor Code
of Conduct](https://olsrr.rsquaredacademy.com/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

## ----echo=FALSE, message=FALSE------------------------------------------------
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)

## ----qqresid, fig.width=5, fig.height=5, fig.align='center'-------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_qq(model)

## ----normtest-----------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_normality(model)

## ----corrtest-----------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_correlation(model)

## ----rvsfplot, fig.width=5, fig.height=5, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_fit(model)

## ----residhist, fig.width=5, fig.height=5, fig.align='center'-----------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_hist(model)

## ----echo=FALSE, message=FALSE------------------------------------------------
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)

## ----viftol-------------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_vif_tol(model)

## ----cindex-------------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_eigen_cindex(model)

## ----colldiag-----------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_coll_diag(model)

## ----rfsplot, fig.width=10, fig.height=5, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_fit_spread(model)

## ----partcor------------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_correlations(model)

## ----obspred, fig.width=5, fig.height=5, fig.align='center'-------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_obs_fit(model)

## ----lackfit------------------------------------------------------------------
model <- lm(mpg ~ disp, data = mtcars)
ols_pure_error_anova(model)

## ----diagpanel, fig.width=10, fig.height=10, fig.align='center'---------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_diagnostics(model)

## ----rvsrplot, fig.width=5, fig.height=5, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_resid_regressor(model, 'drat')

## ----avplot, fig.width=10, fig.height=10, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_added_variable(model)

## ----cplusr, fig.width=10, fig.height=10, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_comp_plus_resid(model)

## ----echo=FALSE, message=FALSE------------------------------------------------
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)

## ----regress------------------------------------------------------------------
ols_regress(mpg ~ disp + hp + wt + qsec, data = mtcars)

## ----rvsfplot, fig.width=5, fig.height=5, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_fit(model)

## ----dfbpanel, fig.width=7, fig.height=7, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_dfbetas(model)

## ----rfsplot, fig.width=10, fig.height=5, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_fit_spread(model)

## ----bp1----------------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model)

## ----colldiag-----------------------------------------------------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_coll_diag(model)

## ----stepwise1----------------------------------------------------------------
# stepwise regression
model <- lm(y ~ ., data = surgical)
ols_step_both_p(model)

## ----stepwise2, fig.width=10, fig.height=15, fig.align='center'---------------
model <- lm(y ~ ., data = surgical)
k <- ols_step_both_p(model)
plot(k)

## ----stepaicb1----------------------------------------------------------------
# stepwise aic backward regression
model <- lm(y ~ ., data = surgical)
k <- ols_step_backward_aic(model)
k

## ----stepaicb2, fig.width=5, fig.height=5, fig.align='center'-----------------
model <- lm(y ~ ., data = surgical)
k <- ols_step_backward_aic(model)
plot(k)

## ----echo=FALSE, message=FALSE------------------------------------------------
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)

## ----ckdbp, fig.width=7, fig.height=5, fig.align='center'---------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_cooksd_bar(model)

## ----ckchart, fig.width=5, fig.height=5, fig.align='center'-------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_cooksd_chart(model)

## ----dfbpanel, fig.width=7, fig.height=7, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_dfbetas(model)

## ----dfitsplot, fig.width=5, fig.height=5, fig.align='center'-----------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_dffits(model)

## ----srplot, fig.width=5, fig.height=5, fig.align='center'--------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_stud(model)

## ----srchart, fig.width=5, fig.height=5, fig.align='center'-------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_stand(model)

## ----studlev, fig.width=7, fig.height=5, fig.align='center'-------------------
model <- lm(read ~ write + math + science, data = hsb)
ols_plot_resid_lev(model)

## ----dsrvsp, fig.width=7, fig.height=5, fig.align='center'--------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_stud_fit(model)

## ----hadiplot, fig.width=5, fig.height=5, fig.align='center'------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_hadi(model)

## ----potres, fig.width=5, fig.height=5, fig.align='center'--------------------
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_pot(model)

#' Best subsets regression
#'
#' Select the subset of predictors that do the best at meeting some
#' well-defined objective criterion, such as having the largest R2 value or the
#' smallest MSE, Mallow's Cp or AIC. The default metric used for selecting the
#' model is R2 but the user can choose any of the other available metrics.
#'
#' @param model An object of class \code{lm}.
#' @param max_order Maximum subset order.
#' @param include Character or numeric vector; variables to be included in selection process.
#' @param exclude Character or numeric vector; variables to be excluded from selection process.
#' @param metric Metric to select model.
#' @param x An object of class \code{ols_step_best_subset}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#' @param ... Other inputs.
#'
#' @return \code{ols_step_best_subset} returns an object of class \code{"ols_step_best_subset"}.
#' An object of class \code{"ols_step_best_subset"} is a list containing the following:
#'
#' \item{metrics}{selection metrics}
#'
#' @references
#' Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_step_best_subset(model)
#' ols_step_best_subset(model, metric = "adjr")
#' ols_step_best_subset(model, metric = "cp")
#'
#' # maximum subset
#' model <- lm(mpg ~ disp + hp + drat + wt + qsec, data = mtcars)
#' ols_step_best_subset(model, max_order = 3)
#'
#' # plot
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' k <- ols_step_best_subset(model)
#' plot(k)
#'
#' # return only models including `qsec`
#' ols_step_best_subset(model, include = c("qsec"))
#'
#' # exclude `hp` from selection process
#' ols_step_best_subset(model, exclude = c("hp"))
#'
#' @export
#'
ols_step_best_subset <- function(model, ...) UseMethod("ols_step_best_subset")

#' @export
#' @rdname ols_step_best_subset
#'
ols_step_best_subset.default <- function(model, max_order = NULL,
                                         include = NULL, exclude = NULL,
                                         metric = c("rsquare", "adjr", "predrsq",
                                                    "cp", "aic", "sbic", "sbc",
                                                    "msep", "fpe", "apc", "hsp"),
                                         ...) {

  check_model(model)
  check_npredictors(model, 2)

  indterms <- coeff_names(model)
  lenterms <- length(indterms)

  if (is.character(include)) {
    npm <- include %in% indterms
    if (!all(npm)) {
      stop(paste(paste(include[!npm], collapse = ", "), "not part of the model and hence cannot be forcibly included. Please verify the variable names."), call. = FALSE)
    }
  }

  if (is.character(exclude)) {
    npm <- exclude %in% indterms
    if (!all(npm)) {
      stop(paste(paste(exclude[!npm], collapse = ", "), "not part of the model and hence cannot be forcibly excluded. Please verify the variable names."), call. = FALSE)
    }
  }

  if (is.numeric(include)) {
    if (any(include > lenterms)) {
      stop(paste0("Index of variable to be included should be between 1 and ", lenterms, "."), call. = FALSE)
    } else {
      include <- indterms[include]
    }
  }

  if (is.numeric(exclude)) {
    if (any(exclude > lenterms)) {
      stop(paste0("Index of variable to be excluded should be between 1 and ", lenterms, "."), call. = FALSE)
    } else {
      exclude <- indterms[exclude]
    }
  }

  nam   <- setdiff(coeff_names(model), exclude)
  nam   <- setdiff(nam, include)
  n     <- length(nam)
  r     <- seq_len(n)
  combs <- list()

  for (i in seq_len(n)) {
    combs[[i]] <- combn(n, r[i])
  }

  if (!is.null(max_order)) {
    check_order(n, max_order)
  }

  if (is.null(max_order)) {
    max_order <- n
  }

  # lc        <- length(combs)
  lc        <- max_order
  varnames  <- model_colnames(model)
  data      <- model$model
  colas     <- unname(unlist(lapply(combs, ncol)))
  response  <- varnames[1]
  predicts  <- list()

  if (is.null(include)) {
    k <- 1
  } else {
    predicts[[1]] <- include
    k <- 2
  }

  for (i in seq_len(lc)) {
    for (j in seq_len(colas[i])) {
      predicts[[k]] <- c(include, nam[combs[[i]][, j]])
      k <- k + 1
    }
  }

  # if(!is.null(include)) {
  #   y <- grep(include, predicts)
  #   predicts <- predicts[y]
  # }

  len_elig <- length(predicts)

  mcount    <- 0
  rsq       <- list()
  adjr      <- list()
  cp        <- list()
  aic       <- list()
  sbic      <- list()
  sbc       <- list()
  mse       <- list()
  gmsep     <- list()
  jp        <- list()
  pc        <- list()
  sp        <- list()
  press     <- list()
  predrsq   <- list()
  preds     <- list()
  lpreds    <- c()

  for (i in seq_len(len_elig)) {
    predictors        <- predicts[[i]]
    lp                <- length(predictors)
    out               <- ols_regress(paste(response, "~", paste(predictors, collapse = " + ")), data = data)
    mcount            <- mcount + 1
    lpreds[mcount]    <- lp
    rsq[[mcount]]     <- out$rsq
    adjr[[mcount]]    <- out$adjr
    cp[[mcount]]      <- ols_mallows_cp(out$model, model)
    aic[[mcount]]     <- ols_aic(out$model)
    sbic[[mcount]]    <- ols_sbic(out$model, model)
    sbc[[mcount]]     <- ols_sbc(out$model)
    gmsep[[mcount]]   <- ols_msep(out$model)
    jp[[mcount]]      <- ols_fpe(out$model)
    pc[[mcount]]      <- ols_apc(out$model)
    sp[[mcount]]      <- ols_hsp(out$model)
    predrsq[[mcount]] <- ols_pred_rsq(out$model)
    preds[[mcount]]   <- paste(predictors, collapse = " ")
  }

  ui <- data.frame(
    n          = lpreds,
    predictors = unlist(preds),
    rsquare    = unlist(rsq),
    adjr       = unlist(adjr),
    predrsq    = unlist(predrsq),
    cp         = unlist(cp),
    aic        = unlist(aic),
    sbic       = unlist(sbic),
    sbc        = unlist(sbc),
    msep       = unlist(gmsep),
    fpe        = unlist(jp),
    apc        = unlist(pc),
    hsp        = unlist(sp),
    stringsAsFactors = F
  )

  metrics <- match.arg(metric)
  sorted <- c()
  l <- split(ui, ui$n)

  if (metrics == "rsquare" || metrics == "adjr" || metrics == "predrsq") {
    temp <- lapply(l, function(x) x[order(x[[metrics]], decreasing = TRUE), ][1, ])
  } else {
    temp <- lapply(l, function(x) x[order(x[[metrics]]), ][1, ])
  }

  sorted <- do.call(rbind, temp)
  mindex <- seq_len(nrow(sorted))
  sorted <- cbind(mindex, sorted)
  result <- list(metrics = sorted)
  class(result) <- c("ols_step_best_subset")

  return(result)
  # return(list(nam, combs, colas, predicts))

}

#' @export
#'
print.ols_step_best_subset <- function(x, ...) {
  print_best_subset(x)
}

#' @export
#' @rdname ols_step_best_subset
#'
plot.ols_step_best_subset <- function(x, model = NA, print_plot = TRUE, ...) {

  z <- x$metrics

  d <- data.frame(mindex = z$mindex, rsquare = z$rsquare, adjr = z$adjr,
                  cp = z$cp, aic = z$aic, sbic = z$sbic, sbc = z$sbc)

  p1 <- best_subset_plot(d, "rsquare")
  p2 <- best_subset_plot(d, "adjr", title = "Adj. R-Square")
  p3 <- best_subset_plot(d, "cp", title = "C(p)")
  p4 <- best_subset_plot(d, "aic", title = "AIC")
  p5 <- best_subset_plot(d, "sbic", title = "SBIC")
  p6 <- best_subset_plot(d, "sbc", title = "SBC")

  myplots <- list(plot_1 = p1, plot_2 = p2, plot_3 = p3,
                  plot_4 = p4, plot_5 = p5, plot_6 = p6)

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2, top = "Best Subset Regression")
  } else {
    return(myplots)
  }

}

#' Best subset plot
#'
#' Generate plots for best subset regression.
#'
#' @importFrom ggplot2 geom_line theme element_blank
#'
#' @param d A data.frame.
#' @param title Plot title.
#'
#' @noRd
#'
best_subset_plot <- function(d, var, title = "R-Square") {

  d1 <- d[, c("mindex", var)]
  colnames(d1) <- c("a", "b")

  ggplot(d1, aes(x = a, y = b)) +
    geom_line(color = "blue") +
    geom_point(color = "blue", shape = 1, size = 2) +
    xlab("") +
    ylab("") +
    ggtitle(title) +
    theme(axis.ticks = element_blank())

}

model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)

test_that("best subsets selection output matches the expected result", {
  k <- ols_step_best_subset(model)
  pred_exp <- c("x4", "x1 x2", "x1 x2 x4", "x1 x2 x3 x4")
  expect_equal(k$metrics$mindex, c(1, 2, 3, 4))
  expect_equal(k$metrics$predictors, pred_exp, ignore_attr = TRUE)

  k2 <- ols_step_best_subset(model, max_order = 3)
  preds_selected <- c("x4", "x1 x2", "x1 x2 x4")
  expect_equal(k2$metrics$mindex, c(1, 2, 3))
  expect_equal(k2$metrics$predictors, preds_selected, ignore_attr = TRUE)
})


test_that("best subsets regression returns the appropriate error", {
  expect_error(ols_step_best_subset(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_best_subset(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_best_subset(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_best_subset(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")

  error_msg <- "Maximum subset order should be less than or equal to the number of predictors in the specified model."
  expect_error(ols_step_best_subset(model, max_order = 5), error_msg)
})

test_that("output from best subsets regression matches the expected output when variables are locked in", {

  k <- ols_step_best_subset(model, include = c("x3"))
  pred_exp <- c("x3", "x3 x4", "x3 x1 x2", "x3 x1 x2 x4")
  expect_equal(k$metrics$mindex, c(1, 2, 3, 4))
  expect_equal(k$metrics$predictors, pred_exp, ignore_attr = TRUE)

  k <- ols_step_best_subset(model, include = c(3))
  pred_exp <- c("x3", "x3 x4", "x3 x1 x2", "x3 x1 x2 x4")
  expect_equal(k$metrics$mindex, c(1, 2, 3, 4))
  expect_equal(k$metrics$predictors, pred_exp, ignore_attr = TRUE)
})

test_that("output from best subsets regression matches the expected output when variables are locked out", {

  k <- ols_step_best_subset(model, exclude = c("x1"))
  pred_exp <- c("x4", "x3 x4", "x2 x3 x4")
  expect_equal(k$metrics$mindex, c(1, 2, 3))
  expect_equal(k$metrics$predictors, pred_exp, ignore_attr = TRUE)

  k <- ols_step_best_subset(model, exclude = c(1))
  pred_exp <- c("x4", "x3 x4", "x2 x3 x4")
  expect_equal(k$metrics$mindex, c(1, 2, 3))
  expect_equal(k$metrics$predictors, pred_exp, ignore_attr = TRUE)
})

test_that("output from best subsets regression is as expected", {
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  expect_snapshot(ols_step_best_subset(model))
})

test_that("output from best subsets regression is as expected when using different metric", {
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  expect_snapshot(ols_step_best_subset(model, metric = "aic"))
})

#' ANOVA
#'
#' ANOVA for regression model.
#'
#' @param model An object of class \code{lm}.
#' @param ... Other arguments.
#'
#' @return \code{ols_anova} returns an object of class \code{"ols_anova"}.
#' An object of class \code{"ols_anova"} is a list containing the following
#' components:
#'
#' \item{error_df}{residual degrees of freedom}
#' \item{model_df}{regression degrees of freedom}
#' \item{total_df}{total degrees of freedom}
#' \item{ess}{error sum of squares}
#' \item{rss}{regression sum of squares}
#' \item{tss}{total sum of squares}
#' \item{rms}{regression mean square}
#' \item{ems}{error mean square}
#' \item{f}{f statistis}
#' \item{p}{p-value for \code{f}}
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#'
#' # anova
#' ols_anova(model)
#'
#' @export
#'
ols_anova <- function(model, ...) UseMethod("ols_anova")

#' @export
#'
ols_anova.default <- function(model, ...) {
  anovam   <- anova(model)
  n        <- length(anovam$Df)
  error_df <- anovam$Df[n]
  model_df <- sum(anovam$Df) - error_df
  total_df <- sum(anovam$Df)
  ess      <- anovam$`Sum Sq`[n]
  tss      <- sum(anovam$`Sum Sq`)
  rss      <- tss - ess
  rms      <- rss / model_df
  ems      <- ess / error_df
  f        <- rms / ems
  p        <- pf(f, model_df, error_df, lower.tail = F)

  result <-
    list(
      error_df = error_df,
      model_df = model_df,
      total_df = total_df,
      ess      = ess,
      rss      = rss,
      tss      = tss,
      rms      = rms,
      ems      = ems,
      f        = f,
      p        = p
    )

  class(result) <- "ols_anova"

  return(result)
}

#' @export
#'
print.ols_anova <- function(x, ...) {
  print_anova(x)
}

test_that("output from anova is as expected", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  result <- ols_anova(model)
  expect_equal(result$error_df, 28)
  expect_equal(result$model_df, 3)
  expect_equal(result$total_df, 31)
  expect_equal(round(result$ess, 3), 194.991)
  expect_equal(round(result$tss, 3), 1126.047)
  expect_equal(round(result$rss, 3), 931.057)
  expect_equal(round(result$rms, 3), 310.352)
  expect_equal(round(result$ems, 3), 6.964)
  expect_equal(round(result$f, 3), 44.566)
  expect_equal(round(result$p, 3), 0)
})

test_that("ols_anova prints the expected output", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  expect_snapshot(ols_anova(model))
})

#' Residual histogram
#'
#' Histogram of residuals for detecting violation of normality assumption.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_hist(model)
#'
#' @family residual diagnostics
#'
#' @importFrom stats dnorm sd
#' @importFrom graphics hist
#'
#' @export
#'
ols_plot_resid_hist <- function(model, print_plot = TRUE) {

  check_model(model)

  k  <- histdata(model)
  h  <- hist(k$resid, plot = FALSE)
  l  <- histn(k$resid, h)
  d2 <- data.frame(x = l$xfit, y = l$yfit)
  d  <- data.frame(x = k$resid)

  p <-
    ggplot(d, aes(x = x)) +
    geom_histogram(bins = 6, color = "black", fill = "#ADD8E6") +
    geom_line(data = d2, aes(x = x, y = y), color = "#0000A0", linewidth = 1.2) +
    xlab("Residuals") + 
    ggtitle("Residual Histogram")

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

histdata <- function(model) {

  resid <- residuals(model)
  minx  <- min(resid) - 1
  maxx  <- max(resid) + 1
  list(resid = resid, minx = minx, maxx = maxx)

}

histn <- function(resid, h) {

  xfit  <- seq(min(resid), max(resid), length = 80)
  yfit  <- dnorm(xfit, mean = mean(resid), sd = sd(resid))
  yfit1 <- yfit * diff(h$mids[1:2]) * length(resid)
  list(xfit = xfit, yfit = yfit1)

}

#' Stepwise hierarchical selection
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering predictors based on p values, in a stepwise hierarchical manner
#' until there is no variable left to enter any more.
#'
#' @param model An object of class \code{lm}; the model should include all
#'   candidate predictor variables.
#' @param p_value p value; variables with p value less/higher than \code{p_val} will
#'   enter/exit the model.
#' @param forward Logical; if \code{TRUE}, performs forward selection else backward.
#' @param progress Logical; if \code{TRUE}, will display variable selection progress.
#' @param details Logical; if \code{TRUE}, will print the regression result at
#'   each step.
#' @param steps Number of steps after which the stepwise procedures should stop.
#'
#' @examples
#' # forward hierarchical selection
#' model <- lm(y ~ ., data = surgical)
#' ols_step_hierarchical(model)
#'
#' # backward hierarchical selection
#' model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + alc_mod + age + gender, data = surgical)
#' ols_step_hierarchical(model, forward = FALSE)
#'
#' # steps
#' model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
#' ols_step_hierarchical(model, steps = 2)
#'
#' @keywords internal
#'
#' @noRd
#'
ols_step_hierarchical <- function(model, p_value = 0.1, forward = TRUE, progress = FALSE, details = FALSE, steps = NULL) {

  if (forward) {
    ols_step_hierarchical_forward(model, p_value, progress, details, steps)
  } else {
    ols_step_hierarchical_backward(model, p_value, progress, details, steps)
  }

}

ols_step_hierarchical_forward <- function(model, p_value = 0.1, progress = FALSE, details = FALSE, steps = NULL) {

  if (details) {
    progress <- FALSE
  }

  l        <- model$model
  nam      <- colnames(attr(model$terms, "factors"))
  n        <- ncol(l)
  response <- names(model$model)[1]
  mlen_p   <- length(nam)
  preds    <- c()
  pvals    <- c()
  aicvals  <- c()
  rsqvals  <- c()
  arsqvals <- c()
  p_val    <- c()
  all_pred <- nam
  cterms   <- all_pred

  if (progress || details) {
    ols_candidate_terms(nam, "forward")
  }

  step     <- 0
  rsq      <- c()
  adjrsq   <- c()
  aic      <- c()
  sbic     <- c()
  cp       <- c()
  sbc      <- c()
  rmse     <- c()

  base_model <- lm(paste(response, "~", 1), data = l)
  rsq_base   <- summary(base_model)$r.squared

  if (details) {
    ols_rsquared_init(NULL, "r2", response, rsq_base)
  }

  if (progress) {
    ols_progress_init("forward")
  }

  if (!is.null(steps)) {
    mlen_p <- steps
  }

  for (i in seq_len(mlen_p)) {
    predictors  <- c(preds, all_pred[i])
    m           <- lm(paste(response, "~", paste(predictors, collapse = " + ")), l)
    m_sum       <- Anova(m)
    pvals       <- c(pvals, m_sum$`Pr(>F)`[i])
    aicvals[i]  <- ols_aic(m)
    rsqvals[i]  <- summary(m)$r.squared
    arsqvals[i] <- summary(m)$adj.r.squared

    if (details) {
      d <- data.frame(predictor = predictors, p_val = pvals, rsq = rsqvals,
                       arsq = arsqvals, aic = aicvals)
      d <- d[order(d$p_val), ]
      ols_stepwise_table(d, predictors, pvals, rsqvals, arsqvals, aicvals)
    }

    if (pvals[i] <= p_value) {
      preds  <- c(preds, cterms[i])
      p_val  <- c(p_val, pvals[i])
      step   <- step + 1
      fr     <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), l)
      rsq    <- c(rsq, fr$rsq)
      adjrsq <- c(adjrsq, fr$adjr)
      aic    <- c(aic, ols_aic(fr$model))
      sbc    <- c(sbc, ols_sbc(fr$model))
      sbic   <- c(sbic, ols_sbic(fr$model, model))
      cp     <- c(cp, ols_mallows_cp(fr$model, model))
      rmse   <- c(rmse, fr$rmse)

      if (details) {
        rsq1 <- tail(rsq, n = 1)
        ols_rsquared_selected("r2", step, preds, response, rsq1)
      }

      if (progress) {
        ols_progress_display(preds, "others")
      }

    } else {
      if (progress) {
        ols_stepwise_break("forward")
      }
      break
    }

  }

  if (details) {
    ols_stepwise_vars(preds, "forward")
  }

  final_model <- lm(paste(response, "~", paste(preds, collapse = " + ")), data = l)

  metrics <- data.frame(step       = seq_len(step),
                        variable   = preds,
                        r2         = rsq,
                        adj_r2     = adjrsq,
                        aic        = aic,
                        sbc        = sbc,
                        sbic       = sbic,
                        mallows_cp = cp,
                        rmse       = rmse)

  result <- list(metrics = metrics,
                 model   = final_model,
                 others  = list(base_model = base_model,
                                full_model = model))

  class(result) <- "ols_step_forward_p"

  return(result)

}

ols_step_hierarchical_backward <- function(model, p_value = 0.1, progress = FALSE, details = FALSE, steps = NULL) {

  if (details) {
    progress <- FALSE
  }

  l        <- model$model
  nam      <- colnames(attr(model$terms, "factors"))
  n        <- ncol(l)
  response <- names(model$model)[1]
  mlen_p   <- length(nam)
  preds    <- c()
  cterms   <- nam

  if (progress || details) {
    ols_candidate_terms(nam, "backward")
  }

  step     <- 0
  rsq      <- c()
  adjrsq   <- c()
  aic      <- c()
  sbic     <- c()
  cp       <- c()
  sbc      <- c()
  rmse     <- c()

  rsq_base   <- summary(model)$r.squared

  if (details) {
    ols_rsquared_init(NULL, "r2", response, rsq_base)
  }

  if (progress) {
    ols_progress_init("backward")
  }

  for (i in rev(seq_len(mlen_p))) {
    predictors  <- cterms[1:i]
    m           <- lm(paste(response, "~", paste(predictors, collapse = " + ")), l)
    m_sum       <- Anova(m)
    pvals       <- m_sum$`Pr(>F)`[1:i]
    p_vals      <- pvals[i]

    if (details) {
      d <- data.frame(predictors = predictors, p_val = pvals)
      ols_stepwise_table_p(d, predictors, pvals)
    }

    if (p_vals >= p_value) {
      preds  <- c(preds, cterms[i])
      step   <- step + 1
      rpred  <- setdiff(cterms, preds)
      fr     <- ols_regress(paste(response, "~", paste(rpred, collapse = " + ")), l)
      rsq    <- c(rsq, fr$rsq)
      adjrsq <- c(adjrsq, fr$adjr)
      aic    <- c(aic, ols_aic(fr$model))
      sbc    <- c(sbc, ols_sbc(fr$model))
      sbic   <- c(sbic, ols_sbic(fr$model, model))
      cp     <- c(cp, ols_mallows_cp(fr$model, model))
      rmse   <- c(rmse, fr$rmse)

      if (progress) {
        ols_progress_display(preds, "others")
      }

      if (details) {
        rsq1   <- tail(rsq, n = 1)
        ols_stepwise_details(step, preds, rpred, response, rsq1, "removed", "rsq")
      }

      if (!is.null(steps)) {
        if (step == steps) {
          break
        }
      }

    } else {

      if (progress || details) {
        ols_stepwise_break(direction = "backward")
      }

      break
    }

  }

  pterms <- setdiff(nam, preds)

  if (details) {
    ols_stepwise_vars(preds, "backward")
  }

  final_model <- lm(paste(response, "~", paste(pterms, collapse = " + ")), data = l)

  metrics <- data.frame(step       = seq_len(step),
                        variable   = preds,
                        r2         = rsq,
                        adj_r2     = adjrsq,
                        aic        = aic,
                        sbic       = sbic,
                        sbc        = sbc,
                        mallows_cp = cp,
                        rmse       = rmse)

  result <- list(metrics = metrics,
                 model   = final_model,
                 others  = list(full_model = model))

  class(result) <- "ols_step_backward_p"

  return(result)

}

#' Leverage
#'
#' @description
#' The leverage of an observation is based on how much the observation's value
#' on the predictor variable differs from the mean of the predictor variable.
#' The greater an observation's leverage, the more potential it has to be an
#' influential observation.
#'
#' @param model An object of class \code{lm}.
#'
#' @return Leverage of the model.
#'
#' @references
#' Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_leverage(model)
#'
#' @family influence measures
#'
#' @importFrom stats hatvalues
#'
#' @export
#'
ols_leverage <- function(model) {
  check_model(model)
  unname(hatvalues(model))  
}


#' Hadi's influence measure
#'
#' @description
#' Measure of influence based on the fact that influential observations in
#' either the response variable or in the predictors or both.
#'
#' @param model An object of class \code{lm}.
#'
#' @return Hadi's measure of the model.
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_hadi(model)
#'
#' @family influence measures
#'
#' @export
#'
ols_hadi <- function(model) {

  check_model(model)

  potential <- hadipot(model)
  residual  <- hadires(model)
  hi        <- potential + residual

  list(hadi      = hi,
       potential = potential,
       residual  = residual)

}

hadipot <- function(model) {

  lev <- ols_leverage(model)
  pii <- 1 - lev
  lev / pii

}

hadires <- function(model) {

  pii    <- 1 - ols_leverage(model)
  q      <- model$rank
  p      <- q - 1
  aov_m  <- anova(model)
  j      <- length(aov_m$Df)
  den    <- sqrt(aov_m[j, 2])
  dii    <- (model$residuals / den) ^ 2    
  first  <- (p + 1) / pii
  second <- dii / (1 - dii)

  first * second

}


#' PRESS
#'
#' @description
#' PRESS (prediction sum of squares) tells you how well the model will predict
#' new data.
#'
#' @param model An object of class \code{lm}.
#'
#' @details
#' The prediction sum of squares (PRESS) is the sum of squares of the prediction
#' error. Each fitted to obtain the predicted value for the ith observation. Use
#' PRESS to assess your model's predictive ability. Usually, the smaller the
#' PRESS value, the better the model's predictive ability.
#'
#' @return Predicted sum of squares of the model.
#'
#' @references
#' Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_press(model)
#'
#' @family influence measures
#'
#' @export
#'
ols_press <- function(model) {

  check_model(model)
  lev <- ols_leverage(model)
  k   <- 1 - lev
  sum((residuals(model) / k) ^ 2) 
  
}

#' Predicted rsquare
#'
#' @description
#' Use predicted rsquared to determine how well the model predicts responses for
#' new observations. Larger values of predicted R2 indicate models of greater
#' predictive ability.
#'
#' @param model An object of class \code{lm}.
#'
#' @return Predicted rsquare of the model.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_pred_rsq(model)
#'
#' @family influence measures
#'
#' @export
#'
ols_pred_rsq <- function(model) {

  check_model(model)
  tss  <- sum(anova(model)[[2]])
  prts <- ols_press(model) / tss
  1 - prts

}

model <- lm(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)

test_that("leverage matches hat from influence.measures", {
  s1 <- round(ols_leverage(model), 4)
  s2 <- unname(round(influence.measures(model)[[1]][, 10], 4))
  expect_equal(s1, s2)
})


test_that("hadi measure matches the textbook example", {
  model <- lm(Nitrogen ~ ComIndl, data = rivers)
  act <- unname(round(ols_hadi(model)$hadi, 2))
  exp <- c(
    0.06, 0.07, 0.58, 0.77, 2.04, 0.10, 0.60, 0.37, 0.07, 0.08,
    0.13, 0.14, 0.15, 0.13, 0.16, 0.09, 0.12, 0.09, 0.19, 0.11
  )
  expect_equal(act, exp)
})

test_that("predicted rsquared output matches the expected result", {
  model <- lm(y ~ x4, data = cement)
  act <- round(ols_pred_rsq(model), 3)
  expect_equal(act, 0.56)

  model <- lm(y ~ x2, data = cement)
  act <- round(ols_pred_rsq(model), 3)
  expect_equal(act, 0.557)

  model <- lm(y ~ x1 + x2, data = cement)
  act <- round(ols_pred_rsq(model), 3)
  expect_equal(act, 0.965)

  model <- lm(y ~ x1 + x4, data = cement)
  act <- round(ols_pred_rsq(model), 3)
  expect_equal(act, 0.955)

  model <- lm(y ~ x1 + x2 + x4, data = cement)
  act <- round(ols_pred_rsq(model), 3)
  expect_equal(act, 0.969)

  model <- lm(y ~ x1 + x2 + x3, data = cement)
  act <- round(ols_pred_rsq(model), 3)
  expect_equal(act, 0.967)

  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  act <- round(ols_pred_rsq(model), 3)
  expect_equal(act, 0.959)
})


test_that("output from ols_test_outlier is as expected", {

  model <- lm(y ~ ., data = stepdata)
  act   <- ols_test_outlier(model)
  expect_equal(round(act$studentized_residual, 2), 4.33)
  expect_equal(round(act$unadjusted_p_val, 2), 0)
  expect_equal(round(act$bonferroni_p_val, 2), 0.3)
})

#' Residual fit spread plot
#'
#' Plot to detect non-linearity, influential observations and outliers.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details
#' Consists of side-by-side quantile plots of the centered fit and the
#' residuals. It shows how much variation in the data is explained by the fit
#' and how much remains in the residuals. For inappropriate models, the spread
#' of the residuals in such a plot is often greater than the spread of the
#' centered fit.
#'
#' @references
#' Cleveland, W. S. (1993). Visualizing Data. Summit, NJ: Hobart Press.
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#'
#' # residual fit spread plot
#' ols_plot_resid_fit_spread(model)
#'
#' # fit mean plot
#' ols_plot_fm(model)
#'
#' # residual spread plot
#' ols_plot_resid_spread(model)
#'
#' @importFrom stats ecdf
#' @importFrom gridExtra grid.arrange
#'
#' @export
#'
ols_plot_resid_fit_spread <- function(model, print_plot = TRUE) {

  p1 <- ols_plot_fm(model, FALSE)
  p2 <- ols_plot_resid_spread(model, FALSE)

  if (print_plot) {
    grid.arrange(p1, p2, ncol = 2, top = "Residual Fit Spread Plot")
  } else {
    return(list(fm_plot = p1, rsd_plot = p2))
  }

}

#' @rdname ols_plot_resid_fit_spread
#' @export
#'
ols_plot_fm <- function(model, print_plot = TRUE) {

  check_model(model)

  p <- ols_plot_rfs(model)

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}


#' @rdname ols_plot_resid_fit_spread
#' @export
#'
ols_plot_resid_spread <- function(model, print_plot = TRUE) {

  check_model(model)

  p <- ols_plot_rfs(model, fit_mean = FALSE)

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}


ols_plot_rfs <- function(model, fit_mean = TRUE) {

  if (fit_mean) {
    d <- ols_prep_rfsplot_fmdata(model)  
    y_lab <- "Fit - Mean"
  } else {
    d <- ols_prep_rfsplot_rsdata(model)
    y_lab <- "Residual"
  }

  ymin <- min(d$y) + (0.25 * min(d$y))
  ymax <- max(d$y) + (0.25 * max(d$y))

  p <-
    ggplot(d, aes(x = x, y = y)) +
    geom_point(shape = 1, color = "blue") +
    xlab("Proportion Less") +
    ylab(y_lab) +
    ggtitle("Residual Fit Spread Plot") +
    xlim(c(-0.2, 1.2)) +
    ylim(c(ymin, ymax))

  return(p)

}

ols_step_forward <- function(model, metric = c("aic", "sbic", "sbc", "rsq", "adjrsq"), include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {

  if (details) {
    progress <- FALSE
  }

  check_inputs(model, include, exclude, progress, details)

  criteria <- match.arg(metric)
  response <- names(model$model)[1]
  l        <- model$model
  indterms <- coeff_names(model)
  lenterms <- length(indterms)

  if (is.numeric(include)) {
    include <- indterms[include]
  }

  if (is.numeric(exclude)) {
    exclude <- indterms[exclude]
  }

  lockterm <- c(include, exclude)
  nam      <- setdiff(indterms, lockterm)
  all_pred <- nam
  mlen_p   <- length(all_pred)
  preds    <- include

  if (progress || details) {
    ols_candidate_terms(nam, "forward")
    if (interactive()) {
      Sys.sleep(0.5)  
    }
  }
    
  aics  <- c()
  bics  <- c()
  sbics <- c()
  rsq   <- c()
  arsq  <- c()

  base_model  <- ols_base_model(include, response, l)
  base_metric <- ols_base_criteria(base_model, criteria)

  if (details) {
    ols_base_model_stats(response, include, criteria, base_metric)
    if (interactive()) {
      Sys.sleep(0.5)  
    }
  }

  for (i in seq_len(mlen_p)) {

    predictors <- c(include, all_pred[i])
    k <- ols_regress(paste(response, "~", paste(predictors, collapse = " + ")), data = l)

    aics[i]  <- ols_aic(k$model)
    bics[i]  <- ols_sbc(k$model)
    sbics[i] <- ols_sbic(k$model, model)
    rsq[i]   <- k$rsq
    arsq[i]  <- k$adjr
  }

  da <- data.frame(predictors = all_pred, aics = aics, bics = bics, sbics = sbics, rsq = rsq, arsq = arsq)
  da2 <- ols_sort_metrics(da, criteria)

  if (details) {
    ols_stepwise_metrics(da2, criteria, da2$predictors, aics, bics, sbics, rsq, arsq, "add")
    if (interactive()) {
      Sys.sleep(0.5)  
    }
  }

  mat  <- switch(criteria,
    aic    = aics,
    sbc    = bics,
    sbic   = sbics,
    rsq    = rsq,
    adjrsq = arsq)

  minc <- ols_threshold(mat, criteria)
  crit <- ols_f_criteria(criteria, mat, minc, base_metric)

  if (crit) {
    laic     <- aics[minc]
    lbic     <- bics[minc]
    lsbic    <- sbics[minc]
    lrsq     <- rsq[minc]
    larsq    <- arsq[minc]
    preds    <- c(preds, all_pred[minc])
    lpreds   <- length(preds) - length(include)
    all_pred <- all_pred[-minc]
    len_p    <- length(all_pred)
    step     <- 1

    if (progress) {
      ols_progress_init("forward")
      if (interactive()) {
        Sys.sleep(0.5)  
      }
      ols_progress_display(preds, "others")
      if (interactive()) {
        Sys.sleep(0.5)  
      }
    }
  } else {
    ols_stepwise_break("forward")
    if (interactive()) {
      Sys.sleep(0.5)  
    }
  }

  if (!is.null(steps)) {
    mlen_p <- steps
  }

  while (step < mlen_p) {

    aics  <- c()
    bics  <- c()
    sbics <- c()
    rsq   <- c()
    arsq  <- c()
    mo    <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), data = l)
    met   <- ols_base_criteria(mo$model, criteria)

    if (details) {
      ols_stepwise_details(step, preds, preds, response, met, "added", criteria)
      if (interactive()) {
        Sys.sleep(0.5)  
      }
    }

    for (i in seq_len(len_p)) {

      predictors <- c(preds, all_pred[i])
      k          <- ols_regress(paste(response, "~", paste(predictors, collapse = " + ")), data = l)
      aics[i]    <- ols_aic(k$model)
      bics[i]    <- ols_sbc(k$model)
      sbics[i]   <- ols_sbic(k$model, model)
      rsq[i]     <- k$rsq
      arsq[i]    <- k$adjr
    l}

    if (details) {
      da  <- data.frame(predictors = all_pred, aics = aics, bics = bics, sbics = sbics, rsq = rsq, arsq = arsq)
      da2 <- ols_sort_metrics(da, criteria)
      ols_stepwise_metrics(da2, criteria, da2$predictors, aics, bics, sbics, rsq, arsq, "add")
      if (interactive()) {
        Sys.sleep(0.5)  
      }
    }

    mat  <- switch(criteria,
      aic    = aics,
      sbc    = bics,
      sbic   = sbics,
      rsq    = rsq,
      adjrsq = arsq)

    faic  <- switch(criteria,
      aic    = laic,
      sbc    = lbic,
      sbic   = lsbic,
      rsq    = lrsq,
      adjrsq = larsq)

    minaic <- ols_threshold(mat, criteria)
    crit   <- ols_next_criteria(criteria, mat, minaic, faic, lpreds)

    if (crit) {

      preds    <- c(preds, all_pred[minaic])
      minc     <- aics[minaic]
      mbic     <- bics[minaic]
      msbic    <- sbics[minaic]
      mrsq     <- rsq[minaic]
      marsq    <- arsq[minaic]
      laic     <- c(laic, minc)
      lbic     <- c(lbic, mbic)
      lsbic    <- c(lsbic, msbic)
      lrsq     <- c(lrsq, mrsq)
      larsq    <- c(larsq, marsq)
      lpreds   <- length(preds) - length(include)
      all_pred <- all_pred[-minaic]
      len_p    <- length(all_pred)
      step     <- step + 1

      if (progress) {
        ols_progress_display(preds, "others")
        if (interactive()) {
          Sys.sleep(0.5)  
        }
      }
    } else {
      if (progress || details) {
        ols_stepwise_break("forward")
        if (interactive()) {
          Sys.sleep(0.5)  
        }
      }
      break
    }
  }

  if (details) {
    ols_stepwise_vars(preds, "forward")
    if (interactive()) {
      Sys.sleep(0.5)  
    }
  }

  final_model <- lm(paste(response, "~", paste(preds, collapse = " + ")), data = l)

  if (is.null(include)) {
    var_selected <- preds
  } else {
    var_selected <- preds[-seq_len(length(include))]
  }

  indicator <- switch(criteria,
    aic    = "aic",
    sbc    = "sbc",
    sbic   = "sbic",
    rsq    = "r2",
    adjrsq = "adj_r2")

  metrics     <- data.frame(step     = seq_len(step),
                            variable = var_selected,
                            r2       = lrsq,
                            adj_r2   = larsq,
                            aic      = laic,
                            sbc      = lbic,
                            sbic     = lsbic)

  out <- list(metrics = metrics,
              model   = final_model,
              others  = list(base_model = base_model,
                             criteria   = indicator,
                             direction  = "forward",
                             full_model = model))

  return(out)
}

ols_step_backward <- function(model, metric = c("aic", "sbic", "sbc", "rsq", "adjrsq"), include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {

  if (details) {
    progress <- FALSE
  }

  check_inputs(model, include, exclude, progress, details)

  criteria <- match.arg(metric)
  response <- names(model$model)[1]
  l        <- model$model
  indterms <- coeff_names(model)
  lenterms <- length(indterms)

  if (is.numeric(include)) {
    include <- indterms[include]
  }

  if (is.numeric(exclude)) {
    exclude <- indterms[exclude]
  }

  nam   <- setdiff(indterms, exclude)
  cterm <- setdiff(nam, include)
  preds <- nam

  if (progress || details) {
    ols_candidate_terms(nam, "backward")
  }
    
  mi    <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), data = l)
  laic  <- ols_aic(model)
  lbic  <- ols_sbc(model)
  lsbic <- ols_sbic(model, model)
  lrsq  <- mi$rsq
  larsq <- mi$adjr

  base_metric <- ols_base_criteria(model, criteria)

  if (details) {
    ols_base_model_stats(response, preds, criteria, base_metric)
  }

  ilp   <- length(preds)
  end   <- FALSE
  step  <- 0
  rpred <- c()
  aics  <- c()
  bics  <- c()
  sbics <- c()
  rsq   <- c()
  arsq  <- c()

  for (i in seq_len(ilp)) {

    predictors <- preds[-i]

    m <- ols_regress(paste(response, "~", paste(predictors, collapse = " + ")), data = l)

    aics[i]  <- ols_aic(m$model)
    bics[i]  <- ols_sbc(m$model)
    sbics[i] <- ols_sbic(m$model, model)
    rsq[i]   <- m$rsq
    arsq[i]  <- m$adjr
  }

  da  <- data.frame(predictors = preds, aics = aics, bics = bics, sbics = sbics, rsq = rsq, arsq = arsq)
  da2 <- ols_sort_metrics(da, criteria)

  mat  <- switch(criteria,
               aic    = aics,
               sbc    = bics,
               sbic   = sbics,
               rsq    = rsq,
               adjrsq = arsq)

  if (grepl("rsq", criteria)) {
    da3 <- cbind(loc = order(-mat), da2)  
  } else {
    da3 <- cbind(loc = order(mat), da2)
  }
 

  if (details) {
    ols_stepwise_metrics(da2, criteria, da2$predictors, aics, bics, sbics, rsq, arsq, "remove")
  }

  linc <- length(include)

  if (progress) {
    ols_progress_init("backward")
  }

  while (!end) {

    for (i in seq_len(linc)) {
      lnam <- which(da3$predictors == include[i])
      da3 <- da3[-lnam, ]
    }

    minc <- da3$loc[1]
    crit <- ols_f_criteria(criteria, mat, minc, base_metric)

    if (crit) {

      rpred <- c(rpred, preds[minc])
      preds <- preds[-minc]
      ilp   <- length(preds)
      step  <- step + 1
      base_metric <- mat[minc]

      mi <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), data = l)

      laic   <- c(laic, mi$aic)
      lbic   <- c(lbic, mi$sbc)
      lsbic  <- c(lsbic, mi$sbic)
      lrsq   <- c(lrsq, mi$rsq)
      larsq  <- c(larsq, mi$adjr)
      aics   <- c()
      bics   <- c()
      sbics  <- c()
      rsq    <- c()
      arsq   <- c()

      if (progress) {
        ols_progress_display(rpred, "others")
      }

      for (i in seq_len(ilp)) {

        predictors <- preds[-i]

        m <- ols_regress(paste(response, "~", paste(predictors, collapse = " + ")), data = l)

        aics[i]  <- ols_aic(m$model)
        bics[i]  <- ols_sbc(m$model)
        sbics[i] <- ols_sbic(m$model, model)
        rsq[i]   <- m$rsq
        arsq[i]  <- m$adjr
      }

      da  <- data.frame(predictors = preds, aics = aics, bics = bics, sbics = sbics, rsq = rsq, arsq = arsq)
      da2 <- ols_sort_metrics(da, criteria)

      mat  <- switch(criteria,
               aic    = aics,
               sbc    = bics,
               sbic   = sbics,
               rsq    = rsq,
               adjrsq = arsq)

      if (grepl("rsq", criteria)) {
        da3 <- cbind(loc = order(-mat), da2)  
      } else {
        da3 <- cbind(loc = order(mat), da2)
      }

      if (details) {
        ols_stepwise_details(step, rpred, preds, response, base_metric, "removed", criteria)
        ols_stepwise_metrics(da2, criteria, da2$predictors, aics, bics, sbics, rsq, arsq, "remove")
      }

      if (!is.null(steps)) {
        if (step == steps) {
          end <- TRUE
          if (progress || details) {
            ols_stepwise_break("backward")
          }
        }
      }
    } else {
      end <- TRUE
      if (progress || details) {
        ols_stepwise_break("backward")
      }
    }

  }


  if (details) {
    ols_stepwise_vars(rpred, "backward")
  }

  indicator <- switch(criteria,
    aic    = "aic",
    sbc    = "sbc",
    sbic   = "sbic",
    rsq    = "r2",
    adjrsq = "adj_r2")

  final_model <- lm(paste(response, "~", paste(preds, collapse = " + ")), data = l)

  metrics     <- data.frame(step     = seq_len(step),
                            variable = rpred,
                            r2       = tail(lrsq,  n = step),
                            adj_r2   = tail(larsq, n = step),
                            aic      = tail(laic,  n = step),
                            sbc      = tail(lbic,  n = step),
                            sbic     = tail(lsbic,  n = step))

  out <- list(metrics = metrics,
              model   = final_model,
              others  = list(full_model = model,
                             criteria   = indicator,
                             direction  = "backward"))

  return(out)
}

ols_step_both <- function(model, metric = c("aic", "sbic", "sbc", "rsq", "adjrsq"), include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {

  if (details) {
    progress <- FALSE
  }

  check_inputs(model, include, exclude, progress, details)

  criteria <- match.arg(metric)
  response <- names(model$model)[1]
  l        <- model$model
  nam      <- coeff_names(model)
  indterms <- nam
  lenterms <- length(indterms)
  len_inc  <- length(include) + 1

  if (is.numeric(include)) {
    include <- indterms[include]
  }

  if (is.numeric(exclude)) {
    exclude <- indterms[exclude]
  }

  if (progress || details) {
    ols_candidate_terms(nam, "both")
  }

  lockterm   <- c(include, exclude)
  predictors <- setdiff(nam, lockterm)
  mlen_p     <- length(predictors)
  tech       <- c("addition", "removal")

  base_model <- ols_base_model(include, response, l)
  bmetric <- ols_base_criteria(base_model, criteria)

  if (details) {
    ols_base_model_stats(response, include, criteria, bmetric)
  }

  step      <- 0
  all_step  <- 0
  preds     <- include
  var_index <- c()
  method    <- c()
  laic      <- c()
  lbic      <- c()
  lsbic     <- c()
  lrsq      <- c()
  larsq     <- c()

  if (progress) {
    ols_progress_init("both")
  }
  
  if (!is.null(steps)) {
    mlen_p <- steps
  }

  while (step < mlen_p) {

    aics  <- c()
    bics  <- c()
    sbics <- c()
    rsq   <- c()
    arsq  <- c()
    lpds  <- length(predictors)

    for (i in seq_len(lpds)) {

      predn <- c(preds, predictors[i])

      m <- ols_regress(paste(response, "~", paste(predn, collapse = " + ")), data = l)

      aics[i]  <- ols_aic(m$model)
      bics[i]  <- ols_sbc(m$model)
      sbics[i] <- ols_sbic(m$model, model)
      rsq[i]   <- m$rsq
      arsq[i]  <- m$adjr
    }

    da <- data.frame(predictors = predictors, aics = aics, bics = bics, sbics = sbics, rsq = rsq, arsq = arsq)

    if (details) {
      ols_stepwise_metrics(da, criteria, predictors, aics, bics, sbics, rsq, arsq, "add")
    }

    mat  <- switch(criteria,
    aic    = aics,
    sbc    = bics,
    sbic   = sbics,
    rsq    = rsq,
    adjrsq = arsq)

    minc <- ols_threshold(mat, criteria)
    crit <- ols_f_criteria(criteria, mat, minc, bmetric)

    if (crit) {
      bmetric    <- mat[minc]
      preds      <- c(preds, predictors[minc])
      predictors <- predictors[-minc]
      lpds       <- length(predictors)
      method     <- c(method, tech[1])
      lpreds     <- length(preds)
      var_index  <- c(var_index, preds[lpreds])
      step       <- step + 1
      all_step   <- all_step + 1
      maic       <- aics[minc]
      mbic       <- bics[minc]
      msbic      <- sbics[minc]
      mrsq       <- rsq[minc]
      marsq      <- arsq[minc]
      laic       <- c(laic, maic)
      lbic       <- c(lbic, mbic)
      lsbic      <- c(lsbic, msbic)
      lrsq       <- c(lrsq, mrsq)
      larsq      <- c(larsq, marsq)

      if (progress) {
        ols_progress_display(preds, "both", "added")
      }

      kaic  <- switch(criteria,
          aic    = maic,
          sbc    = mbic,
          sbic   = msbic,
          rsq    = mrsq,
          adjrsq = marsq)

      if (details) {
        ols_stepwise_details(all_step, preds, preds, response, kaic, "added", criteria)
      }

      if (lpreds > 1) {

        aics  <- c()
        bics  <- c()
        sbics <- c()
        rsq   <- c()
        arsq  <- c()
        j     <- 1

        for (i in len_inc:lpreds) {

          preda <- preds[-i]

          m <- ols_regress(paste(response, "~", paste(preda, collapse = " + ")), data = l)

          aics[j]  <- ols_aic(m$model)
          bics[j]  <- ols_sbc(m$model)
          sbics[j] <- ols_sbic(m$model, model)
          rsq[j]   <- m$rsq
          arsq[j]  <- m$adjr

          j <- j + 1
        }

        da <- data.frame(predictors = preds[len_inc: lpreds], aics = aics, bics = bics, sbics = sbics, rsq = rsq, arsq = arsq)

        if (details) {
          ols_stepwise_metrics(da, criteria, da$predictors, aics, bics, sbics, rsq, arsq, "remove")
        }

        mat  <- switch(criteria,
          aic    = aics,
          sbc    = bics,
          sbic   = sbics,
          rsq    = rsq,
          adjrsq = arsq)

        faic  <- switch(criteria,
          aic    = laic,
          sbc    = lbic,
          sbic   = lsbic,
          rsq    = lrsq,
          adjrsq = larsq)

        minc2 <- ols_threshold(mat, criteria)
        crit  <- ols_next_criteria(criteria, mat, minc2, faic, all_step)

        if (crit) {
          bmetric   <- mat[minc2]
          maic      <- aics[minc2]
          mbic      <- bics[minc2]
          msbic     <- sbics[minc2]
          mrsq      <- rsq[minc2]
          marsq     <- arsq[minc2]
          laic      <- c(laic, maic)
          lbic      <- c(lbic, mbic)
          lsbic     <- c(lsbic, msbic)
          lrsq      <- c(lrsq, mrsq)
          larsq     <- c(larsq, marsq)
          var_index <- c(var_index, preds[minc2 + length(include)])
          method    <- c(method, tech[2])
          all_step  <- all_step + 1

          temp <- preds[minc2 + length(include)]

          if (progress) {
            ols_progress_display(temp, "both", "removed")
          }

          preds <- preds[-(minc2 + length(include))]
          lpreds <- length(preds)

          kaic  <- switch(criteria,
            aic    = maic,
            sbc    = mbic,
            sbic   = msbic,
            rsq    = mrsq,
            adjrsq = marsq)

          if (details) {
            ols_stepwise_details(all_step, temp, preds, response, kaic, "removed", criteria)
          }
        }
      } else {
        preds <- preds
        all_step <- all_step
      }
    } else {
      if (progress || details) {
        ols_stepwise_break("both")
      }
      break
    }
  }

  if (details) {
    ols_stepwise_vars(preds, "both")
  }

  indicator <- switch(criteria,
    aic    = "aic",
    sbc    = "sbc",
    sbic   = "sbic",
    rsq    = "r2",
    adjrsq = "adj_r2")

  final_model <- lm(paste(response, "~", paste(preds, collapse = " + ")), data = l)

  metrics     <- data.frame(step     = seq_len(all_step),
                            variable = var_index,
                            method   = method,
                            r2       = lrsq,
                            adj_r2   = larsq,
                            aic      = laic,
                            sbc      = lbic,
                            sbic     = lsbic)

  out <- list(metrics = metrics,
              model   = final_model,
              others   = list(base_model = base_model,
                              criteria   = indicator,
                              direction  = "both",
                              full_model = model))

  return(out)
}

#' Model formula
#'
#' Returns the model formula.
#'
#' @param model An object of class \code{lm}.
#'
#' @return An object of class \code{formula}.
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_get_formula(model)
#'
#' @family model info
#'
#' @export
#'
ols_get_formula <- function(model) {
  formula(model)
}

#' Interaction terms
#'
#' Returns interaction terms present in the model.
#'
#' @param model An object of class \code{lm}.
#'
#' @return \code{Character} vector or \code{NULL}.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl * hp * disp + gear * drat, data = mtcars)
#' ols_get_interaction_terms(model)
#'
#' @export
#'
ols_get_interaction_terms <- function(model) {
  terms <- model$terms
  i <- attr(terms, 'order')
  if (any(i > 1)) {
    iterms <- i > 1
    return(attr(terms, 'term.labels')[iterms])
  } else {
    return(NULL)
  }
}

#' Model variables
#' 
#' Return model variables.
#'
#' @param model An object of class \code{lm}.
#'
#' @return List with response and predictors.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl * hp * disp + gear * drat, data = mtcars)
#' ols_get_variables(model)
#'
#' @export
#'
ols_get_variables <- function(model) {
  vars  <- names(model$model)
  n     <- length(vars)
  resp  <- vars[1]
  if (n == 1) {
    preds <- NULL
  } else {
    preds <- vars[2:n] 
  }
  list(response = resp, predictors = preds)
}

#' Model data
#' 
#' Returns data used in model.
#' 
#' @param model An object of class \code{lm}.
#'
#' @return An object of class \code{data.frame}.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl * hp * disp + gear * drat, data = mtcars)
#' ols_get_data(model)
#'
#' @export
#'
ols_get_data <- function(model) {
  model$model
}


#' Degrees of freedom
#' 
#' Returns residual degrees of freedom.
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Residual degrees of freedom.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl * hp * disp + gear * drat, data = mtcars)
#' ols_get_df(model)
#'
#' @export
#'
ols_get_df <- function(model) {
  model$df.residual
}


#' Model intercept
#' 
#' Returns the value of the model intercept. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return The value of the intercept. 
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_intercept(model)
#'
#' model <- lm(mpg ~ 0 + wt + cyl + hp + disp, data = mtcars)
#' ols_get_intercept(model)
#'
#' @export
#'
ols_get_intercept <- function(model) {
  if (ols_has_intercept(model)) {
    return(model$coefficients[[1]])
  }
  return(NULL)
}

#' Model matrix
#' 
#' Returns the design matrix. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_model_matrix(model)
#'
#' @export
#'
ols_get_model_matrix <- function(model) {
  model.matrix(model)
}

#' Fitted values
#' 
#' Returns the fitted values. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Fitted values. 
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_predicted(model)
#'
#' @export
#'
ols_get_predicted <- function(model) {
  as.numeric(model$fitted.values)
}

#' Model residuals 
#' 
#' Returns the residuals from the model. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Residuals from the regression model. 
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_residuals(model)
#'
#' @export
#'
ols_get_residuals <- function(model) {
  as.numeric(model$residuals)
}

#' Sigma 
#' 
#' Returns the residual standard deviation.
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Residual standard deviation.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_sigma(model)
#'
#' @export
#'
ols_get_sigma <- function(model) {
  summary(model)$sigma
}

#' Variance covariance matrix
#' 
#' Returns the variance covariance matrix from the model. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Variance covariance matrix. 
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_vcov(model)
#'
#' @importFrom stats vcov
#'
#' @export
#'
ols_get_vcov <- function(model) {
  vcov(model)
}

#' Deviance 
#' 
#' Returns the model deviance. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Model deviance. 
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_deviance(model)
#'
#' @export
#'
ols_get_deviance <- function(model) {
  deviance(model)
}

#' Model parameters 
#' 
#' Returns the coefficients. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return A \code{data.frame} with model coefficients.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_parameters(model)
#'
#' @export
#'
ols_get_parameters <- function(model) {
  params <- model$coefficients
  data.frame(Parameter = names(params), Estimate = unname(params))
}

#' Predictor variable data
#' 
#' Returns the data of all model predictors.
#' 
#' @param model An object of class \code{lm}.
#'
#' @return A \code{data.frame} with model predictors.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_predictors(model)
#'
#' @export
#'
ols_get_predictors <- function(model) {
  model$model[, -1]
}

#' Response variable data
#' 
#' Returns the values of the response variable of a model.
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Values of the response variable.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_response(model)
#'
#' @export
#'
ols_get_response <- function(model) {
  model$model[, 1]
}

#' Model function call 
#' 
#' Returns the model's function call. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Model's function call. 
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_get_call(model)
#'
#' @export
#'
ols_get_call <- function(model) {
  model$call
}

#' Observations 
#' 
#' Returns the number of observation from a model. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Number of observations.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_count_obs(model)
#'
#' @export
#'
ols_count_obs <- function(model) {
  nrow(model$model)
}

#' Parameters count
#'
#' Returns the number of parameters from a model. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return Number of parameters.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_count_parameters(model)
#'
#' @export
#'
ols_count_parameters <- function(model) {
  nrow(ols_get_parameters(model))
}

#' Model intercept check 
#' 
#' Checks if model has an intercept. 
#' 
#' @param model An object of class \code{lm}.
#'
#' @return \code{TRUE} if \code{model} has an intercept, \code{FALSE} otherwise.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_has_intercept(model)
#'
#' model <- lm(mpg ~ 0 + wt + cyl + hp + disp + gear + drat, data = mtcars)
#' ols_has_intercept(model)
#'
#' @export
#'
ols_has_intercept <- function(model) {
  as.logical(attr(model$terms, "intercept"))
}

#' Model object info
#'
#' Comprehensive information from model objects.
#' 
#' @param model An object of class \code{lm}.
#'
#' @return A list with information about model objects.
#'
#' @examples
#' model <- lm(mpg ~ wt + cyl * hp * disp + gear * drat, data = mtcars)
#' ols_model_info(model)
#'
#' @export
#'
ols_model_info <- function(model) {
  return(
    list(
      call               = ols_get_call(model),
      degrees_of_freedom = ols_get_df(model),
      deviance           = ols_get_deviance(model),
      formula            = ols_get_formula(model),
      has_intercept      = ols_has_intercept(model),
      interaction_terms  = ols_get_interaction_terms(model),
      intercept_value    = ols_get_intercept(model),
      obs                = ols_count_obs(model),
      parameters         = ols_get_parameters(model),
      sigma              = ols_get_sigma(model),
      variables          = ols_get_variables(model)
    )
  )
}

test_that("the model forumula is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- as.formula(mpg ~ disp + hp + wt)
  got   <- ols_get_formula(model)
  expect_equal(want, got)
})

test_that("interaction terms are returned", {
  model <- lm(mpg ~ wt * cyl + vs * hp * gear + carb, data = mtcars)
  want  <- c("wt:cyl", "vs:hp", "vs:gear",  "hp:gear", "vs:hp:gear")
  got   <- ols_get_interaction_terms(model)
  expect_equal(want, got)
})

test_that("NULL is returned in absence of interaction terms", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- NULL
  got   <- ols_get_interaction_terms(model)
  expect_null(want)
})

test_that("predictors does not return null when model has at least one predictor", {
  model <- lm(mpg ~ wt * cyl + vs * hp * gear + carb, data = mtcars)
  want  <- c("wt", "cyl", "vs", "hp", "gear", "carb")
  got   <- ols_get_variables(model)$predictors
  expect_equal(want, got)
})

test_that("predictors returns null in case of intercept only model", {
  model <- lm(mpg ~ 1, data = mtcars)
  want  <- NULL
  got   <- ols_get_variables(model)
  expect_null(want)
})

test_that("response is not null", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- c("mpg")
  got   <- ols_get_variables(model)$response
  expect_equal(want, got)
})

test_that("model data is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- c("mpg", "disp", "hp", "wt")
  got   <- names(ols_get_data(model))
  expect_equal(want, got)
})

test_that("residual degrees of freedom is returned", {
  model <- lm(mpg ~ wt * cyl + vs * hp * gear + carb, data = mtcars)
  want  <- 20
  got   <- ols_get_df(model)
  expect_equal(want, got)
})

test_that("intercept is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- 37.11
  got   <- round(ols_get_intercept(model), 2)
  expect_equal(want, got)
})


test_that("NULL is returned in the absence of intercept", {
  model <- lm(mpg ~ 0 + disp + hp + wt, data = mtcars)
  want  <- NULL
  got   <- ols_get_intercept(model)
  expect_null(want)
})

test_that("model matrix is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- c("(Intercept)", "disp", "hp", "wt")
  got   <- colnames(ols_get_model_matrix(model))
  expect_equal(want, got)
})

test_that("fitted values are returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- 32
  got   <- length(ols_get_predicted(model))
  expect_equal(want, got)
})

test_that("residuals are returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- 32
  got   <- length(ols_get_residuals(model))
  expect_equal(want, got)
})

test_that("sigma is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- 2.64
  got   <- round(ols_get_sigma(model), 2)
  expect_equal(want, got)
})

test_that("variance covariance matrix is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- c("(Intercept)", "disp", "hp", "wt")
  got   <- colnames(ols_get_vcov(model))
  expect_equal(want, got)
})

test_that("deviance is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- 194.99
  got   <- round(ols_get_deviance(model), 2)
  expect_equal(want, got)
})

test_that("parameters and estimates are returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- c("(Intercept)", "disp", "hp", "wt")
  got   <- ols_get_parameters(model)[, 1]
  expect_equal(want, got)
})

test_that("predictor data is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- c("disp", "hp", "wt")
  got   <- names(ols_get_predictors(model))
  expect_equal(want, got)
})

test_that("response data is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- c(21.0, 21.0, 22.8, 21.4, 18.7, 18.1)
  got   <- head(ols_get_response(model))
  expect_equal(want, got)
})

test_that("model call is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- as.call(lm(mpg ~ disp + hp + wt, data = mtcars)$call)
  got   <- ols_get_call(model)
  expect_equal(want, got)
})

test_that("number of observations is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- 32
  got   <- ols_count_obs(model)
  expect_equal(want, got)
})

test_that("number of parameters is returned", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  want  <- 4
  got   <- ols_count_parameters(model)
  expect_equal(want, got)
})

test_that("model has intercept", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  got   <- ols_has_intercept(model)
  expect_true(got)
})

test_that("model has no intercept", {
  model <- lm(mpg ~ 0 + disp + hp + wt, data = mtcars)
  got   <- ols_has_intercept(model)
  expect_false(got)
})

test_that("model info is correct", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  got   <- ols_model_info(model)
  expect_equal(got$degrees_of_freedom, 28)
  expect_true(got$has_intercept)
  expect_null(got$interaction_terms)
  expect_equal(got$obs, 32)
})

.onAttach <- function(...) {

  if (!interactive() || stats::runif(1) > 0.1) return()

  pkgs          <- utils::available.packages()
  cran_version  <- package_version(pkgs["olsrr", "Version"])
  local_version <- utils::packageVersion("olsrr")
  behind_cran   <- cran_version > local_version

  tips <- c(
    "Learn more about olsrr at https://github.com/rsquaredacademy/olsrr/.",
    "Use suppressPackageStartupMessages() to eliminate package startup messages.",
    "Need help getting started with regression models? Visit: https://www.rsquaredacademy.com",
    "Check out our interactive app for quick data exploration. Visit: https://apps.rsquaredacademy.com/."
  )

  tip <- sample(tips, 1)

  if (interactive()) {
    if (behind_cran) {
      msg <- c("A new version of olsrr is available with bug fixes and new features.")
      packageStartupMessage(msg, "\nWould you like to install it?")
      if (utils::menu(c("Yes", "No")) == 1) {
        utils::update.packages("olsrr")
      }
    } else {
      packageStartupMessage(paste(strwrap(tip), collapse = "\n"))
    }
  }

}

#' Standardized residual chart
#'
#' Chart for identifying outliers.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 2.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details
#' Standardized residual (internally studentized) is the residual divided by
#' estimated standard deviation.
#'
#' @return \code{ols_plot_resid_stand} returns  a list containing the
#' following components:
#'
#' \item{outliers}{a \code{data.frame} with observation number and \code{standardized resiudals} that
#' exceed \code{threshold}} for classifying an observation as an outlier
#' \item{threshold}{\code{threshold} for classifying an observation as an outlier}
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_stand(model)
#' ols_plot_resid_stand(model, threshold = 3)
#'
#' @importFrom stats rstandard
#'
#' @seealso [ols_plot_resid_stud()]
#'
#' @export
#'
ols_plot_resid_stand <- function(model, threshold = NULL, print_plot = TRUE) {

  check_model(model)

  if (is.null(threshold)) {
    threshold <- 2
  }

  d <- ols_prep_srchart_data(model, threshold)
  f <- d[d$color == "outlier", c("obs", "sdres")]
  colnames(f) <- c("observation", "stand_resid")

  p <-
    ggplot(d, aes(x = obs, y = sdres, label = txt, ymin = 0, ymax = sdres)) +
    geom_linerange(colour = "blue") +
    geom_point(shape = 1, colour = "blue") +
    geom_text(hjust = -0.2, nudge_x = 0.15, size = 3, family = "serif",
              fontface = "italic", colour = "darkred", na.rm = TRUE) +
    geom_hline(yintercept = 0, colour = "gray") +
    geom_hline(yintercept = c(threshold, -threshold), colour = "red")

  p <-
    p +
    annotate("text", x = Inf, y = Inf, hjust = 1.5, vjust = 2, family = "serif",
             fontface = "italic", colour = "darkred",
             label = paste0("Threshold: abs(", threshold, ")"))

  p <-
    p +
    xlab("Observation") +
    ylab("Standardized Residuals") +
    ggtitle("Standardized Residuals Chart")

  if (print_plot) {
    suppressWarnings(print(p))
  } else {
    return(list(plot = p, outliers = f, threshold = threshold))
  }

}

#' @title Residual vs regressors plot for shiny app
#'
#' @description Graph to determine whether we should add a new predictor to the
#' model already containing other predictors. The residuals from the model is
#' regressed on the new predictor and if the plot shows non random pattern,
#' you should consider adding the new predictor to the model.
#'
#' @param model An object of class \code{lm}.
#' @param data A \code{data.frame} or \code{tibble}.
#' @param variable Character; new predictor to be added to the \code{model}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' rvsr_plot_shiny(model, mtcars, 'drat')
#'
#' @export
#'
rvsr_plot_shiny <- function(model, data, variable, print_plot = TRUE) {

  check_model(model)
  d    <- ols_prep_rvsrplot_data(model)
  v    <- l(deparse(substitute(variable)))
  xvar <- data[variable]
  k    <- data.frame(x = xvar, y = model$residuals)
  
  colnames(k) <- c("x", "y")

  p <-
    ggplot(k, aes(x = x, y = y)) +
    geom_point(shape = 1, colour = "blue") +
    xlab(paste(variable)) + ylab("Residual") +
    ggtitle(paste("Residual vs", variable)) +
    geom_hline(yintercept = 0, colour = "red")

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

#' DFFITS plot
#'
#' Plot for detecting influential observations using DFFITs.
#'
#' @param model An object of class \code{lm}.
#' @param size_adj_threshold logical; if \code{TRUE} (the default), size 
#' adjusted threshold is used to determine influential observations.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a 
#' plot object.
#'
#' @details
#' DFFIT - difference in fits, is used to identify influential data points. It
#' quantifies the number of standard deviations that the fitted value changes
#' when the ith data point is omitted.
#'
#' Steps to compute DFFITs:
#'
#' \itemize{
#'   \item Delete observations one at a time.
#'   \item Refit the regression model on remaining \eqn{n - 1} observations
#'   \item examine how much all of the fitted values change when the ith observation is deleted.
#' }
#'
#' An observation is deemed influential if the absolute value of its DFFITS value is greater than:
#' \deqn{2\sqrt((p + 1) / (n - p -1))}
#'
#' A size-adjusted cutoff recommended by Belsley, Kuh, and Welsch is 
#' \deqn{2\sqrt(p / n)} and is used by default in **olsrr**.
#' 
#' where \code{n} is the number of observations and \code{p} is the number of predictors including intercept.
#'
#' @return \code{ols_plot_dffits} returns  a list containing the
#' following components:
#'
#' \item{outliers}{a \code{data.frame} with observation number and \code{DFFITs} that exceed \code{threshold}}
#' \item{threshold}{\code{threshold} for classifying an observation as an outlier}
#'
#' @references
#' Belsley, David A.; Kuh, Edwin; Welsh, Roy E. (1980). Regression
#' Diagnostics: Identifying Influential Data and Sources of Collinearity.
#'
#' Wiley Series in Probability and Mathematical Statistics.
#' New York: John Wiley & Sons. ISBN 0-471-05856-4.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_plot_dffits(model)
#' ols_plot_dffits(model, size_adj_threshold = FALSE)
#'
#' @seealso [ols_plot_dfbetas()]
#'
#' @importFrom stats dffits
#'
#' @export
#'
ols_plot_dffits <- function(model, size_adj_threshold = TRUE, print_plot = TRUE) {

  check_model(model)
  data <- ols_prep_dffits_plot_data(model, size_adj_threshold)

  p <-
    ggplot(data$d, aes(x = obs, y = dbetas, label = txt, ymin = 0, ymax = data$dfm)) +
    geom_linerange(colour = "blue") +
    geom_hline(yintercept = c(0, data$dft, -data$dft), colour = "red") +
    geom_point(colour = "blue", shape = 1) +
    geom_text(hjust = -0.2, nudge_x = 0.15, size = 3, family = "serif",
              fontface = "italic", colour = "darkred", na.rm = TRUE)

  # annotations
  p <-
    p +
    annotate("text", x = Inf, y = Inf, hjust = 1.5, vjust = 2,
             family = "serif", fontface = "italic", colour = "darkred",
             label = paste("Threshold:", round(data$dft, 2)))

  # guides
  p <-
    p +
    xlab("Observation") +
    ylab("DFFITS") +
    ggtitle(paste("Influence Diagnostics for", data$title))

  if (print_plot) {
    suppressWarnings(print(p))
  } else {
    return(list(plot = p, outliers = data$f, threshold = round(data$dft, 2)))
  }

}


ols_prep_dffits_plot_data <- function(model, size_adj_threshold = TRUE) {

  dffitsm    <- unlist(dffits(model))
  k          <- model_n_coeffs(model)
  n          <- model_rows(model)
  
  # threshold
  normal_t   <- 2 * sqrt((k + 1) / (n - k - 1))
  adj_t      <- sqrt(k / n) * 2
  dffits_t   <- ifelse(size_adj_threshold, adj_t, normal_t)

  title      <- names(model.frame(model))[1]
  dfits_data <- data.frame(obs = seq_len(n), dbetas = dffitsm)
  d          <- ols_prep_dfbeta_data(dfits_data, dffits_t)
  f          <- ols_prep_dfbeta_outliers(d)
  colnames(f) <- c("observation", "dffits")

  list(d = d, dft = dffits_t, title = title, f = f, dfm = dffitsm)
}

#' @title Launch shiny app
#' @description Launches shiny app for interactive model building.
#' @examples
#' \dontrun{
#' ols_launch_app()
#' }
#' @export
#'
ols_launch_app <- function() {
	xplorerr::app_linear_regression()
}
 

#' Residual QQ plot
#'
#' Graph for detecting violation of normality assumption.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_qq(model)
#'
#' @family residual diagnostics
#'
#' @importFrom stats qqnorm qqline
#'
#' @export
#'
ols_plot_resid_qq <- function(model, print_plot = TRUE) {

  check_model(model)

  resid <- residuals(model)
  y     <- quantile(resid[!is.na(resid)], c(0.25, 0.75))
  x     <- qnorm(c(0.25, 0.75))
  slope <- diff(y) / diff(x)
  int   <- y[1L] - slope * x[1L]
  d     <- data.frame(x = resid)

  p <-
    ggplot(d, aes(sample = x)) +
    geom_abline(slope = slope, intercept = int, color = "red") +
    stat_qq(color = "blue") +
    xlab("Theoretical Quantiles") +
    ylab("Sample Quantiles") +
    ggtitle("Normal Q-Q Plot")

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

#' Cooks' D bar plot
#'
#' @description
#' Bar Plot of cook's distance to detect observations that strongly influence
#' fitted values of the model.
#'
#' @param model An object of class \code{lm}.
#' @param type An integer between 1 and 5 selecting one of the 5 methods for
#'     computing the threshold.
#' @param threshold Threshold for detecting outliers.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a
#'     plot object.
#'
#' @details
#' Cook's distance was introduced by American statistician R Dennis Cook in
#' 1977. It is used to identify influential data points. It depends on both the
#' residual and leverage i.e it takes it account both the \emph{x} value and
#' \emph{y} value of the observation.
#'
#' Steps to compute Cook's distance:
#'
#' \itemize{
#'   \item Delete observations one at a time.
#'   \item Refit the regression model on remaining \eqn{n - 1} observations
#'   \item examine how much all of the fitted values change when the ith observation is deleted.
#' }
#'
#' A data point having a large cook's d indicates that the data point strongly
#' influences the fitted values. There are several methods/formulas to compute
#' the threshold used for detecting or classifying observations as outliers and
#' we list them below.
#'
#' \itemize{
#'   \item \strong{Type 1} : 4 / n
#'   \item \strong{Type 2} : 4 / (n - k - 1)
#'   \item \strong{Type 3} : ~1
#'   \item \strong{Type 4} : 1 / (n - k - 1)
#'   \item \strong{Type 5} : 3 * mean(Vector of cook's distance values)
#' }
#'
#' where \strong{n} and \strong{k} stand for
#'
#' \itemize{
#'   \item \strong{n}: Number of observations
#'   \item \strong{k}: Number of predictors
#' }
#'
#' @return \code{ols_plot_cooksd_bar} returns  a list containing the
#' following components:
#'
#' \item{outliers}{a \code{data.frame} with observation number and \code{cooks distance} that exceed \code{threshold}}
#' \item{threshold}{\code{threshold} for classifying an observation as an outlier}
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_cooksd_bar(model)
#' ols_plot_cooksd_bar(model, type = 4)
#' ols_plot_cooksd_bar(model, threshold = 0.2)
#'
#' @importFrom stats cooks.distance
#' @importFrom ggplot2 geom_bar coord_flip ylim geom_hline geom_label
#'
#' @seealso [ols_plot_cooksd_chart()]
#'
#' @export
#'
ols_plot_cooksd_bar <- function(model, type = 1, threshold = NULL, print_plot = TRUE) {

  check_model(model)

  k <- ols_prep_cdplot_data(model, type)
  d <- ols_prep_outlier_obs(k)
  f <- ols_prep_cdplot_outliers(k)
  
  if (is.null(threshold)) {
      threshold <- k$ts
  }

  y_max <- max(k$maxx, threshold)
  y_lim <- y_max + (y_max * 0.1)

  # geoms
  p <-
    ggplot(d, aes(x = obs, y = cd, label = txt)) +
    geom_bar(width = 0.5, stat = "identity", aes(fill = fct_color)) +
    geom_text(hjust = -0.2, nudge_x = 0.05, size = 2, na.rm = TRUE) +
    geom_hline(yintercept = 0) +
    geom_hline(yintercept = threshold, colour = "red")

  # annotations
  p <-
    p +
    annotate("text", x = Inf, y = Inf, hjust = 1.2, vjust = 2,
      family = "serif", fontface = "italic", colour = "darkred",
      label = paste("Threshold:", round(threshold, 3)))

  # scales
  p <-
    p +
    scale_fill_manual(values = c("blue", "red"))

  # guides
  p <-
    p +
    labs(fill = "Observation") +
    xlab("Observation") +
    ylab("Cook's D") +
    ggtitle("Cook's D Bar Plot") +
    ylim(0, y_lim)

  if (print_plot) {
    suppressWarnings(print(p))
  } else {
    return(list(plot = p, outliers = f, threshold = threshold))
  }

}

#' Correlation test for normality
#'
#' @description
#' Correlation between observed residuals and expected residuals under normality.
#'
#' @param model An object of class \code{lm}.
#'
#' @return Correlation between fitted regression model residuals and expected
#' values of residuals.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_test_correlation(model)
#'
#' @importFrom stats qnorm
#'
#' @family residual diagnostics
#'
#' @export
#'
ols_test_correlation <- function(model) {

  check_model(model)
  corrout(model)

}

corrout <- function(model) {

  n      <- model_rows(model)
  stderr <- summary(model)[[6]]
  h      <- ka(seq_len(n), stderr = stderr, n = n)
  out    <- sort(residuals(model))

  cor(h, out)

}

ka <- function(k, stderr, n) {
  stderr * qnorm((k - 0.375) / (n + 0.25))
}


#' Test for normality
#'
#' Test for detecting violation of normality assumption.
#'
#' @param y A numeric vector or an object of class \code{lm}.
#' @param ... Other arguments.
#'
#' @return \code{ols_test_normality} returns an object of class \code{"ols_test_normality"}.
#' An object of class \code{"ols_test_normality"} is a list containing the
#' following components:
#'
#' \item{kolmogorv}{kolmogorv smirnov statistic}
#' \item{shapiro}{shapiro wilk statistic}
#' \item{cramer}{cramer von mises statistic}
#' \item{anderson}{anderson darling statistic}
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_test_normality(model)
#'
#' @importFrom stats ks.test shapiro.test
#' @importFrom goftest cvm.test
#' @importFrom nortest ad.test
#'
#' @family residual diagnostics
#'
#' @export
#'
ols_test_normality <- function(y, ...) UseMethod("ols_test_normality")

#' @export
#'
ols_test_normality.default <- function(y, ...) {

  if (!is.numeric(y)) {
    stop("y must be numeric")
  }

  ks  <- ks.test(y, "pnorm", mean(y), sd(y))
  sw  <- shapiro.test(y)
  cvm <- cvm.test(y)
  ad  <- ad.test(y)

  result <- list(kolmogorv = ks, shapiro = sw, cramer = cvm, anderson = ad)

  class(result) <- "ols_test_normality"
  return(result)
}

#' @export
#' @rdname ols_test_normality
#'
ols_test_normality.lm <- function(y, ...) {

  if (!all(class(y) == "lm")) {
    stop("Please specify a OLS linear regression model.", call. = FALSE)
  }

  ols_test_normality.default(residuals(y))

}

#' @export
#'
print.ols_test_normality <- function(x, ...) {
  print_norm_test(x)
}

test_that("output from normality test is as expected", {
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  expect_snapshot(ols_test_normality(model))
})

test_that("output from ols_corr_test is as expected", {
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  expect_equal(round(ols_test_correlation(model), 3), 0.97)
})

#' Part and partial correlations
#'
#' Zero-order, part and partial correlations.
#'
#' @param model An object of class \code{lm}.
#'
#' @details
#' \code{ols_correlations()} returns the relative importance of independent
#' variables in determining response variable. How much each variable uniquely
#' contributes to rsquare over and above that which can be accounted for by the
#' other predictors? Zero order correlation is the Pearson correlation
#' coefficient between the dependent variable and the independent variables.
#' Part correlations indicates how much rsquare will decrease if that variable
#' is removed from the model and partial correlations indicates amount of
#' variance in response variable, which is not estimated by the other
#' independent variables in the model, but is estimated by the specific
#' variable.
#'
#' @return \code{ols_correlations} returns an object of class \code{"ols_correlations"}.
#' An object of class \code{"ols_correlations"} is a data frame containing the
#' following components:
#'
#' \item{Zero-order}{zero order correlations}
#' \item{Partial}{partial correlations}
#' \item{Part}{part correlations}
#'
#' @references
#' Morrison, D. F. 1976. Multivariate statistical methods. New York: McGraw-Hill.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_correlations(model)
#'
#' @importFrom stats cor
#'
#' @export
#'
ols_correlations <- function(model) UseMethod("ols_correlations")

#' @export
#'
ols_correlations.default <- function(model) {

  check_model(model)

  result <- corout(model, corm2(model))
  class(result) <- c("ols_correlations", "data.frame")
  return(result)

}

#' @export
#'
print.ols_correlations <- function(x, ...) {
  print_correlations(x)
}

corout <- function(model, r2) {

  r1               <- summary(model)$r.squared
  mdata            <- cordata(model)
  cor_mdata        <- cmdata(mdata)
  n                <- ncol(mdata)
  ksign            <- corsign(model)
  n2               <- n - 1
  parts            <- ksign * sqrt(r1 - r2)
  partials         <- parts / sqrt(1 - r2)
  result           <- data.frame(cor_mdata, partials, parts)
  rownames(result) <- names(ksign)
  colnames(result) <- c("Zero-order", "Partial", "Part")

  return(result)

}

corm2 <- function(model) {

  mdata <- cordata(model)
  n     <- ncol(mdata)
  r2    <- c()

  for (i in 2:n) {
    out <- rtwo(i, mdata)
    r2  <- c(r2, out)
  }

  return(r2)
}

cordata <- function(model) {
  ols_prep_avplot_data(model)
}

cmdata <- function(mdata) {
  cor(mdata)[-1, 1]
}

rtwo <- function(i, mdata) {
  dat <- mdata[, c(-1, -i), drop = FALSE]
  summary(lm(mdata[[1]] ~ ., data = dat))[[8]]
}

corsign <- function(model) {
  sign(summary(model)$coefficients[-1, 1])
}

model <- lm(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)

test_that("correlations output match", {
  data <- c(
    -0.848, -0.776, -0.868, 0.681, 0.419, 0.151, -0.256, -0.569,
    0.289, 0.264, 0.059, -0.103, -0.269, 0.117, 0.106
  )
  expt <- ols_correlations(model) %>% unlist(use.names = FALSE)

  expect_equal(round(expt, 3), data)
})

#' Diagnostics panel
#'
#' Panel of plots for regression diagnostics.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_plot_diagnostics(model)
#'
#' @importFrom ggplot2 xlim stat_qq geom_histogram geom_line theme element_blank
#' @importFrom stats quantile
#'
#' @export
#'
ols_plot_diagnostics <- function(model, print_plot = TRUE) {

  check_model(model)

  p1  <- ols_plot_resid_fit(model)
  p2  <- ols_plot_resid_stud_fit(model)
  p3  <- ols_plot_resid_lev(model)
  p4  <- ols_plot_resid_qq(model)
  p5  <- ols_plot_obs_fit(model) 
  p6  <- ols_plot_cooksd_chart(model)
  p7  <- ols_plot_fm(model)
  p8  <- ols_plot_resid_spread(model)
  p9  <- ols_plot_resid_hist(model)
  p10 <- ols_plot_resid_box(model)

  result <- list(
    plot_1 = p1, plot_2 = p2, plot_3 = p3, plot_4 = p4,
    plot_5 = p5, plot_6 = p6, plot_7 = p7, plot_8 = p8,
    plot_9 = p9, plot_10 = p10
  )

  if (print_plot) {
    marrangeGrob(result, nrow = 2, ncol = 2, top = "Regression Diagnostics")
  } else {
    return(result)
  }

}

#' All possible regression
#'
#' @description
#' Fits all regressions involving one regressor, two regressors, three
#' regressors, and so on. It tests all possible subsets of the set of potential
#' independent variables.
#'
#' @param model An object of class \code{lm}.
#' @param max_order Maximum subset order.
#' @param x An object of class \code{ols_step_all_possible}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#' @param ... Other arguments.
#'
#' @return \code{ols_step_all_possible} returns an object of class \code{"ols_step_all_possible"}.
#' An object of class \code{"ols_step_all_possible"} is a data frame containing the
#' following components:
#'
#' \item{mindex}{model index}
#' \item{n}{number of predictors}
#' \item{predictors}{predictors in the model}
#' \item{rsquare}{rsquare of the model}
#' \item{adjr}{adjusted rsquare of the model}
#' \item{rmse}{root mean squared error of the model}
#' \item{predrsq}{predicted rsquare of the model}
#' \item{cp}{mallow's Cp}
#' \item{aic}{akaike information criteria}
#' \item{sbic}{sawa bayesian information criteria}
#' \item{sbc}{schwarz bayes information criteria}
#' \item{msep}{estimated MSE of prediction, assuming multivariate normality}
#' \item{fpe}{final prediction error}
#' \item{apc}{amemiya prediction criteria}
#' \item{hsp}{hocking's Sp}
#'
#' @references
#' Mendenhall William and  Sinsich Terry, 2012, A Second Course in Statistics Regression Analysis (7th edition).
#' Prentice Hall
#'
#' @examples
#' model <- lm(mpg ~ disp + hp, data = mtcars)
#' k <- ols_step_all_possible(model)
#' k
#'
#' # plot
#' plot(k)
#' 
#' # maximum subset
#' model <- lm(mpg ~ disp + hp + drat + wt + qsec, data = mtcars)
#' ols_step_all_possible(model, max_order = 3)
#'
#' @importFrom utils combn
#'
#' @export
#'
ols_step_all_possible <- function(model, ...) UseMethod("ols_step_all_possible")

#' @export
#' @rdname ols_step_all_possible
#'
ols_step_all_possible.default <- function(model, max_order = NULL, ...) {

  check_model(model)
  check_npredictors(model, 2)

  metrics <- allpos_helper(model, max_order)

  ui <- data.frame(
    n          = metrics$lpreds,
    predictors = unlist(metrics$preds),
    rsquare    = unlist(metrics$rsq),
    adjr       = unlist(metrics$adjrsq),
    rmse       = unlist(metrics$rmse),
    predrsq    = unlist(metrics$predrsq),
    cp         = unlist(metrics$cp),
    aic        = unlist(metrics$aic),
    sbic       = unlist(metrics$sbic),
    sbc        = unlist(metrics$sbc),
    msep       = unlist(metrics$msep),
    fpe        = unlist(metrics$fpe),
    apc        = unlist(metrics$apc),
    hsp        = unlist(metrics$hsp),
    stringsAsFactors = F
  )

  sorted <- c()

  for (i in seq_len(metrics$lc)) {
    temp   <- ui[metrics$q[i]:metrics$t[i], ]
    temp   <- temp[order(temp$rsquare, decreasing = TRUE), ]
    sorted <- rbind(sorted, temp)
  }

  mindex <- seq_len(nrow(sorted))
  sorted <- cbind(mindex, sorted)
  out    <- list(result = sorted)
  class(out) <- c("ols_step_all_possible")

  return(out)
}

#' @export
#'
print.ols_step_all_possible <- function(x, ...) {

  n <- max(x$result$mindex)
  k <- data.frame(x$result)[, c(1:5, 8)]
  names(k) <- c("Index", "N", "Predictors", "R-Square", "Adj. R-Square", "Mallow's Cp")
  print(k)

}

#' @export
#' @rdname ols_step_all_possible
#'
plot.ols_step_all_possible <- function(x, model = NA, print_plot = TRUE, ...) {

  k <- x$result
  d <- data.frame(index   = k$mindex, 
                  n       = k$n, 
                  rsquare = k$rsquare, 
                  adjr    = k$adjr,
                  cp      = k$cp, 
                  aic     = k$aic, 
                  sbic    = k$sbic, 
                  sbc     = k$sbc)

  d$cps <- abs(d$n - d$cp)

  p1 <- all_possible_plot(d, "rsquare", title = "R-Square")
  p2 <- all_possible_plot(d, "adjr", title = "Adj. R-Square")
  p3 <- all_possible_plot(d, "cps", title = "Cp")
  p4 <- all_possible_plot(d, "aic", title = "AIC")
  p5 <- all_possible_plot(d, "sbic", title = "SBIC")
  p6 <- all_possible_plot(d, "sbc", title = "SBC")

  myplots <- list(plot_1 = p1, plot_2 = p2, plot_3 = p3,
                  plot_4 = p4, plot_5 = p5, plot_6 = p6)

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2, top = "All Possible Regression")
  } else {
    return(myplots)
  }

}

#' All possible regression plot
#'
#' Generate plots for best subset regression.
#'
#' @importFrom ggplot2 ggtitle scale_shape_manual scale_size_manual scale_color_manual ggtitle geom_text
#'
#' @param d1 A data.frame.
#' @param d2 A data.frame.
#' @param title Plot title.
#'
#' @noRd
#'
all_possible_plot <- function(d, var, title = "R-Square") {

  d1    <- d[, c("n", var)]
  colnames(d1) <- c("x", "y")
  maxs  <- all_pos_maxs(d, var, title)
  lmaxs <- all_pos_lmaxs(maxs)
  index <- all_pos_index(d, var, title)

  d2 <- data.frame(x = lmaxs, y = maxs, tx = index, shape = 6, size = 4)

  ggplot(d1, aes(x = x, y = y)) + 
    geom_point(color = "blue", size = 2) +
    geom_point(data = d2, aes(x = x, y = y, shape = factor(shape),
      color = factor(shape), size = factor(size))) +
    geom_text(data = d2, aes(label = tx), hjust = 0, nudge_x = 0.1) +
    scale_shape_manual(values = c(2), guide = "none") +
    scale_size_manual(values = c(4), guide = "none") +
    scale_color_manual(values = c("red"), guide = "none") +
    xlab("") + 
    ylab("") + 
    ggtitle(title)

}

all_pos_maxs <- function(d, var, title = "R-Square") {

  if (title == "R-Square" | title == "Adj. R-Square") {
    as.numeric(lapply(split(d[[var]], d$n), max))
  } else {
    as.numeric(lapply(split(d[[var]], d$n), min))
  }

}

all_pos_lmaxs <- function(maxs) {

  seq_len(length(maxs))

}

all_pos_index <- function(d, var, title = "R-Square") {

  index   <- c()

  if (title == "R-Square" | title == "Adj. R-Square") {
    n <- as.numeric(lapply(split(d[[var]], d$n), max))
    m <- data.frame(n = seq_len(length(n)), maximum = n)
  } else {
    n <- as.numeric(lapply(split(d[[var]], d$n), min))
    m <- data.frame(n = seq_len(length(n)), minimum = n)
  }

  colnames(m) <- c("n", var)
  k  <- split(d[c("index", var)], d$n)

  for (i in m$n) {
    j <- which(part_2(m, var, i) == part_3(k, var, i))
    index[i] <- part_1(k, i)[j]
  }

  return(index)

}


part_1 <- function(k, i) {
  k[[i]]$index
}

part_2 <- function(m, var, i) {
  m[[var]][i]
}

part_3 <- function(k, var, i) {
  k[[i]][[var]]
}


#' All possible regression variable coefficients
#'
#' Returns the coefficients for each variable from each model.
#'
#' @param object An object of class \code{lm}.
#' @param ... Other arguments.
#'
#' @return \code{ols_step_all_possible_betas} returns a \code{data.frame} containing:
#'
#' \item{model_index}{model number}
#' \item{predictor}{predictor}
#' \item{beta_coef}{coefficient for the predictor}
#'
#' @examples
#' \dontrun{
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_step_all_possible_betas(model)
#' }
#'
#' @export
#'
ols_step_all_possible_betas <- function(object, ...) {

  if (!all(class(object) == "lm")) {
    stop("Please specify a OLS linear regression model.", call. = FALSE)
  }

  if (length(object$coefficients) < 3) {
    stop("Please specify a model with at least 2 predictors.", call. = FALSE)
  }

  metrics    <- allpos_helper(object)
  beta_names <- names(metrics$betas)
  mindex     <- seq_len(length(metrics$rsq))
  
  # detect index of intercepts
  intercepts <- grep("(Intercept)", beta_names)

  # increment length of betas  
  lbeta_nam  <- length(beta_names) + 1

  # detect length of betas in each model plus the last model with all variables
  reps       <- c(diff(intercepts), (lbeta_nam - rev(intercepts)[1]))
  m_index    <- rep(mindex, reps)
  beta       <- metrics$betas

  data.frame(model = m_index, predictor = beta_names, beta = beta)

}

#' All possible regression internal
#'
#' Internal function for all possible regression.
#'
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
allpos_helper <- function(model, max_order = NULL) {

  nam <- coeff_names(model)
  n   <- length(nam)
  
  r     <- seq_len(n)
  combs <- list()

  for (i in seq_len(n)) {
    combs[[i]] <- combn(n, r[i])
  }

  if (!is.null(max_order)) {
    check_order(n, max_order)
  }

  if (is.null(max_order)) {
    max_order <- n
  }

  pos_data  <- model$model
  predicts  <- nam
  # lc        <- length(combs)
  lc        <- max_order
  varnames  <- model_colnames(model)
  len_preds <- length(predicts)
  gap       <- len_preds - 1
  space     <- coeff_length(predicts, gap)
  colas     <- unname(unlist(lapply(combs, ncol)))
  response  <- varnames[1]
  p         <- colas
  t         <- cumsum(colas)
  q         <- c(1, t[-lc] + 1)

  mcount    <- 0
  rsq       <- list()
  adjrsq    <- list()
  sigma     <- list()
  predrsq   <- list()
  cp        <- list()
  aic       <- list()
  sbic      <- list()
  sbc       <- list()
  msep      <- list()
  fpe       <- list()
  apc       <- list()
  hsp       <- list()
  preds     <- list()
  lpreds    <- c()
  betas     <- c()

  for (i in seq_len(lc)) {
    for (j in seq_len(colas[i])) {

      predictors <- nam[combs[[i]][, j]]
      lp         <- length(predictors)

      out <- ols_regress(paste(response, "~", paste(predictors, collapse = " + ")), data = pos_data)

      mcount            <- mcount + 1
      lpreds[mcount]    <- lp
      rsq[[mcount]]     <- out$rsq
      adjrsq[[mcount]]  <- out$adjr
      sigma[[mcount]]   <- out$rmse
      predrsq[[mcount]] <- ols_pred_rsq(out$model)
      cp[[mcount]]      <- ols_mallows_cp(out$model, model)
      aic[[mcount]]     <- ols_aic(out$model)
      sbic[[mcount]]    <- ols_sbic(out$model, model)
      sbc[[mcount]]     <- ols_sbc(out$model)
      msep[[mcount]]    <- ols_msep(out$model)
      fpe[[mcount]]     <- ols_fpe(out$model)
      apc[[mcount]]     <- ols_apc(out$model)
      hsp[[mcount]]     <- ols_hsp(out$model)
      preds[[mcount]]   <- paste(predictors, collapse = " ")
      betas             <- append(betas, out$betas)
    }
  }

  result <- list(
    lpreds = lpreds, rsq = rsq, adjrsq = adjrsq, rmse = sigma,
    predrsq = predrsq, cp = cp, aic = aic, sbic = sbic,
    sbc = sbc, msep = msep, fpe = fpe, apc = apc, hsp = hsp,
    preds = preds, lc = lc, q = q, t = t, betas = betas
  )

  return(result)
}

test_that("all possible regression output matches the expected result", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_all_possible(model)
  pred_exp <- c(
    "x4", "x2", "x1", "x3", "x1 x2",
    "x1 x4", "x3 x4", "x2 x3", "x2 x4", "x1 x3",
    "x1 x2 x4", "x1 x2 x3", "x1 x3 x4", "x2 x3 x4",
    "x1 x2 x3 x4"
  )
  expect_equal(k$result$mindex, c(1:15))
  expect_equal(k$result$predictors, pred_exp, ignore_attr = TRUE)

  k2 <- ols_step_all_possible(model, max_order = 3)
  preds_selected <- c(
    "x4", "x2", "x1", "x3", "x1 x2",
    "x1 x4", "x3 x4", "x2 x3", "x2 x4", "x1 x3",
    "x1 x2 x4", "x1 x2 x3", "x1 x3 x4", "x2 x3 x4"
  )
  expect_equal(k2$result$mindex, c(1:14))
  expect_equal(k2$result$predictors, preds_selected, ignore_attr = TRUE)
})

test_that("all possible regression throws error messages", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  error_msg <- "Maximum subset order should be less than or equal to the number of predictors in the specified model."
  expect_error(ols_step_all_possible(model, max_order = 5), error_msg)
})


test_that("output from all subsets regression is as expected", {
  model <- lm(mpg ~ disp + hp, data = mtcars)
  expect_snapshot(ols_step_all_possible(model))
})

test_that("all possible regression betas are as expected", {
  
  model   <- lm(mpg ~ disp + hp + wt, data = mtcars)
  k       <- ols_step_all_possible_betas(model)
  out     <- lapply(split(k$beta, k$predictor), mean)
  actual  <- data.frame(predictor = names(out), beta = as.numeric(out))

  predictor <- c("(Intercept)", "disp", "hp", "wt")
  beta <- c(33.85901073, -0.02255579, -0.03899945, -4.09350456)

  expected <- data.frame(predictor, beta)
  expect_equal(actual$predictor, expected$predictor, ignore_attr = TRUE)
  expect_equal(actual$beta, expected$beta, ignore_attr = TRUE)
})

#' Bartlett test
#'
#' Test if k samples are from populations with equal variances.
#'
#' @param data A \code{data.frame} or \code{tibble}.
#' @param ... Columns in \code{data}.
#' @param group_var Grouping variable.
#'
#' @details
#' Bartlett's test is used to test if variances across samples is equal.
#' It is sensitive to departures from normality. The Levene test
#' is an alternative test that is less sensitive to departures from normality.
#'
#' @return \code{ols_test_bartlett} returns an object of class \code{"ols_test_bartlett"}.
#' An object of class \code{"ols_test_bartlett"} is a list containing the
#' following components:
#'
#' \item{fstat}{f statistic}
#' \item{pval}{p-value of \code{fstat}}
#' \item{df}{degrees of freedom}
#'
#' @references
#' Snedecor, George W. and Cochran, William G. (1989), Statistical Methods,
#' Eighth Edition, Iowa State University Press.
#'
#' @family heteroskedasticity tests
#'
#' @examples
#' # using grouping variable
#' if (require("descriptr")) {
#'   library(descriptr)
#'   ols_test_bartlett(mtcarz, 'mpg', group_var = 'cyl')
#' }
#'
#' # using variables
#' ols_test_bartlett(hsb, 'read', 'write')
#'
#' @importFrom stats pchisq formula
#'
#' @export
#'
ols_test_bartlett <- function(data, ...) UseMethod("ols_test_bartlett")

#' @export
#' @rdname ols_test_bartlett
#'
ols_test_bartlett.default <- function(data, ..., group_var = NULL) {

  groupvar  <- group_var
  varyables <- unlist(list(...))
  fdata     <- data[varyables]
  var_c     <- names(fdata)

  if (is.null(groupvar)) {

    z  <- as.list(fdata)
    ln <- unname(unlist(lapply(z, length)))
    ly <- seq_len(length(z))
  
    if (length(z) < 2) {
      stop("Please specify at least two variables.", call. = FALSE)
    }

    out       <- rep(ly, times = ln)
    fdata     <- unlist(z)
    groupvars <- as.factor(unlist(out))
    g_var     <- NULL

  } else {

    fdata     <- fdata[[1]]
    groupvars <- data[[groupvar]]
    g_var     <- names(data[groupvar])
     
    if (length(fdata) != length(groupvars)) {
      stop("Length of variable and group_var do not match.", call. = FALSE)
    }
  }

  df    <- nlevels(groupvars) - 1
  fstat <- bartlett_fstat(fdata, groupvars)
  pval  <- pchisq(fstat, df, lower.tail = FALSE)

  out <- list(df    = df,
              fstat = fstat,
              g_var = g_var,
              pval  = pval,
              var_c = var_c)

  class(out) <- "ols_test_bartlett"

  return(out)

}

#' @export
#'
print.ols_test_bartlett <- function(x, ...) {
  print_bartlett_test(x)
}


#' Bartlett internal
#'
#' Computes the f statistic for bartlett test.
#'
#' @importFrom stats complete.cases var
#'
#' @noRd
#'
bartlett_fstat <- function(variable, grp_var) {

  n     <- length(variable)
  k     <- nlevels(grp_var)
  comp  <- complete.cases(variable, grp_var)
  vars  <- tapply(variable[comp], grp_var[comp], var)
  lens  <- tapply(variable[comp], grp_var[comp], length)
  v     <- lens - 1
  sumv  <- sum(v)
  isumv <- sum(1 / v)
  c     <- 1 + (1 / (3 * (k - 1))) * (isumv - (1 / sumv))
  n2    <- sum(v * log10(vars))
  l     <- length(vars)
  ps    <- ((lens - 1) * vars) / (n - k)
  pvar  <- sum(ps)
  ((1 / c) * (sumv * log10(pvar) - n2)) * 2.3026

}

test_that("output from the test match the result when using variables", {

  b <- ols_test_bartlett(mtcars, 'mpg', 'disp')

  expect_equal(round(b$fstat, 3), 142.336)
  expect_equal(b$pval, 0)
  expect_equal(b$df, 1)
  expect_equal(b$var_c, c("mpg", "disp"), ignore_attr = TRUE)
  expect_null(b$g_var)

})

test_that("output from test match the result when using grouping variables", {
  if (requireNamespace("descriptr", quietly = TRUE)) {
    b <- ols_test_bartlett(descriptr::mtcarz, 'mpg', group_var = 'vs')

    expect_equal(round(b$fstat, 3), 1.585)
    expect_equal(round(b$pval, 3), 0.208)
    expect_equal(b$df, 1)
    expect_equal(b$var_c, "mpg")
    expect_equal(b$g_var, "vs")
  }
})

test_that("bartlett test throws error messages", {
  expect_error(ols_test_bartlett(mtcars, 'mpg'), "Please specify at least two variables.")
})

test_that("output from bartlett test is as expected when using variables", {
  expect_snapshot(ols_test_bartlett(hsb, 'read', 'write'))
  expect_snapshot(ols_test_bartlett(descriptr::mtcarz, 'mpg', group_var = 'vs'))
})

#' Bonferroni Outlier Test
#'
#' Detect outliers using Bonferroni p values.
#'
#' @param model An object of class \code{lm}.
#' @param cut_off Bonferroni p-values cut off for reporting observations.
#' @param n_max Maximum number of observations to report, default is 10.
#' @param ... Other arguments.
#'
#' @examples
#' # model
#' model <- lm(y ~ ., data = surgical)
#' ols_test_outlier(model)
#'
#' @importFrom stats df.residual pt na.omit
#' @export
#'
ols_test_outlier <- function(model, cut_off = 0.05, n_max = 10, ...) {

  d <- ols_outlier_data(model, cut_off)

  if (d$nf == 0) {
    out <- d$data_bon[which.max(abs(d$data_bon$stud_resid)), ]
  } else {
    if (d$nf < n_max) {
      out <- d$data_co
    } else {
      out <- d$data_co[1:n_max, ]
    }
  }

  colnames(out) <- d$out_names
  return(out)

}

ols_outlier_data <- function(model, cut_off) {
  stud_resid <- rstudent(model)
  df_resid   <- df.residual(model) - 1
  n          <- length(stud_resid)
  p_val      <- pt(q = abs(stud_resid), df = df_resid, lower.tail = FALSE) * 2
  p_bon      <- n * p_val
  data_bon   <- na.omit(data.frame(stud_resid, p_val, p_bon))
  data_bonf  <- data_bon[data_bon$p_bon <= cut_off, ]
  data_co    <- data_bonf[order(-data_bonf$p_bon), ]
  nf         <- nrow(data_co)
  out_names  <- c("studentized_residual", "unadjusted_p_val", "bonferroni_p_val")

  list(nf = nf, data_co = data_co, data_bon = data_bon, out_names = out_names)
}

#' Stepwise regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering and removing predictors based on p values, in a stepwise manner
#' until there is no variable left to enter or remove any more.
#'
#' @param model An object of class \code{lm}; the model should include all
#'   candidate predictor variables.
#' @param p_enter p value; variables with p value less than \code{p_enter} will enter
#'   into the model.
#' @param p_remove p value; variables with p more than \code{p_remove} will be removed
#'   from the model.
#' @param include Character or numeric vector; variables to be included in selection process.
#' @param exclude Character or numeric vector; variables to be excluded from selection process.
#' @param progress Logical; if \code{TRUE}, will display variable selection progress.
#' @param details Logical; if \code{TRUE}, will print the regression result at
#' each step.
#' @param steps Number of steps after which the stepwise procedures should stop.
#' @param x An object of class \code{ols_step_both_p}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#' @param ... Other arguments.
#' @return \code{ols_step_both_p} returns an object of class \code{"ols_step_both_p"}.
#' An object of class \code{"ols_step_both_p"} is a list containing the
#' following components:
#'
#' \item{model}{final model; an object of class \code{lm}}
#' \item{metrics}{selection metrics}
#' \item{beta_pval}{beta and p values of models in each selection step}
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' @examples
#' \dontrun{
#' # stepwise regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_both_p(model)
#'
#' # stepwise regression plot
#' model <- lm(y ~ ., data = surgical)
#' k <- ols_step_both_p(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # final model
#' k$model
#'
#' # include or exclude variables
#' model <- lm(y ~ ., data = stepdata)
#'
#' # force variable to be included in selection process
#' ols_step_both_p(model, include = c("x6"))
#'
#' # use index of variable instead of name
#' ols_step_both_p(model, include = c(6))
#'
#' # force variable to be excluded from selection process
#' ols_step_both_p(model, exclude = c("x1"))
#'
#' # use index of variable instead of name
#' ols_step_both_p(model, exclude = c(1))
#'
#' # steps
#' ols_step_both_p(model, steps = 3)
#'
#' }
#'
#' @family both direction selection_procedures
#'
#' @export
#'
ols_step_both_p <- function(model, ...) UseMethod("ols_step_both_p")

#' @export
#' @rdname ols_step_both_p
#'
ols_step_both_p.default <- function(model, p_enter = 0.1, p_remove = 0.3, include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {

  if (details) {
    progress <- FALSE
  }

  check_inputs(model, include, exclude, progress, details)
  check_values(p_enter, 0, 1)
  check_values(p_remove, 0, 1)

  indterms <- coeff_names(model)
  lenterms <- length(indterms)

  if (is.numeric(include)) {
    include <- indterms[include]
  }

  if (is.numeric(exclude)) {
    exclude <- indterms[exclude]
  }

  response <- names(model$model)[1]
  l        <- model$model
  lockterm <- c(include, exclude)
  nam      <- colnames(attr(model$terms, "factors"))
  cterms   <- setdiff(nam, exclude)

  if (progress || details) {
    ols_candidate_terms(cterms, "both")
  }

  nam      <- setdiff(nam, lockterm)
  all_pred <- nam
  mlen_p   <- length(all_pred)
  tech     <- c("addition", "removal")

  pvalues <- c()
  lbetas  <- c()
  betas   <- c()
  preds   <- include
  pvals   <- c()
  fvals   <- c()
  step    <- 0
  ppos    <- length(include) + 1
  rsq     <- c()
  adjrsq  <- c()
  aic     <- c()
  sbc     <- c()
  sbic    <- c()
  rmse    <- c()
  cp      <- c()
  f       <- c()
  fp      <- c()
  method  <- c()

  base_model <- ols_base_model(include, response, l)
  rsq_base   <- summary(base_model)$r.squared

  if (details) {
    ols_rsquared_init(include, "r2", response, rsq_base)
  }

  for (i in seq_len(mlen_p)) {
    predictors <- c(include, all_pred[i])
    m          <- lm(paste(response, "~", paste(predictors, collapse = " + ")), l)
    m_sum      <- Anova(m)
    fvals[i]   <- m_sum$`F value`[ppos]
    pvals[i]   <- m_sum$`Pr(>F)`[ppos]
  }

  maxf     <- which(fvals == max(fvals, na.rm = TRUE))
  minp     <- which(pvals == min(pvals, na.rm = TRUE))
  len_minp <- length(minp)

  if (len_minp > 1) {
    minp <- minp[1]
    maxf <- maxf[1]
  }

  if (pvals[minp] > p_enter) {
    stop("None of the variables satisfy the criteria for entering the model.", call. = FALSE)
  }

  preds   <- c(preds, all_pred[maxf])
  lpreds  <- length(preds)
  step    <- step + 1
  fr      <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), l)
  rsq     <- c(rsq, fr$rsq)
  adjrsq  <- c(adjrsq, fr$adjr)
  cp      <- c(cp, ols_mallows_cp(fr$model, model))
  aic     <- c(aic, ols_aic(fr$model))
  sbc     <- c(sbc, ols_sbc(fr$model))
  sbic    <- c(sbic, ols_sbic(fr$model, model))
  rmse    <- c(rmse, fr$rmse)
  betas   <- append(betas, fr$betas)
  lbetas  <- append(lbetas, length(fr$betas))
  pvalues <- append(pvalues, fr$pvalues)
  method  <- c(method, tech[1])

  if (progress) {
    ols_progress_init("both")
    ols_progress_display(preds, "both", "added")
  }

  if (details) {
    rsq1 <- tail(rsq, n = 1)
    ols_rsquared_selected("r2", step, preds, response, rsq1)
  }

  all_step  <- step
  var_index <- tail(preds, n = 1)

  if (!is.null(steps)) {
    mlen_p <- steps
  }

  while (step < mlen_p) {

    all_pred <- all_pred[-maxf]
    len_p    <- length(all_pred)

    if (len_p == 0) {
      break
    }

    ppos     <- ppos + length(maxf)
    pvals    <- c()
    fvals    <- c()

    for (i in seq_len(len_p)) {

      predictors <- c(preds, all_pred[i])
      m          <- lm(paste(response, "~", paste(predictors, collapse = " + ")), l)
      m_sum      <- Anova(m)
      fvals[i]   <- m_sum$`F value`[ppos]
      pvals[i]   <- m_sum$`Pr(>F)`[ppos]
    }

    maxf  <- which(fvals == max(fvals, na.rm = TRUE))
    minp  <- pvals[maxf]
    len_minp <- length(minp)

    if (len_minp > 1) {
      minp <- minp[1]
      maxf <- maxf[1]
    }

    if (minp <= p_enter) {

      preds     <- c(preds, all_pred[maxf])
      var_index <- c(var_index, all_pred[maxf])
      method    <- c(method, tech[1])
      lpreds    <- length(preds)
      all_step  <- all_step + 1
      step      <- step + 1
      fr        <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), l)
      rsq       <- c(rsq, fr$rsq)
      adjrsq    <- c(adjrsq, fr$adjr)
      aic       <- c(aic, ols_aic(fr$model))
      sbc       <- c(sbc, ols_sbc(fr$model))
      sbic      <- c(sbic, ols_sbic(fr$model, model))
      cp        <- c(cp, ols_mallows_cp(fr$model, model))
      rmse      <- c(rmse, fr$rmse)
      betas     <- append(betas, fr$betas)
      lbetas    <- append(lbetas, length(fr$betas))
      pvalues   <- append(pvalues, fr$pvalues)

      if (progress) {
        ols_progress_display(preds, "both", "added")
      }

      if (details) {
        rsq1 <- tail(rsq, n = 1)
        ols_rsquared_selected("r2", all_step, preds, response, rsq1)
      }

      # check p value of predictors previously added in above step
      m2      <- lm(paste(response, "~", paste(preds, collapse = " + ")), l)
      m2_sum  <- Anova(m2)
      len_inc <- length(include) + 1
      fvals_r <- m2_sum$`F value`[len_inc:lpreds]
      pvals_r <- m2_sum$`Pr(>F)`[len_inc:lpreds]

      minf    <- which(fvals_r == min(fvals_r, na.rm = TRUE))
      maxp    <- pvals_r[minf]
      len_maxp <- length(maxp)

      if (len_maxp > 1) {
        maxp <- maxp[1]
        minf <- minf[1]
      }

      if (maxp > p_remove) {

        var_index <- c(var_index, preds[minf])
        lvar      <- length(var_index)
        method    <- c(method, tech[2])
        preds     <- preds[-(minf + length(include))]
        all_step  <- all_step + 1
        ppos      <- ppos - length(minf)
        fr        <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), l)
        rsq       <- c(rsq, fr$rsq)
        adjrsq    <- c(adjrsq, fr$adjr)
        aic       <- c(aic, ols_aic(fr$model))
        sbc       <- c(sbc, ols_sbc(fr$model))
        sbic      <- c(sbic, ols_sbic(fr$model, model))
        cp        <- c(cp, ols_mallows_cp(fr$model, model))
        rmse      <- c(rmse, fr$rmse)
        betas     <- append(betas, fr$betas)
        lbetas    <- append(lbetas, length(fr$betas))
        pvalues   <- append(pvalues, fr$pvalues)

        if (progress) {
          ols_progress_display(var_index, "both", "removed")
        }

        if (details) {
          rsq1 <- tail(rsq, n = 1)
          ols_stepwise_details(all_step, var_index, preds, response, rsq1, "removed", "R-Squared")
        }
      } else {
        preds <- preds
        all_step <- all_step
      }
    } else {
      if (progress || details) {
        ols_stepwise_break(direction = "both")
      }
      break
    }
  }

  final_model <- lm(paste(response, "~", paste(preds, collapse = " + ")), data = l)

  metrics     <- data.frame(step       = seq_len(all_step),
                            variable   = var_index,
                            method     = method,
                            r2         = rsq,
                            adj_r2     = adjrsq,
                            aic        = aic,
                            sbc        = sbc,
                            sbic       = sbic,
                            mallows_cp = cp,
                            rmse       = rmse)

  beta_pval <- data.frame(
    model     = rep(seq_len(all_step), lbetas),
    predictor = names(betas),
    beta      = betas,
    pval      = pvalues
  )

  out <- list(metrics = metrics,
              model   = final_model,
              others  = list(base_model = base_model,
                             full_model = model,
                             beta_pval = beta_pval))

  class(out) <- "ols_step_both_p"

  return(out)
}

#' @export
#'
print.ols_step_both_p <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "both")
  } else {
    print("No variables have been added to or removed from the model.")
  }
}

#' @export
#' @rdname ols_step_both_p
#'
plot.ols_step_both_p <- function(x, model = NA, print_plot = TRUE, details = TRUE, ...) {

  p1 <- ols_plot_stepwise(x, "r2", "R-Square", details, "both")
  p2 <- ols_plot_stepwise(x, "adj_r2", "Adj. R-Square", details, "both")
  p3 <- ols_plot_stepwise(x, "aic", "AIC", details, "both")
  p4 <- ols_plot_stepwise(x, "rmse", "RMSE", details, "both")

  myplots <- list(plot_1 = p1, plot_2 = p2, plot_3 = p3, plot_4 = p4)

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2, top = "Stepwise Both Direction Regression")
  } else {
    return(myplots)
  }

}

test_that("output from stepwise regression is as expected", {
  model <- lm(y ~ ., data = surgical)
  expect_snapshot(ols_step_both_p(model))
  expect_snapshot(ols_step_both_p(model, progress = TRUE))
  expect_snapshot(ols_step_both_p(model, details = TRUE))
})

test_that("output from stepwise regression is as expected when steps is specified", {
  model <- lm(y ~ ., data = surgical)
  expect_snapshot(ols_step_both_p(model, steps = 3))
  expect_snapshot(ols_step_both_p(model, progress = TRUE, steps = 3))
  expect_snapshot(ols_step_both_p(model, details = TRUE, steps = 3))
})

#' F test
#'
#' @description
#' Test for heteroskedasticity under the assumption that the errors are
#' independent and identically distributed (i.i.d.).
#'
#' @param model An object of class \code{lm}.
#' @param fitted_values Logical; if TRUE, use fitted values of regression model.
#' @param rhs Logical; if TRUE, specifies that tests for heteroskedasticity be
#'   performed for the right-hand-side (explanatory) variables of the fitted
#'   regression model.
#' @param vars Variables to be used for for heteroskedasticity test.
#' @param ... Other arguments.
#'
#' @return \code{ols_test_f} returns an object of class \code{"ols_test_f"}.
#' An object of class \code{"ols_test_f"} is a list containing the
#' following components:
#'
#' \item{f}{f statistic}
#' \item{p}{p-value of \code{f}}
#' \item{fv}{fitted values of the regression model}
#' \item{rhs}{names of explanatory variables of fitted regression model}
#' \item{numdf}{numerator degrees of freedom}
#' \item{dendf}{denominator degrees of freedom}
#' \item{vars}{variables to be used for heteroskedasticity test}
#' \item{resp}{response variable}
#' \item{preds}{predictors}
#'
#' @references
#' Wooldridge, J. M. 2013. Introductory Econometrics: A Modern Approach. 5th ed. Mason, OH: South-Western.
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#'
#' # using fitted values
#' ols_test_f(model)
#'
#' # using all predictors of the model
#' ols_test_f(model, rhs = TRUE)
#'
#' # using fitted values
#' ols_test_f(model, vars = c('disp', 'hp'))
#'
#' @family heteroskedasticity tests
#'
#' @importFrom stats pf
#'
#' @export
#'
ols_test_f <- function(model, fitted_values = TRUE, rhs = FALSE, vars = NULL, ...) UseMethod("ols_test_f")

#' @export
#'
ols_test_f.default <- function(model, fitted_values = TRUE, rhs = FALSE, vars = NULL, ...) {

  check_model(model)
  check_logic(fitted_values)
  check_logic(rhs)

  if (length(vars) > 0) {
    check_modelvars(model, vars)
    fitted_values <- FALSE
  }

  l    <- ols_prep_avplot_data(model)
  nam  <- names(l)[-1]
  resp <- names(l)[1]
  n    <- nrow(l)

  if (rhs) {
    fitted_values <- FALSE
    k             <- frhs(nam, model, n, l)
    result        <- ftest_result(k)
  } else {
    if (fitted_values) {
      k      <- ffit(model)
      result <- ftest_result(k)
    } else {
      k      <- fvar(n, l, model, vars)
      result <- ftest_result(k)
    }
  }

  out <- list(dendf = result$dendf, f = result$f, fv = fitted_values,
              numdf = result$numdf, p = result$p, preds = nam,
              resp = resp, rhs = rhs, vars = vars)

  class(out) <- "ols_test_f"

  return(out)
}

#' @export
#'
print.ols_test_f <- function(x, ...) {
  print_ftest(x)
}

frhs <- function(nam, model, n, l) {

  np         <- length(nam)
  var_resid  <- (model_rss(model) / n) - 1
  ind        <- (residuals(model) ^ 2) / var_resid
  l          <- cbind(l, ind)
  mdata      <- l[-1]

  summary(lm(ind ~ ., data = mdata))$fstatistic 

}

fvar <- function(n, l, model, vars) {

  var_resid  <- (model_rss(model) / n) - 1
  ind        <- (residuals(model) ^ 2) / var_resid
  mdata      <- l[-1]
  dl         <- mdata[, vars]
  dk         <- as.data.frame(cbind(ind, dl))

  summary(lm(ind ~ ., data = dk))$fstatistic

}

ffit <- function(model) {

  pred         <- fitted(model)
  pred_len     <- length(pred)
  resid        <- model$residuals ^ 2
  avg_resid    <- sum(resid) / pred_len
  scaled_resid <- resid / avg_resid

  summary(lm(scaled_resid ~ pred))$fstatistic

}

ftest_result <- function(k) {

  f     <- k[[1]]
  numdf <- k[[2]]
  dendf <- k[[3]]
  p     <- pf(f, numdf, dendf, lower.tail = F)

  list(f = f, numdf = numdf, dendf = dendf, p = p)

}

test_that("output from ftest is as expected", {
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  expect_snapshot(ols_test_f(model))
  expect_snapshot(ols_test_f(model, rhs = TRUE))
  expect_snapshot(ols_test_f(model, vars = c("disp", "hp")))
})

model <- lm(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)

test_that("when fitted.values == TRUE, fitted values from the regression\n\tare used for the test", {
  k <- ols_test_f(model)

  expect_equal(round(k$f, 3), 1.239)
  expect_equal(round(k$p, 3), 0.275)
  expect_equal(k$numdf, 1)
  expect_equal(k$dendf, 30)

  expect_true(k$fv)
  expect_false(k$rhs)

  expect_null(k$vars)

  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when fitted_values == TRUE and rhs == TRUE, predictors from the\n\tmodel are used for the test", {
  k <- ols_test_f(model, fitted_values = TRUE, rhs = TRUE)

  expect_equal(round(k$f, 3), 0.444)
  expect_equal(round(k$p, 3), 0.814)
  expect_equal(k$numdf, 5)
  expect_equal(k$dendf, 26)

  expect_false(k$fv)
  expect_true(k$rhs)

  expect_null(k$vars)

  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when vars != NULL, variables specified from the are\n\tused for the test", {
  k <- ols_test_f(model, vars = c("disp", "hp"))

  expect_equal(round(k$f, 3), 0.453)
  expect_equal(round(k$p, 3), 0.64)
  expect_equal(k$numdf, 2)
  expect_equal(k$dendf, 29)

  expect_false(k$fv)
  expect_false(k$rhs)

  expect_equal(k$vars, c("disp", "hp"), ignore_attr = TRUE)
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when vars != NULL and rhs == TRUE, predictors in the model are\n\tused for the test", {
  k <- ols_test_f(model, rhs = TRUE, vars = c("disp", "hp"))

  expect_equal(round(k$f, 3), 0.444)
  expect_equal(round(k$p, 3), 0.814)
  expect_equal(k$numdf, 5)
  expect_equal(k$dendf, 26)

  expect_false(k$fv)
  expect_true(k$rhs)

  expect_equal(k$vars, c("disp", "hp"), ignore_attr = TRUE)
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})

fs <- function() {
  x <- rep("  ")
  return(x)
}

fg <- function(x, w) {
  z <- as.character(x)
  y <- format(z, width = w, justify = "right")
  return(y)
}

fl <- function(x, w) {
  x <- as.character(x)
  ret <- format(x, width = w, justify = "left")
  return(ret)
}

fc <- function(x, w) {
  x <- as.character(x)
  ret <- format(x, width = w, justify = "centre")
  return(ret)
}

fw <- function(x, w) {
  z <- format(as.character(x), width = w, justify = "right")
  return(z)
}

formatter_t <- function(x, w) {
  ret <- format(x, width = w, justify = "centre")
  return(ret)
}

formatter_n <- function(x, w) {
  ret <- format(x, nsmall = 3)
  ret1 <- format(ret, width = w, justify = "centre")
  return(ret1)
}

format_cil <- function(x, w) {
  ret <- format(x, nsmall = 3)
  ret1 <- format(ret, width = w, justify = "centre")
  return(ret1)
}

format_ciu <- function(x, w) {
  ret <- format(x, nsmall = 3)
  ret1 <- format(ret, width = w, justify = "centre")
  return(ret1)
}

formats_t <- function() {
  x <- rep("  ")
  return(x)
}

l <- function(x) {
  x <- as.character(x)
  k <- grep("\\$", x)
  if (length(k) == 1) {
    temp <- strsplit(x, "\\$")
    out <- temp[[1]][2]
  } else {
    out <- x
  }
  return(out)
}

null_model_metrics <- function(model, full_model) {

  output <- summary(model)
  anovam <- anova(model)
  aic    <- ols_aic(model)
  sbc    <- ols_sbc(model)
  sbic   <- ols_sbic(model, full_model)
  n      <- length(anovam$Df)
  ess    <- anovam$`Sum Sq`[n]
  tss    <- sum(anovam$`Sum Sq`)
  rss    <- tss - ess
  rsq    <- output$r.squared
  adjr   <- output$adj.r.squared
  rmse   <- sqrt(mean(model$residuals ^ 2))

  list(adjr = adjr, aic = aic, sbc = sbc,  sbic = sbic, ess = ess,
       rsq = rsq, rss = rss, rmse = rmse)

}

max_nchar <- function(char, val, rn = 3, ns = 3) {
  max(nchar(char), nchar(format(round(val, rn), nsmall = ns)))
}

#' Stepwise backward regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' removing predictors based on p values, in a stepwise manner until there is
#' no variable left to remove any more.
#'
#' @param model An object of class \code{lm}; the model should include all
#'   candidate predictor variables.
#' @param p_val p value; variables with p more than \code{p_val} will be removed
#'   from the model.
#' @param include Character or numeric vector; variables to be included in selection process.
#' @param exclude Character or numeric vector; variables to be excluded from selection process.
#' @param hierarchical Logical; if \code{TRUE}, performs hierarchical selection.
#' @param progress Logical; if \code{TRUE}, will display variable selection progress.
#' @param details Logical; if \code{TRUE}, will print the regression result at
#'   each step.
#' @param steps Number of steps after which the stepwise procedures should stop.
#' @param x An object of class \code{ols_step_backward_p}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#' @param ... Other inputs.
#'
#' @return \code{ols_step_backward_p} returns an object of class \code{"ols_step_backward_p"}.
#' An object of class \code{"ols_step_backward_p"} is a list containing the
#' following components:
#'
#' \item{model}{final model; an object of class \code{lm}}
#' \item{metrics}{selection metrics}
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' @examples
#' # stepwise backward regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_backward_p(model)
#'
#' # stepwise backward regression plot
#' model <- lm(y ~ ., data = surgical)
#' k <- ols_step_backward_p(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # final model
#' k$model
#'
#' # include or exclude variables
#' # force variable to be included in selection process
#' ols_step_backward_p(model, include = c("age", "alc_mod"))
#'
#' # use index of variable instead of name
#' ols_step_backward_p(model, include = c(5, 7))
#'
#' # force variable to be excluded from selection process
#' ols_step_backward_p(model, exclude = c("pindex"))
#'
#' # use index of variable instead of name
#' ols_step_backward_p(model, exclude = c(2))
#'
#' # hierarchical selection
#' model <- lm(y ~ bcs + alc_heavy + pindex + age + alc_mod, data = surgical)
#' ols_step_backward_p(model, 0.1, hierarchical = TRUE)
#'
#' # plot
#' k <- ols_step_backward_p(model, 0.1, hierarchical = TRUE)
#' plot(k)
#'
#' # steps
#' ols_step_backward_p(model, steps = 2)
#'
#'
#' @family backward selection procedures
#'
#' @export
#'
ols_step_backward_p <- function(model, ...) UseMethod("ols_step_backward_p")

#' @export
#' @rdname ols_step_backward_p
#'
ols_step_backward_p.default <- function(model, p_val = 0.3, include = NULL, exclude = NULL, hierarchical = FALSE, progress = FALSE, details = FALSE, steps = NULL, ...) {

  if (details) {
    progress <- FALSE
  }

  check_inputs(model, include, exclude, progress, details)
  check_values(p_val, 0, 1)

  indterms <- coeff_names(model)
  lenterms <- length(indterms)

  if (is.numeric(include)) {
    include <- indterms[include]
  }

  if (is.numeric(exclude)) {
    exclude <- indterms[exclude]
  }

  if (hierarchical) {
    ols_step_hierarchical(model, p_val, FALSE, progress, details, steps)
  } else {
    l        <- model$model
    nam      <- colnames(attr(model$terms, "factors"))
    lockterm <- c(include, exclude)
    response <- names(model$model)[1]
    preds    <- setdiff(nam, lockterm)
    cterms   <- c(include, preds)

    if (progress || details) {
      ols_candidate_terms(cterms, "backward")
    }

    end   <- FALSE
    step  <- 0
    rpred <- c()
    rsq   <- c()
    arsq  <- c()
    aic   <- c()
    ess   <- c()
    rss   <- c()
    sbic  <- c()
    sbc   <- c()
    cp    <- c()
    rmse  <- c()
    ppos  <- step + 1 + length(include)

    rsq_base   <- summary(model)$r.squared

    if (details) {
      ols_rsquared_init(indterms, "r2", response, rsq_base)
    }

    if (progress) {
      ols_progress_init("backward")
    }

    while (!end) {
      m     <- lm(paste(response, "~", paste(c(include, cterms), collapse = " + ")), l)
      m_sum <- Anova(m)
      nm    <- length(m_sum$`F value`)
      fvals <- m_sum$`F value`[ppos:nm]
      pvals <- m_sum$`Pr(>F)`
      min_f <- which(fvals == min(fvals, na.rm = TRUE))

      len_minf <- length(min_f)

      if (len_minf > 1) {
        min_f <- min_f[1]
      }

      minf <- min_f + ppos - 1


        if (pvals[minf] > p_val) {

          step   <- step + 1
          rpred  <- c(rpred, cterms[minf])
          cterms <- cterms[-minf]
          lp     <- length(rpred)
          fr     <- ols_regress(paste(response, "~", paste(c(include, cterms), collapse = " + ")), l)
          rsq    <- c(rsq, fr$rsq)
          arsq   <- c(arsq, fr$adjr)
          rss    <- c(rss, fr$rss)
          ess    <- c(ess, fr$ess)
          aic    <- c(aic, ols_aic(fr$model))
          sbc    <- c(sbc, ols_sbc(fr$model))
          sbic   <- c(sbic, ols_sbic(fr$model, model))
          cp     <- c(cp, ols_mallows_cp(fr$model, model))
          rmse   <- c(rmse, fr$rmse)

          if (progress) {
            ols_progress_display(rpred, "others")
          }

          if (details) {
            mypred <- c(include, cterms)
            rsq1   <- tail(rsq, n = 1)
            ols_stepwise_details(step, rpred, mypred, response, rsq1, "removed", "rsq")
          }

          if (!is.null(steps)) {
            if (step == steps) {
              end <- TRUE
              if (progress || details) {
                ols_stepwise_break(direction = "backward")
              }
            }
          }
        } else {
          end <- TRUE
          if (progress || details) {
            ols_stepwise_break(direction = "backward")
          }
        }

    }

    if (details) {
      ols_stepwise_vars(rpred, "backward")
    }

    final_model <- lm(paste(response, "~", paste(c(include, cterms), collapse = " + ")), data = l)

    metrics     <- data.frame(step       = seq_len(step),
                              variable   = rpred,
                              r2         = rsq,
                              adj_r2     = arsq,
                              aic        = aic,
                              sbc        = sbc,
                              sbic       = sbic,
                              mallows_cp = cp,
                              rmse       = rmse)

    out <- list(metrics = metrics,
                model   = final_model,
                others  = list(full_model = model))

    class(out) <- "ols_step_backward_p"

    return(out)
  }

}

#' @export
#'
print.ols_step_backward_p <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "backward")
  } else {
    print("No variables have been removed from the model.")
  }
}



#' @export
#' @rdname ols_step_backward_p
#'
plot.ols_step_backward_p <- function(x, model = NA, print_plot = TRUE, details = TRUE, ...) {

  p1 <- ols_plot_stepwise(x, "r2", "R-Square", details, "backward")
  p2 <- ols_plot_stepwise(x, "adj_r2", "Adjusted R-Square", details, "backward")
  p3 <- ols_plot_stepwise(x, "aic", "Adjusted R-Square", details, "backward")
  p4 <- ols_plot_stepwise(x, "rmse", "RMSE", details, "backward")

  myplots <- list(plot_1 = p1, plot_2 = p2, plot_3 = p3, plot_4 = p4)

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2, top = "Stepwise Backward Regression")
  } else {
    return(myplots)
  }

}

test_that("output from stepwise backward regression is as expected", {
  model <- lm(y ~ ., data = surgical)
  expect_snapshot(ols_step_backward_p(model))
  expect_snapshot(ols_step_backward_p(model, progress = TRUE))
  expect_snapshot(ols_step_backward_p(model, details = TRUE))
})

test_that("output from stepwise backward hierarchical regression", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  expect_snapshot(ols_step_backward_p(model, hierarchical = TRUE))
  expect_snapshot(ols_step_backward_p(model, hierarchical = TRUE, progress = TRUE))
  expect_snapshot(ols_step_backward_p(model, hierarchical = TRUE, details = TRUE))
})

test_that("output from stepwise backward regression matches when steps is specified", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  expect_snapshot(ols_step_backward_p(model, steps = 2))
  expect_snapshot(ols_step_backward_p(model, steps = 2, progress = TRUE))
  expect_snapshot(ols_step_backward_p(model, steps = 2, details = TRUE))
})

test_that("backward elimination output matches the expected result", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_backward_p(model)
  expect_equal(k$metrics$step, 1)
  expect_equal(k$metrics$variable, c("x3"))

  k1 <- ols_step_backward_p(model, include = "x3")
  expect_equal(k1$metrics$step, 1)
  expect_equal(k1$metrics$variable, c("x4"))

  k2 <- ols_step_backward_p(model, p_val = 0.1, exclude = "x4")
  expect_equal(k2$metrics$step, 1)
  expect_equal(k2$metrics$variable, c("x3"))

  k3 <- ols_step_backward_p(model, include = c(3))
  expect_equal(k3$metrics$step, 1)
  expect_equal(k3$metrics$variable, c("x4"))

  k4 <- ols_step_backward_p(model, p_val = 0.1, exclude = c(4))
  expect_equal(k4$metrics$step, 1)
  expect_equal(k4$metrics$variable, c("x3"))

})

test_that("backward hierarchical selection output matches the expected result", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  k <- ols_step_backward_p(model, 0.1, hierarchical = TRUE)
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("alc_mod", "gender", "age", "liver_test"), ignore_attr = TRUE)
})

test_that("backward hierarchical selection stops after a specific number of steps", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  k <- ols_step_backward_p(model, 0.1, hierarchical = TRUE, steps = 2)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("alc_mod", "gender"), ignore_attr = TRUE)
})

test_that("backward elimination output matches the expected result when steps in specified", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  k <- ols_step_backward_p(model, steps = 2)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("alc_mod", "gender"))
})

test_that("step_backward_p returns the appropriate error", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  expect_error(ols_step_backward_p(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_backward_p(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_backward_p(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_backward_p(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

#' Studentized residual plot
#'
#' Graph for identifying outliers.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 3.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details
#' Studentized deleted residuals (or externally studentized residuals) is the
#' deleted residual divided by its estimated standard deviation. Studentized
#' residuals are going to be more effective for detecting outlying Y
#' observations than standardized residuals. If an observation has an externally
#' studentized residual that is larger than 3 (in absolute value) we can call
#' it an outlier.
#'
#' @return \code{ols_plot_resid_stud} returns  a list containing the
#' following components:
#'
#' \item{outliers}{a \code{data.frame} with observation number and \code{studentized residuals} that
#' exceed \code{threshold}} for classifying an observation as an outlier
#' \item{threshold}{\code{threshold} for classifying an observation as an outlier}
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_stud(model)
#' ols_plot_resid_stud(model, threshold = 2)
#'
#' @importFrom ggplot2 scale_fill_manual annotate
#'
#' @seealso [ols_plot_resid_stand()]
#'
#' @export
#'
ols_plot_resid_stud <- function(model, threshold = NULL, print_plot = TRUE) {

  check_model(model)

  if (is.null(threshold)) {
    threshold <- 3
  }

  g           <- ols_prep_srplot_data(model, threshold)
  d           <- g$dsr
  d$txt       <- ifelse(d$color == "outlier", d$obs, NA)
  f           <- d[d$color == "outlier", c("obs", "dsr")]
  colnames(f) <- c("observation", "stud_resid")

  p <-
    ggplot(d, aes(x = obs, y = dsr, label = txt)) +
    geom_bar(width = 0.5, stat = "identity", aes(fill = fct_color)) +
    geom_text(hjust = -0.2, nudge_x = 0.05, size = 2, na.rm = TRUE) +
    geom_hline(yintercept = c(0, g$nseq, g$pseq)) +
    geom_hline(yintercept = c(-threshold, threshold), color = "red")

  p <-
    p +
    annotate("text", x = Inf, y = Inf, hjust = 1.2, vjust = 2,
             family = "serif", fontface = "italic", colour = "darkred",
             label = paste0("Threshold: abs(", threshold, ")"))

  p <-
    p +
    scale_fill_manual(values = c("blue", "red"))

  p <-
    p +
    labs(fill = "Observation") +
    xlab("Observation") +
    ylab("Deleted Studentized Residuals") +
    ggtitle("Studentized Residuals Plot") +
    ylim(g$cminx, g$cmaxx)


  if (print_plot) {
    suppressWarnings(print(p))
  } else {
    return(list(plot = p, outliers = f, threshold = threshold))
  }

}

#' Deleted studentized residual vs fitted values plot
#'
#' @description
#' Plot for detecting violation of assumptions about residuals such as
#' non-linearity, constant variances and outliers. It can also be used to
#' examine model fit.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 2.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details
#' Studentized deleted residuals (or externally studentized residuals) is the
#' deleted residual divided by its estimated standard deviation. Studentized
#' residuals are going to be more effective for detecting outlying Y
#' observations than standardized residuals. If an observation has an externally
#' studentized residual that is larger than 2 (in absolute value) we can call
#' it an outlier.
#'
#' @return \code{ols_plot_resid_stud_fit} returns  a list containing the
#' following components:
#'
#' \item{outliers}{a \code{data.frame} with observation number, fitted values and deleted studentized
#' residuals that exceed the \code{threshold} for classifying observations as
#' outliers/influential observations}
#' \item{threshold}{\code{threshold} for classifying an observation as an outlier/influential observation}
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_plot_resid_stud_fit(model)
#' ols_plot_resid_stud_fit(model, threshold = 3)
#'
#' @seealso [ols_plot_resid_lev()], [ols_plot_resid_stand()],
#'   [ols_plot_resid_stud()]
#'
#' @importFrom stats fitted rstudent
#'
#' @export
#'
ols_plot_resid_stud_fit <- function(model, threshold = NULL, print_plot = TRUE) {

  check_model(model)

  if (is.null(threshold)) {
    threshold <- 2
  }

  k           <- ols_prep_dsrvf_data(model, threshold)
  d           <- k$ds
  d$txt       <- ifelse(d$color == "outlier", d$obs, NA)
  f           <- d[d$color == "outlier", c("obs", "pred", "dsr")]
  colnames(f) <- c("observation", "fitted_values", "del_stud_resid")

  p <-
    ggplot(d, aes(x = pred, y = dsr, label = txt)) +
    geom_point(aes(colour = fct_color)) +
    geom_hline(yintercept = c(-threshold, threshold), colour = "red") +
    geom_text(hjust = -0.2, nudge_x = 0.15, size = 3, family = "serif",
              fontface = "italic", colour = "darkred", na.rm = TRUE)

  p <-
    p +
    annotate("text", x = Inf, y = Inf, hjust = 1.5, vjust = 2,
             family = "serif", fontface = "italic", colour = "darkred",
             label = paste0("Threshold: abs(", threshold, ")"))

  p <-
    p +
    scale_color_manual(values = c("blue", "red"))

  p <-
    p +
    labs(color = "Observation") +
    xlab("Predicted Value") +
    ylab("Deleted Studentized Residual") +
    ggtitle("Deleted Studentized Residual vs Predicted Values") +
    ylim(k$cminx, k$cmaxx)

  if (print_plot) {
    suppressWarnings(print(p))
  } else {
    return(list(plot = p, outliers = f, threshold = threshold))
  }

}

#' DFBETAs panel
#'
#' Panel of plots to detect influential observations using DFBETAs.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details
#' DFBETA measures the difference in each parameter estimate with and without
#' the influential point. There is a DFBETA for each data point i.e if there are
#' n observations and k variables, there will be \eqn{n * k} DFBETAs. In
#' general, large values of DFBETAS indicate observations that are influential
#' in estimating a given parameter. Belsley, Kuh, and Welsch recommend 2 as a
#' general cutoff value to indicate influential observations and
#' \eqn{2/\sqrt(n)} as a size-adjusted cutoff.
#'
#' @return list; \code{ols_plot_dfbetas} returns a list of \code{data.frame} (for intercept and each predictor)
#' with the observation number and DFBETA of observations that exceed the threshold for classifying
#' an observation as an outlier/influential observation.
#'
#' @references
#' Belsley, David A.; Kuh, Edwin; Welsh, Roy E. (1980). Regression
#' Diagnostics: Identifying Influential Data and Sources of Collinearity.
#'
#' Wiley Series in Probability and Mathematical Statistics.
#' New York: John Wiley & Sons. pp. ISBN 0-471-05856-4.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_plot_dfbetas(model)
#'
#' @seealso [ols_plot_dffits()]
#'
#' @importFrom stats dfbetas
#'
#' @export
#'
ols_plot_dfbetas <- function(model, print_plot = TRUE) {

  check_model(model)

  dfb       <- dfbetas(model)
  n         <- nrow(dfb)
  np        <- ncol(dfb)
  threshold <- 2 / sqrt(n)

  myplots   <- list()
  outliers  <- list()

  for (i in seq_len(np)) {

    dbetas  <- dfb[, i]
    df_data <- data.frame(obs = seq_len(n), dbetas = dbetas)
    d       <- ols_prep_dfbeta_data(df_data, threshold)
    f       <- ols_prep_dfbeta_outliers(d)

    p <-
      eval(
        substitute(
          ggplot(d, aes(x = obs, y = dbetas, label = txt, ymin = 0, ymax = dbetas)) +
          geom_linerange(colour = "blue") +
          geom_hline(yintercept = c(0, threshold, -threshold), colour = "red") +
          geom_point(colour = "blue", shape = 1) +
          geom_text(hjust = -0.2, nudge_x = 0.15, size = 2, family = "serif",
                    fontface = "italic", colour = "darkred", na.rm = TRUE) +
          annotate("text", x = Inf, y = Inf, hjust = 1.5, vjust = 2,
                   family = "serif", fontface = "italic", colour = "darkred",
                   label = paste("Threshold:", round(threshold, 2))) +
          xlab("Observation") +
          ylab("DFBETAS") +
          ggtitle(paste("Influence Diagnostics for", colnames(dfb)[i])),
          list(i = i)
        )
      )

    myplots[[i]]  <- p
    outliers[[i]] <- f

  }

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2)
  } else {
    return(list(plots = myplots, outliers = outliers))
  }

}

#' Stepwise forward regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering predictors based on p values, in a stepwise manner until there is
#' no variable left to enter any more.
#'
#' @param model An object of class \code{lm}; the model should include all
#'   candidate predictor variables.
#' @param p_val p value; variables with p value less than \code{p_val} will
#'   enter into the model
#' @param include Character or numeric vector; variables to be included in selection process.
#' @param exclude Character or numeric vector; variables to be excluded from selection process.
#' @param hierarchical Logical; if \code{TRUE}, performs hierarchical selection.
#' @param progress Logical; if \code{TRUE}, will display variable selection progress.
#' @param details Logical; if \code{TRUE}, will print the regression result at
#'   each step.
#' @param steps Number of steps after which the stepwise procedures should stop.
#' @param x An object of class \code{ols_step_forward_p}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#' @param ... Other arguments.
#'
#' @return \code{ols_step_forward_p} returns an object of class \code{"ols_step_forward_p"}.
#' An object of class \code{"ols_step_forward_p"} is a list containing the
#' following components:
#'
#' \item{model}{final model; an object of class \code{lm}}
#' \item{metrics}{selection metrics}
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
#'
#' @examples
#' # stepwise forward regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_forward_p(model)
#'
#' # stepwise forward regression plot
#' model <- lm(y ~ ., data = surgical)
#' k <- ols_step_forward_p(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # final model
#' k$model
#'
#' # include or exclude variables
#' # force variable to be included in selection process
#' ols_step_forward_p(model, include = c("age", "alc_mod"))
#'
#' # use index of variable instead of name
#' ols_step_forward_p(model, include = c(5, 7))
#'
#' # force variable to be excluded from selection process
#' ols_step_forward_p(model, exclude = c("pindex"))
#'
#' # use index of variable instead of name
#' ols_step_forward_p(model, exclude = c(2))
#'
#' # hierarchical selection
#' model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test, data = surgical)
#' ols_step_forward_p(model, 0.1, hierarchical = TRUE)
#'
#' # plot
#' k <- ols_step_forward_p(model, 0.1, hierarchical = TRUE)
#' plot(k)
#'
#' # steps
#' ols_step_forward_p(model, steps = 3)
#'
#' @importFrom stats qt
#' @importFrom car Anova
#'
#' @family forward selection procedures
#'
#' @export
#'
ols_step_forward_p <- function(model, ...) UseMethod("ols_step_forward_p")

#' @export
#' @rdname ols_step_forward_p
#'
ols_step_forward_p.default <- function(model, p_val = 0.3, include = NULL, exclude = NULL, hierarchical = FALSE, progress = FALSE, details = FALSE, steps = NULL, ...) {

  if (details) {
    progress <- FALSE
  }

  check_inputs(model, include, exclude, progress, details)
  check_values(p_val, 0, 1)

  indterms <- coeff_names(model)
  lenterms <- length(indterms)

  if (is.numeric(include)) {
    include <- indterms[include]
  }

  if (is.numeric(exclude)) {
    exclude <- indterms[exclude]
  }

  if (hierarchical) {
    ols_step_hierarchical(model, p_val, TRUE, progress, details, steps)
  } else {
    l        <- model$model
    nam      <- colnames(attr(model$terms, "factors"))
    lockterm <- c(include, exclude)
    cterms   <- setdiff(nam, exclude)

    if (progress || details) {
      ols_candidate_terms(cterms, "forward")
    }

    nam      <- setdiff(nam, lockterm)
    n        <- ncol(l)
    response <- names(model$model)[1]
    all_pred <- nam
    mlen_p   <- length(all_pred)

    step     <- 0
    ppos     <- step + 1 + length(include)
    preds    <- include
    pvals    <- c()
    fvals    <- c()
    aicvals  <- c()
    rsqvals  <- c()
    arsqvals <- c()
    rsq      <- c()
    adjrsq   <- c()
    aic      <- c()
    sbic     <- c()
    cp       <- c()
    sbc      <- c()
    rmse     <- c()

    base_model <- ols_base_model(include, response, l)
    rsq_base   <- summary(base_model)$r.squared

    if (details) {
      ols_rsquared_init(include, "r2", response, rsq_base)
    }

    for (i in seq_len(mlen_p)) {
      predictors  <- c(include, all_pred[i])
      m           <- lm(paste(response, "~", paste(predictors, collapse = " + ")), l)
      m_sum       <- Anova(m)
      fvals[i]    <- m_sum$`F value`[ppos]
      pvals[i]    <- m_sum$`Pr(>F)`[ppos]
      aicvals[i]  <- ols_aic(m)
      rsqvals[i]  <- summary(m)$r.squared
      arsqvals[i] <- summary(m)$adj.r.squared
    }

    if (details) {
      d <- data.frame(predictor = all_pred, p_val = pvals, rsq = rsqvals,
                       arsq = arsqvals, aic = aicvals)
      d <- d[order(d$p_val), ]
      ols_stepwise_table(d, all_pred, pvals, rsqvals, arsqvals, aicvals)
    }

    maxf     <- which(fvals == max(fvals, na.rm = TRUE))
    minp     <- which(pvals == min(pvals, na.rm = TRUE))
    len_minp <- length(minp)

    if (len_minp > 1) {
      minp <- minp[1]
    }

    if (pvals[minp] > p_val) {
      stop("None of the variables satisfy the criteria for entering the model.", call. = FALSE)
    } else {
      if (progress) {
        ols_progress_init("forward")
      }
    }

    preds  <- c(preds, all_pred[maxf])
    lpreds <- length(preds)

    if (lpreds > 1) {

      for (i in seq_len(lpreds)) {

        step   <- step + 1
        npreds <- preds[1:i]
        fr     <- ols_regress(paste(response, "~", paste(npreds, collapse = " + ")), l)
        rsq    <- c(rsq, fr$rsq)
        adjrsq <- c(adjrsq, fr$adjr)
        aic    <- c(aic, ols_aic(fr$model))
        sbc    <- c(sbc, ols_sbc(fr$model))
        sbic   <- c(sbic, ols_sbic(fr$model, model))
        cp     <- c(cp, ols_mallows_cp(fr$model, model))
        rmse   <- c(rmse, fr$rmse)
        rsq1   <- tail(rsq, n = 1)

        if (progress) {
          ols_progress_display(npreds, "others")
        }

        if (details) {
          ols_rsquared_selected("r2", step, npreds, response, rsq1)
        }

      }

    } else {

      step   <- step + 1
      fr     <- ols_regress(paste(response, "~", preds), l)
      rsq    <- c(rsq, fr$rsq)
      adjrsq <- c(adjrsq, fr$adjr)
      aic    <- c(aic, ols_aic(fr$model))
      sbc    <- c(sbc, ols_sbc(fr$model))
      sbic   <- c(sbic, ols_sbic(fr$model, model))
      cp     <- c(cp, ols_mallows_cp(fr$model, model))
      rmse   <- c(rmse, fr$rmse)
      rsq1   <- tail(rsq, n = 1)

      if (progress) {
        ols_progress_display(preds, "others")
      }

      if (details) {
        ols_rsquared_selected("r2", step, preds, response, rsq1)
      }

    }

    if (!is.null(steps)) {
      mlen_p <- steps
    }

    while (step < mlen_p) {

      all_pred <- all_pred[-maxf]
      len_p    <- length(all_pred)

      if (len_p == 0) {
        break
      }

      ppos     <- ppos + length(maxf)
      pvals    <- c()
      fvals    <- c()
      aicvals  <- c()
      rsqvals  <- c()
      arsqvals <- c()

      for (i in seq_len(len_p)) {

        predictors  <- c(preds, all_pred[i])
        m           <- lm(paste(response, "~", paste(predictors, collapse = " + ")), l)
        m_sum       <- Anova(m)
        fvals[i]    <- m_sum$`F value`[ppos]
        pvals[i]    <- m_sum$`Pr(>F)`[ppos]
        aicvals[i]  <- ols_aic(m)
        rsqvals[i]  <- summary(m)$r.squared
        arsqvals[i] <- summary(m)$adj.r.squared
      }

      if (details) {
        d <- data.frame(predictor = all_pred, p_val = pvals, rsq = rsqvals,
                         arsq = arsqvals, aic = aicvals)
        d <- d[order(d$p_val), ]
        ols_stepwise_table(d, all_pred, pvals, rsqvals, arsqvals, aicvals)
      }

      maxf  <- which(fvals == max(fvals, na.rm = TRUE))
      minp  <- pvals[maxf]
      len_minp <- length(minp)

      if (len_minp > 1) {
        minp <- minp[1]
        maxf <- maxf[1]
      }

      if (minp <= p_val) {

        step   <- step + 1
        preds  <- c(preds, all_pred[maxf])
        lpreds <- length(preds)
        fr     <- ols_regress(paste(response, "~", paste(preds, collapse = " + ")), l)
        rsq    <- c(rsq, fr$rsq)
        adjrsq <- c(adjrsq, fr$adjr)
        aic    <- c(aic, ols_aic(fr$model))
        sbc    <- c(sbc, ols_sbc(fr$model))
        sbic   <- c(sbic, ols_sbic(fr$model, model))
        cp     <- c(cp, ols_mallows_cp(fr$model, model))
        rmse   <- c(rmse, fr$rmse)
        rsq1   <- tail(rsq, n = 1)

        if (progress) {
          ols_progress_display(preds, "others")
        }

        if (details) {
          ols_rsquared_selected("r2", step, preds, response, rsq1)
        }
      } else {
        if (progress || details) {
          ols_stepwise_break(direction = "forward")
        }
        break
      }
    }

    prsq <- c(rsq[1], diff(rsq))

    if (details) {
      ols_stepwise_vars(preds, "forward")
    }

    final_model <- lm(paste(response, "~", paste(preds, collapse = " + ")), data = l)

    metrics     <- data.frame(step       = seq_len(step),
                              variable   = preds,
                              r2         = rsq,
                              adj_r2     = adjrsq,
                              aic        = aic,
                              sbc        = sbc,
                              sbic       = sbic,
                              mallows_cp = cp,
                              rmse       = rmse)

    out <- list(metrics = metrics,
                model   = final_model,
                others  = list(base_model = base_model,
                               full_model = model))

    class(out) <- "ols_step_forward_p"

    return(out)
  }

}

#' @export
#'
print.ols_step_forward_p <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "forward")
  } else {
    print("No variables have been added to the model.")
  }
}

#' @export
#' @rdname ols_step_forward_p
#'
plot.ols_step_forward_p <- function(x, model = NA, print_plot = TRUE, details = TRUE, ...) {

  a <- NULL

  p1 <- ols_plot_stepwise(x, "r2", "R-Square", details, "forward")
  p2 <- ols_plot_stepwise(x, "adj_r2", "Adjusted R-Square", details, "forward")
  p3 <- ols_plot_stepwise(x, "aic", "AIC", details, "forward")
  p4 <- ols_plot_stepwise(x, "rmse", "RMSE", details, "forward")

  myplots <- list(plot_1 = p1, plot_2 = p2, plot_3 = p3, plot_4 = p4)

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2, top = "Stepwise Forward Regression")
  } else {
    return(myplots)
  }

}

test_that("output from stepwise forward regression is as expected", {
  model <- lm(y ~ ., data = surgical)
  expect_snapshot(ols_step_forward_p(model))
  expect_snapshot(ols_step_forward_p(model, progress = TRUE))
  expect_snapshot(ols_step_forward_p(model, details = TRUE))
})

test_that("output from stepwise forward hierarchical regression", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  expect_snapshot(ols_step_forward_p(model, 0.1, hierarchical = TRUE))
  expect_snapshot(ols_step_forward_p(model, 0.1, hierarchical = TRUE, progress = TRUE))
  expect_snapshot(ols_step_forward_p(model, 0.1, hierarchical = TRUE, details = TRUE))
})

test_that("output from stepwise forward regression is as expected when steps is specified", {
  model <- lm(y ~ ., data = surgical)
  expect_snapshot(ols_step_forward_p(model, steps = 2))
  expect_snapshot(ols_step_forward_p(model, progress = TRUE, steps = 2))
  expect_snapshot(ols_step_forward_p(model, details = TRUE, steps = 2))
})

test_that("forward selection output matches the expected result", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_forward_p(model)
  expect_equal(k$metrics$step, 1:3)
  expect_equal(k$metrics$variable, c("x4", "x1", "x2"), ignore_attr = TRUE)

  k1 <- ols_step_forward_p(model, include = c("x3"))
  expect_equal(k1$metrics$step, 1:3)
  expect_equal(k1$metrics$variable, c("x3", "x4", "x1"), ignore_attr = TRUE)

  k2 <- ols_step_forward_p(model, exclude = c("x4"))
  expect_equal(k2$metrics$step, 1:3)
  expect_equal(k2$metrics$variable, c("x2", "x1", "x3"), ignore_attr = TRUE)

  k3 <- ols_step_forward_p(model, include = c(3))
  expect_equal(k3$metrics$step, 1:3)
  expect_equal(k3$metrics$variable, c("x3", "x4", "x1"), ignore_attr = TRUE)

  k4 <- ols_step_forward_p(model, exclude = c(4))
  expect_equal(k4$metrics$step, 1:3)
  expect_equal(k4$metrics$variable, c("x2", "x1", "x3"), ignore_attr = TRUE)
})

test_that("forward hierarchical selection output matches the expected result", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  k <- ols_step_forward_p(model, 0.1, hierarchical = TRUE)
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("bcs", "alc_heavy", "pindex", "enzyme_test"), ignore_attr = TRUE)
})

test_that("forward hierarchical selection stops after a specific number of steps", {
  model <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
  k <- ols_step_forward_p(model, 0.1, hierarchical = TRUE, steps = 2)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("bcs", "alc_heavy"), ignore_attr = TRUE)
})

test_that("stepwise forward regression stops after a specific number of steps", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_forward_p(model, steps = 2)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("x4", "x1"), ignore_attr = TRUE)
})


test_that("step_forward_p returns the appropriate error", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  expect_error(ols_step_forward_p(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_forward_p(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_forward_p(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_forward_p(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

#' Model Performance
#'
#' @description
#' Evaluate performance of regression models.
#'
#' @param model An object of class \code{lm}.
#' @param ... Other arguments.
#'
#' @return \code{ols_model_performance} returns an object of class
#' \code{"ols_model_performance"}. An object of class \code{"ols_regress"} is a
#' list containing the following components:
#'
#' \item{r}{square root of rsquare, correlation between observed and predicted values of dependent variable}
#' \item{rsq}{coefficient of determination or r-square}
#' \item{adjr}{adjusted rsquare}
#' \item{prsq}{predicted rsquare}
#' \item{aic}{akaike information criteria}
#' \item{sbc}{bayesian information criteria}
#' \item{sbic}{sawa bayesian information criteria}
#' \item{rmse}{root mean squared error}
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#'
#' # model performance
#' ols_model_performance(model)
#'
#' @export
#'
ols_model_performance <- function(model, ...) UseMethod("ols_model_performance")

#' @export
#'
ols_model_performance.default <- function(model, ...) {
  rsq  <- summary(model)$r.squared
  r    <- sqrt(rsq)
  adjr <- summary(model)$adj.r.squared
  rmse <- sqrt(mean(model$residuals ^ 2))
  aic  <- ols_aic(model)
  sbc  <- ols_sbc(model)
  sbic <- ols_sbic(model, model)
  prsq <- ols_pred_rsq(model)

  result <- list(
    r    = r,
    rsq  = rsq,
    adjr = adjr,
    prsq = prsq,
    aic  = aic,
    sbc  = sbc,
    sbic = sbic,
    rmse = rmse
  )

  class(result) <- "ols_model_performance"

  return(result)

}

#' @export
#'
print.ols_model_performance <- function(x, ...) {
  print_model_perf(x)
}

#' Compare model performance
#'
#' @description
#' Compare performance of regression models using multiple metrics.
#'
#' @param ... Multiple objects of class \code{lm}.
#'
#' @examples
#' model_1 <- lm(mpg ~ disp + hp, data = mtcars)
#' model_2 <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' model_3 <- lm(mpg ~ disp + hp + wt + qsec + + drat, data = mtcars)
#' ols_compare_model_performance(model_1, model_2, model_3)
#'
#' @export
#'
ols_compare_model_performance <- function(...) UseMethod("ols_compare_model_performance")

#' @export
#'
ols_compare_model_performance.lm <- function(...) {

  models <- list(...)
  n <- length(models)
  model_names <- sapply(substitute(list(...)), deparse)[-1]
  result <- data.frame(matrix(ncol = 9, nrow = 0))

  for (i in seq_len(n)) {
    result <-
      rbind(
        result,
        unlist(olsrr::ols_model_performance(models[[i]]))[-1]
      )
  }

  sigma <- vector(mode = "numeric", length = n)
  for (i in seq_len(n)) {
    sigma[i] = summary(models[[i]])$sigma
  }

  result <- cbind(cbind(model_names, result), sigma)

  col_names <- c("name", "r2", "r2_adj", "r2_pred", "aic", "sbc", "sbic",
    "rmse", "sigma")
  colnames(result) <- col_names

  class(result) <- c("ols_compare_model_performance", "data.frame")

  return(result)
}

test_that("model performance works as expected", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  perf  <- ols_model_performance(model)
  expect_equal(round(perf$r, 3), 0.909)
  expect_equal(round(perf$rsq, 3), 0.827)
  expect_equal(round(perf$adjr, 3), 0.808)
  expect_equal(round(perf$prsq, 3), 0.768)
  expect_equal(round(perf$aic, 3), 158.643)
  expect_equal(round(perf$sbc, 3), 165.972)
  expect_equal(round(perf$sbic, 3), 68.933)
  expect_equal(round(perf$rmse, 3), 2.468)
})

test_that("output from model performance is as expected", {
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  expect_snapshot(ols_model_performance(model))
})

test_that("model comparison works as expected", {
  model_1 <- lm(mpg ~ disp + hp, data = mtcars)
  model_2 <- lm(mpg ~ disp + hp + wt, data = mtcars)
  model_3 <- lm(mpg ~ disp + hp + wt + qsec + + drat, data = mtcars)
  result <- ols_compare_model_performance(model_1, model_2, model_3)
  expect_equal(ncol(result), 9)
  expect_equal(nrow(result), 3)
  expect_equal(result[, 1], c("model_1", "model_2", "model_3"))
  expect_equal(round(result$sigma, 3), c(3.127, 2.639, 2.558))
})

#' Studentized residuals vs leverage plot
#'
#' Graph for detecting outliers and/or observations with high leverage.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 2.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(read ~ write + math + science, data = hsb)
#' ols_plot_resid_lev(model)
#' ols_plot_resid_lev(model, threshold = 3)
#'
#' @importFrom ggplot2 geom_vline
#'
#' @seealso [ols_plot_resid_stud_fit()], [ols_plot_resid_lev()]
#'
#' @export
#'
ols_plot_resid_lev <- function(model, threshold = NULL, print_plot = TRUE) {

  check_model(model)

  if (is.null(threshold)) {
    threshold <- 2
  }

  data <- ols_prep_residlev_data(model, threshold)

  p <-
    ggplot(data$d, aes(leverage, rstudent, label = txt)) +
    geom_point(shape = 1, aes(colour = fct_color)) +
    geom_text(data = data$d2, vjust = -1, size = 3, family = "serif",
              fontface = "italic", colour = "darkred") +
    geom_hline(yintercept = c(threshold, -threshold), colour = "maroon") +
    geom_vline(xintercept = data$g$lev_thrsh, colour = "maroon")

  p <-
    p +
    annotate("text", x = Inf, y = Inf, hjust = 1.2, vjust = 2,
             family = "serif", fontface = "italic", colour = "darkred",
             label = data$ann_lev) +
    annotate("text", x = Inf, y = Inf, hjust = 1.5, vjust = 4.5,
             family = "serif", fontface = "italic", colour = "darkred",
             label = data$ann_out)

  p <-
    p +
    scale_color_manual(values = c("blue", "red", "green", "violet"))

  p <-
    p +
    labs(colour = "Observation") +
    xlab("Leverage") +
    ylab("RStudent") +
    ggtitle(data$title) +
    xlim(data$g$minx, data$g$maxx) +
    ylim(data$g$miny, data$g$maxy)

  if (print_plot) {
    suppressWarnings(print(p))
  } else {
    return(list(plot = p, leverage  = data$f, threshold = data$g$lev_thrsh))
  }

}

ols_prep_residlev_data <- function(model, threshold) {
  resp        <- names(model.frame(model))[1]
  title       <- paste("Outlier and Leverage Diagnostics for", resp)
  g           <- ols_prep_rstudlev_data(model, threshold)
  ann_paste   <- round(g$lev_thrsh, 3)
  ann_lev     <- paste("Leverage Threshold:", ann_paste)
  ann_out     <- paste("Outlier Threshold:", threshold)
  d           <- g$levrstud
  d$txt       <- ifelse(d$color == "normal", NA, d$obs)
  f           <- d[d$color == "outlier", c("obs", "leverage", "rstudent")]
  colnames(f) <- c("observation", "leverage", "stud_resid")
  d2          <- d[!is.na(d$txt), ]

  list(d = d, d2 = d2, f = f, g = g, title = title, ann_lev = ann_lev, ann_out = ann_out)
}

#' Akaike information criterion
#'
#' @description Akaike information criterion for model selection.
#'
#' @param model An object of class \code{lm}.
#'
#' @param method A character vector; specify the method to compute AIC. Valid
#'   options include R, STATA and SAS.
#'
#' @param corrected Logical; if \code{TRUE}, returns corrected akaike information criterion for SAS method.
#'
#' @details
#' AIC provides a means for model selection. Given a collection of models for
#' the data, AIC estimates the quality of each model, relative to each of the
#' other models. R and STATA use loglikelihood to compute AIC. SAS uses residual
#' sum of squares. Below is the formula in each case:
#'
#' \emph{R & STATA}
#' \deqn{AIC = -2(loglikelihood) + 2p}
#'
#' \emph{SAS}
#' \deqn{AIC = n * ln(SSE / n) + 2p}
#'
#' \emph{corrected}
#' \deqn{AIC = n * ln(SSE / n) + ((n * (n + p)) / (n - p - 2))}
#'
#' where \emph{n} is the sample size and \emph{p} is the number of model parameters including intercept.
#'
#' @return Akaike information criterion of the model.
#'
#' @references
#' Akaike, H. (1969). “Fitting Autoregressive Models for Prediction.” Annals of the Institute of Statistical
#' Mathematics 21:243–247.
#'
#' Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
#' New York: John Wiley & Sons.
#'
#' @examples
#' # using R computation method
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_aic(model)
#'
#' # using STATA computation method
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_aic(model, method = 'STATA')
#'
#' # using SAS computation method
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_aic(model, method = 'SAS')
#'
#' # corrected akaike information criterion
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_aic(model, method = 'SAS', corrected = TRUE)
#'
#' @family model selection criteria
#'
#' @importFrom stats logLik
#'
#' @export
#'
ols_aic <- function(model, method = c("R", "STATA", "SAS"), corrected = FALSE) {

  check_model(model)

  method <- match.arg(method)

  if (method != "SAS") {
    corrected <- FALSE
  }

  n      <- model_rows(model)
  p      <- model_n_coeffs(model)

  if (method == "R") {

    lk <- logLik(model)
    -2 * lk[1] + 2 * (p + 1)

  } else if (method == "STATA") {

    lk <- logLik(model)
    -2 * lk[1] + 2 * p

  } else {

    sse <- model_rss(model)

    if (corrected) {
      (n * log(sse / n)) + ((n * (n + p)) / (n - p - 2))
    } else {
      (n * log(sse / n)) + (2 * p)
    }
    
  } 

}


#' Bayesian information criterion
#'
#' Bayesian information criterion for model selection.
#'
#' @param model An object of class \code{lm}.
#' @param method A character vector; specify the method to compute BIC. Valid
#'   options include R, STATA and SAS.
#'
#' @details
#' SBC provides a means for model selection. Given a collection of models for
#' the data, SBC estimates the quality of each model, relative to each of the
#' other models. R and STATA use loglikelihood to compute SBC. SAS uses residual
#' sum of squares. Below is the formula in each case:
#'
#' \emph{R & STATA}
#' \deqn{AIC = -2(loglikelihood) + ln(n) * 2p}
#'
#' \emph{SAS}
#' \deqn{AIC = n * ln(SSE / n) + p * ln(n)}
#'
#' where \emph{n} is the sample size and \emph{p} is the number of model parameters including intercept.
#'
#' @return The bayesian information criterion of the model.
#'
#' @references
#' Schwarz, G. (1978). “Estimating the Dimension of a Model.” Annals of Statistics 6:461–464.
#'
#' Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
#' New York: John Wiley & Sons.
#'
#' @examples
#' # using R computation method
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_sbc(model)
#'
#' # using STATA computation method
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_sbc(model, method = 'STATA')
#'
#' # using SAS computation method
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_sbc(model, method = 'SAS')
#'
#' @family model selection criteria
#'
#' @export
#'
ols_sbc <- function(model, method = c("R", "STATA", "SAS")) {

  check_model(model)

  method <- match.arg(method)
  n      <- model_rows(model)
  p      <- model_n_coeffs(model)

  if (method == "R") {

    lk <- logLik(model)
    -2 * lk[1] + log(n) * (p + 1)

  } else if (method == "STATA") {

    lk <- logLik(model)
    -2 * lk[1] + log(n) * p

  } else {

    sse <- model_rss(model)
    n * log(sse / n) + p * log(n)

  } 
}



#' Sawa's bayesian information criterion
#'
#' @description Sawa's bayesian information criterion for model selection.
#'
#' @param model An object of class \code{lm}.
#' @param full_model An object of class \code{lm}.
#'
#' @details
#' Sawa (1978) developed a model selection criterion that was derived from a
#' Bayesian modification of the AIC criterion. Sawa's Bayesian Information
#' Criterion (BIC) is a function of the number of observations n, the SSE, the
#' pure error variance fitting the full model, and the number of independent
#' variables including the intercept.
#'
#' \deqn{SBIC = n * ln(SSE / n) + 2(p + 2)q - 2(q^2)}
#'
#' where \eqn{q = n(\sigma^2)/SSE}, \emph{n} is the sample size, \emph{p} is the number of model parameters including intercept
#' \emph{SSE} is the residual sum of squares.
#' @return Sawa's Bayesian Information Criterion
#'
#' @references
#' Sawa, T. (1978). “Information Criteria for Discriminating among Alternative Regression Models.” Econometrica
#' 46:1273–1282.
#'
#' Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
#' New York: John Wiley & Sons.
#'
#' @examples
#' full_model <- lm(mpg ~ ., data = mtcars)
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_sbic(model, full_model)
#'
#' @family model selection criteria
#'
#' @export
#'
ols_sbic <- function(model, full_model) {

  check_model(model)
  check_model(full_model)

  n <- model_rows(model)
  p <- anova_coeffs(model)
  r <- full_model_coeffs(full_model)
  q <- (q1(full_model, r) / (q2(model, p))) * (n)

  sbicout(model, n, p, q)

}

#' Extract mean square
#'
#' Extracts the mean square from \code{anova()}.
#'
#' @param full_model An object of class \code{lm}.
#' @param r A vector of length 1.
#'
#' @keywords internal
#'
#' @noRd
#'
q1 <- function(full_model, r) {
  anova(full_model)[[3]][r]
}

#' Extract sum of squares
#'
#' Extracts the sum of squares from \code{anova()}.
#'
#' @param full_model An object of class \code{lm}.
#' @param p A vector of length 1.
#'
#' @keywords internal
#'
#' @noRd
#'
q2 <- function(model, p) {
  anova(model)[[2]][p] 
}

#' SBIC internal
#'
#' Returns sawa's bayesian information criterion.
#'
#' @param model An object of class \code{lm}.
#' @param n A vector of length 1.
#' @param p A vector of length 1.
#' @param q A vector of length 1.
#'
#' @keywords internal
#'
#' @noRd
#'
sbicout <- function(model, n, p, q) {

  a <- (2 * (p + 2) * q)
  b <- (2 * (q ^ 2))
  (log(model_rss(model) / n) * n) + a - b

}



#' Mallow's Cp
#'
#' @description Mallow's Cp.
#'
#' @param model An object of class \code{lm}.
#' @param fullmodel An object of class \code{lm}.
#'
#' @details
#' Mallows' Cp statistic estimates the size of the bias that is introduced into
#' the predicted responses by having an underspecified model. Use Mallows' Cp
#' to choose between multiple regression models. Look for models where
#' Mallows' Cp is small and close to the number of predictors in the model plus
#' the constant (p).
#'
#' @return Mallow's Cp of the model.
#'
#' @references
#' Hocking, R. R. (1976). “The Analysis and Selection of Variables in a Linear Regression.” Biometrics
#' 32:1–50.
#'
#' Mallows, C. L. (1973). “Some Comments on Cp.” Technometrics 15:661–675.
#'
#' @examples
#' full_model <- lm(mpg ~ ., data = mtcars)
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_mallows_cp(model, full_model)
#'
#' @family model selection criteria
#'
#' @export
#'
ols_mallows_cp <- function(model, fullmodel) {

  check_model(model)
  check_model(fullmodel)

  n <- model_rows(model)
  p <- model_n_coeffs(model)
  q <- model_n_coeffs(fullmodel)

  mcpout(model, fullmodel, n, p, q)

}

#' Mallow's Cp internal
#'
#' Computes Mallow's Cp.
#'
#' @param model An object of class \code{lm}.
#' @param fullmodel An object of class \code{lm}.
#' @param n A numeric vector of length 1.
#' @param p A numeric vector of length 1.
#' @param q A numeric vector of length 1.
#'
#' @keywords internal
#'
#' @noRd
#'
mcpout <- function(model, fullmodel, n, p, q) {

  sse <- model_rss(model)
  sec <- (n - (2 * p))
  mse <- rev(anova(fullmodel)[[3]])[1]
  
  (sse / mse) - sec

}



#' MSEP
#'
#' Estimated error of prediction, assuming multivariate normality.
#'
#' @param model An object of class \code{lm}.
#'
#' @details
#' Computes the estimated mean square error of prediction assuming that both
#' independent and dependent variables are multivariate normal.
#'
#' \deqn{MSE(n + 1)(n - 2) / n(n - p - 1)}
#'
#' where \eqn{MSE = SSE / (n - p)}, n is the sample size and p is the number of
#' predictors including the intercept
#'
#' @return Estimated error of prediction of the model.
#'
#' @references
#' Stein, C. (1960). “Multiple Regression.” In Contributions to Probability and Statistics: Essays in Honor
#' of Harold Hotelling, edited by I. Olkin, S. G. Ghurye, W. Hoeffding, W. G. Madow, and H. B. Mann,
#' 264–305. Stanford, CA: Stanford University Press.
#'
#' Darlington, R. B. (1968). “Multiple Regression in Psychological Research and Practice.” Psychological
#' Bulletin 69:161–182.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_msep(model)
#'
#' @family model selection criteria
#'
#' @export
#'
ols_msep <- function(model) {

  check_model(model)
  sepout(model)

}

#' MSEP internal
#'
#' Computes the estimated mean square error of prediction.
#'
#' @param model An object of class \code{lm}.
#'
#' @keywords internal
#'
#' @noRd
#'
sepout <- function(model) {

  n   <- model_rows(model)
  p   <- anova_coeffs(model)
  mse <- anova(model)[[2]][p]
  num <- ((n + 1) * (n - 2)) * mse
  den <- n * (n - p - 1)

  num / den

}

#' Final prediction error
#'
#' @description Estimated mean square error of prediction.
#'
#' @param model An object of class \code{lm}.
#'
#' @details
#' Computes the estimated mean square error of prediction for each model
#' selected assuming that the values of the regressors are fixed and that the
#' model is correct.
#'
#' \deqn{MSE((n + p) / n)}
#'
#' where \eqn{MSE = SSE / (n - p)}, n is the sample size and p is the number of predictors including the intercept
#'
#' @return Final prediction error of the model.
#'
#' @references
#' Akaike, H. (1969). “Fitting Autoregressive Models for Prediction.” Annals of the Institute of Statistical
#' Mathematics 21:243–247.
#'
#' Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
#' New York: John Wiley & Sons.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_fpe(model)
#'
#' @family model selection criteria
#'
#' @export
#'
ols_fpe <- function(model) {

  check_model(model)
  jpout(model)

}

#' Final prediction error internal
#'
#' Computes the final prediction error.
#'
#' @param model An object of class \code{lm}.
#'
#' @keywords internal
#'
#' @noRd
#'
jpout <- function(model) {

  n   <- model_rows(model)
  p   <- anova_coeffs(model)
  mse <- anova(model)[[3]][p]
  ((n + p) / n) * mse

}


#' @title Amemiya's prediction criterion
#'
#' @description Amemiya's prediction error.
#'
#' @param model An object of class \code{lm}.
#'
#' @details
#' Amemiya's Prediction Criterion penalizes R-squared more heavily than does
#' adjusted R-squared for each addition degree of freedom used on the
#' right-hand-side of the equation.  The lower the better for this criterion.
#'
#' \deqn{((n + p) / (n - p))(1 - (R^2))}
#'
#' where \emph{n} is the sample size, \emph{p} is the number of predictors including the intercept and
#' \emph{R^2} is the coefficient of determination.
#'
#' @return Amemiya's prediction error of the model.
#'
#' @references
#' Amemiya, T. (1976). Selection of Regressors. Technical Report 225, Stanford University, Stanford, CA.
#'
#' Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
#' New York: John Wiley & Sons.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_apc(model)
#'
#' @family model selection criteria
#'
#' @export
#'
ols_apc <- function(model) {

  check_model(model)
  pcout(model)

}

#' Amemiya internal
#'
#' Computes the Amemiya prediction error.
#'
#' @param model An object of class \code{lm}.
#'
#' @keywords internal
#'
#' @noRd
#'
pcout <- function(model) {

  n   <- model_rows(model)
  p   <- anova_coeffs(model)
  rse <- summary(model)[[8]]
  ((n + p) / (n - p)) * (1 - rse)

}


#' @title Hocking's Sp
#'
#' @description Average prediction mean squared error.
#'
#' @param model An object of class \code{lm}.
#'
#' @details Hocking's Sp criterion is an adjustment of the residual sum of
#' Squares. Minimize this criterion.
#'
#' \deqn{MSE / (n - p - 1)}
#'
#' where \eqn{MSE = SSE / (n - p)}, n is the sample size and p is the number of predictors including the intercept
#'
#' @return Hocking's Sp of the model.
#'
#' @references
#' Hocking, R. R. (1976). “The Analysis and Selection of Variables in a Linear Regression.” Biometrics
#' 32:1–50.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_hsp(model)
#'
#' @family model selection criteria
#'
#' @export
#'
ols_hsp <- function(model) {

  check_model(model)
  spout(model)

}

#' Hocking's internal
#'
#' Computes the Hocking's Sp statistic.
#'
#' @param model An object of class \code{lm}.
#'
#' @keywords internal
#'
#' @noRd
#'
spout <- function(model) {

  n   <- model_rows(model)
  p   <- anova_coeffs(model)
  mse <- anova(model)[[3]][p]
    
  mse / (n - p - 1)

}

#' Model data rows
#'
#' Returns the number of rows in the data used in the model.
#'
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
model_rows <- function(model) {
  nrow(model.frame(model))
}

#' Model Coefficients
#'
#' Returns the number of coefficients in the model.
#'
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
model_n_coeffs <- function(model) {
  length(model$coefficients)
}

#' Residual sum of squares
#'
#' Returns the residual sum of squares.
#'
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
model_rss <- function(model) {
  sum(residuals(model) ^ 2)
}

#' Coefficients
#'
#' Returns the number of coefficients in the model
#'   using `anova()` including the residual sum of squares.
#'
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
anova_coeffs <- function(model) {
  length(anova(model)[[1]])
}

#' Number of columns
#'
#' Returns the number of columns in the data.
#'
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
full_model_coeffs <- function(model) {
  length(model.frame(model)) 
}

fullmodel <- lm(y ~ x1 + x2 + x3 + x4, data = cement)

test_that("aic output matches expected result", {
  expect_equal(ols_aic(fullmodel, method = "R"), AIC(fullmodel))
  expect_equal(ols_aic(fullmodel, method = "STATA"), 63.83669)
  expect_equal(round(ols_aic(fullmodel, method = "SAS"), 3), 26.944)
  expect_equal(round(ols_aic(fullmodel, method = "SAS", TRUE), 3), 55.944)

  model <- lm(y ~ x1 + x2 + x4, data = cement)
  expect_equal(ols_aic(model, method = "R"), AIC(model))
  expect_equal(round(ols_aic(model, method = "STATA"), 4), 61.8663)

  model <- lm(y ~ x2 + x4, data = cement)
  expect_equal(ols_aic(model, method = "R"), AIC(model))
  expect_equal(round(ols_aic(model, method = "STATA"), 4), 97.5217)
})


test_that("sbc output matches expected result", {
  expect_equal(ols_sbc(fullmodel, method = "R"), BIC(fullmodel))
  expect_equal(round(ols_sbc(fullmodel, method = "STATA"), 4), 66.6614)
  expect_equal(round(ols_sbc(fullmodel, method = "SAS"), 4), 29.769)

  model <- lm(y ~ x1 + x2 + x4, data = cement)
  expect_equal(ols_sbc(model, method = "R"), BIC(model))
  expect_equal(round(ols_sbc(model, method = "STATA"), 4), 64.1261)

  model <- lm(y ~ x2 + x4, data = cement)
  expect_equal(ols_sbc(model, method = "R"), BIC(model))
  expect_equal(round(ols_sbc(model, method = "STATA"), 4), 99.2166)
})

test_that("apc output matches expected result", {
  expect_equal(round(ols_apc(fullmodel), 3), 0.04)

  model <- lm(y ~ x1, data = cement)
  expect_equal(round(ols_apc(model), 3), 0.636)

  model <- lm(y ~ x4, data = cement)
  expect_equal(round(ols_apc(model), 3), 0.444)

  model <- lm(y ~ x1 + x2 + x4, data = cement)
  expect_equal(round(ols_apc(model), 3), 0.033)

  model <- lm(y ~ x1 + x3 + x4, data = cement)
  expect_equal(round(ols_apc(model), 3), 0.035)

  model <- lm(y ~ x2 + x4, data = cement)
  expect_equal(round(ols_apc(model), 3), 0.512)
})


test_that("mallow's cp output matches expected result", {
  model <- lm(y ~ x1, data = cement)
  act <- round(ols_mallows_cp(model, fullmodel), 2)
  exp <- 202.55
  expect_equal(act, exp)

  model <- lm(y ~ x4, data = cement)
  act <- round(ols_mallows_cp(model, fullmodel), 2)
  exp <- 138.73
  expect_equal(act, exp)

  model <- lm(y ~ x1 + x2, data = cement)
  act <- round(ols_mallows_cp(model, fullmodel), 2)
  exp <- 2.68
  expect_equal(act, exp)

  model <- lm(y ~ x2 + x4, data = cement)
  act <- round(ols_mallows_cp(model, fullmodel), 2)
  exp <- 138.23
  expect_equal(act, exp)

  model <- lm(y ~ x1 + x2 + x4, data = cement)
  act <- round(ols_mallows_cp(model, fullmodel), 2)
  exp <- 3.02
  expect_equal(act, exp)

  model <- lm(y ~ x1 + x3 + x4, data = cement)
  act <- round(ols_mallows_cp(model, fullmodel), 2)
  exp <- 3.50
  expect_equal(act, exp)
})

#' Added variable plot data
#'
#' Data for generating the added variable plots.
#'
#' @param model An object of class \code{lm}.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_prep_avplot_data(model)
#'
#' @importFrom stats model.frame residuals as.formula
#'
#' @export
#'
ols_prep_avplot_data <- function(model) {

  m1 <- as.data.frame(model.frame(model))[1]
  m2 <- as.data.frame(model.matrix(model))[, -1]
  as.data.frame(cbind(m1, m2))

}

#' Regress predictor on other predictors
#'
#' Regress a predictor in the model on all the other predictors.
#'
#' @param data A \code{data.frame}.
#' @param i A numeric vector (indicates the predictor in the model).
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' data <- ols_prep_avplot_data(model)
#' ols_prep_regress_x(data, 1)
#'
#' @importFrom stats lsfit
#'
#' @export
#'
ols_prep_regress_x <- function(data, i) {

  x <- remove_columns(data, i)
  y <- select_columns(data, i)
  lsfit(x, y)$residuals

}

#' Regress y on other predictors
#'
#' Regress y on all the predictors except the ith predictor.
#'
#' @param data A \code{data.frame}.
#' @param i A numeric vector (indicates the predictor in the model).
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' data <- ols_prep_avplot_data(model)
#' ols_prep_regress_y(data, 1)
#'
#' @export
#'
ols_prep_regress_y <- function(data, i) {

  x <- remove_columns(data, i)
  y <- select_columns(data)
  lsfit(x, y)$residuals

}

#' Cooks' D plot data
#'
#' Prepare data for cook's d bar plot.
#'
#' @param model An object of class \code{lm}.
#' @param type An integer between 1 and 5 selecting one of the 6 methods for computing the threshold.
#'
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_prep_cdplot_data(model)
#'
#' @export
#'
ols_prep_cdplot_data <- function(model, type = 1) {

  cooksd    <- cooks.distance(model)
  n         <- length(cooksd)
  obs       <- seq_len(n)
  ckd       <- data.frame(obs = obs, cd = cooksd)
  ts        <- ols_cooks_ts(model, type)
  cooks_max <- max(cooksd)

  ckd$color     <- ifelse(ckd$cd >= ts, "outlier", "normal")
  ckd$fct_color <- ordered(factor(ckd$color), levels = c("normal", "outlier"))
  maxx          <- cooks_max * 0.01 + cooks_max
  
  list(ckd = ckd, maxx = maxx, ts = ts)

}

ols_cooks_ts <- function(model, type = 1) {

  cooksd <- cooks.distance(model)
  n      <- length(cooksd)
  k      <- length(model$coefficients) - 1

  switch(type,
         `1` = (4 / n),
         `2` = (4 / (n - k - 1)),
         `3` = (1),
         `4` = (1 / (n - k - 1)),
         `5` = (3 * mean(cooksd)))

}

#' Cooks' D outlier observations
#'
#' Identify outliers in cook's d plot.
#'
#' @param k Cooks' d bar plot data.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' k <- ols_prep_cdplot_data(model)
#' ols_prep_outlier_obs(k)
#'
#' @export
#'
ols_prep_outlier_obs <- function(k) {

  data <- k$ckd
  data$txt <- ifelse(data$color == "outlier", data$obs, NA)
  return(data)

}

#' Cooks' d outlier data
#'
#' Outlier data for cook's d bar plot.
#'
#' @param k Cooks' d bar plot data.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' k <- ols_prep_cdplot_data(model)
#' ols_prep_cdplot_outliers(k)
#'
#' @export
#'
ols_prep_cdplot_outliers <- function(k) {

  result <- k$ckd[k$ckd$color == "outlier", c("obs", "cd")]
  names(result) <- c("observation", "cooks_distance")
  return(result)

}

#' DFBETAs plot data
#'
#' Prepares the data for dfbetas plot.
#'
#' @param d A \code{tibble} or \code{data.frame} with dfbetas.
#' @param threshold The threshold for outliers.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' dfb <- dfbetas(model)
#' n <- nrow(dfb)
#' threshold <- 2 / sqrt(n)
#' dbetas  <- dfb[, 1]
#' df_data <- data.frame(obs = seq_len(n), dbetas = dbetas)
#' ols_prep_dfbeta_data(df_data, threshold)
#'
#' @export
#'
ols_prep_dfbeta_data <- function(d, threshold) {

  d$color     <- ifelse(((d$dbetas >= threshold) | (d$dbetas <= -threshold)), c("outlier"), c("normal"))
  d$fct_color <- ordered(factor(d$color), levels = c("normal", "outlier"))
  d$txt       <- ifelse(d$color == "outlier", d$obs, NA)

  return(d)

}

#' DFBETAs plot outliers
#'
#' Data for identifying outliers in dfbetas plot.
#'
#' @param d A \code{tibble} or \code{data.frame}.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' dfb <- dfbetas(model)
#' n <- nrow(dfb)
#' threshold <- 2 / sqrt(n)
#' dbetas  <- dfb[, 1]
#' df_data <- data.frame(obs = seq_len(n), dbetas = dbetas)
#' d <- ols_prep_dfbeta_data(df_data, threshold)
#' ols_prep_dfbeta_outliers(d)
#'
#' @export
#'
ols_prep_dfbeta_outliers <- function(d) {

  d[d$color == "outlier", c("obs", "dbetas")]

}


#' Deleted studentized residual plot data
#'
#' Generates data for deleted studentized residual vs fitted plot.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 2.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_prep_dsrvf_data(model)
#' ols_prep_dsrvf_data(model, threshold = 3)
#'
#' @export
#'
ols_prep_dsrvf_data <- function(model, threshold = NULL) {

  pred    <- fitted(model)
  dsresid <- unname(rstudent(model))
  n       <- length(dsresid)
  ds      <- data.frame(obs = seq_len(n), dsr = dsresid)

  if (is.null(threshold)) {
    threshold <- 2
  }

  ds$color     <- ifelse((abs(ds$dsr) >= threshold), "outlier", "normal")
  ds$fct_color <- ordered(factor(ds$color), levels = c("normal", "outlier"))

  ds2 <- data.frame(obs       = seq_len(n),
                    pred      = pred,
                    dsr       = ds$dsr,
                    color     = ds$color,
                    fct_color = ds$fct_color)

  minx  <- min(ds2$dsr) - 1
  maxx  <- max(ds2$dsr) + 1
  cminx <- ifelse(minx < -threshold, minx, (-threshold - 0.5))
  cmaxx <- ifelse(maxx > threshold, maxx, (threshold + 0.5))

  list(ds = ds2, cminx = cminx, cmaxx = cmaxx)

}


#' Residual fit spread plot data
#'
#' Data for generating residual fit spread plot.
#'
#' @param model An object of class \code{lm}.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_prep_rfsplot_fmdata(model)
#' ols_prep_rfsplot_rsdata(model)
#'
#' @export
#'
ols_prep_rfsplot_fmdata<- function(model) {

  predicted <- fitted(model)
  pred_m    <- mean(predicted)
  y         <- predicted - pred_m
  percenti  <- ecdf(y)
  x         <- percenti(y)

  data.frame(x, y)

}

#' @rdname ols_prep_rfsplot_fmdata
#' @export
#'
ols_prep_rfsplot_rsdata <- function(model) {

  y         <- residuals(model)
  residtile <- ecdf(y)
  x         <- residtile(y)

  data.frame(x, y)

}

#' Residual vs regressor plot data
#'
#' Data for generating residual vs regressor plot.
#'
#' @param model An object of class \code{lm}.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_prep_rvsrplot_data(model)
#'
#' @export
#'
ols_prep_rvsrplot_data <- function(model) {

  np     <- length(coefficients(model)) - 1
  dat    <- model.frame(model)[, -1]
  pnames <- names(coefficients(model))[-1]

  list(np = np, dat = dat, pnames = pnames)

}

#' Studentized residual vs leverage plot data
#'
#' Generates data for studentized resiudual vs leverage plot.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 2.
#'
#' @examples
#' model <- lm(read ~ write + math + science, data = hsb)
#' ols_prep_rstudlev_data(model)
#' ols_prep_rstudlev_data(model, threshold = 3)
#'
#'
#' @export
#'
ols_prep_rstudlev_data <- function(model, threshold = NULL) {

  if (is.null(threshold)) {
    threshold <- 2
  }

  leverage  <- unname(hatvalues(model))
  rstudent  <- unname(rstudent(model))
  k         <- length(coefficients(model))
  n         <- nrow(model.frame(model))
  lev_thrsh <- (2 * k) / n
  rst_thrsh <- threshold
  miny      <- min(rstudent) - 3
  maxy      <- max(rstudent) + 3
  minx      <- min(leverage)
  maxx      <- ifelse((max(leverage) > lev_thrsh), max(leverage), (lev_thrsh + 0.05))
  levrstud  <- data.frame(obs = seq_len(n), leverage, rstudent)

  levrstud$color1 <- ifelse((levrstud$leverage < lev_thrsh & abs(levrstud$rstudent) < rst_thrsh), "normal", NA)
  levrstud$color2 <- ifelse((levrstud$leverage > lev_thrsh & abs(levrstud$rstudent) < rst_thrsh), "leverage", NA)
  levrstud$color3 <- ifelse((levrstud$leverage < lev_thrsh & abs(levrstud$rstudent) > rst_thrsh), "outlier", NA)
  levrstud$color4 <- ifelse((levrstud$leverage > lev_thrsh & abs(levrstud$rstudent) > rst_thrsh), "outlier & leverage", NA)

  d1 <- levrstud[!is.na(levrstud$color1), c("obs", "leverage", "rstudent", "color1")]
  d2 <- levrstud[!is.na(levrstud$color2), c("obs", "leverage", "rstudent", "color2")]
  d3 <- levrstud[!is.na(levrstud$color3), c("obs", "leverage", "rstudent", "color3")]
  d4 <- levrstud[!is.na(levrstud$color4), c("obs", "leverage", "rstudent", "color4")]

  colnames(d1) <- c("obs", "leverage", "rstudent", "color")
  colnames(d2) <- c("obs", "leverage", "rstudent", "color")
  colnames(d3) <- c("obs", "leverage", "rstudent", "color")
  colnames(d4) <- c("obs", "leverage", "rstudent", "color")

  out <- rbind(d1, d2, d3, d4)
  out$fct_color <- ordered(factor(out$color), levels = c("normal", "leverage", "outlier", "outlier & leverage"))
  levdata <- out[order(out$obs), ]

  list(levrstud  = levdata,
       lev_thrsh = lev_thrsh,
       minx      = minx,
       miny      = miny,
       maxx      = maxx,
       maxy      = maxy
  )

}

#' Studentized residual plot data
#'
#' Generates data for studentized residual plot.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 3.
#'
#' @examples
#' model <- lm(read ~ write + math + science, data = hsb)
#' ols_prep_srplot_data(model)
#'
#' @export
#'
ols_prep_srplot_data<- function(model, threshold = NULL) {

  if (is.null(threshold)) {
    threshold <- 3
  }

  dstud <- unname(rstudent(model))
  obs   <- seq_len(length(dstud))
  dsr   <- data.frame(obs = obs, dsr = dstud)
  dsr$color     <- ifelse((abs(dsr$dsr) >= threshold), "outlier", "normal")
  dsr$fct_color <- ordered(factor(dsr$color), levels = c("normal", "outlier"))

  cminxx <- floor(min(dsr$dsr) - 1)
  cmaxxx <- ceiling(max(dsr$dsr) + 1)
  cminx  <- ifelse(cminxx > -threshold, -threshold, cminxx)
  cmaxx  <- ifelse(cmaxxx < threshold, threshold, cmaxxx)
  nseq   <- seq_len(abs(cminx + 1)) * -1
  pseq   <- seq_len(cmaxx - 1)

  list(cminx = cminx,
       cmaxx = cmaxx,
       nseq  = nseq,
       pseq  = pseq,
       dsr   = dsr)

}

#' Standardized residual chart data
#'
#' Generates data for standardized residual chart.
#'
#' @param model An object of class \code{lm}.
#' @param threshold Threshold for detecting outliers. Default is 2.
#'
#' @examples
#' model <- lm(read ~ write + math + science, data = hsb)
#' ols_prep_srchart_data(model)
#' ols_prep_srchart_data(model, threshold = 3)
#'
#' @export
#'
ols_prep_srchart_data <- function(model, threshold = NULL) {

  if (is.null(threshold)) {
    threshold <- 2
  }

  sdres         <- rstandard(model)
  sdres_out     <- abs(sdres) > threshold
  outlier       <- sdres[sdres_out]
  obs           <- seq_len(length(sdres))

  out           <- data.frame(obs = obs, sdres = sdres)
  out$color     <- ifelse(((out$sdres >= threshold) | (out$sdres <= -threshold)), c("outlier"), c("normal"))
  out$fct_color <- ordered( factor(out$color), levels = c("normal", "outlier"))
  out$txt       <- ifelse(out$color == "outlier", out$obs, NA)

  return(out)

}

remove_columns <- function(data, i) {
	as.matrix(data[, c(-1, -i)])
}

select_columns <- function(data, i = 1) {
	as.matrix(data[, i])
}

test_that("output from text formatting matches the expected result", {
  expect_equal(fg(3, 10), "         3", ignore_attr = TRUE)
  expect_equal(fw(3, 10), "         3", ignore_attr = TRUE)
  expect_equal(fs(), "  ", ignore_attr = TRUE)
  expect_equal(fl(3, 10), "3         ", ignore_attr = TRUE)
  expect_equal(fc(3, 10), "    3     ", ignore_attr = TRUE)
  expect_equal(formatter_t(3, 10), "         3", ignore_attr = TRUE)
  expect_equal(formatter_n(3.7589, 10), "  3.7589  ", ignore_attr = TRUE)
  expect_equal(format_cil(3, 10), "  3.000   ", ignore_attr = TRUE)
  expect_equal(format_ciu(3, 10), "  3.000   ", ignore_attr = TRUE)
  expect_equal(formats_t(), "  ", ignore_attr = TRUE)
})

test_that("output from l matches the expected result", {
  expect_equal(l(deparse(substitute(mtcars$mpg))), "mpg")
  expect_equal(l(deparse(substitute(mpg))), "mpg")
  expect_equal(l(deparse(substitute(mtcars@mpg))), "mtcars@mpg")
})

test_that("output from cpout matches the expected result", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- cpout(model)
  expect_equal(k$e, residuals(model), ignore_attr = TRUE)
  expect_equal(k$mc, model$coefficients[-1], ignore_attr = TRUE)
  expect_equal(as.data.frame(k$data), model.frame(model)[-1], ignore_attr = TRUE)
  expect_equal(k$lmc, length(model$coefficients[-1]), ignore_attr = TRUE)
  expect_equal(k$nam, names(model.frame(model)[-1]), ignore_attr = TRUE)
  expect_equal(k$indvar, names(model.frame(model)[1]), ignore_attr = TRUE)
})

test_that("output from cpdata matches the expected result", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- cpout(model)
  actual <- cpdata(mtcars, k$mc, k$e, 2)
  expect_equal(
    actual[[1]],
    c(
      6, 6, 4, 6, 8, 6, 8, 4, 4, 6, 6, 8, 8, 8, 8, 8, 8, 4, 4, 4, 4, 8, 8,
      8, 8, 4, 4, 4, 8, 6, 8, 4
    ), ignore_attr = TRUE
  )
  expect_equal(actual[[2]], c(
    -2.9203498, -2.0330346, -2.3992081,
    0.6259587, 0.3873286, -1.3037619,
    -1.2315502, 1.2850510, 0.3095797, -1.4791933, -2.8791933, 0.9161240,
    0.6330370, -1.2929796, -0.6093035, 0.2660135, 4.4825264, 5.5492117,
    -0.4647590, 5.5258185, -2.8364793, -2.5541073, -3.7860046, -2.1732729,
    2.2677477, -0.4740448, -0.9675111, 3.2790314, -2.2139222, -0.8851877,
    2.2406110, -2.1513153
  ), ignore_attr = TRUE)
})

test_that("output from cdplot matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- ols_prep_cdplot_data(model)
  expect_equal(round(k$maxx, 3), 0.398)
  expect_equal(k$ts, 0.125)
})

test_that("output from dpred matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- ols_prep_dsrvf_data(model)
  expect_equal(round(k$cmaxx, 3), 3.516)
  expect_equal(round(k$cminx, 3), -2.44)
})


test_that("output from ka matches the expected result", {
  expect_equal(round(ka(10, 0.69, 30), 3), -0.326)
})

test_that("output from corrout matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- corrout(model)
  expect_equal(round(k, 3), 0.960)
})

test_that("output from obspred matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- obspred(model)
  expect_equal(mean(k[[1]]), mean(k[[2]]))
})

test_that("output from hadio matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- hadio(model, 1)
  expect_equal(round(mean(k), 3), 0.412)
})

test_that("output from rsdata matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- ols_prep_rfsplot_rsdata(model)
  expect_equal(round(mean(k$x), 3), 0.516)
  expect_equal(round(max(k$y), 3), 5.688)
})

test_that("output from fmdata matches the expected output", {

  skip_on_cran()

  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- ols_prep_rfsplot_fmdata(model)
  expect_equal(round(mean(k$x), 3), 0.517)
  expect_equal(round(max(k$y), 3), 10.635)
})

test_that("output from rvspdata matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- rvspdata(model)
  expect_equal(round(mean(k$predicted), 3), 20.091)
  expect_equal(round(median(k$resid), 3), -0.506)
})

test_that("output from rvsrdata matches the expected output", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- ols_prep_rvsrplot_data(model)
  expect_equal(k$np, 4)
  expect_equal(k$pnames, c("disp", "hp", "wt", "drat"), ignore_attr = TRUE)
  expect_equal(round(mean(k$dat[[1]]), 3), 230.722)
})

test_that("output from srdata matches the expected result", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- ols_prep_srplot_data(model)
  expect_equal(k$cmaxx, 4)
  expect_equal(k$cminx, -3)
  expect_equal(k$nseq, c(-1, -2), ignore_attr = TRUE)
  expect_equal(k$pseq, c(1, 2, 3), ignore_attr = TRUE)
})

test_that("output from histdata matches the expected result", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- histdata(model)
  expect_equal(round(k$maxx, 3), 6.688)
  expect_equal(round(k$minx, 3), -4.508)
})


test_that("output from rstudlev matches the expected result", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  k <- ols_prep_rstudlev_data(model)
  expect_equal(k$lev_thrsh, 0.3125)
  expect_equal(round(k$minx, 3), 0.046)
  expect_equal(round(k$miny, 3), -4.44)
  expect_equal(round(k$maxx, 3), 0.508)
  expect_equal(round(k$maxy, 3), 5.516)
})

test_that("output from corrout matches the expected result", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  expect_equal(round(corrout(model), 3), 0.96)
})

test_that("output from advarx matches the expected output", {
  
  actual   <- ols_prep_regress_x(mtcars, 1)
  expected <- unname(residuals(lm(mpg ~., data = mtcars)))
  
  expect_equal(actual, expected, ignore_attr = TRUE)
})

#' Surgical Unit Data Set
#'
#' A dataset containing data about survival of patients undergoing liver operation.
#'
#' @format A data frame with 54 rows and 9 variables:
#' \describe{
#'   \item{bcs}{blood clotting score}
#'   \item{pindex}{prognostic index}
#'   \item{enzyme_test}{enzyme function test score}
#'   \item{liver_test}{liver function test score}
#'   \item{age}{age, in years}
#'   \item{gender}{indicator variable for gender (0 = male, 1 = female)}
#'   \item{alc_mod}{indicator variable for history of alcohol use (0 = None, 1 = Moderate)}
#'   \item{alc_heavy}{indicator variable for history of alcohol use (0 = None, 1 = Heavy)}
#'   \item{y}{Survival Time}
#' }
#'
#' @source Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
"surgical"

#' Cooks' D chart
#'
#' @description
#' Chart of cook's distance to detect observations that strongly influence
#' fitted values of the model.
#'
#' @param model An object of class \code{lm}.
#' @param type An integer between 1 and 5 selecting one of the 6 methods for computing the threshold.
#' @param threshold Threshold for detecting outliers.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details
#' Cook's distance was introduced by American statistician R Dennis Cook in
#' 1977. It is used to identify influential data points. It depends on both the
#' residual and leverage i.e it takes it account both the \emph{x} value and
#' \emph{y} value of the observation.
#'
#' Steps to compute Cook's distance:
#'
#' \itemize{
#'   \item Delete observations one at a time.
#'   \item Refit the regression model on remaining \eqn{n - 1} observations
#'   \item exmine how much all of the fitted values change when the ith observation is deleted.
#' }
#'
#' A data point having a large cook's d indicates that the data point strongly
#' influences the fitted values. There are several methods/formulas to compute
#' the threshold used for detecting or classifying observations as outliers
#' and we list them below.
#'
#' \itemize{
#'   \item \strong{Type 1} : 4 / n
#'   \item \strong{Type 2} : 4 / (n - k - 1)
#'   \item \strong{Type 3} : ~1
#'   \item \strong{Type 4} : 1 / (n - k - 1)
#'   \item \strong{Type 5} : 3 * mean(Vector of cook's distance values)
#' }
#'
#' where \strong{n} and \strong{k} stand for
#'
#' \itemize{
#'   \item \strong{n}: Number of observations
#'   \item \strong{k}: Number of predictors
#' }
#'
#' @return \code{ols_plot_cooksd_chart} returns  a list containing the
#' following components:
#'
#' \item{outliers}{a \code{data.frame} with observation number and \code{cooks distance} that exceed \code{threshold}}
#' \item{threshold}{\code{threshold} for classifying an observation as an outlier}
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_cooksd_chart(model)
#' ols_plot_cooksd_chart(model, type = 4)
#' ols_plot_cooksd_chart(model, threshold = 0.2)
#'
#' @importFrom ggplot2 geom_linerange
#'
#' @seealso [ols_plot_cooksd_bar()]
#'
#' @export
#'
ols_plot_cooksd_chart <- function(model, type = 1, threshold = NULL, print_plot = TRUE) {

  check_model(model)

  k <- ols_prep_cdplot_data(model, type)
  d <- ols_prep_outlier_obs(k)
  f <- ols_prep_cdplot_outliers(k)

  if (is.null(threshold)) {
      threshold <- k$ts
  }

  p <- 
    ggplot(d, aes(x = obs, y = cd, label = txt, ymin = min(cd), ymax = cd)) +
    geom_linerange(colour = "blue") + 
    geom_point(shape = 1, colour = "blue") +
    geom_text(vjust = -1, size = 3, family = "serif", fontface = "italic",
              colour = "darkred", na.rm = TRUE) +
    geom_hline(yintercept = threshold, colour = "red") 

  # annotations
  p <- 
    p +
    annotate("text", x = Inf, y = Inf, hjust = 1.2, vjust = 2,
      family = "serif", fontface = "italic", colour = "darkred",
      label = paste("Threshold:", round(threshold, 3)))

  # guides
  p <-
    p +
    xlab("Observation") +
    ylab("Cook's D") + 
    ggtitle("Cook's D Chart") 
    

  if (print_plot) {
    suppressWarnings(print(p))
  } else {
    return(list(plot = p, outliers = f, threshold = threshold))
  }

}

print_reg <- function(data) {

  # width
  # prints
  a <- c("R", "R-Squared", "Adj. R-Squared", "Pred R-Squared", "MAE")
  b <- c(
    format(round(data$r, 3), nsmall = 3), format(round(data$rsq, 3), nsmall = 3), format(round(data$adjr, 3), nsmall = 3), 
    format(round(data$prsq, 3), nsmall = 3), format(round(data$mae, 3), nsmall = 3)
  )
  d <- c("RMSE", "MSE", "Coef. Var", "AIC", "SBC")
  e <- c(
    format(round(data$rmse, 3), nsmall = 3), format(round(data$mse, 3), nsmall = 3), format(round(data$cv, 3), nsmall = 3),
    format(round(data$aic, 3), nsmall = 3), format(round(data$sbc, 3), nsmall = 3)
  )

  w1 <- max(nchar(a))
  w2 <- max(nchar(format(b, nsmall = 3)))
  w3 <- max(nchar(d))
  w4 <- max(nchar(format(e, nsmall = 3)))
  w5 <- sum(w1, w2, w3, w4, 28)
  nw <- length(b)

  # model summary
  cat(fc("Model Summary", w5), "\n")
  cat(rep("-", w5), sep = "", "\n")
  for (i in seq_len(nw)) {
    cat(fl(a[i], w1), fs(), fs(), fs(), fg(b[i], w2), fs(), fs(), fl(d[i], w3), fs(), fs(), fs(), fg(e[i], w4), "\n")
  }
  cat(rep("-", w5), sep = "", "\n")
  cat(" RMSE: Root Mean Square Error", "\n", "MSE: Mean Square Error", "\n", "MAE: Mean Absolute Error", "\n", "AIC: Akaike Information Criteria", "\n", "SBC: Schwarz Bayesian Criteria", "\n\n")

  # anova
  w7 <- nchar("Regression")
  w8 <- max(
    nchar("Squares"), nchar(format(round(data$rss, 3), nsmall = 3)), nchar(format(round(data$ess, 3), nsmall = 3)),
    nchar(format(round(data$tss, 3), nsmall = 3))
  )
  w9 <- max(
    nchar("DF"), nchar(format(round(data$model_df, 3), nsmall = 3)), nchar(format(round(data$error_df, 3), nsmall = 3)),
    nchar(format(round(data$total_df, 3), nsmall = 3))
  )
  w10 <- max(nchar("Mean Square"), nchar(format(round(data$rms, 3), nsmall = 3)), nchar(format(round(data$ems, 3), nsmall = 3)))
  w11 <- max(nchar("F"), nchar(format(round(data$f, 3), nsmall = 3)))
  w12 <- max(nchar("Sig."), nchar(format(round(data$p, 3), nsmall = 4)))
  w <- sum(w7, w8, w9, w10, w11, w12, 21)

  p <- format(round(data$p, 4), nsmall = 4)

  # ANOVA
  cat(fc("ANOVA", w), "\n")
  cat(rep("-", w), sep = "", "\n")
  cat(fg("", w7), fs(), fg("Sum of", w8), fs(), fg("", w9), fs(), fg("", w10), fs(), fg("", w11), fs(), fg("", w12), "\n")
  cat(fg("", w7), fs(), fg("Squares", w8), fs(), fg("DF", w9), fs(), fg("Mean Square", w10), fs(), fc("F", w11), fs(), fg("Sig.", w12), "\n")
  cat(rep("-", w), sep = "", "\n")
  cat(
    fl("Regression", w7), fs(), fg(format(round(data$rss, 3), nsmall = 3), w8), fs(), fg(round(data$model_df, 3), w9), fs(),
    fg(format(round(data$rms, 3), nsmall = 3), w10), fs(), fg(round(data$f, 3), w11), fs(), fg(p, w12), "\n"
  )
  cat(
    fl("Residual", w7), fs(), fg(format(round(data$ess, 3), nsmall = 3), w8), fs(), fg(round(data$error_df, 3), w9), fs(),
    fg(format(round(data$ems, 3), nsmall = 3), w10), fs(), fg("", w11), fs(), fg("", w12), "\n"
  )
  cat(
    fl("Total", w7), fs(), fg(format(round(data$tss, 3), nsmall = 3), w8), fs(), fg(round(data$total_df, 3), w9), fs(), fg("", w10), fs(),
    fg("", w11), fs(), fg("", w12), "\n"
  )
  cat(rep("-", w), sep = "", "\n\n")

  # coefficients
  w13 <- max(nchar(data$title), nchar(data$mvars))
  w14 <- max(nchar("Beta"), nchar(format(round(data$betas, 3), nsmall = 3)))
  w15 <- max(nchar("Std. Error"), nchar(format(round(data$std_errors, 3), nsmall = 3)))
  w16 <- max(nchar("Std. Beta"), nchar(format(round(data$sbetas, 3), nsmall = 3)))
  w17 <- max(nchar("t"), nchar(format(round(data$tvalues, 3), nsmall = 3)))
  w18 <- max(nchar("Sig."), nchar(format(round(data$pvalues, 3), nsmall = 3)))
  w19 <- max(nchar("lower"), nchar(format(round(data$conf_lm[, 1], 3), nsmall = 3)))
  w20 <- max(nchar("upper"), nchar(format(round(data$conf_lm[, 2], 3), nsmall = 3)))
  w21 <- sum(w13, w14, w15, w16, w17, w18, w19, w20, 29)

  k <- length(data$mvars)

  sb <- c("", format(round(data$sbetas, 3), nsmall = 3))

  cat(fc("Parameter Estimates", w21), "\n")
  cat(rep("-", w21), sep = "", "\n")
  cat(
    fg(data$title, w13), fs(), fg("Beta", w14), fs(), fg("Std. Error", w15), fs(), fg("Std. Beta", w16), fs(),
    fc("t", w17), fs(), fc("Sig", w18), fs(), fg("lower", w19), fs(), fg("upper", w20), "\n"
  )
  cat(rep("-", w21), sep = "", "\n")
  for (i in seq_len(k)) {
    cat(
      fg(data$mvars[i], w13), fs(), fg(format(round(data$betas[i], 3), nsmall = 3), w14),
      fs(), fg(format(round(data$std_errors[i], 3), nsmall = 3), w15), fs(), fg(sb[i], w16),
      fs(), fg(format(round(data$tvalues[i], 3), nsmall = 3), w17), fs(),
      fg(format(round(data$pvalues[i], 3), nsmall = 3), w18), fs(),
      fg(as.vector(format(round(data$conf_lm[, 1], 3), nsmall = 3))[i], w19), fs(),
      fg(as.vector(format(round(data$conf_lm[, 2], 3), nsmall = 3))[i], w20), "\n"
    )
  }
  cat(rep("-", w21), sep = "", "\n")
}


print_correlations <- function(data) {

  # number of rows
  nr <- nrow(data)
  vars <- rownames(data)
  cols <- colnames(data)

  # widths
  w1 <- max(nchar("Variable"), nchar(vars))
  w2 <- nchar(cols[1])
  w3 <- nchar(cols[2])
  w4 <- max(nchar(cols[3]), nchar(format(round(data[, 3], 3), nsmall = 3)))
  w <- sum(w1, w2, w3, w4, 12)

  # print
  cat(fc("Correlations", w), "\n")
  cat(rep("-", w), sep = "", "\n")
  cat(fl("Variable", w1), fs(), fc("Zero Order", w2), fs(), fc("Partial", w3), fs(), fc("Part", w4), "\n")
  cat(rep("-", w), sep = "", "\n")
  for (i in seq_len(nr)) {
    cat(
      fl(vars[i], w1), fs(), fg(format(round(data[i, 1], 3), nsmall = 3), w2), fs(),
      fg(format(round(data[i, 2], 3), nsmall = 3), w3), fs(),
      fg(format(round(data[i, 3], 3), nsmall = 3), w4), "\n"
    )
  }
  cat(rep("-", w), sep = "", "\n")
}


print_ftest <- function(data) {

  # width
  w1 <- max(
    nchar(data$numdf), nchar(data$dendf),
    nchar(format(data$f, nsmall = 3)),
    nchar(format(data$p, nsmall = 3))
  )

  w <- w1 + 16

  cat("\n F Test for Heteroskedasticity\n")
  cat(" ", rep("-", 29), sep = "", "\n")
  cat(" Ho: Variance is homogenous\n", "Ha: Variance is not homogenous\n\n")


  if (data$fv == TRUE) {
    cat(" Variables: fitted values of", data$resp, "\n\n")
  } else if (data$rhs == TRUE) {
    cat(" Variables:", data$preds, "\n\n")
  } else {
    cat(" Variables:", data$vars, "\n\n")
  }

  cat(format("Test Summary", width = w, justify = "centre"), "\n")
  cat(" ", rep("-", w), sep = "", "\n")
  cat(
    " Num DF     =   ", data$numdf, "\n",
    "Den DF     =   ", data$dendf, "\n",
    "F          =   ", format(data$f, nsmall = 3), "\n",
    "Prob > F   =   ", format(data$p, nsmall = 3), "\n"
  )
}


print_bp_test <- function(data) {
  cat("\n", "Breusch Pagan Test for Heteroskedasticity\n")
  cat(
    " ", rep("-", 41), sep = "", "\n",
    format(" Ho: the variance is constant", width = 41, justify = "left"), "\n",
    format(" Ha: the variance is not constant", width = 41, justify = "left"), "\n\n"
  )


  if (data$fv) {
    if (data$rhs) {
      if (data$multiple) {
        w1 <- max(nchar(data$preds), 13)
        w2 <- max(nchar("chi2"), nchar(format(data$bp, nsmall = 4)))
        w3 <- max(nchar(length(data$preds)), nchar("df"))
        w4 <- max(nchar(format(data$p, nsmall = 4)))
        w <- sum(w1, w2, w3, w4, 13)

        ldp <- length(data$preds)
        ldp2 <- ldp + 1

        w5 <- 11 + sum(nchar(data$preds)) + ldp - 1

        cat(format("Data", width = w5, justify = "centre"), "\n")
        cat(" ", rep("-", w5), sep = "", "\n")
        cat(
          " Response :", data$resp, "\n",
          "Variables:", data$preds, "\n\n"
        )


        if (data$padj == "bonferroni") {
          cat(" ", format("Test Summary (Bonferroni p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        } else if (data$padj == "sidak") {
          cat(" ", format("Test Summary (Sidak p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        } else if (data$padj == "holm") {
          cat(" ", format("Test Summary (Holm's p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        } else {
          cat(" ", format("Test Summary (Unadjusted p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        }

        bp <- format(data$bp, nsmall = 4)
        p <- format(data$p, nsmall = 4)

        cat(
          " ", format("Variable", width = w1, justify = "left"), fs(),
          format("chi2", width = w2, justify = "centre"), fs(),
          format("df", width = w3, justify = "centre"), fs(),
          format("p", width = w4, justify = "centre"), "\n"
        )
        cat(" ", rep("-", w), sep = "", "\n")
        for (i in seq_len(ldp)) {
          cat(
            " ", format(data$preds[i], width = w1, justify = "left"), fs(),
            format(bp[i], width = w2, justify = "centre"), fs(),
            format("1", width = w3, justify = "right"), fs(),
            format(p[i], width = w4, justify = "centre"), "\n"
          )
        }
        cat(" ", rep("-", w), sep = "", "\n")
        cat(
          " ", format("simultaneous", width = w1, justify = "left"), fs(),
          format(bp[ldp2], width = w2, justify = "centre"), fs(),
          format(ldp, width = w3, justify = "right"), fs(),
          format(p[ldp2], width = w4, justify = "centre"), "\n"
        )
        cat(" ", rep("-", w), sep = "", "\n")
      } else {
        ldp <- length(data$preds)

        w1 <- 11 + sum(nchar(data$preds)) + ldp - 1

        w3 <- max(
          nchar(ldp),
          nchar(format(data$bp, nsmall = 4)),
          nchar(format(data$p, nsmall = 4))
        )

        w <- w3 + 19


        cat(format("Data", width = w1, justify = "centre"), "\n")
        cat(" ", rep("-", w1), sep = "", "\n")
        cat(
          " Response :", data$resp, "\n",
          "Variables:", data$preds, "\n\n"
        )

        cat(format("Test Summary", width = w, justify = "centre"), "\n")
        cat(" ", rep("-", w), sep = "", "\n")
        cat(
          " DF            =   ", ldp, "\n",
          "Chi2          =   ", format(data$bp, nsmall = 4), "\n",
          "Prob > Chi2   =   ", format(data$p, nsmall = 4), "\n"
        )
      }
    } else {
      w1 <- nchar(data$resp)
      w2 <- w1 + 28

      w3 <- max(
        nchar("1"),
        nchar(format(data$bp, nsmall = 4)),
        nchar(format(data$p, nsmall = 4))
      )

      w <- w3 + 19


      cat(format("Data", width = w2, justify = "centre"), "\n")
      cat(" ", rep("-", w2), sep = "", "\n")
      cat(
        " Response :", data$resp, "\n",
        "Variables: fitted values of", data$resp, "\n\n"
      )

      cat(format("Test Summary", width = w, justify = "centre"), "\n")
      cat(" ", rep("-", w), sep = "", "\n")
      cat(
        " DF            =   ", 1, "\n",
        "Chi2          =   ", format(data$bp, nsmall = 4), "\n",
        "Prob > Chi2   =   ", format(data$p, nsmall = 4), "\n"
      )
    }
  } else {
    if (data$multiple) {
      if (data$rhs) {
        w1 <- max(nchar(data$preds), 13)
        w2 <- max(nchar("chi2"), nchar(format(data$bp, nsmall = 4)))
        w3 <- max(nchar(length(data$preds)), nchar("df"))
        w4 <- max(nchar(format(data$p, nsmall = 4)))
        w <- sum(w1, w2, w3, w4, 13)

        ldp <- length(data$preds)
        ldp2 <- ldp + 1

        w5 <- 11 + sum(nchar(data$preds)) + ldp - 1

        cat(format("Data", width = w5, justify = "centre"), "\n")
        cat(" ", rep("-", w5), sep = "", "\n")
        cat(
          " Response :", data$resp, "\n",
          "Variables:", data$preds, "\n\n"
        )


        if (data$padj == "bonferroni") {
          cat(" ", format("Test Summary (Bonferroni p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        } else if (data$padj == "sidak") {
          cat(" ", format("Test Summary (Sidak p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        } else if (data$padj == "holm") {
          cat(" ", format("Test Summary (Holm's p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        } else {
          cat(" ", format("Test Summary (Unadjusted p values)", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
        }

        bp <- format(data$bp, nsmall = 4)
        p <- format(data$p, nsmall = 4)

        cat(
          " ", format("Variable", width = w1, justify = "left"), fs(),
          format("chi2", width = w2, justify = "centre"), fs(),
          format("df", width = w3, justify = "centre"), fs(),
          format("p", width = w4, justify = "centre"), "\n"
        )
        cat(" ", rep("-", w), sep = "", "\n")
        for (i in seq_len(ldp)) {
          cat(
            " ", format(data$preds[i], width = w1, justify = "left"), fs(),
            format(bp[i], width = w2, justify = "centre"), fs(),
            format("1", width = w3, justify = "right"), fs(),
            format(p[i], width = w4, justify = "centre"), "\n"
          )
        }
        cat(" ", rep("-", w), sep = "", "\n")
        cat(
          " ", format("simultaneous", width = w1, justify = "left"), fs(),
          format(bp[ldp2], width = w2, justify = "centre"), fs(),
          format(ldp, width = w3, justify = "right"), fs(),
          format(p[ldp2], width = w4, justify = "centre"), "\n"
        )
        cat(" ", rep("-", w), sep = "", "\n")
      } else {
        if (length(data$vars) > 1) {
          w1 <- max(nchar(data$vars), 13)
          w2 <- max(nchar("chi2"), nchar(format(data$bp, nsmall = 4)))
          w3 <- max(nchar(length(data$vars)), nchar("df"))
          w4 <- max(nchar(format(data$p, nsmall = 4)))
          w <- sum(w1, w2, w3, w4, 13)

          ldp <- length(data$vars)
          ldp2 <- ldp + 1

          w5 <- 11 + sum(nchar(data$vars)) + ldp - 1

          cat(format("Data", width = w5, justify = "centre"), "\n")
          cat(" ", rep("-", w5), sep = "", "\n")
          cat(
            " Response :", data$resp, "\n",
            "Variables:", data$vars, "\n\n"
          )


          if (data$padj == "bonferroni") {
            cat(" ", format("Test Summary (Bonferroni p values)", width = w, justify = "centre"), "\n")
            cat(" ", rep("-", w), sep = "", "\n")
          } else if (data$padj == "sidak") {
            cat(" ", format("Test Summary (Sidak p values)", width = w, justify = "centre"), "\n")
            cat(" ", rep("-", w), sep = "", "\n")
          } else if (data$padj == "holm") {
            cat(" ", format("Test Summary (Holm's p values)", width = w, justify = "centre"), "\n")
            cat(" ", rep("-", w), sep = "", "\n")
          } else {
            cat(" ", format("Test Summary (Unadjusted p values)", width = w, justify = "centre"), "\n")
            cat(" ", rep("-", w), sep = "", "\n")
          }

          bp <- format(data$bp, nsmall = 4)
          p <- format(data$p, nsmall = 4)

          cat(
            " ", format("Variable", width = w1, justify = "left"), fs(),
            format("chi2", width = w2, justify = "centre"), fs(),
            format("df", width = w3, justify = "centre"), fs(),
            format("p", width = w4, justify = "centre"), "\n"
          )
          cat(" ", rep("-", w), sep = "", "\n")
          for (i in seq_len(ldp)) {
            cat(
              " ", format(data$vars[i], width = w1, justify = "left"), fs(),
              format(bp[i], width = w2, justify = "centre"), fs(),
              format("1", width = w3, justify = "right"), fs(),
              format(p[i], width = w4, justify = "centre"), "\n"
            )
          }
          cat(" ", rep("-", w), sep = "", "\n")
          cat(
            " ", format("simultaneous", width = w1, justify = "left"), fs(),
            format(bp[ldp2], width = w2, justify = "centre"), fs(),
            format(ldp, width = w3, justify = "right"), fs(),
            format(p[ldp2], width = w4, justify = "centre"), "\n"
          )
          cat(" ", rep("-", w), sep = "", "\n")
        } else {
          ldp <- length(data$vars)

          w1 <- 11 + sum(nchar(data$vars)) + ldp - 1

          w3 <- max(
            nchar(ldp),
            nchar(format(data$bp, nsmall = 4)),
            nchar(format(data$p, nsmall = 4))
          )

          w <- w3 + 19


          cat(format("Data", width = w1, justify = "centre"), "\n")
          cat(" ", rep("-", w1), sep = "", "\n")
          cat(
            " Response :", data$resp, "\n",
            "Variables:", data$vars, "\n\n"
          )

          cat(format("Test Summary", width = w, justify = "centre"), "\n")
          cat(" ", rep("-", w), sep = "", "\n")
          cat(
            " DF            =   ", ldp, "\n",
            "Chi2          =   ", format(data$bp, nsmall = 4), "\n",
            "Prob > Chi2   =   ", format(data$p, nsmall = 4), "\n"
          )
        }
      }
    } else {
      if (data$rhs) {
        ldp <- length(data$preds)

        w1 <- 11 + sum(nchar(data$preds)) + ldp - 1

        w3 <- max(
          nchar(ldp),
          nchar(format(data$bp, nsmall = 4)),
          nchar(format(data$p, nsmall = 4))
        )

        w <- w3 + 19


        cat(format("Data", width = w1, justify = "centre"), "\n")
        cat(" ", rep("-", w1), sep = "", "\n")
        cat(
          " Response :", data$resp, "\n",
          "Variables:", data$preds, "\n\n"
        )

        cat(format("Test Summary", width = w, justify = "centre"), "\n")
        cat(" ", rep("-", w), sep = "", "\n")
        cat(
          " DF            =   ", ldp, "\n",
          "Chi2          =   ", format(data$bp, nsmall = 4), "\n",
          "Prob > Chi2   =   ", format(data$p, nsmall = 4), "\n"
        )
      } else {
        lvars <- length(data$vars)

        w1 <- 11 + sum(nchar(data$vars)) + lvars - 1

        w3 <- max(
          nchar(lvars),
          nchar(format(data$bp, nsmall = 4)),
          nchar(format(data$p, nsmall = 4))
        )

        w <- w3 + 19


        cat(format("Data", width = w1, justify = "centre"), "\n")
        cat(" ", rep("-", w1), sep = "", "\n")
        cat(
          " Response :", data$resp, "\n",
          "Variables:", data$vars, "\n\n"
        )

        cat(format("Test Summary", width = w, justify = "centre"), "\n")
        cat(" ", rep("-", w), sep = "", "\n")
        cat(
          " DF            =   ", lvars, "\n",
          "Chi2          =   ", format(data$bp, nsmall = 4), "\n",
          "Prob > Chi2   =   ", format(data$p, nsmall = 4), "\n"
        )
      }
    }
  }
}


print_bartlett_test <- function(data) {

  # width
  w1 <- max(
    nchar(data$df),
    nchar(format(data$fstat, nsmall = 3)),
    nchar(format(data$pval, nsmall = 3))
  )

  w <- w1 + 19

  # variable names
  if (is.null(data$g_var)) {
    ln <- length(data$var_c)
    w2 <- sum(nchar(unlist(lapply(data$var_c, l))))
    w3 <- w2 + ln + 10
  } else {
    w2 <- max(nchar(l(data$var_c)), nchar(l(data$g_var)))
    w3 <- w2 + 19
  }



  cat("\n", format(
    " Bartlett's Test of Homogenity of Variances",
    width = 48, justify = "centre"
  ), "\n")
  cat(
    rep("-", 48), sep = "", "\n",
    "Ho: Variances are equal across groups\n",
    "Ha: Variances are unequal for atleast two groups\n\n"
  )

  if (is.null(data$g_var)) {
    cat(format("Data", width = w3, justify = "centre"), "\n")
    cat(" ", rep("-", w3), sep = "", "\n")
    cat(" Variables:", unlist(lapply(data$var_c, l)), "\n\n")
  }

  cat(format("Test Summary", width = w, justify = "centre"), "\n")
  cat(" ", rep("-", w), sep = "", "\n")
  cat(
    " DF            =   ", data$df, "\n",
    "Chi2          =   ", format(data$fstat, nsmall = 3), "\n",
    "Prob > Chi2   =   ", format(data$pval, nsmall = 3), "\n"
  )
}


print_score_test <- function(data) {

  # width
  w1 <- max(
    nchar(data$df),
    nchar(format(data$score, nsmall = 3)),
    nchar(format(data$p, nsmall = 3))
  )

  w <- w1 + 19

  cat("\n Score Test for Heteroskedasticity\n")
  cat(" ", rep("-", 33), sep = "", "\n")
  cat(" Ho: Variance is homogenous\n", "Ha: Variance is not homogenous\n\n")


  if (data$rhs == TRUE) {
    cat(" Variables:", data$preds, "\n\n")
  } else if (data$fv == TRUE) {
    cat(" Variables: fitted values of", data$resp, "\n\n")
  } else {
    cat(" Variables:", data$preds, "\n\n")
  }

  cat(format("Test Summary", width = w, justify = "centre"), "\n")
  cat(" ", rep("-", w), sep = "", "\n")
  cat(
    " DF            =   ", data$df, "\n",
    "Chi2          =   ", format(data$score, nsmall = 3), "\n",
    "Prob > Chi2   =   ", format(data$p, nsmall = 3), "\n"
  )
}

print_best_subset <- function(data) {
  w1 <- 11
  w2 <- max(nchar(data$metrics$predictors))
  w <- sum(w1, w2, 4)

  w3 <- nchar("Model")
  w4 <- nchar("R-Square")
  w5 <- max(nchar("Pred"), nchar(format(round(data$metrics$predrsq, 4), nsmall = 4)))
  w6 <- max(nchar("C(p)"), nchar(format(round(data$metrics$cp, 4), nsmall = 4)))
  w7 <- max(nchar("AIC"), nchar(format(round(data$metrics$aic, 4), nsmall = 4)))
  w8 <- max(nchar("SBIC"), nchar(format(round(data$metrics$sbic, 4), nsmall = 4)))
  w9 <- max(nchar("SBC"), nchar(format(round(data$metrics$sbc, 4), nsmall = 4)))
  w10 <- max(nchar("MSEP"), nchar(format(round(data$metrics$msep, 4), nsmall = 4)))
  w11 <- max(nchar("FPE"), nchar(format(round(data$metrics$fpe, 4), nsmall = 4)))
  w12 <- max(nchar("HSP"), nchar(format(round(data$metrics$hsp, 4), nsmall = 4)))
  w13 <- max(nchar("APC"), nchar(format(round(data$metrics$apc, 4), nsmall = 4)))

  v <- sum(w3, w4, w4, w4, w6, w7, w8, w9, w10, w11, w12, w13, 44)

  cat(format("Best Subsets Regression", width = w, justify = "centre"))
  cat("\n", rep("-", w), sep = "", "\n")
  cat("Model Index    Predictors")
  cat("\n", rep("-", w), sep = "", "\n")
  for (i in data$metrics$mindex) {
    cat(
      format(as.character(data$metrics$mindex[i]), width = w1, justify = "centre"), fs(),
      format(data$metrics$predictors[i], width = w2), "\n"
    )
  }
  cat(rep("-", w), sep = "", "\n\n")

  cat(format("Subsets Regression Summary", width = v, justify = "centre"))
  cat("\n", rep("-", v), sep = "", "\n")
  cat(
    format("", width = w3, justify = "centre"), fs(), 
    format("", width = w4, justify = "centre"), fs(),
    format("Adj.", width = w4, justify = "centre"), fs(), 
    format("Pred", width = w4, justify = "centre"), fs(),
    format("", width = w6, justify = "centre"), fs(), 
    format("", width = w7, justify = "centre"), fs(),
    format("", width = w8, justify = "centre"), fs(), 
    format("", width = w9, justify = "centre"), fs(),
    format("", width = w10, justify = "centre"), fs(), 
    format("", width = w11, justify = "centre"), fs(),
    format("", width = w12, justify = "centre"), fs(), 
    format("", width = w13, justify = "centre"), "\n"
  )
  cat(
    format("Model", width = w3, justify = "centre"), fs(), 
    format("R-Square", width = w4, justify = "centre"), fs(),
    format("R-Square", width = w4, justify = "centre"), fs(), 
    format("R-Square", width = w4, justify = "centre"), fs(),
    format("C(p)", width = w6, justify = "centre"), fs(), 
    format("AIC", width = w7, justify = "centre"), fs(),
    format("SBIC", width = w8, justify = "centre"), fs(), 
    format("SBC", width = w9, justify = "centre"), fs(),
    format("MSEP", width = w10, justify = "centre"), fs(), 
    format("FPE", width = w11, justify = "centre"), fs(),
    format("HSP", width = w12, justify = "centre"), fs(), 
    format("APC", width = w13, justify = "centre")
  )
  cat("\n", rep("-", v), sep = "", "\n")
  for (i in data$metrics$mindex) {
    cat(
      format(as.character(data$metrics$mindex[i]), width = w3, justify = "centre"), fs(),
      format(round(data$metrics$rsquare[i], 4), nsmall = 4, width = w4, justify = "centre"), fs(),
      format(round(data$metrics$adjr[i], 4), nsmall = 4, width = w4, justify = "centre"), fs(),
      format(round(data$metrics$predrsq[i], 4), width = w4, justify = "centre"), fs(),
      format(round(data$metrics$cp[i], 4), nsmall = 4, width = w6, justify = "centre"), fs(),
      format(round(data$metrics$aic[i], 4), nsmall = 4, width = w7, justify = "centre"), fs(),
      format(round(data$metrics$sbic[i], 4), nsmall = 4, width = w8, justify = "centre"), fs(),
      format(round(data$metrics$sbc[i], 4), nsmall = 4, width = w9, justify = "centre"), fs(),
      format(round(data$metrics$msep[i], 4), nsmall = 4, width = w10, justify = "centre"), fs(),
      format(round(data$metrics$fpe[i], 4), nsmall = 4, width = w11, justify = "centre"), fs(),
      format(round(data$metrics$hsp[i], 4), nsmall = 4, width = w12, justify = "centre"), fs(),
      format(round(data$metrics$apc[i], 4), nsmall = 4, width = w13, justify = "centre"), "\n"
    )
  }
  cat(rep("-", v), sep = "", "\n")
  cat(
    "AIC: Akaike Information Criteria", "\n", "SBIC: Sawa's Bayesian Information Criteria", "\n",
    "SBC: Schwarz Bayesian Criteria", "\n", "MSEP: Estimated error of prediction, assuming multivariate normality",
    "\n", "FPE: Final Prediction Error", "\n", "HSP: Hocking's Sp", "\n", "APC: Amemiya Prediction Criteria",
    "\n\n"
  )
}

print_norm_test <- function(data) {

  # width
  w1 <- 18
  w2 <- 14
  w3 <- 7
  w <- sum(w1, w2, w3, 8)

  # vectors
  tests <- c(
    "Shapiro-Wilk", "Kolmogorov-Smirnov", "Cramer-von Mises",
    "Anderson-Darling"
  )
  stats <- c(
    data$shapiro$statistic, data$kolmogorv$statistic,
    data$cramer$statistic, data$anderson$statistic
  )
  pvals <- c(
    data$shapiro$p.value, data$kolmogorv$p.value,
    data$cramer$p.value, data$anderson$p.value
  )
  n <- length(stats)

  # print
  cat(rep("-", w), sep = "", "\n")
  cat(
    format("Test", width = w1, justify = "centre"), fs(), format("Statistic", width = w2, justify = "centre"),
    fs(), format("pvalue", width = 7, justify = "centre"), "\n"
  )
  cat(rep("-", w), sep = "", "\n")
  for (i in seq_len(n)) {
    cat(
      format(tests[i], width = w1), fs(), format(as.character(round(stats[i], 4)), width = w2, justify = "centre"),
      fs(), format(round(pvals[i], 4), nsmall = 4, width = 7, justify = "centre"), "\n"
    )
  }
  cat(rep("-", w), sep = "", "\n")
}



print_pure_error_anova <- function(data) {
  wt <- max(nchar(data$resp), nchar(data$preds)) + 13

  cat(format("Lack of Fit F Test", width = wt, justify = "centre"), "\n")
  cat(rep("-", wt), sep = "", "\n")
  cat("Response :  ", data$resp, "\n")
  cat("Predictor:  ", data$preds, "\n\n")

  # widths
  w1 <- max(nchar(data$resp), 12)
  w2 <- max(nchar("DF"), nchar(data$df_rss), nchar(data$df_ess), nchar(data$df_lof), nchar(data$df_error))
  w3 <- max(
    nchar("Sum Sq"), nchar(format(data$rss, nsmall = 2)), nchar(format(data$ess, nsmall = 2)),
    nchar(format(data$lackoffit, nsmall = 2)), nchar(format(data$pure_error, nsmall = 2))
  )
  w4 <- max(
    nchar("Mean Sq"), nchar(format(data$rms, nsmall = 2)), nchar(format(data$ems, nsmall = 2)),
    nchar(format(data$lms, nsmall = 2)), nchar(format(data$pms, nsmall = 2))
  )
  w5 <- max(nchar("F Value"), nchar(format(data$rf, nsmall = 2)), nchar(format(data$lf, nsmall = 2)))
  w6 <- max(nchar("Pr(>F)"), nchar(format(data$pr, nsmall = 2)), nchar(format(data$pl, nsmall = 2)))
  w <- sum(w1, w2, w3, w4, w5, w6, 20)

  cat(format("Analysis of Variance Table", width = w, justify = "centre"), "\n")
  cat(rep("-", w), sep = "", "\n")
  cat(
    fc("", w = w1), fs(), fc("DF", w = w2), fs(), fc("Sum Sq", w = w3), fs(), fc("Mean Sq", w = w4),
    fs(), fc("F Value", w = w5), fs(), fc("Pr(>F)", w = w6), "\n"
  )
  cat(rep("-", w), sep = "", "\n")
  cat(
    fl(data$preds, w = w1), fs(), fg(data$df_rss, w = w2), fs(), format(data$rss, nsmall = 2, width = w3), fs(),
    format(data$rms, nsmall = 2, width = w4), fs(), format(data$rf, nsmall = 2, width = w5), fs(),
    format(data$pr, nsmall = 2, width = w6), "\n"
  )
  cat(
    fl("Residual", w = w1), fs(), fg(data$df_ess, w = w2), fs(), format(data$ess, nsmall = 2, width = w3), fs(),
    format(data$ems, nsmall = 2, width = w4), fs(), format("", nsmall = 2, width = w5), fs(),
    format("", nsmall = 2, width = w6), "\n"
  )
  cat(
    fl(" Lack of fit", w = w1), fs(), fg(data$df_lof, w = w2), fs(), format(data$lackoffit, nsmall = 2, width = w3), fs(),
    format(data$lms, nsmall = 2, width = w4), fs(), format(data$lf, nsmall = 2, width = w5), fs(),
    format(data$pl, nsmall = 2, width = w6), "\n"
  )
  cat(
    fl(" Pure Error", w = w1), fs(), fg(data$df_error, w = w2), fs(), format(data$pure_error, nsmall = 2, width = w3), fs(),
    format(data$pms, nsmall = 2, width = w4), fs(), format("", nsmall = 2, width = w5), fs(),
    format("", nsmall = 2, width = w6), "\n"
  )
  cat(rep("-", w), sep = "", "\n")
}

print_step_rsquared <- function(data) {

  if (data$others$direction == "forward") {
    print_step_output(data, "forward")
  } else if (data$others$direction == "backward") {
    print_step_output(data, "backward")
  } else {
    print_step_output(data, "both")
  }

}

print_model_perf <- function(data) {
  a <- c("R", "R-Squared", "Adj. R-Squared", "Pred R-Squared")
  b <- c(
    format(round(data$r, 3), nsmall = 3),
    format(round(data$rsq, 3), nsmall = 3),
    format(round(data$adjr, 3), nsmall = 3),
    format(round(data$prsq, 3), nsmall = 3)
  )
  d <- c("AIC", "SBC", "SBIC", "RMSE")
  e <- c(
    format(round(data$aic, 3), nsmall = 3),
    format(round(data$sbc, 3), nsmall = 3),
    format(round(data$sbic, 3), nsmall = 3),
    format(round(data$rmse, 3), nsmall = 3)
  )

  w1 <- max(nchar(a))
  w2 <- max(nchar(format(b, nsmall = 3)))
  w3 <- max(nchar(d))
  w4 <- max(nchar(format(e, nsmall = 3)))
  w5 <- sum(w1, w2, w3, w4, 15)
  nw <- length(b)

  # model summary
  cat(fc("Model Performance", w5), "\n")
  cat(rep("-", w5), sep = "", "\n")
  for (i in seq_len(nw)) {
    cat(fl(a[i], w1), fs(), fg(b[i], w2), fs(), fs(), fl(d[i], w3),
      fs(), fg(e[i], w4), "\n")
  }
  cat(rep("-", w5), sep = "", "\n")
  cat("  AIC: Akaike Information Criteria", "\n",
    " SBC: Schwarz Bayesian Criteria", "\n",
    "SBIC: Sawa's Bayesian Criteria", "\n",
    "RMSE: Root Mean Square Error", "\n\n")

}

print_anova <- function(data) {
  w7 <- nchar("Regression")

  w8 <- max(
    nchar("Squares"), nchar(format(round(data$rss, 3), nsmall = 3)),
    nchar(format(round(data$ess, 3), nsmall = 3)),
    nchar(format(round(data$tss, 3), nsmall = 3))
  )

  w9 <- max(
    nchar("DF"), nchar(format(round(data$model_df, 3), nsmall = 3)),
    nchar(format(round(data$error_df, 3), nsmall = 3)),
    nchar(format(round(data$total_df, 3), nsmall = 3))
  )

  w10 <- max(
    nchar("Mean Square"),
    nchar(format(round(data$rms, 3), nsmall = 3)),
    nchar(format(round(data$ems, 3), nsmall = 3))
  )

  w11 <- max(nchar("F"), nchar(format(round(data$f, 3), nsmall = 3)))
  w12 <- max(nchar("Sig."), nchar(format(round(data$p, 3), nsmall = 4)))
  w <- sum(w7, w8, w9, w10, w11, w12, 21)

  p <- format(round(data$p, 4), nsmall = 4)

  cat(fc("ANOVA", w), "\n")
  cat(rep("-", w), sep = "", "\n")

  cat(fg("", w7), fs(), fg("Sum of", w8), fs(), fg("", w9), fs(), fg("", w10),
    fs(), fg("", w11), fs(), fg("", w12), "\n")

  cat(fg("", w7), fs(), fg("Squares", w8), fs(), fg("DF", w9), fs(),
    fg("Mean Square", w10), fs(), fc("F", w11), fs(), fg("Sig.", w12), "\n")

  cat(rep("-", w), sep = "", "\n")

  cat(
    fl("Regression", w7), fs(), fg(format(round(data$rss, 3), nsmall = 3), w8),
    fs(), fg(round(data$model_df, 3), w9), fs(),
    fg(format(round(data$rms, 3), nsmall = 3), w10), fs(),
    fg(round(data$f, 3), w11), fs(), fg(p, w12), "\n"
  )

  cat(
    fl("Residual", w7), fs(), fg(format(round(data$ess, 3), nsmall = 3), w8),
    fs(), fg(round(data$error_df, 3), w9), fs(),
    fg(format(round(data$ems, 3), nsmall = 3), w10), fs(), fg("", w11), fs(),
    fg("", w12), "\n"
  )

  cat(
    fl("Total", w7), fs(), fg(format(round(data$tss, 3), nsmall = 3), w8), fs(),
    fg(round(data$total_df, 3), w9), fs(), fg("", w10), fs(),
    fg("", w11), fs(), fg("", w12), "\n"
  )

  cat(rep("-", w), sep = "", "\n\n")

}

#' Added variable plots
#'
#' @description
#' Added variable plot provides information about the marginal importance of a
#' predictor variable, given the other predictor variables already in
#' the model. It shows the marginal importance of the variable in reducing the
#' residual variability.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details The added variable plot was introduced by Mosteller and Tukey
#' (1977). It enables us to visualize the regression coefficient of a new
#' variable being considered to be included in a model. The plot can be
#' constructed for each predictor variable.
#'
#' Let us assume we want to test the effect of adding/removing variable \emph{X} from a
#' model. Let the response variable of the model be \emph{Y}
#'
#' Steps to construct an added variable plot:
#'
#' \itemize{
#'   \item Regress \emph{Y} on all variables other than \emph{X} and store the residuals (\emph{Y} residuals).
#'   \item Regress \emph{X} on all the other variables included in the model (\emph{X} residuals).
#'   \item Construct a scatter plot of \emph{Y} residuals and \emph{X} residuals.
#' }
#'
#' What do the \emph{Y} and \emph{X} residuals represent? The \emph{Y} residuals represent the part
#' of \strong{Y} not explained by all the variables other than X. The \emph{X} residuals
#' represent the part of \strong{X} not explained by other variables. The slope of the line
#' fitted to the points in the added variable plot is equal to the regression
#' coefficient when \strong{Y} is regressed on all variables including \strong{X}.
#'
#' A strong linear relationship in the added variable plot indicates the increased
#' importance of the contribution of \strong{X} to the model already containing the
#' other predictors.
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_added_variable(model)
#'
#' @importFrom stats lm
#' @importFrom ggplot2 ggplot aes geom_point xlab ylab stat_smooth
#' @importFrom gridExtra marrangeGrob
#'
#' @seealso [ols_plot_resid_regressor()], [ols_plot_comp_plus_resid()]
#'
#' @export
#'
ols_plot_added_variable <- function(model, print_plot = TRUE) {

  check_model(model)

  data    <- ols_prep_avplot_data(model)
  xnames  <- colnames(data)
  nl      <- length(xnames)
  myplots <- list()

  for (i in 2:nl) {

    x <- ols_prep_regress_x(data, i)
    y <- ols_prep_regress_y(data, i)
    d <- data.frame(x, y)

    p <-
      eval(
        substitute(
          ggplot(d, aes(x = x, y = y)) +
            geom_point(colour = "blue", size = 2) +
            stat_smooth(method = "lm", se = FALSE) +
            xlab(paste(xnames[i], " | Others")) +
            ylab(paste(xnames[1], " | Others")),
          list(i = i)
        )
      )

    j <- i - 1
    myplots[[j]] <- p

  }

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2, top = "Added Variable Plots")
  } else {
    return(myplots)
  }

}

#' Score test
#'
#' @description
#' Test for heteroskedasticity under the assumption that the errors are
#' independent and identically distributed (i.i.d.).
#'
#' @param model An object of class \code{lm}.
#' @param fitted_values Logical; if TRUE, use fitted values of regression model.
#' @param rhs Logical; if TRUE, specifies that tests for heteroskedasticity be
#'   performed for the right-hand-side (explanatory) variables of the fitted
#'   regression model.
#' @param vars Variables to be used for for heteroskedasticity test.
#'
#' @return \code{ols_test_score} returns an object of class \code{"ols_test_score"}.
#' An object of class \code{"ols_test_score"} is a list containing the
#' following components:
#'
#' \item{score}{f statistic}
#' \item{p}{p value of \code{score}}
#' \item{df}{degrees of freedom}
#' \item{fv}{fitted values of the regression model}
#' \item{rhs}{names of explanatory variables of fitted regression model}
#' \item{resp}{response variable}
#' \item{preds}{predictors}
#'
#' @references
#' Breusch, T. S. and Pagan, A. R. (1979) A simple test for heteroscedasticity and random coefficient variation. Econometrica 47, 1287–1294.
#'
#' Cook, R. D. and Weisberg, S. (1983) Diagnostics for heteroscedasticity in regression. Biometrika 70, 1–10.
#'
#' Koenker, R. 1981. A note on studentizing a test for heteroskedasticity. Journal of Econometrics 17: 107–112.
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#'
#' # using fitted values of the model
#' ols_test_score(model)
#'
#' # using predictors from the model
#' ols_test_score(model, rhs = TRUE)
#'
#' # specify predictors from the model
#' ols_test_score(model, vars = c('disp', 'wt'))
#'
#' @family heteroskedasticity tests
#'
#' @export
#'
ols_test_score <- function(model, fitted_values = TRUE, rhs = FALSE,
                           vars = NULL) UseMethod("ols_test_score")

#' @export
#'
ols_test_score.default <- function(model, fitted_values = TRUE, rhs = FALSE,
                                   vars = NULL) {

  check_model(model)
  check_logic(fitted_values)
  check_logic(rhs)

  if (length(vars) > 0) {
    check_modelvars(model, vars)
    fitted_values <- FALSE
  }

  resp <- names(model.frame(model))[1]

  if (rhs) {
    fitted_values <- FALSE
    d <- rhsout(model)
  } else {
    if (fitted_values) {
      d <- fitout(model, resp)
    } else {
      d <- varout(model, vars)
    }
  }

  out <- list(df = d$np, fv = fitted_values, p = d$p, preds = d$preds,
              resp = resp, rhs = rhs, score = d$score)

  class(out) <- "ols_test_score"
  return(out)

}

#' @export
#'
print.ols_test_score <- function(x, ...) {
  print_score_test(x, ...)
}

rhsout <- function(model) {

  l         <- ols_prep_avplot_data(model)
  nam       <- names(l)[-1]
  var_resid <- residual_var(model, nrow(l))
  ind       <- ind_score(model, var_resid)
  score     <- rhs_score(l, ind, nrow(l))
  p         <- pchisq(score, length(nam), lower.tail = F)

  list(score = score, preds = nam, np = length(nam), p = p)

}

fitout <- function(model, resp) {

  score <- fit_score(model)
  p     <- pchisq(score, 1, lower.tail = F)

  list(score = score, preds = paste("fitted values of", resp), np = 1, p = p)

}

varout <- function(model, vars) {

  score <- var_score(model, vars)
  nd    <- ncol(score_data(model, vars)) - 1
  p     <- pchisq(score, nd, lower.tail = F)

  list(score = score, preds = vars, np = nd, p = p)

}

residual_var <- function(model, n) {
  sum(residuals(model) ^ 2) / n
}

ind_score <- function(model, var_resid) {

  vresid <- var_resid - 1
  (residuals(model) ^ 2) / vresid

}

rhs_score <- function(l, ind, n) {

  ind <- data.frame(ind = ind)
  (summary(lm(ind ~ ., data = cbind(l, ind)[, -1]))$r.squared) * n

}

fit_score <- function(model) {

  pred         <- fitted(model)
  scaled_resid <- resid_scaled(model, pred)
  l            <- ols_prep_avplot_data(model)
  n            <- nrow(l)

  (summary(lm(scaled_resid ~ pred))$r.squared) * n

}

resid_scaled <- function(model, pred) {

  resid     <- residuals(model) ^ 2
  avg_resid <- sum(resid) / length(pred)
  resid / avg_resid

}

var_score <- function(model, vars) {

  n <- nrow(ols_prep_avplot_data(model)) 
  (summary(lm(ind ~ ., data = score_data(model, vars)))$r.squared) * n
    
}

score_data <- function(model, vars) {

  l              <- ols_prep_avplot_data(model)
  var_resid      <- residual_var(model, nrow(l))
  ind            <- as.data.frame(ind_score(model, var_resid)) 
  colnames(ind)  <- c("ind")

  cbind(l[, vars], ind)
}

model <- lm(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)

test_that("when fitted.values == TRUE, fitted values from the regression\n\tare used for the test", {
  b <- ols_test_score(model)

  expect_equal(round(b$score, 3), 1.269)
  expect_equal(round(b$p, 3), 0.26)
  expect_equal(b$df, 1)
  expect_true(b$fv)
  expect_false(b$rhs)
  expect_match(b$preds, "fitted values of mpg")
  expect_match(b$resp, "mpg")
})

test_that("when fitted.values == TRUE and rhs == TRUE, predictors from the\n\tmodel are used for the test", {
  b <- ols_test_score(model, rhs = TRUE)

  expect_equal(round(b$score, 3), 2.516)
  expect_equal(round(b$p, 3), 0.774)
  expect_equal(b$df, 5)
  expect_false(b$fv)
  expect_true(b$rhs)
  expect_equal(b$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
  expect_match(b$resp, "mpg")
})


test_that("when vars != NULL, variables specified from the are\n\tused for the test", {
  b <- ols_test_score(model, vars = c("disp", "hp"))

  expect_equal(round(b$score, 3), 0.969)
  expect_equal(round(b$p, 3), 0.616)
  expect_equal(b$df, 2)
  expect_false(b$fv)
  expect_false(b$rhs)
  expect_equal(b$preds, c("disp", "hp"), ignore_attr = TRUE)
  expect_match(b$resp, "mpg")
})


test_that("when vars != NULL and rhs == TRUE, predictors in the model are\n\tused for the test", {
  b <- ols_test_score(model, rhs = TRUE, vars = c("disp", "hp"))

  expect_equal(round(b$score, 3), 2.516)
  expect_equal(round(b$p, 3), 0.774)
  expect_equal(b$df, 5)
  expect_false(b$fv)
  expect_true(b$rhs)
  expect_equal(b$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
  expect_match(b$resp, "mpg")
})

test_that("when fitted.values == TRUE, fitted vals from the reg are used", {
  model <- lm(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)
  expect_snapshot(ols_test_score(model))
  expect_snapshot(ols_test_score(model, rhs = TRUE))
  expect_snapshot(ols_test_score(model, vars = c("disp", "hp")))
  expect_snapshot(ols_test_score(model, rhs = TRUE, vars = c("disp", "hp")))
})

#' Lack of fit F test
#'
#' Assess how much of the error in prediction is due to lack of model fit.
#'
#' @param model An object of class \code{lm}.
#' @param ... Other parameters.
#'
#' @details
#' The residual sum of squares resulting from a regression can be decomposed
#' into 2 components:
#'
#' \itemize{
#'   \item Due to lack of fit
#'   \item Due to random variation
#' }
#'
#' If most of the error is due to lack of fit and not just random error, the
#' model should be discarded and a new model must be built.
#'
#' @note The lack of fit F test works only with simple linear regression.
#' Moreover, it is important that the data contains repeat observations i.e.
#' replicates for at least one of the values of the predictor x. This
#' test generally only applies to datasets with plenty of replicates.
#'
#' @return \code{ols_pure_error_anova} returns an object of class
#' \code{"ols_pure_error_anova"}. An object of class \code{"ols_pure_error_anova"} is a
#' list containing the following components:
#'
#' \item{lackoffit}{lack of fit sum of squares}
#' \item{pure_error}{pure error sum of squares}
#' \item{rss}{regression sum of squares}
#' \item{ess}{error sum of squares}
#' \item{total}{total sum of squares}
#' \item{rms}{regression mean square}
#' \item{ems}{error mean square}
#' \item{lms}{lack of fit mean square}
#' \item{pms}{pure error mean square}
#' \item{rf}{f statistic}
#' \item{lf}{lack of fit f statistic}
#' \item{pr}{p-value of f statistic}
#' \item{pl}{p-value pf lack of fit f statistic}
#' \item{mpred}{\code{data.frame} containing data for the response and predictor of the \code{model}}
#' \item{df_rss}{regression sum of squares degrees of freedom}
#' \item{df_ess}{error sum of squares degrees of freedom}
#' \item{df_lof}{lack of fit degrees of freedom}
#' \item{df_error}{pure error degrees of freedom}
#' \item{final}{data.frame; contains computed values used for the lack of fit f test}
#' \item{resp}{character vector; name of \code{response variable}}
#' \item{preds}{character vector; name of \code{predictor variable}}
#'
#' @references
#' Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
#'
#' @examples
#' model <- lm(mpg ~ disp, data = mtcars)
#' ols_pure_error_anova(model)
#'
#' @importFrom stats coefficients
#'
#' @export
#'
ols_pure_error_anova <- function(model, ...) UseMethod("ols_pure_error_anova")

#' @export
#'
ols_pure_error_anova.default <- function(model, ...) {

  check_model(model)

  ln <- length(coefficients(model))

  if (ln > 2) {
    stop("Lack of fit F test is available only for simple linear regression.", call. = FALSE)
  }

  k <- peanova(model)

  result <- list(lackoffit = k$lackoffit, pure_error = k$pure_error, rss = k$rss,
    ess = k$ess, total = k$total, rms = k$rms, ems = k$ems, lms = k$lms,
    pms = k$pms, rf = k$rf, lf = k$lf, pr = k$pr, pl = k$pl, mpred = k$mpred,
    df_rss = k$df_rss, df_ess = k$df_ess, df_lof = k$df_lof,
    df_error = k$df_error, final = k$final, resp = k$resp, preds = k$preds)

  class(result) <- "ols_pure_error_anova"

  return(result)

}

#' @export
#'
print.ols_pure_error_anova <- function(x, ...) {
  print_pure_error_anova(x)
}

peanova <- function(model) {

  n         <- model_rows(model)
  nd        <- pred_table_length(model)
  comp      <- pea_data(model)
  final     <- comp$result
  mean_pred <- comp$mean_pred
  pred_name <- comp$pred_name
  dep_name  <- comp$resp
  lackoffit <- sum(final$lfit)
  
  random_error <- sum(final$rerror)

  rss      <- rss_model(model)
  ess      <- sum(lackoffit, random_error)
  total    <- sum(ess, rss)
  df_rss   <- 1
  df_lof   <- nd - 2
  df_error <- n - nd
  df_ess   <- sum(df_lof, df_error)
  rms      <- rss / df_rss
  ems      <- ess / df_ess
  lms      <- lackoffit / df_lof
  pms      <- random_error / df_error
  rf       <- rms / pms
  lf       <- lms / pms
  pr       <- pf(rf, df_rss, df_ess, lower.tail = F)
  pl       <- pf(lf, df_lof, df_error, lower.tail = F)

  list(
    lackoffit = lackoffit, pure_error = random_error, rss = rss, ess = ess,
    total = total, rms = rms, ems = ems, lms = lms, pms = pms, rf = rf, lf = lf,
    pr = pr, pl = pl, mpred = mean_pred, df_rss = df_rss, df_ess = df_ess,
    df_lof = df_lof, df_error = df_error, final = final, resp = dep_name,
    preds = pred_name
  )

}

pea_data <- function(model) {

  data      <- model.frame(model)
  pred_name <- names(data)[2]
  resp      <- names(data)[1]
  pred_u    <- pred_table(model)
  mean_pred <- predictor_mean(data, pred_name, resp)
  mean_rep  <- replicate_mean(mean_pred, pred_u)
  result    <- pea_data_comp(data, model, mean_rep)

  list(pred_name = pred_name,
       resp      = resp,
       result    = result,
       mean_pred = mean_pred)

}

pred_table <- function(model) {
  table(model.frame(model)[[2]])
}

pred_table_length <- function(model) {
  length(pred_table(model))
}

predictor_mean <- function(data, pred_name, resp) {

  d   <- split(data[[resp]], data[[pred_name]])
  out <- lapply(d, mean)
  data.frame(n = names(out), mean = as.numeric(out))

}

rss_model <- function(model) {
  anova(model)[1, 2]
}


replicate_mean <- function(mean_pred, pred_u) {

  out <- as.data.frame(rep(mean_pred[[2]], times = pred_u))
  colnames(out) <- c("ybar")
  return(out)

}


pea_data_comp <- function(data, model, mean_rep) {

  data$yhat      <- fitted(model)
  colnames(data) <- c("y", "pred", "yhat")
  data           <- data[order(data$pred), ]
  data           <- cbind(data, mean_rep)
  data$lfit      <- (data$ybar - data$yhat) ^ 2
  data$rerror    <- (data$y - data$ybar) ^ 2

  return(data)

}

test_that("output from pure_error_anova matches expected result", {
  model <- lm(mpg ~ hp, data = mtcars)
  k <- ols_pure_error_anova(model)
  act <- k$lackoffit
  exp <- 430.32
  expect_equal(round(act, 2), exp)

  act <- k$pure_error
  exp <- 17.36
  expect_equal(round(act, 2), exp)

  model <- lm(mpg ~ disp, data = mtcars)
  k <- ols_pure_error_anova(model)
  act <- k$lackoffit
  exp <- 304.28
  expect_equal(round(act, 2), exp)

  act <- k$pure_error
  exp <- 12.88
  expect_equal(round(act, 2), exp)
})

#' Collinearity diagnostics
#'
#' @description
#' Variance inflation factor, tolerance, eigenvalues and condition indices.
#'
#' @param model An object of class \code{lm}.
#'
#' @details
#' Collinearity implies two variables are near perfect linear combinations of
#' one another. Multicollinearity involves more than two variables. In the
#' presence of multicollinearity, regression estimates are unstable and have
#' high standard errors.
#'
#' \emph{Tolerance}
#'
#' Percent of variance in the predictor that cannot be accounted for by other predictors.
#'
#' Steps to calculate tolerance:
#'
#' \itemize{
#'   \item Regress the kth predictor on rest of the predictors in the model.
#'   \item Compute \eqn{R^2} - the coefficient of determination from the regression in the above step.
#'   \item \eqn{Tolerance = 1 - R^2}
#' }
#'
#' \emph{Variance Inflation Factor}
#'
#' Variance inflation factors measure the inflation in the variances of the parameter estimates due to
#' collinearities that exist among the predictors. It is a measure of how much the variance of the estimated
#' regression coefficient \eqn{\beta_k}  is inflated by the existence of correlation among the predictor variables
#' in the model. A VIF of 1 means that there is no correlation among the kth predictor and the remaining predictor
#' variables, and hence the variance of \eqn{\beta_k} is not inflated at all. The general rule of thumb is that VIFs
#' exceeding 4 warrant further investigation, while VIFs exceeding 10 are signs of serious multicollinearity
#' requiring correction.
#'
#' Steps to calculate VIF:
#'
#' \itemize{
#'   \item Regress the kth predictor on rest of the predictors in the model.
#'   \item Compute \eqn{R^2} - the coefficient of determination from the regression in the above step.
#'   \item \eqn{Tolerance = 1 / 1 - R^2 = 1 / Tolerance}
#' }
#'
#' \emph{Condition Index}
#'
#' Most multivariate statistical approaches involve decomposing a correlation matrix into linear
#' combinations of variables. The linear combinations are chosen so that the first combination has
#' the largest possible variance (subject to some restrictions), the second combination
#' has the next largest variance, subject to being uncorrelated with the first, the third has the largest
#' possible variance, subject to being uncorrelated with the first and second, and so forth. The variance
#' of each of these linear combinations is called an eigenvalue. Collinearity is spotted by finding 2 or
#' more variables that have large proportions of variance (.50 or more) that correspond to large condition
#' indices. A rule of thumb is to label as large those condition indices in the range of 30 or larger.
#'
#'
#' @return \code{ols_coll_diag} returns an object of class \code{"ols_coll_diag"}.
#' An object of class \code{"ols_coll_diag"} is a list containing the
#' following components:
#'
#' \item{vif_t}{tolerance and variance inflation factors}
#' \item{eig_cindex}{eigen values and condition index}
#'
#' @references
#' Belsley, D. A., Kuh, E., and Welsch, R. E. (1980). Regression Diagnostics: Identifying Influential Data and
#' Sources of Collinearity. New York: John Wiley & Sons.
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
#'
#' # vif and tolerance
#' ols_vif_tol(model)
#'
#' # eigenvalues and condition indices
#' ols_eigen_cindex(model)
#'
#' # collinearity diagnostics
#' ols_coll_diag(model)
#'
#' @export
#'
ols_coll_diag <- function(model) UseMethod("ols_coll_diag")

#' @export
#'
ols_coll_diag.default <- function(model) {

  check_model(model)

  vift    <- ols_vif_tol(model)
  eig_ind <- ols_eigen_cindex(model)
  result  <- list(eig_cindex = eig_ind, vif_t = vift)

  class(result) <- "ols_coll_diag"
  return(result)

}

#' @export
#'
print.ols_coll_diag <- function(x, ...) {
  cat("Tolerance and Variance Inflation Factor\n")
  cat("---------------------------------------\n")
  print(x$vif_t)
  cat("\n\n")
  cat("Eigenvalue and Condition Index\n")
  cat("------------------------------\n")
  print(x$eig_cindex)
}

#' @rdname ols_coll_diag
#' @export
#'
ols_vif_tol <- function(model) {

  check_model(model)
  vt <- viftol(model)
  data.frame(Variables = vt$nam, Tolerance = vt$tol, VIF = vt$vifs)

}

#' @rdname ols_coll_diag
#' @export
#'
ols_eigen_cindex <- function(model) {

  check_model(model)

  x      <- as.data.frame(model.matrix(model))
  e      <- evalue(x)$e
  cindex <- cindx(e)
  pv     <- pveindex(evalue(x)$pvdata)
  out    <- data.frame(Eigenvalue = cbind(e, cindex, pv))
  
  colnames(out) <- c("Eigenvalue", "Condition Index", colnames(evalue(x)$pvdata))
  return(out)

}


evalue <- function(x) {

  y              <- x
  colnames(y)[1] <- "intercept"
  z              <- scale(y, scale = T, center = F)
  tu             <- t(z) %*% z

  e <- eigen(tu / diag(tu))$values
  list(e = e, pvdata = z)

}


cindx <- function(e) {
  sqrt(e[1] / e) 
}

pveindex <- function(z) {

  svdx  <- svd(z)
  svdxd <- svdx$d

  phi_diag <- diag(1 / svdxd)
  phi      <- svdx$v %*% phi_diag
  ph       <- t(phi ^ 2)
  diag_sum <- diag(rowSums(ph, dims = 1))
  
  prop.table(ph %*% diag_sum, margin = 2)

}


fmrsq <- function(nam, data, i) {
  
  fm <- as.formula(paste0("`", nam[i], "` ", "~ ."))
  m1 <- summary(lm(fm, data = data))$r.squared

  1 - m1

}


#' @description Computes vif and tolerance
#'
#' @noRd
#'
viftol <- function(model) {

  m   <- as.data.frame(model.matrix(model))[, -1]
  nam <- names(m)
  p   <- length(model$coefficients) - 1
  tol <- c()

  for (i in seq_len(p)) {
    tol[i] <- fmrsq(nam, m, i)
  }

  vifs <- 1 / tol
  list(nam = names(m), tol = tol, vifs = vifs)

}

#' Hadi plot
#'
#' @description
#' Hadi's measure of influence based on the fact that influential observations
#' can be present in either the response variable or in the predictors or both.
#' The plot is used to detect influential observations based on Hadi's measure.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_hadi(model)
#'
#' @seealso [ols_plot_resid_pot()]
#'
#' @export
#'
ols_plot_hadi <- function(model, print_plot = TRUE) {

  check_model(model)

  hdi  <- unname(ols_hadi(model)$hadi)
  obs  <- seq_len(length(hdi))
  d    <- data.frame(obs = obs, hdi = hdi)

  p <-
    ggplot(d, aes(obs, hdi, ymin = min(hdi), ymax = hdi)) +
    geom_linerange(colour = "blue") +
    geom_point(shape = 1, colour = "blue")

  p <-
    p +
    xlab("Observation") +
    ylab("Hadi's Measure") +
    ggtitle("Hadi's Influence Measure")

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

#' Residual box plot
#'
#' Box plot of residuals to examine if residuals are normally distributed.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_box(model)
#'
#' @family residual diagnostics
#'
#' @importFrom stats residuals
#' @importFrom ggplot2 geom_boxplot theme element_blank
#'
#' @export
#'
ols_plot_resid_box <- function(model, print_plot = TRUE) {

  check_model(model)

  resid <- residuals(model)
  d     <- data.frame(resid = resid)

  p <-
    ggplot(d, aes(x = factor(0), y = resid)) +
    geom_boxplot(outlier.color = "green", outlier.size = 3, fill = "grey80",
                 colour = "#3366FF")

  p <-
    p +
    xlab(" ") +
    ylab("Residuals") +
    ggtitle("Residual Box Plot")

  p <-
    p +
    theme(axis.text.x = element_blank())

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

#' Breusch pagan test
#'
#' @description
#' Test for constant variance. It assumes that the error terms are normally
#' distributed.
#'
#' @param model An object of class \code{lm}.
#' @param fitted.values Logical; if TRUE, use fitted values of regression model.
#' @param rhs Logical; if TRUE, specifies that tests for heteroskedasticity be
#' performed for the right-hand-side (explanatory) variables of the fitted
#' regression model.
#' @param multiple Logical; if TRUE, specifies that multiple testing be performed.
#' @param p.adj Adjustment for p value, the following options are available:
#' bonferroni, holm, sidak and none.
#' @param vars Variables to be used for heteroskedasticity test.
#'
#' @details
#' Breusch Pagan Test was introduced by Trevor Breusch and Adrian Pagan in 1979.
#' It is used to test for heteroskedasticity in a linear regression model.
#' It test whether variance of errors from a regression is dependent on the
#' values of a independent variable.
#'
#' \itemize{
#' \item Null Hypothesis: Equal/constant variances
#' \item Alternative Hypothesis: Unequal/non-constant variances
#' }
#'
#' Computation
#'
#' \itemize{
#'   \item Fit a regression model
#'   \item Regress the squared residuals from the above model on the independent variables
#'   \item Compute \eqn{nR^2}. It follows a chi square distribution with p -1 degrees of
#'         freedom, where p is the number of independent variables, n is the sample size and
#' 				 \eqn{R^2} is the coefficient of determination from the regression in step 2.
#' }
#'
#' @return \code{ols_test_breusch_pagan} returns an object of class \code{"ols_test_breusch_pagan"}.
#' An object of class \code{"ols_test_breusch_pagan"} is a list containing the
#' following components:
#'
#' \item{bp}{breusch pagan statistic}
#' \item{p}{p-value of \code{bp}}
#' \item{fv}{fitted values of the regression model}
#' \item{rhs}{names of explanatory variables of fitted regression model}
#' \item{multiple}{logical value indicating if multiple tests should be performed}
#' \item{padj}{adjusted p values}
#' \item{vars}{variables to be used for heteroskedasticity test}
#' \item{resp}{response variable}
#' \item{preds}{predictors}
#'
#' @references
#' T.S. Breusch & A.R. Pagan (1979), A Simple Test for Heteroscedasticity and
#' Random Coefficient Variation. Econometrica 47, 1287–1294
#'
#' Cook, R. D.; Weisberg, S. (1983). "Diagnostics for Heteroskedasticity in Regression". Biometrika. 70 (1): 1–10.
#'
#' @family heteroskedasticity tests
#'
#' @examples
#' # model
#' model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
#'
#' # use fitted values of the model
#' ols_test_breusch_pagan(model)
#'
#' # use independent variables of the model
#' ols_test_breusch_pagan(model, rhs = TRUE)
#'
#' # use independent variables of the model and perform multiple tests
#' ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE)
#'
#' # bonferroni p value adjustment
#' ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'bonferroni')
#'
#' # sidak p value adjustment
#' ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'sidak')
#'
#' # holm's p value adjustment
#' ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'holm')
#'
#' @importFrom stats anova
#'
#' @export
#'
ols_test_breusch_pagan <- function(model, fitted.values = TRUE, rhs = FALSE, multiple = FALSE,
                        p.adj = c("none", "bonferroni", "sidak", "holm"), vars = NA) UseMethod("ols_test_breusch_pagan")

#' @export
#'
ols_test_breusch_pagan.default <- function(model, fitted.values = TRUE, rhs = FALSE, multiple = FALSE,
                                p.adj = c("none", "bonferroni", "sidak", "holm"), vars = NA) {

  check_model(model)
  check_logic(fitted.values)
  check_logic(rhs)
  check_logic(multiple)

  suppressWarnings(
    if (!is.na(vars[1])) {
      check_modelvars(model, vars)
      fitted.values <- FALSE
    }
  )

  method <- match.arg(p.adj)
  p.adj  <- match.arg(p.adj)

  check_options(p.adj)

  l <- ols_prep_avplot_data(model)
  n <- nrow(l)
  response   <- names(l)[1]
  predictors <- names(l)[-1]

  if (fitted.values) {
    vars <- NA

    if (rhs) {
      if (multiple) {

        start <- bp_case_one(l, model)
        loops <- bp_case_loop(start$np, start$nam, start$l)
        inter <- bp_case_inter(start$l, start$np, loops$tstat)
        bp    <- inter$bp
        p     <- bp_case_adj(method, loops$pvals, start$np, inter$ps)

      } else {
        result <- bp_case_2(l, model)
        bp     <- result$bp
        p      <- pchisq(bp, df = result$df, lower.tail = FALSE)
      }
    } else {
      bp <- bp_case_3(model)
      p  <- pchisq(bp, df = 1, lower.tail = FALSE)
    }
  } else {
    if (multiple) {
      if (rhs) {
        start <- bp_case_one(l, model)
        loops <- bp_case_loop(start$np, start$nam, start$l)
        inter <- bp_case_inter(start$l, start$np, loops$tstat)
        bp    <- inter$bp
        p     <- bp_case_adj(method, loops$pvals, start$np, inter$ps)

      } else {
        len_vars <- length(vars)

        if (len_vars > 1) {
          start   <- bp_case_one(l, model)
          len_var <- length(vars)
          loops   <- bp_case_loop(len_var, vars, start$l)
          inter   <- bp_case_5_inter(l, model, vars, loops$tstat)
          bp      <- inter$bp
          p       <- bp_case_adj(method, loops$pvals, inter$np, inter$ps)
        } else {
          result  <- bp_case_7(l, model, vars)
          bp      <- result$bp
          p       <- pchisq(bp, df = result$df, lower.tail = FALSE)
        }
      }
    } else {
      if (rhs) {
        result <- bp_case_6(l, model)
        bp     <- result$bp
        p      <- pchisq(bp, df = result$df, lower.tail = FALSE)
      } else {
        result <- bp_case_7(l, model, vars)
        bp     <- result$bp
        p      <- pchisq(bp, df = result$df, lower.tail = FALSE)
      }
    }
  }

  out <- list(bp = bp, fv = fitted.values, multiple = multiple, p = p,
              padj = method, preds = predictors, resp = response,
              rhs = rhs, vars = vars)

  class(out) <- "ols_test_breusch_pagan"

  return(out)
}

#' @export
#'
print.ols_test_breusch_pagan <- function(x, ...) {
  print_bp_test(x)
}

#' @description
#' Computes breusch pagan statistics and degrees of freedom when:
#' * fit = TRUE
#' * rhs = TRUE
#' * multiple = TRUE
#'
#' @param l A \code{data.frame} created using `avplots_data()`.
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
bp_case_2 <- function(l, model) {

  n         <- model_rows(model)
  var_resid <- residual_var(model, n)
  ind       <- ind_bp(model, var_resid)
  df        <- ncol(l[, -1])
  l         <- cbind(l[, -1], ind)
  bp        <- bp_model(l)

  list(bp = bp, df = df)

}

#' @description
#' Computes breusch pagan statistics and degrees of freedom when:
#' * fit = TRUE
#' * rhs = FALSE
#'
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
bp_case_3 <- function(model) {

  `Sum Sq`     <- NULL
  pred         <- fitted(model)
  scaled_resid <- resid_scaled(model, pred)

  (anova(lm(scaled_resid ~ pred))$`Sum Sq`[1]) / 2

}

#' @description
#' Computes breusch pagan statistics and degrees of freedom when:
#' * fit = FALSE
#' * rhs = TRUE
#' * multiple = FALSE
#'
#' @param l A \code{data.frame} created using `avplots_data()`.
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
bp_case_6 <- function(l, model) {

  n         <- nrow(l)
  var_resid <- residual_var(model, n)
  ind       <- ind_bp(model, var_resid)
  np        <- length(names(l)[-1])
  bp        <- bp_model(cbind(l, ind)[, -1])
  
  list(bp = bp, df = np)

}

#' @description
#' Computes breusch pagan statistics and degrees of freedom when:
#' * fit = FALSE
#' * rhs = FALSE
#' * multiple = FALSE
#'
#' @param l A \code{data.frame} created using `avplots_data()`.
#' @param model An object of class \code{lm}.
#' @param vars Variables to be used for the test.
#'
#' @noRd
#'
bp_case_7 <- function(l, model, vars) {

  n         <- nrow(l)
  var_resid <- residual_var(model, n)
  ind       <- ind_bp(model, var_resid)
  l         <- cbind(l[, vars], ind)
  bp        <- bp_model(l)
  nd        <- ncol(l) - 1

  list(bp = bp, df = nd)

}

#' @description Fit model using the columns in the data set.
#'
#' @param l A \code{data.frame} created using `avplots_data()`.
#'
#' @noRd
#'
bp_model <- function(l) {
  bp_fit(lm(ind ~ ., data = l))
}

bp_fit <- function(l) {
  (sum(fitted(l) ^ 2)) / 2
}

ind_bp <- function(model, var_resid) {
  data.frame(ind = ((residuals(model) ^ 2) / var_resid) - 1)
}

#' @description
#' Computes breusch pagan statistics and degrees of freedom when:
#' * fit = TRUE
#' * rhs = TRUE
#' * multiple = TRUE
#'
#' @param l A \code{data.frame} created using `avplots_data()`.
#' @param model An object of class \code{lm}.
#'
#' @noRd
#'
bp_case_one <- function(l, model) {

  nam       <- names(l)[-1]
  np        <- length(nam)
  n         <- nrow(l)
  var_resid <- residual_var(model, n)
  ind       <- ind_bp(model, var_resid)
  l         <- cbind(l, ind)

  list(np = np, nam = nam, l = l)

}

#' @description
#' Computes breusch pagan statistic and p values when multiple = TRUE
#'
#' @noRd
#'
bp_case_loop <- function(np, nam, l) {

  tstat <- c()
  pvals <- c()

  for (i in seq_len(np)) {

    form     <- as.formula(paste("ind ~", nam[i]))
    tstat[i] <- bp_fit(lm(form, data = l))
    pvals[i] <- pchisq(tstat[i], df = 1, lower.tail = FALSE)

  }

  list(tstat = tstat, pvals = pvals)

}

#' @description
#' Computes breusch pagan statistic and p values when multiple = FALSE.
#'
#' @noRd
#'
bp_case_inter <- function(l, np, tstat) {

  comp <- bp_fit(lm(ind ~ ., data = l[, -1])) 
  ps   <- pchisq(comp, df = np, lower.tail = FALSE)
  bp   <- c(tstat, comp)

  list(bp = bp, ps = ps)

}

#' @description Computes adjusted p values.
#'
#' @noRd
#'
bp_case_adj <- function(method, pvals, np, ps) {

  if (method == "bonferroni") {

    bpvals <- pmin(1, pvals * np)
    p      <- c(bpvals, ps)

  } else if (method == "sidak") {

    spvals <- pmin(1, 1 - (1 - pvals) ^ np)
    p      <- c(spvals, ps)

  } else if (method == "holm") {

    j          <- rev(seq_len(length(pvals)))
    h          <- order(order(pvals))
    pvals_sort <- sort(pvals) * j
    pholms     <- pmin(1, pvals_sort)[h] 
    p          <- c(pholms, ps)

  } else {

    p <- c(pvals, ps)

  }

  return(p)

}

#' @description Computes breusch pagan statistic and p values when:
#' * fit = FALSE
#' * rhs = FALSE
#' * multiple = FALSE
#'
#' @noRd
#'
bp_case_5_inter <- function(l, model, vars, tstat) {

  n         <- nrow(l)
  var_resid <- residual_var(model, n)
  ind       <- ind_bp(model, var_resid)
  l         <- cbind(l[, -1][, vars], ind)
  np        <- ncol(l) - 1
  ps        <- pchisq(q = bp_model(l), df = np, lower.tail = FALSE)
  bp        <- c(tstat, bp_model(l))

  list(bp = bp, ps = ps, np = np)

}

model <- lm(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)

test_that("when fitted.values == TRUE, fitted values from the regression\n\tare used for the test", {
  k <- ols_test_breusch_pagan(model)

  expect_equal(round(k$bp, 3), 1.256)
  expect_equal(round(k$p, 3), 0.263)

  expect_true(k$fv)
  expect_false(k$rhs)
  expect_false(k$multiple)

  expect_output(k$vars, NA)

  expect_match(k$padj, "none")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})

test_that("when rhs == TRUE, predictors from the regression\n\tare used for the test", {
  k <- ols_test_breusch_pagan(model, rhs = TRUE)

  expect_equal(round(k$bp, 3), 2.489)
  expect_equal(round(k$p, 3), 0.778)

  expect_true(k$fv)
  expect_true(k$rhs)
  expect_false(k$multiple)

  expect_output(k$vars, NA)

  expect_match(k$padj, "none")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})

test_that("when rhs == TRUE and multiple == TRUE, multiple p values are returned", {
  k <- ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE)

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.7749, 0.7751, 1.2903, 2.4890), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(0.3365, 0.3817, 0.3787, 0.3786, 0.2560, 0.7781), ignore_attr = TRUE)

  expect_true(k$fv)
  expect_true(k$rhs)
  expect_true(k$multiple)

  expect_output(k$vars, NA)

  expect_match(k$padj, "none")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})

test_that("when fitted.value == FALSE and rhs == TRUE and multiple == TRUE, multiple p values are returned", {
  k <- ols_test_breusch_pagan(model, fitted.values = FALSE, rhs = TRUE, multiple = TRUE)

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.7749, 0.7751, 1.2903, 2.4890), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(0.3365, 0.3817, 0.3787, 0.3786, 0.2560, 0.7781), ignore_attr = TRUE)

  expect_false(k$fv)
  expect_true(k$rhs)
  expect_true(k$multiple)

  expect_output(k$vars, NA)

  expect_match(k$padj, "none")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when rhs == TRUE, multiple == TRUE and p.adj == 'bonferroni'
	bonferroni adjusted p values are returned", {
  k <- ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = "bonferroni")

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.7749, 0.7751, 1.2903, 2.4890), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 0.7781), ignore_attr = TRUE)

  expect_true(k$fv)
  expect_true(k$rhs)
  expect_true(k$multiple)

  expect_output(k$vars, NA)

  expect_match(k$padj, "bonferroni")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when rhs == TRUE, multiple == TRUE and p.adj == 'holm',
	bonferroni adjusted p values are returned", {
  k <- ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = "holm")

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.7749, 0.7751, 1.2903, 2.4890), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(1.0000, 0.3817, 0.7574, 1.0000, 1.0000, 0.7781), ignore_attr = TRUE)

  expect_true(k$fv)
  expect_true(k$rhs)
  expect_true(k$multiple)

  expect_output(k$vars, NA)

  expect_match(k$padj, "holm")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when rhs == TRUE, multiple == TRUE and p.adj == 'sidak',
	bonferroni adjusted p values are returned", {
  k <- ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = "sidak")

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.7749, 0.7751, 1.2903, 2.4890), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(0.8714, 0.9096, 0.9074, 0.9074, 0.7720, 0.7781), ignore_attr = TRUE)

  expect_true(k$fv)
  expect_true(k$rhs)
  expect_true(k$multiple)

  expect_output(k$vars, NA)

  expect_match(k$padj, "sidak")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})

test_that("when vars != NA, variables specified are used for the test", {
  k <- ols_test_breusch_pagan(model, vars = c("disp"))

  expect_equal(round(k$bp, 4), 0.9237)
  expect_equal(round(k$p, 4), 0.3365)

  expect_false(k$fv)
  expect_false(k$rhs)
  expect_false(k$multiple)

  expect_match(k$vars, "disp")
  expect_match(k$padj, "none")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})

test_that("when rhs == TRUE and vars != NA, variables specified\n\tused for the test", {
  k <- ols_test_breusch_pagan(model, vars = c("disp", "hp"), rhs = TRUE)

  expect_equal(round(k$bp, 4), 2.489)
  expect_equal(round(k$p, 4), 0.7781)

  expect_false(k$fv)
  expect_true(k$rhs)
  expect_false(k$multiple)

  expect_equal(k$vars, c("disp", "hp"), ignore_attr = TRUE)

  expect_match(k$padj, "none")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when rhs == FALSE, multiple == TRUE and vars != NA,\n\tvariables specified are used for the test", {
  k <- ols_test_breusch_pagan(model, multiple = TRUE, rhs = FALSE, vars = c("disp", "hp"))

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.9588), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(0.3365, 0.3817, 0.6192), ignore_attr = TRUE)

  expect_false(k$fv)
  expect_false(k$rhs)
  expect_true(k$multiple)

  expect_equal(k$vars, c("disp", "hp"), ignore_attr = TRUE)

  expect_match(k$padj, "none")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when multiple == TRUE and vars != NA and p.adj == 'bonferroni',
	variables specified are used for the test", {
  k <- ols_test_breusch_pagan(
    model, multiple = TRUE, vars = c("disp", "hp"),
    p.adj = "bonferroni"
  )

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.9588), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(0.6730, 0.7634, 0.6192), ignore_attr = TRUE)

  expect_false(k$fv)
  expect_false(k$rhs)
  expect_true(k$multiple)

  expect_equal(k$vars, c("disp", "hp"), ignore_attr = TRUE)

  expect_match(k$padj, "bonferroni")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when multiple == TRUE and vars != NA and p.adj == 'sidak',
	variables specified are used for the test", {
  k <- ols_test_breusch_pagan(model, multiple = TRUE, vars = c("disp", "hp"), p.adj = "sidak")


  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.9588), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(0.5598, 0.6177, 0.6192), ignore_attr = TRUE)

  expect_false(k$fv)
  expect_false(k$rhs)
  expect_true(k$multiple)

  expect_equal(k$vars, c("disp", "hp"), ignore_attr = TRUE)

  expect_match(k$padj, "sidak")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})


test_that("when multiple == TRUE and vars != NA and p.adj == 'holm',
	variables specified are used for the test", {
  k <- ols_test_breusch_pagan(
    model, multiple = TRUE,
    vars = c("disp", "hp"), p.adj = "holm"
  )

  expect_equal(round(k$bp, 4), c(0.9237, 0.7652, 0.9588), ignore_attr = TRUE)
  expect_equal(round(k$p, 4), c(0.6730, 0.3817, 0.6192), ignore_attr = TRUE)

  expect_false(k$fv)
  expect_false(k$rhs)
  expect_true(k$multiple)

  expect_equal(k$vars, c("disp", "hp"), ignore_attr = TRUE)

  expect_match(k$padj, "holm")
  expect_match(k$resp, "mpg")
  expect_equal(k$preds, c("disp", "hp", "wt", "drat", "qsec"), ignore_attr = TRUE)
})

test_that("when fitted.values == TRUE, fitted vals from reg are used", {
  model <- lm(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)
  expect_snapshot(ols_test_breusch_pagan(model))
  expect_snapshot(ols_test_breusch_pagan(model, rhs = TRUE))
  expect_snapshot(ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE))
  expect_snapshot(ols_test_breusch_pagan(model, fitted.values = FALSE, rhs = TRUE, multiple = TRUE))
  expect_snapshot(ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = "bonferroni"))
  expect_snapshot(ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = "holm"))
  expect_snapshot(ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = "sidak"))
  expect_snapshot(ols_test_breusch_pagan(model, vars = c("disp")))
  expect_snapshot(ols_test_breusch_pagan(model, multiple = TRUE, rhs = FALSE, vars = c("disp", "hp")))
  expect_snapshot(ols_test_breusch_pagan(model, multiple = TRUE, vars = c("disp", "hp"), p.adj = "bonferroni"))
  expect_snapshot(ols_test_breusch_pagan(model, multiple = TRUE, vars = c("disp", "hp"), p.adj = "sidak"))
  expect_snapshot(ols_test_breusch_pagan(model, multiple = TRUE, vars = c("disp", "hp"), p.adj = "holm"))
  expect_snapshot(ols_test_breusch_pagan(model, multiple = TRUE, rhs = FALSE, vars = c("disp")))
})

#' Potential residual plot
#'
#' @description
#' Plot to aid in classifying unusual observations as high-leverage points,
#' outliers, or a combination of both.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_pot(model)
#'
#' @seealso [ols_plot_hadi()]
#'
#' @export
#'
ols_plot_resid_pot <- function(model, print_plot = TRUE) {

  check_model(model)
  d <- data.frame(res = hadio(model, 3), pot = hadio(model, 2))

  p <-
    ggplot(d, aes(x = res, y = pot)) +
    geom_point(colour = "blue", shape = 1) +
    xlab("Residual") +
    ylab("Potential") +
    ggtitle("Potential-Residual Plot")

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

hadio <- function(model, n) {
  ols_hadi(model)[[n]]
}

#' Residual vs fitted plot
#'
#' @description
#' Scatter plot of residuals on the y axis and fitted values on the
#' x axis to detect non-linearity, unequal error variances, and outliers.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details Characteristics of a well behaved residual vs fitted plot:
#'
#' \itemize{
#'   \item The residuals spread randomly around the 0 line indicating that the relationship is linear.
#'   \item The residuals form an approximate horizontal band around the 0 line indicating homogeneity of error variance.
#'   \item No one residual is visibly away from the random pattern of the residuals indicating that there are no outliers.
#' }
#'
#' @family residual diagnostics
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_fit(model)
#'
#' @export
#'
ols_plot_resid_fit <- function(model, print_plot = TRUE) {

  check_model(model)
  d <- rvspdata(model)
  p <-
    ggplot(d, aes(x = predicted, y = resid)) +
    geom_point(shape = 1, colour = "blue") +
    geom_hline(yintercept = 0, colour = "red") +
    xlab("Fitted Value") +
    ylab("Residual") +
    ggtitle("Residual vs Fitted Values")

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

rvspdata <- function(model) {
  resid     <- residuals(model)
  predicted <- fitted(model)
  data.frame(predicted = predicted, resid = resid)
}

#' Observed vs fitted values plot
#'
#' Plot of observed vs fitted values to assess the fit of the model.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @details
#' Ideally, all your points should be close to a regressed diagonal line. Draw
#' such a diagonal line within your graph and check out where the points lie. If
#' your model had a high R Square, all the points would be close to this
#' diagonal line. The lower the R Square, the weaker the Goodness of fit of your
#' model, the more foggy or dispersed your points are from this diagonal line.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_obs_fit(model)
#'
#' @importFrom stats fitted.values
#' @importFrom ggplot2 geom_abline geom_segment
#'
#' @export
#'
ols_plot_obs_fit <- function(model, print_plot = TRUE) {

  check_model(model)

  oname <- names(model.frame(model))[1]
  d     <- obspred(model)
  df    <- data.frame(x1 = min(d$x), x2 = max(d$x), y1 = min(d$y), y2 = max(d$y))

  p <-
    ggplot(d, aes(x = x, y = y)) +
    geom_point(color = "blue", shape = 1) +
    geom_segment(data = df, aes(x = x1, y = y1, xend = x2,
      yend = y2), colour = "red") +
    geom_abline(intercept = 0, slope = 1, color = "blue")

  p <-
    p +
    ylab("Fitted Value") +
    xlab(paste(oname)) +
    ggtitle(paste("Actual vs Fitted for", oname))

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

obspred <- function(model) {

  y <- fitted(model)
  x <- model.frame(model)[[1]]
  data.frame(x, y)

}

#' Stepwise AIC forward regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering predictors based on akaike information criterion, in a stepwise
#' manner until there is no variable left to enter any more.
#'
#' @param model An object of class \code{lm}.
#' @param include Character or numeric vector; variables to be included in selection process.
#' @param exclude Character or numeric vector; variables to be excluded from selection process.
#' @param progress Logical; if \code{TRUE}, will display variable selection progress.
#' @param details Logical; if \code{TRUE}, will print the regression result at
#'   each step.
#' @param steps Number of steps after which the stepwise procedures should stop.
#' @param x An object of class \code{ols_step_forward_*}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#' @param digits Number of decimal places to display.
#' @param ... Other arguments.
#'
#' @return List containing the following components:
#'
#' \item{model}{final model; an object of class \code{lm}}
#' \item{metrics}{selection metrics}
#' \item{others}{list; info used for plotting and printing}
#'
#' @references
#' Venables, W. N. and Ripley, B. D. (2002) Modern Applied Statistics with S. Fourth edition. Springer.
#'
#' @examples
#' # stepwise forward regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_forward_aic(model)
#'
#' # stepwise forward regression plot
#' k <- ols_step_forward_aic(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # extract final model
#' k$model
#'
#' # include or exclude variables
#' # force variable to be included in selection process
#' ols_step_forward_aic(model, include = c("age"))
#'
#' # use index of variable instead of name
#' ols_step_forward_aic(model, include = c(5))
#'
#' # force variable to be excluded from selection process
#' ols_step_forward_aic(model, exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_aic(model, exclude = c(4))
#'
#' # include & exclude variables in the selection process
#' ols_step_forward_aic(model, include = c("age"), exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_aic(model, include = c(5), exclude = c(4))
#'
#' # steps
#' ols_step_forward_aic(model, steps = 2)
#'
#' @family forward selection procedures
#'
#' @export
#'
ols_step_forward_aic <- function(model, ...) UseMethod("ols_step_forward_aic")

#' @export
#' @rdname ols_step_forward_aic
#'
ols_step_forward_aic.default <- function(model, include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {
  out <- ols_step_forward(model, "aic", include, exclude, progress, details, steps)
  class(out) <- "ols_step_forward_aic"
  return(out)
}

#' @export
#'
print.ols_step_forward_aic <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "forward")
  } else {
    print("No variables have been added to the model.")
  }
}

#' @rdname ols_step_forward_aic
#' @export
#'
plot.ols_step_forward_aic <- function(x, print_plot = TRUE, details = TRUE, digits = 3, ...) {

  p <- ols_stepaic_plot(x, details, digits)

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

#' Stepwise SBC forward regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering predictors based on schwarz bayesian criterion, in a stepwise
#' manner until there is no variable left to enter any more.
#'
#' @inheritParams ols_step_forward_aic
#'
#' @inherit ols_step_forward_aic return references
#'
#' @examples
#' # stepwise forward regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_forward_sbc(model)
#'
#' # stepwise forward regression plot
#' k <- ols_step_forward_sbc(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # extract final model
#' k$model
#'
#' # include or exclude variables
#' # force variable to be included in selection process
#' ols_step_forward_sbc(model, include = c("age"))
#'
#' # use index of variable instead of name
#' ols_step_forward_sbc(model, include = c(5))
#'
#' # force variable to be excluded from selection process
#' ols_step_forward_sbc(model, exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_sbc(model, exclude = c(4))
#'
#' # include & exclude variables in the selection process
#' ols_step_forward_sbc(model, include = c("age"), exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_sbc(model, include = c(5), exclude = c(4))
#'
#' # steps
#' ols_step_forward_sbc(model, steps = 2)
#' 
#' @family forward selection procedures
#'
#' @export
#'
ols_step_forward_sbc <- function(model, ...) UseMethod("ols_step_forward_sbc")

#' @rdname ols_step_forward_sbc
#' @export
#'
ols_step_forward_sbc.default <- function(model, include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {
  out <- ols_step_forward(model, "sbc", include, exclude, progress, details, steps)
  class(out) <- "ols_step_forward_sbc"
  return(out)
}

#' @export
#'
print.ols_step_forward_sbc <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "forward")
  } else {
    print("No variables have been removed from the model.")
  }
}

#' @rdname ols_step_forward_sbc
#' @export
#'
plot.ols_step_forward_sbc <- function(x, print_plot = TRUE, details = TRUE, digits = 3, ...) {
  p <- ols_stepaic_plot(x, details, digits)
  if (print_plot) {
    print(p)
  } else {
    return(p)
  }
}

#' Stepwise SBIC forward regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering predictors based on sawa bayesian criterion, in a stepwise
#' manner until there is no variable left to enter any more.
#'
#' @inheritParams ols_step_forward_aic
#' @inherit ols_step_forward_aic return references
#'
#' @examples
#' # stepwise forward regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_forward_sbic(model)
#'
#' # stepwise forward regression plot
#' k <- ols_step_forward_sbic(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # extract final model
#' k$model
#'
#' # include or exclude variables
#' # force variable to be included in selection process
#' ols_step_forward_sbic(model, include = c("age"))
#'
#' # use index of variable instead of name
#' ols_step_forward_sbic(model, include = c(5))
#'
#' # force variable to be excluded from selection process
#' ols_step_forward_sbic(model, exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_sbic(model, exclude = c(4))
#'
#' # include & exclude variables in the selection process
#' ols_step_forward_sbic(model, include = c("age"), exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_sbic(model, include = c(5), exclude = c(4))
#'
#' # steps
#' ols_step_forward_sbic(model, steps = 2)
#' 
#' @family forward selection procedures
#'
#' @export
#'
ols_step_forward_sbic <- function(model, ...) UseMethod("ols_step_forward_sbic")

#' @export
#' @rdname ols_step_forward_sbic
#'
ols_step_forward_sbic.default <- function(model, include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {
  out <- ols_step_forward(model, "sbic", include, exclude, progress, details, steps)
  class(out) <- "ols_step_forward_sbic"
  return(out)
}

#' @export
#'
print.ols_step_forward_sbic <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "forward")
  } else {
    print("No variables have been removed from the model.")
  }
}

#' @rdname ols_step_forward_sbic
#' @export
#'
plot.ols_step_forward_sbic <- function(x, print_plot = TRUE, details = TRUE, digits = 3, ...) {
  p <- ols_stepaic_plot(x, details, digits)
  if (print_plot) {
    print(p)
  } else {
    return(p)
  }
}

#' Stepwise R-Squared forward regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering predictors based on r-squared, in a stepwise manner until there
#' is no variable left to enter any more.
#'
#' @inheritParams ols_step_forward_aic
#' @inherit ols_step_forward_aic return references
#'
#' @examples
#' # stepwise forward regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_forward_r2(model)
#'
#' # stepwise forward regression plot
#' k <- ols_step_forward_r2(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # extract final model
#' k$model
#'
#' # include or exclude variables
#' # force variable to be included in selection process
#' ols_step_forward_r2(model, include = c("age"))
#'
#' # use index of variable instead of name
#' ols_step_forward_r2(model, include = c(5))
#'
#' # force variable to be excluded from selection process
#' ols_step_forward_r2(model, exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_r2(model, exclude = c(4))
#'
#' # include & exclude variables in the selection process
#' ols_step_forward_r2(model, include = c("age"), exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_r2(model, include = c(5), exclude = c(4))
#'
#' # steps
#' ols_step_forward_r2(model, steps = 2)
#' 
#' @family forward selection procedures
#'
#' @export
#'
ols_step_forward_r2 <- function(model, ...) UseMethod("ols_step_forward_r2")

#' @rdname ols_step_forward_r2
#' @export
#'
ols_step_forward_r2.default <- function(model, include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {
  out <- ols_step_forward(model, "rsq", include, exclude, progress, details, steps)
  class(out) <- "ols_step_forward_r2"
  return(out)
}

#' @export
#'
print.ols_step_forward_r2 <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "forward")
  } else {
    print("No variables have been removed from the model.")
  }
}

#' @rdname ols_step_forward_r2
#' @export
#'
plot.ols_step_forward_r2 <- function(x, print_plot = TRUE, details = TRUE, digits = 3, ...) {
  p <- ols_stepaic_plot(x, details, digits)
  if (print_plot) {
    print(p)
  } else {
    return(p)
  }
}

#' Stepwise Adjusted R-Squared forward regression
#'
#' @description
#' Build regression model from a set of candidate predictor variables by
#' entering predictors based on adjusted r-squared, in a stepwise
#' manner until there is no variable left to enter any more.
#'
#' @inheritParams ols_step_forward_aic
#' @inherit ols_step_forward_aic return references
#'
#' @examples
#' # stepwise forward regression
#' model <- lm(y ~ ., data = surgical)
#' ols_step_forward_adj_r2(model)
#'
#' # stepwise forward regression plot
#' k <- ols_step_forward_adj_r2(model)
#' plot(k)
#'
#' # selection metrics
#' k$metrics
#'
#' # extract final model
#' k$model
#'
#' # include or exclude variables
#' # force variable to be included in selection process
#' ols_step_forward_adj_r2(model, include = c("age"))
#'
#' # use index of variable instead of name
#' ols_step_forward_adj_r2(model, include = c(5))
#'
#' # force variable to be excluded from selection process
#' ols_step_forward_adj_r2(model, exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_adj_r2(model, exclude = c(4))
#'
#' # include & exclude variables in the selection process
#' ols_step_forward_adj_r2(model, include = c("age"), exclude = c("liver_test"))
#'
#' # use index of variable instead of name
#' ols_step_forward_adj_r2(model, include = c(5), exclude = c(4))
#'
#' # steps
#' ols_step_forward_adj_r2(model, steps = 2)
#' 
#' @family forward selection procedures
#'
#' @export
#'
ols_step_forward_adj_r2 <- function(model, ...) UseMethod("ols_step_forward_adj_r2")

#' @rdname ols_step_forward_adj_r2
#' @export
#'
ols_step_forward_adj_r2.default <- function(model, include = NULL, exclude = NULL, progress = FALSE, details = FALSE, steps = NULL, ...) {
  out <- ols_step_forward(model, "adjrsq", include, exclude, progress, details, steps)
  class(out) <- "ols_step_forward_adj_r2"
  return(out)
}

#' @export
#'
print.ols_step_forward_adj_r2 <- function(x, ...) {
  if (length(x$metrics$step) > 0) {
    print_step_output(x, "forward")
  } else {
    print("No variables have been removed from the model.")
  }
}

#' @rdname ols_step_forward_adj_r2
#' @export
#'
plot.ols_step_forward_adj_r2 <- function(x, print_plot = TRUE, details = TRUE, digits = 3, ...) {
  p <- ols_stepaic_plot(x, details, digits)
  if (print_plot) {
    print(p)
  } else {
    return(p)
  }
}

model <- lm(y ~ ., data = stepdata)

test_that("output from rsquared forward regression is as expected", {
  expect_snapshot(ols_step_forward_r2(model))
  expect_snapshot(ols_step_forward_r2(model, progress = TRUE))
  expect_snapshot(ols_step_forward_r2(model, details = TRUE))
})

test_that("output from adjusted rsquared forward regression is as expected", {
  expect_snapshot(ols_step_forward_adj_r2(model))
  expect_snapshot(ols_step_forward_adj_r2(model, progress = TRUE))
  expect_snapshot(ols_step_forward_adj_r2(model, details = TRUE))
})

test_that("output from rsquared backward regression is as expected", {
  expect_snapshot(ols_step_backward_r2(model))
  expect_snapshot(ols_step_backward_r2(model, progress = TRUE))
  expect_snapshot(ols_step_backward_r2(model, details = TRUE))
})

test_that("output from adjusted rsquared backward regression is as expected", {
  expect_snapshot(ols_step_backward_adj_r2(model))
  expect_snapshot(ols_step_backward_adj_r2(model, progress = TRUE))
  expect_snapshot(ols_step_backward_adj_r2(model, details = TRUE))
})

test_that("output from rsquared both direction regression is as expected", {
  expect_snapshot(ols_step_both_r2(model))
  expect_snapshot(ols_step_both_r2(model, progress = TRUE))
  expect_snapshot(ols_step_both_r2(model, details = TRUE))
})

test_that("output from adjusted rsquared both direction regression is as expected", {
  expect_snapshot(ols_step_both_adj_r2(model))
  expect_snapshot(ols_step_both_adj_r2(model, progress = TRUE))
  expect_snapshot(ols_step_both_adj_r2(model, details = TRUE))
})

model <- lm(y ~ ., data = stepdata)

test_that("output from forward regression matches the expected output", {  
  k <- ols_step_forward_r2(model)
  expect_equal(k$metrics$step, 1:6)
  expect_equal(k$metrics$variable, c("x6", "x1", "x3", "x2", "x4", "x5"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 2), c(33473.30, 32931.76, 31912.72, 29304.30, 29302.80, 29304.73), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.691, 0.700, 0.715, 0.750, 0.750, 0.750), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.691, 0.700, 0.715, 0.750, 0.750, 0.750), ignore_attr = TRUE)
})

test_that("output from forward regression matches the expected output when variables are locked in", {

  k <- ols_step_forward_adj_r2(model, include = "x5")
  expect_equal(k$metrics$step, 1:5)
  expect_equal(k$metrics$variable, c("x6", "x1", "x3", "x2", "x4"), ignore_attr = TRUE)

  k <- ols_step_forward_adj_r2(model, include = c(5))
  expect_equal(k$metrics$step, 1:5)
  expect_equal(k$metrics$variable, c("x6", "x1", "x3", "x2", "x4"), ignore_attr = TRUE)
})

test_that("output from forward regression matches the expected output when variables are locked out", {

  k <- ols_step_forward_adj_r2(model, include = "x6")
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("x1", "x3", "x2", "x4"), ignore_attr = TRUE)

  k <- ols_step_forward_adj_r2(model, include = c(6))
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("x1", "x3", "x2", "x4"), ignore_attr = TRUE)
})

test_that("output from backward regression matches the expected output", {  
  k <- ols_step_backward_adj_r2(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("x6", "x5"), ignore_attr = TRUE)
})

test_that("output from stepwise regression matches the expected output", {  
  k <- ols_step_both_adj_r2(model)
  expect_equal(k$metrics$step, 1:6)
  expect_equal(k$metrics$variable, c("x6", "x1", "x3", "x2", "x6", "x4"), ignore_attr = TRUE)
  expect_equal(k$metrics$method, c("addition", "addition", "addition", "addition", "removal", "addition"), ignore_attr = TRUE)
})


test_that("step rsquared returns the appropriate error", {
  model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
  expect_error(ols_step_forward_r2(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_forward_r2(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_forward_r2(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_forward_r2(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
  expect_error(ols_step_backward_r2(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_backward_r2(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_backward_r2(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_backward_r2(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
  expect_error(ols_step_both_r2(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_both_r2(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_both_r2(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_both_r2(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

model <- lm(mpg ~ disp + hp + wt, data = mtcars)

test_that("residual histogram plot is as expected", {
  skip_on_cran()
  # skip_on_ci()
  p <- ols_plot_resid_hist(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("ggplot2 histogram", p$plot)
})

test_that("hadi plot is as expected", {
  skip_on_cran()
  p <- ols_plot_hadi(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("hadi plot", p$plot)
})

test_that("observed vs predicted plot is as expected", {
  # skip_if(getRversion() > '4.0.3')
  skip_on_cran()
  p <- ols_plot_obs_fit(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("ovsp plot", p$plot)
})

test_that("potential residual plot is as expected", {
  skip_on_cran()
  p <- ols_plot_resid_pot(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("potential residual plot", p$plot)
})

test_that("residual box plot is as expected", {
  skip_on_cran()
  p <- ols_plot_resid_box(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("residual box plot", p$plot)
})

test_that("residual fit spread plot 1 is as expected", {
  skip_on_cran()
  p <- ols_plot_resid_spread(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("residual fit spread plot", p$plot)
})

test_that("residual fit spread plot 2 is as expected", {
  skip_on_cran()
  p <- ols_plot_fm(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("residual fit spread plot 2", p$plot)
})

test_that("residual qq plot is as expected", {
  # skip_if(getRversion() > '4.0.3')
  skip_on_cran()
  p <- ols_plot_resid_qq(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("residual qq plot", p$plot)
})

test_that("residual vs fitted plot is as expected", {
  skip_on_cran()
  p <- ols_plot_resid_fit(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("residual vs fitted plot", p$plot)
})

test_that("residual vs regressor plot is as expected", {
  skip_on_cran()
  p <- ols_plot_resid_regressor(model, 'drat', print_plot = FALSE)
  vdiffr::expect_doppelganger("residual vs regressor plot", p$plot)
})

test_that("cooks d bar plot is as expected", {
  skip_on_cran()
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)

  p <- ols_plot_cooksd_bar(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar plot", p$plot)

  p1 <- ols_plot_cooksd_bar(model, threshold = 0.2, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar plot threshold", p1$plot)
  
  p2 <- ols_plot_cooksd_bar(model, type = 2, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar plot type 2", p2$plot)

  p3 <- ols_plot_cooksd_bar(model, type = 3, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar plot type 3", p3$plot)

  p4 <- ols_plot_cooksd_bar(model, type = 4, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar plot type 4", p4$plot)

  p5 <- ols_plot_cooksd_bar(model, type = 5, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar plot type 5", p5$plot)
})

test_that("cooks d bar chart is as expected", {
  skip_on_cran()
  p <- ols_plot_cooksd_chart(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar chart", p$plot)

  p1 <- ols_plot_cooksd_chart(model, threshold = 0.2, print_plot = FALSE)
  vdiffr::expect_doppelganger("cooks d bar chart threshold", p1$plot)
})

test_that("dffits plot is as expected", {
  skip_on_cran()
  p <- ols_plot_dffits(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("dffits plot", p$plot)
})

test_that("dffits plot is as expected when not using size adjusted cutoff", {
  skip_on_cran()
  p <- ols_plot_dffits(model, size_adj_threshold = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("dffits cutoff plot", p$plot)
})

test_that("deleted studentized residual vs fitted plot is as expected", {
  skip_on_cran()
  p <- ols_plot_resid_stud_fit(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("dsrvsp plot", p$plot)
})

test_that("residual vs regressor shiny plot is as expected", {
  skip_on_cran()
  p <- rvsr_plot_shiny(model, mtcars, "drat", print_plot = FALSE)
  vdiffr::expect_doppelganger("residual vs regressor shiny plot", p$plot)
})

test_that("residual fit spread plot is as expected", {
  skip_on_cran()
  p <- ols_plot_resid_fit_spread(model, print_plot = TRUE)
  vdiffr::expect_doppelganger("fm_plot", p$fm_plot)
  vdiffr::expect_doppelganger("rsd_plot", p$rsd_plot)
})

test_that("response profile plot is as expected", {
  skip_on_cran()
  p <- ols_plot_response(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("resp viz dot plot", p$dot_plot)
  vdiffr::expect_doppelganger("resp viz trend plot", p$trend_plot)
  vdiffr::expect_doppelganger("resp viz histogram", p$histogram)
  vdiffr::expect_doppelganger("resp viz boxplot", p$boxplot)
})

test_that("stepAIC backward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = surgical)
  p <- plot(ols_step_backward_aic(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("stepaic backward regression plot", p$plot)

  p1 <- plot(ols_step_backward_aic(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("stepaic backward regression plot without details", p1$plot)
})

test_that("stepAIC forward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = surgical)
  p <- plot(ols_step_forward_aic(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("stepaic forward regression plot", p$plot)

  p1 <- plot(ols_step_forward_aic(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("stepaic forward regression plot without details", p1$plot)
})

test_that("stepAIC both direction regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = surgical)
  
  p1 <- plot(ols_step_both_aic(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("stepaic both regression plot details", p1$plot)

  p2 <- plot(ols_step_both_aic(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("stepaic both regression plot", p2$plot)
})

test_that("added variable plot is as expected", {
  skip_on_cran()
  p <- ols_plot_added_variable(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("avplot_1", p[[1]])
  vdiffr::expect_doppelganger("avplot_2", p[[2]])
  vdiffr::expect_doppelganger("avplot_3", p[[3]])
})

test_that("all possible regression plots are as expected", {
  # skip_if(getRversion() > '4.0.3')
  skip_on_cran()
  model <- lm(mpg ~ disp + hp, data = mtcars)
  k     <- ols_step_all_possible(model)
  p     <- plot(k, print_plot = FALSE)
  vdiffr::expect_doppelganger("allplot_1", p$plot_1)
  vdiffr::expect_doppelganger("allplot_2", p$plot_2)
  vdiffr::expect_doppelganger("allplot_3", p$plot_3)
  vdiffr::expect_doppelganger("allplot_4", p$plot_4)
  vdiffr::expect_doppelganger("allplot_5", p$plot_5)
  vdiffr::expect_doppelganger("allplot_6", p$plot_6)
})

test_that("best subset regression plots are as expected", {
  skip_on_cran()
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  k     <- ols_step_best_subset(model)
  p     <- plot(k, print_plot = FALSE)
  vdiffr::expect_doppelganger("bestplot_1", p$plot_1)
  vdiffr::expect_doppelganger("bestplot_2", p$plot_2)
  vdiffr::expect_doppelganger("bestplot_3", p$plot_3)
  vdiffr::expect_doppelganger("bestplot_4", p$plot_4)
  vdiffr::expect_doppelganger("bestplot_5", p$plot_5)
  vdiffr::expect_doppelganger("bestplot_6", p$plot_6)
})

test_that("dfbetas plot is as expected", {
  skip_on_cran()
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  p <- ols_plot_dfbetas(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("dfbetas_1", p$plots[[1]])
  vdiffr::expect_doppelganger("dfbetas_2", p$plots[[2]])
  vdiffr::expect_doppelganger("dfbetas_3", p$plots[[3]])
  vdiffr::expect_doppelganger("dfbetas_4", p$plots[[4]])
  vdiffr::expect_doppelganger("dfbetas_5", p$plots[[5]])
})

test_that("diagnostics panel is as expected", {
  # skip_if(getRversion() > '4.0.3')
  skip_on_cran()
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  p <- ols_plot_diagnostics(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("diag_1", p[[1]])
  vdiffr::expect_doppelganger("diag_2", p[[2]])
  vdiffr::expect_doppelganger("diag_3", p[[3]])
  vdiffr::expect_doppelganger("diag_4", p[[4]])
  vdiffr::expect_doppelganger("diag_5", p[[5]])
  vdiffr::expect_doppelganger("diag_6", p[[6]])
  vdiffr::expect_doppelganger("diag_7", p[[7]])
  vdiffr::expect_doppelganger("diag_8", p[[8]])
  vdiffr::expect_doppelganger("diag_9", p[[9]])
  vdiffr::expect_doppelganger("diag_10", p[[10]])
})


test_that("fitted line plot is as expected", {
  # skip_if(getRversion() > '4.0.3')
  skip_on_cran()
  p <- ols_plot_reg_line(mtcars$mpg, mtcars$disp, print_plot = FALSE)
  vdiffr::expect_doppelganger("reg_line_plot", p$plot)
})

test_that("residual plus component plot is as expected", {
  skip_on_cran()
  model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
  p     <- ols_plot_comp_plus_resid(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("rpcplot_1", p[[1]])
  vdiffr::expect_doppelganger("rpcplot_2", p[[2]])
  vdiffr::expect_doppelganger("rpcplot_3", p[[3]])
  vdiffr::expect_doppelganger("rpcplot_4", p[[4]])
})

test_that("rstud vs lev plot is as expected", {
  skip_on_cran()
  model <- lm(read ~ write + math + science, data = hsb)
  p     <- ols_plot_resid_lev(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("rslev_1", p$plot)
})

test_that("standardized residual plot is as expected", {
  skip_on_cran()
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  p     <- ols_plot_resid_stand(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("stanres_1", p$plot)
})

test_that("studentized residual plot is as expected", {
  skip_on_cran()
  model <- lm(mpg ~ disp + hp + wt, data = mtcars)
  p     <- ols_plot_resid_stud(model, print_plot = FALSE)
  vdiffr::expect_doppelganger("studres_1", p$plot)
})

test_that("stepwise backward regression plots are as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = surgical)
  k     <- ols_step_backward_p(model)
  p     <- plot(k, print_plot = FALSE)
  vdiffr::expect_doppelganger("step_back_1", p$plot_1)
  vdiffr::expect_doppelganger("step_back_2", p$plot_2)
  vdiffr::expect_doppelganger("step_back_3", p$plot_3)
  vdiffr::expect_doppelganger("step_back_4", p$plot_4)
})

test_that("stepwise forward regression plots are as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = surgical)
  k     <- ols_step_forward_p(model)
  p     <- plot(k, print_plot = FALSE)
  vdiffr::expect_doppelganger("step_forward_1", p$plot_1)
  vdiffr::expect_doppelganger("step_forward_2", p$plot_2)
  vdiffr::expect_doppelganger("step_forward_3", p$plot_3)
  vdiffr::expect_doppelganger("step_forward_4", p$plot_4)
})

test_that("stepwise both regression plots are as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = surgical)
  k     <- ols_step_both_p(model)
  p     <- plot(k, print_plot = FALSE)
  vdiffr::expect_doppelganger("step_both_1", p$plot_1)
  vdiffr::expect_doppelganger("step_both_2", p$plot_2)
  vdiffr::expect_doppelganger("step_both_3", p$plot_3)
  vdiffr::expect_doppelganger("step_both_4", p$plot_4)
})


test_that("rsquared forward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_forward_r2(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("rsquared forward regression plot details", p1$plot)

  p2 <- plot(ols_step_forward_r2(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("rsquared forward regression plot", p2$plot)
})

test_that("adjusted rsquared forward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_forward_adj_r2(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("adjusted rsquared forward regression plot details", p1$plot)

  p2 <- plot(ols_step_forward_adj_r2(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("adjusted rsquared forward regression plot", p2$plot)
})

test_that("adjusted rsquared backward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_backward_adj_r2(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("adjusted rsquared backward regression plot details", p1$plot)

  p2 <- plot(ols_step_backward_adj_r2(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("adjusted rsquared backward regression plot", p2$plot)
})

test_that("rsquared both direction regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_both_r2(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("rsquared both direction regression plot details", p1$plot)

  p2 <- plot(ols_step_both_r2(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("rsquared both direction regression plot", p2$plot)
})

test_that("adjusted rsquared both direction regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_both_adj_r2(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("adjusted rsquared both direction regression plot details", p1$plot)

  p2 <- plot(ols_step_both_adj_r2(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("adjusted rsquared both direction regression plot", p2$plot)
})

test_that("sbc forward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_forward_sbc(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("sbc forward regression plot details", p1$plot)

  p2 <- plot(ols_step_forward_sbc(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("sbc forward regression plot", p2$plot)
})

test_that("sbc backward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_backward_sbc(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("sbc backward regression plot details", p1$plot)

  p2 <- plot(ols_step_backward_sbc(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("sbc backward regression plot", p2$plot)
})

test_that("sbc both direction regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_both_sbc(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("sbc both direction regression plot details", p1$plot)

  p2 <- plot(ols_step_both_sbc(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("sbc both direction regression plot", p2$plot)
})

test_that("sbic forward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_forward_sbic(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("sbic forward regression plot details", p1$plot)

  p2 <- plot(ols_step_forward_sbic(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("sbic forward regression plot", p2$plot)
})

test_that("sbic backward regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_backward_sbic(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("sbic backward regression plot details", p1$plot)

  p2 <- plot(ols_step_backward_sbic(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("sbic backward regression plot", p2$plot)
})

test_that("sbic both direction regression plot is as expected", {
  skip_on_cran()
  model <- lm(y ~ ., data = stepdata)
  
  p1 <- plot(ols_step_both_sbic(model), print_plot = FALSE)
  vdiffr::expect_doppelganger("sbic both direction regression plot details", p1$plot)

  p2 <- plot(ols_step_both_sbic(model), details = FALSE, print_plot = FALSE)
  vdiffr::expect_doppelganger("sbic both direction regression plot", p2$plot)
})

model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)

test_that("output from stepaic_forward matches the expected output", {  
  k <- ols_step_forward_aic(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "hp"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 3), c(166.029, 156.652), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.753, 0.827), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.745, 0.815), ignore_attr = TRUE)
})

test_that("output from stepaic_forward matches when steps is specified", {  
  model <- lm(y ~ ., data = surgical)
  k <- ols_step_forward_aic(model, steps = 2)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("liver_test", "alc_heavy"), ignore_attr = TRUE)
})

test_that("output from stepaic_forward matches the expected output when variables are locked in", {

  k <- ols_step_forward_aic(model, include = c("disp"))
  expect_equal(k$metrics$step, 1:3)
  expect_equal(k$metrics$variable, c("wt", "hp", "drat"), ignore_attr = TRUE)

  h <- ols_step_forward_aic(model, include = c(1))
  expect_equal(h$metrics$step, 1:3)
  expect_equal(h$metrics$variable, c("wt", "hp", "drat"), ignore_attr = TRUE)
})

test_that("output from stepaic_forward matches the expected output when variables are locked out", {

  k <- ols_step_forward_aic(model, exclude = c("hp"))
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "disp"), ignore_attr = TRUE)

  h <- ols_step_forward_aic(model, exclude = c(2))
  expect_equal(h$metrics$step, 1:2)
  expect_equal(h$metrics$variable, c("wt", "disp"), ignore_attr = TRUE)
})

test_that("output from stepaic_forward matches the expected output when variables are locked in and out", {

  k <- ols_step_forward_aic(model, include = c("disp"), exclude = c("hp"))
  expect_equal(k$metrics$step, 1)
  expect_equal(k$metrics$variable, c("wt"), ignore_attr = TRUE)

  h <- ols_step_forward_aic(model, include = c(1), exclude = c(2))
  expect_equal(h$metrics$step, 1)
  expect_equal(h$metrics$variable, c("wt"), ignore_attr = TRUE)
})


test_that("stepaic_forward returns the appropriate error", {
  expect_error(ols_step_forward_aic(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_forward_aic(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_forward_aic(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_forward_aic(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

test_that("output from stepsbc_forward matches the expected output", {  
  k <- ols_step_forward_sbc(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "hp"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 3), c(166.029, 156.652), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.753, 0.827), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.745, 0.815), ignore_attr = TRUE)
})

test_that("output from stepsbc_forward matches the expected output when variables are locked in", {

  k <- ols_step_forward_sbc(model, include = c("disp"))
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "hp"), ignore_attr = TRUE)

  h <- ols_step_forward_sbc(model, include = c(1))
  expect_equal(h$metrics$step, 1:2)
  expect_equal(h$metrics$variable, c("wt", "hp"), ignore_attr = TRUE)
})

test_that("output from stepsbc_forward matches the expected output when variables are locked out", {

  k <- ols_step_forward_sbc(model, exclude = c("hp"))
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "disp"), ignore_attr = TRUE)

  h <- ols_step_forward_sbc(model, exclude = c(2))
  expect_equal(h$metrics$step, 1:2)
  expect_equal(h$metrics$variable, c("wt", "disp"), ignore_attr = TRUE)
})

test_that("output from stepsbc_forward matches the expected output when variables are locked in and out", {

  k <- ols_step_forward_sbc(model, include = c("disp"), exclude = c("hp"))
  expect_equal(k$metrics$step, 1)
  expect_equal(k$metrics$variable, c("wt"), ignore_attr = TRUE)

  h <- ols_step_forward_sbc(model, include = c(1), exclude = c(2))
  expect_equal(h$metrics$step, 1)
  expect_equal(h$metrics$variable, c("wt"), ignore_attr = TRUE)
})


test_that("stepsbc_forward returns the appropriate error", {
  expect_error(ols_step_forward_sbc(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_forward_sbc(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_forward_sbc(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_forward_sbc(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

test_that("output from stepsbic_forward matches the expected output", {  
  k <- ols_step_forward_sbic(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "hp"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 3), c(166.029, 156.652), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.753, 0.827), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.745, 0.815), ignore_attr = TRUE)
})

test_that("output from stepsbic_forward matches the expected output when variables are locked in", {

  k <- ols_step_forward_sbic(model, include = c("disp"))
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "hp"), ignore_attr = TRUE)

  h <- ols_step_forward_sbic(model, include = c(1))
  expect_equal(h$metrics$step, 1:2)
  expect_equal(h$metrics$variable, c("wt", "hp"), ignore_attr = TRUE)
})

test_that("output from stepsbic_forward matches the expected output when variables are locked out", {

  k <- ols_step_forward_sbic(model, exclude = c("hp"))
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("wt", "disp"), ignore_attr = TRUE)

  h <- ols_step_forward_sbic(model, exclude = c(2))
  expect_equal(h$metrics$step, 1:2)
  expect_equal(h$metrics$variable, c("wt", "disp"), ignore_attr = TRUE)
})

test_that("output from stepsbic_forward matches the expected output when variables are locked in and out", {

  k <- ols_step_forward_sbic(model, include = c("disp"), exclude = c("hp"))
  expect_equal(k$metrics$step, 1)
  expect_equal(k$metrics$variable, c("wt"), ignore_attr = TRUE)

  h <- ols_step_forward_sbic(model, include = c(1), exclude = c(2))
  expect_equal(h$metrics$step, 1)
  expect_equal(h$metrics$variable, c("wt"), ignore_attr = TRUE)
})


test_that("stepsbic_forward returns the appropriate error", {
  expect_error(ols_step_forward_sbic(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_forward_sbic(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_forward_sbic(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_forward_sbic(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

model <- lm(y ~ ., data = surgical)

test_that("output from stepAIC forward regression is as expected", {
  expect_snapshot(ols_step_forward_aic(model))
  expect_snapshot(ols_step_forward_aic(model, progress = TRUE))
  expect_snapshot(ols_step_forward_aic(model, details = TRUE))
  expect_snapshot(ols_step_forward_aic(model, include = "pindex", details = TRUE))
})

test_that("output from stepSBC forward regression is as expected", {
  expect_snapshot(ols_step_forward_sbc(model))
  expect_snapshot(ols_step_forward_sbc(model, progress = TRUE))
  expect_snapshot(ols_step_forward_sbc(model, details = TRUE))
  expect_snapshot(ols_step_forward_sbc(model, include = "pindex", details = TRUE))
})

test_that("output from stepSBIC forward regression is as expected", {
  expect_snapshot(ols_step_forward_sbic(model))
  expect_snapshot(ols_step_forward_sbic(model, progress = TRUE))
  expect_snapshot(ols_step_forward_sbic(model, details = TRUE))
  expect_snapshot(ols_step_forward_sbic(model, include = "pindex", details = TRUE))
})

#' Ordinary least squares regression
#'
#' @description Ordinary least squares regression.
#'
#' @param object An object of class "formula" (or one that can be coerced to
#'   that class): a symbolic description of the model to be fitted or class
#'   \code{lm}.
#'
#' @param ... Other inputs.
#'
#' @return \code{ols_regress} returns an object of class \code{"ols_regress"}.
#' An object of class \code{"ols_regress"} is a list containing the following
#' components:
#'
#' \item{r}{square root of rsquare, correlation between observed and predicted values of dependent variable}
#' \item{rsq}{coefficient of determination or r-square}
#' \item{adjr}{adjusted rsquare}
#' \item{rmse}{root mean squared error}
#' \item{cv}{coefficient of variation}
#' \item{mse}{mean squared error}
#' \item{mae}{mean absolute error}
#' \item{aic}{akaike information criteria}
#' \item{sbc}{bayesian information criteria}
#' \item{sbic}{sawa bayesian information criteria}
#' \item{prsq}{predicted rsquare}
#' \item{error_df}{residual degrees of freedom}
#' \item{model_df}{regression degrees of freedom}
#' \item{total_df}{total degrees of freedom}
#' \item{ess}{error sum of squares}
#' \item{rss}{regression sum of squares}
#' \item{tss}{total sum of squares}
#' \item{rms}{regression mean square}
#' \item{ems}{error mean square}
#' \item{f}{f statistis}
#' \item{p}{p-value for \code{f}}
#' \item{n}{number of predictors including intercept}
#' \item{betas}{betas; estimated coefficients}
#' \item{sbetas}{standardized betas}
#' \item{std_errors}{standard errors}
#' \item{tvalues}{t values}
#' \item{pvalues}{p-value of \code{tvalues}}
#' \item{df}{degrees of freedom of \code{betas}}
#' \item{conf_lm}{confidence intervals for coefficients}
#' \item{title}{title for the model}
#' \item{dependent}{character vector; name of the dependent variable}
#' \item{predictors}{character vector; name of the predictor variables}
#' \item{mvars}{character vector; name of the predictor variables including intercept}
#' \item{model}{input model for \code{ols_regress}}
#'
#' @section Interaction Terms:
#' If the model includes interaction terms, the standardized betas
#' are computed after scaling and centering the predictors.
#'
#' @references https://www.ssc.wisc.edu/~hemken/Stataworkshops/stdBeta/Getting%20Standardized%20Coefficients%20Right.pdf
#'
#' @examples
#' ols_regress(mpg ~ disp + hp + wt, data = mtcars)
#'
#' # if model includes interaction terms set iterm to TRUE
#' ols_regress(mpg ~ disp * wt, data = mtcars, iterm = TRUE)
#'
#' @export
#'
ols_regress <- function(object, ...) UseMethod("ols_regress")

#' @export
#'
ols_regress.default <- function(object, data, conf.level = 0.95,
                                iterm = FALSE, title = "model", ...) {
  if (missing(data)) {
    stop("data missing", call. = FALSE)
  }

  if (!is.numeric(conf.level)) {
    stop("conf.level must be numeric", call. = FALSE)
  }

  if ((conf.level < 0) | (conf.level > 1)) {
    stop("conf.level must be between 0 and 1", call. = FALSE)
  }

  check_logic(iterm)

  if (!is.character(title)) {
    stop(paste(title, "is not a string, Please specify a string as title."), call. = FALSE)
  }

  # detect if model formula includes interaction terms
  if (inherits(object, "formula")) {
    detect_iterm <- grepl(object, pattern = "\\*")[3]
  } else {
    detect_iterm <-  grepl(object, pattern = "\\*")
  }

  # set interaction to TRUE if formula contains interaction terms
  if (detect_iterm) {
    iterm <- TRUE
  }

  result <- reg_comp(object, data, conf.level, iterm, title)
  class(result) <- "ols_regress"
  return(result)
}

#' @rdname ols_regress
#' @export
#'
ols_regress.lm <- function(object, ...) {

  check_model(object)

  formula <- formula(object)
  data    <- object$model

  ols_regress.default(object = formula, data = data)

}

#' @export
#'
print.ols_regress <- function(x, ...) {
  print_reg(x)
}

model <- ols_regress(mpg ~ disp + hp + wt + drat + qsec, data = mtcars)

test_that("regress returns all the model validation metrics", {
  sb  <- c(disp = 0.179, hp = -0.234, wt = -0.712, drat = 0.179, qsec = 0.190)
  lm1 <- c("(Intercept)" = -6.00372864, disp = -0.01427933, hp = -0.05201291, wt = -6.94137515, drat = -0.67585793, qsec = -0.30404593)
  lm2 <- c("(Intercept)" = 39.07086783, disp = 0.03171968, hp = 0.01081675, wt = -1.82955263, drat = 4.70740704, qsec = 1.58434573)

  expect_equal(model$r, 0.9213657)
  expect_equal(round(model$cv, 3), 12.732)
  expect_equal(round(model$mae, 3), 1.84)
  expect_equal(round(model$prsq, 4), 0.7666)
  expect_equal(round(model$sbetas, 3), sb)
  expect_equal(model$conf_lm[, 1], lm1)
  expect_equal(model$conf_lm[, 2], lm2)
})

test_that("If model includes interaction terms, ols_regress scales and centers\n          predictors before computing standardized betas", {
  model_inter <- ols_regress(price ~ weight * displacement, data = auto, iterm = TRUE)
  actual   <- round(model_inter$sbetas, 2)
  expected <- c(weight = 0.57, displacement = -0.15, `weight:displacement` = 0.38)
  expect_equal(actual, expected, ignore_attr = TRUE)
})

test_that("ols_regress returns error messages", {
  expect_error(ols_regress(mpg ~ disp + hp + wt), "data missing")
  expect_error(ols_regress(mpg ~ disp + hp + wt, data = mtcars, conf.level = "0.95"), "conf.level must be numeric")
  expect_error(ols_regress(mpg ~ disp + hp + wt, data = mtcars, conf.level = 1.95), "conf.level must be between 0 and 1")
  expect_error(ols_regress(mpg ~ disp + hp + wt, data = mtcars, title = 1), "1 is not a string, Please specify a string as title.")
})

#' Simple linear regression line
#'
#' @description
#' Plot to demonstrate that the regression line always passes  through mean of
#' the response and predictor variables.
#'
#' @param response Response variable.
#' @param predictor Predictor variable.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' ols_plot_reg_line(mtcars$mpg, mtcars$disp)
#'
#' @importFrom ggplot2 labs geom_smooth
#'
#' @export
#'
ols_plot_reg_line <- function(response, predictor, print_plot = TRUE) {

  data <- ols_prep_regline_data(response, predictor)

  p <-
    ggplot(data$d, aes(x = x, y = y)) +
    geom_point(fill = "blue") +
    geom_point(data = data$d2, aes(x = x, y = y), color = "red", shape = 2, size = 3) +
    geom_smooth(method = "lm", se = FALSE)

  p <-
    p +
    labs(title = "Regression Line") +
    xlab(paste0(data$preds)) +
    ylab(paste0(data$resp))

  if (print_plot) {
    print(p)
  } else {
    return(p)
  }

}

ols_prep_regline_data <- function(response, predictor) {
  resp        <- l(deparse(substitute(response)))
  preds       <- l(deparse(substitute(predictor)))
  m_predictor <- round(mean(predictor), 2)
  m_response  <- round(mean(response), 2)
  d2          <- data.frame(x = m_predictor, y = m_response)
  d           <- data.frame(x = predictor, y = response)

  list(d = d, d2 = d2, preds = preds, resp = resp)
}

check_model <- function(model) {
  if (!any(class(model) == "lm")) {
    model_name <- deparse(substitute(model))
    stop(paste0("`", model_name, "` must be an object of class `lm`."), call. = FALSE)
  }
}

check_order <- function(n, max_order) {
  if (max_order > n) {
    stop("Maximum subset order should be less than or equal to the number of predictors in the specified model.", call. = FALSE)
  }
}

check_logic <- function(logic) {

  lval        <- is.logical(logic)
  logic_name  <- deparse(substitute(logic))

  if (lval != TRUE) {
    stop(paste0("`", logic_name, '` must be a flag (TRUE or FALSE).'), call. = FALSE)
  }

}

check_options <- function(option) {

  default_options <- c("none", "bonferroni", "sidak", "holm")
  valid           <- any(default_options == option)
  option_class    <- class(option)
  option_name     <- deparse(substitute(option))

  if (valid != TRUE) {
    err_message <- paste0("`", option_name, "` can take the following values only:\n* `none`\n* `bonferroni`\n* `sidak`\n* `holm`")
    stop(err_message, call. = FALSE)
  }
}


check_values <- function(value, lower, upper) {

  valid       <- (value >= lower && value <= upper)
  value_class <- class(value)
  value_name  <- deparse(substitute(value))

  if (valid != TRUE) {
    err_message <- paste0("`", value_name, "` must be between ", lower, " and ", upper, ".")
    stop(err_message, call. = FALSE)
  }

}

check_npredictors <- function(model, min) {

  n <- length(coefficients(model))

  if (n < min) {
    err_message <- paste0("For stepwise selection, model should include at least `", min, "` predictors.")
    stop(err_message, call. = FALSE)
  }

}


check_lfit <- function(model) {

  n     <- length(coefficients(model))
  preds <- n - 1

  if (n > 2) {
    err_message <- paste0("Lack of fit F test is available only for models with a single predictor.")
    stop(err_message, call. = FALSE)
  }

}

check_modelvars <- function(model, vars) {

  fmla  <- deparse(formula(model))
  k     <- vars %in% names(model$coefficients)
  wvars <- vars[!k]
  nvars <- length(wvars)

  if (nvars > 0) {
    err_message <- paste0("Model `(", fmla, ")` should include the predictor(s) `", paste0(wvars, collapse = ", "), "`.")
    stop(err_message, call. = FALSE)
  }

}

#'  Response variable profile
#'
#' Panel of plots to explore and visualize the response variable.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_response(model)
#'
#' @importFrom stats model.frame model.response
#' @importFrom ggplot2 geom_dotplot geom_histogram
#'
#' @export
#'
ols_plot_response <- function(model, print_plot = TRUE) {

  check_model(model)

  d  <- ols_response_data(model)
  p1 <- ols_response_dotplot(d$d1, d$nam)
  p2 <- ols_response_trend(d$d2, d$nam)
  p3 <- ols_response_hist(d$d1, d$nam)
  p4 <- ols_response_box(d$d1, d$nam)

  if (print_plot) {
    grid.arrange(p1, p2, p3, p4, ncol = 2, top = "Response Diagnostics")
  } else {
    return(
      list(dot_plot   = p1,
           trend_plot = p2,
           histogram  = p3,
           boxplot    = p4)
      )
  }

}

ols_response_data <- function(model) {
  nam  <- names(model.frame(model))
  pred <- model.response(model.frame(model))
  xval <- seq_len(length(pred))
  d1   <- data.frame(x = pred)
  d2   <- data.frame(x = xval, y = pred)

  list(nam = nam, d1 = d1, d2 = d2)
}

ols_response_dotplot <- function(d, nam) {
  ggplot(d, aes(x = x)) +
    geom_dotplot(binwidth = 1, fill = "blue") +
    xlab(nam[1]) +
    ggtitle(paste("Dot Plot of", nam[1]))
}

ols_response_trend <- function(d, nam) {
  ggplot(d, aes(x = x, y = y)) +
    geom_point(color = "blue") +
    geom_line(color = "blue") +
    xlab("Observation") +
    ylab(nam[1]) +
    ggtitle(paste("Trend Plot of", nam[1]))
}

ols_response_hist <- function(d, nam) {
  ggplot(d, aes(x = x)) +
    geom_histogram(bins = 5, color = "black", fill = "blue") +
    xlab(nam[1]) +
    ggtitle(paste("Histogram of", nam[1]))
}

ols_response_box <- function(d, nam) {
  ggplot(d, aes(x = factor(0), y = x)) +
    geom_boxplot(fill = "blue") +
    xlab("") +
    ylab(nam[1]) +
    ggtitle(paste("Boxplot of", nam[1])) +
    theme(axis.text.x = element_blank())
}

#' @importFrom stats model.matrix confint.lm
#' @importFrom stats as.formula deviance
reg_comp <- function(formula, data, conf.level = 0.95, iterm, title = "model") {

  model      <- lm(formula = formula, data = data)
  nam        <- names(model.frame(model))
  response   <- nam[1]
  predictors <- nam[-1]
  output     <- summary(model)
  anovam     <- anova(model)
  dep        <- model.frame(model)[1]
  rsq        <- output$r.squared
  r          <- sqrt(rsq)
  adjr       <- output$adj.r.squared
  rmse       <- sqrt(mean(model$residuals ^ 2))
  cv         <- (output$sigma / mean(dep[[1]])) * 100
  mae        <- mean(abs(residuals(model)))
  aic        <- ols_aic(model)
  sbc        <- ols_sbc(model)
  sbic       <- ols_sbic(model, model)
  prsq       <- ols_pred_rsq(model)
  n          <- length(anovam$Df)
  mse        <- mean(model$residuals ^ 2)
  error_df   <- anovam$Df[n]
  model_df   <- sum(anovam$Df) - error_df
  total_df   <- sum(anovam$Df)
  ess        <- anovam$`Sum Sq`[n]
  tss        <- sum(anovam$`Sum Sq`)
  rss        <- tss - ess
  rms        <- rss / model_df
  ems        <- ess / error_df
  f          <- rms / ems
  p          <- pf(f, model_df, error_df, lower.tail = F)

  # standardised betas
  if (iterm) {

    data_scaled <- data

    is_num    <- unname(unlist(lapply(data_scaled, is.numeric)))
    num_scale <- as.data.frame(lapply(data_scaled[, is_num], scale))
    newdata   <- cbind(data_scaled[, !is_num], num_scale)
    model2    <- lm(formula, data = newdata)
    output2   <- summary(model2)
    b         <- output2$coef[-1, 1]
    g         <- as.data.frame(model.matrix(model2)[, -1])
    sx        <- sapply(g, sd)
    sy        <- sapply(model2$model[1], sd)
    sbeta     <- b * sx / sy
    sbetas    <- sbeta

  } else {

    b        <- output$coef[-1, 1]
    g        <- as.data.frame(model.matrix(model)[, -1])
    sx       <- sapply(g, sd)
    sy       <- sapply(model$model[1], sd)
    sbeta    <- b * sx / sy
    sbetas   <- sbeta

  }

  betas      <- coefficients(model)
  std_errors <- output$coefficients[, 2]
  tvalues    <- output$coefficients[, 3]
  pvalues    <- output$coefficients[, 4]
  df         <- rep(1, n)
  conf_lm    <- confint.lm(model, level = conf.level)
  mvars      <- names(model$coefficients)


  result <- list(
    r          = r,
    rsq        = rsq,
    adjr       = adjr,
    rmse       = rmse,
    cv         = cv,
    mse        = mse,
    mae        = mae,
    aic        = aic,
    sbc        = sbc,
    sbic       = sbic,
    prsq       = prsq,
    error_df   = error_df,
    model_df   = model_df,
    total_df   = total_df,
    ess        = ess,
    rss        = rss,
    tss        = tss,
    rms        = rms,
    ems        = ems,
    f          = f,
    p          = p,
    n          = n,
    betas      = betas,
    sbetas     = sbetas,
    std_errors = std_errors,
    tvalues    = tvalues,
    pvalues    = pvalues,
    df         = df,
    conf_lm    = conf_lm,
    title      = title,
    dependent  = response,
    predictors = predictors,
    mvars      = mvars,
    model      = model
  )

  return(result)
}

check_inputs <- function(model, include, exclude, progress, details) {

  check_model(model)
  check_npredictors(model, 2)

  indterms <- coeff_names(model)
  check_terms(include, indterms)
  check_terms(exclude, indterms, include = FALSE)

  check_logic(progress)
  check_logic(details)

}

check_terms <- function(clude, indterms, include = TRUE) {

  if (!include) {
    process <- "excluded"
  } else {
    process <- "included"
  }

  if (is.character(clude)) {
    npm <- clude %in% indterms
    if (!all(npm)) {
      stop(
        paste0(
          paste(clude[!npm], collapse = ", "),
          " not part of the model and hence cannot be forcibly ", process, ". Please verify the variable names."),
        call. = FALSE)
    }
  }

  lenterms <- length(indterms)

  if (is.numeric(clude)) {
    if (any(clude > lenterms)) {
      stop(paste0("Index of variable to be ", process, " should be between 1 and ", lenterms, "."), call. = FALSE)
    }
  }

}

ols_base_model <- function(include, response, data) {
  if (is.null(include)) {
    lm(paste(response, "~", 1), data = data)
  } else {
    lm(paste(response, "~", paste(include, collapse = " + ")), data = data)
  }
}

ols_candidate_terms <- function(cterms = NULL, direction = c("forward", "backward", "both")) {
  method <- match.arg(direction)

  if (method == "forward") {
    title <- "Forward Selection Method"
    width <- 24
  } else if (method == "backward") {
    title <- "Backward Elimination Method"
    width <- 27
  } else {
    title <- "Stepwise Selection Method"
    width <- 25
  }

  cat(format(title, justify = "left", width = width), "\n")
  cat(rep("-", width), sep = "", "\n\n")
  cat(format("Candidate Terms:", justify = "left", width = 16), "\n\n")
  for (i in seq_len(length(cterms))) {
    cat(paste0(i, ". ", cterms[i]), "\n")
    if (interactive()) {
      Sys.sleep(0.3)
    }
  }
  cat("\n")

}

ols_base_model_stats <- function(response, include, criteria, aic) {

  mat  <- switch(criteria,
    aic    = "AIC    ",
    sbc    = "SBC    ",
    sbic   = "SBIC   ",
    rsq    = "R2     ",
    adjrsq = "Adj. R2")

  cat("\n")
  cat("Step     => 0", "\n")
  if (interactive()) {
    Sys.sleep(0.3)
  }

  if (is.null(include)) {
    cat("Model    =>", paste(response, "~", 1, "\n"))
  } else {
    cat("Model    =>", paste(response, "~", paste(include, collapse = " + "), "\n"))
  }

  if (interactive()) {
    Sys.sleep(0.3)
  }

  cat(paste0(mat, "  =>"), aic, "\n\n")

  if (interactive()) {
    Sys.sleep(0.3)
  }
  cat("Initiating stepwise selection...", "\n\n")

}

ols_progress_init <- function(direction = c("forward", "backward", "both")) {

  method <- match.arg(direction)

  if (method == "forward") {
    display <- "Entered:"
  } else if (method == "backward") {
    display <- "Removed:"
  } else {
    display <- "Added/Removed:"
  }

  cat("\n")
  cat(paste0("Variables ", display), "\n\n")

}

ols_progress_display <- function(preds, direction = c("others", "both"), type = c("added", "removed")) {

  method <- match.arg(direction)
  base   <- paste("=>", tail(preds, n = 1))

  if (method == "others") {
    cat(base, "\n")
  } else {
    cat(paste(base, type), "\n")
  }

  if (interactive()) {
    Sys.sleep(0.3)
  }

}

ols_stepwise_details <- function(step, rpred, preds, response, aic, type = c("added", "removed"), metric = "AIC") {

  cat("Step     =>", step, "\n")
  
  if (interactive()) {
    Sys.sleep(0.3)
  }

  if (type == "added") {
    cat("Added    =>", tail(rpred, n = 1), "\n")
  } else {
    cat("Removed  =>", tail(rpred, n = 1), "\n")
  }

  if (interactive()) {
    Sys.sleep(0.3)
  }

  cat("Model    =>", paste(response, "~", paste(preds, collapse = " + "), "\n"))

  if (interactive()) {
    Sys.sleep(0.3)
  }

  mat  <- switch(metric,
    aic    = "AIC    ",
    sbc    = "SBC    ",
    sbic   = "SBIC   ",
    rsq    = "R2     ",
    adjrsq = "Adj. R2")

  cat(paste0(mat, "  =>"), round(aic, 5), "\n\n")

}

ols_stepwise_metrics <- function(df, metric = c("aic", "sbc", "sbic", "rsq", "adjrsq"), predictors, aics, bics, sbics, rsq, arsq, method = c("add", "remove")) {

  type <- match.arg(metric)

  w1 <- max(nchar("Predictor"), nchar(predictors))
  w2 <- 2
  w3 <- max(nchar("AIC"), nchar(format(round(aics, 3), nsmall = 3)))
  w4 <- max(nchar("SBC"), nchar(format(round(bics, 3), nsmall = 3)))
  w5 <- max(nchar("SBIC"), nchar(format(round(sbics, 3), nsmall = 3)))
  w6 <- max(nchar("R2"), nchar(format(round(rsq, 5), nsmall = 5)))
  w7 <- max(nchar("Adj. R2"), nchar(format(round(arsq, 5), nsmall = 5)))
  w  <- sum(w1, w2, w3, w4, w5, w6, w7, 24)
  ln <- length(aics)

  if (method == "add") {
    cat(format("Table: Adding New Variables", justify = "centre", width = w), "\n")
  } else {
    cat(format("Table: Removing Existing Variables", justify = "centre", width = w), "\n")
  }

  cat(rep("-", w), sep = "", "\n")

  cat(
    fl("Predictor", w1), fs(),
    fc("DF", w2), fs(),
    fc("AIC", w3), fs(),
    fc("SBC", w4), fs(),
    fc("SBIC", w5), fs(),
    fc("R2", w6), fs(),
    fc("Adj. R2", w7), "\n")

  cat(rep("-", w), sep = "", "\n")

  for (i in seq_len(ln)) {
    cat(
      fl(df[i, 1], w1), fs(),
      fg(1, w2), fs(),
      fg(format(round(df[i, 2], 3), nsmall = 3), w3), fs(),
      fg(format(round(df[i, 3], 3), nsmall = 3), w4), fs(),
      fg(format(round(df[i, 4], 3), nsmall = 3), w5), fs(),
      fg(format(round(df[i, 5], 5), nsmall = 5), w6), fs(),
      fg(format(round(df[i, 6], 5), nsmall = 5), w7), "\n")
  }

  cat(rep("-", w), sep = "", "\n\n")

}

ols_stepwise_table <- function(df, predictors, p_val, rsq, arsq, aics) {

  w1 <- max(nchar("Predictor"), nchar(predictors))
  w2 <- max(nchar("Pr(>|t|)"), nchar(format(round(p_val, 5), nsmall = 5)))
  w3 <- max(nchar("R-Squared"), nchar(format(round(rsq, 3), nsmall = 3)))
  w4 <- max(nchar("Adj. R-Squared"), nchar(format(round(arsq, 3), nsmall = 3)))
  w5 <- max(nchar("AIC"), nchar(format(round(aics, 3), nsmall = 3)))
  w  <- sum(w1, w2, w3, w4, w5, 16)
  ln <- length(aics)

  cat(format("Selection Metrics Table", justify = "centre", width = w), "\n")

  cat(rep("-", w), sep = "", "\n")

  cat(
    fl("Predictor", w1), fs(),
    fc("Pr(>|t|)", w2), fs(),
    fc("R-Squared", w3), fs(),
    fc("Adj. R-Squared", w4), fs(),
    fc("AIC", w5), "\n")

  cat(rep("-", w), sep = "", "\n")

  for (i in seq_len(ln)) {
    cat(
      fl(df[i, 1], w1), fs(),
      fg(format(round(df[i, 2], 5), nsmall = 5), w2), fs(),
      fg(format(round(df[i, 3], 3), nsmall = 3), w3), fs(),
      fg(format(round(df[i, 4], 3), nsmall = 3), w4), fs(),
      fg(format(round(df[i, 5], 3), nsmall = 3), w5), "\n")
  }

  cat(rep("-", w), sep = "", "\n\n")

}

ols_stepwise_table_p <- function(df, predictors, p_val) {

  w1 <- max(nchar("Predictor"), nchar(predictors))
  w2 <- max(nchar("Pr(>|t|)"), nchar(format(round(p_val, 5), nsmall = 5)))
  w  <- sum(w1, w2, 4)
  ln <- length(predictors)

  cat(format("Significance Table", justify = "centre", width = w), "\n")

  cat(rep("-", w), sep = "", "\n")

  cat(
    fl("Predictor", w1), fs(),
    fc("Pr(>|t|)", w2), "\n")

  cat(rep("-", w), sep = "", "\n")

  for (i in seq_len(ln)) {
    cat(
      fl(df[i, 1], w1), fs(),
      fg(format(round(df[i, 2], 5), nsmall = 5), w2), "\n")
  }

  cat(rep("-", w), sep = "", "\n\n")

}

ols_stepwise_break <- function(direction = c("forward", "backward", "both")) {

  method <- match.arg(direction)

  if (method == "forward") {
    op <- "added."
  } else if (method == "backward") {
    op <- "removed."
  } else {
    op <- "added or removed."
  }

  cat("\n")
  cat(paste("No more variables to be", op))
  cat("\n")

}

ols_stepwise_vars <- function(preds, direction = c("forward", "backward", "both")) {

  method <- match.arg(direction)

  if (method == "forward" || method == "both") {
    op <- "Selected:"
  } else {
    op <- "Removed:"
  }

  cat("\n")
  if (length(preds) > 0) {
    cat(paste("Variables", op), "\n\n")
    for (i in seq_len(length(preds))) {
      cat(paste("=>", preds[i]), "\n")
      if (interactive()) {
        Sys.sleep(0.3)
      }
    }
  }

}

ols_stepaic_plot <- function(x, details = TRUE, digits = 3) {

  pred <- ols_step_plot_text(x, x$others$direction, details, x$others$criteria, digits)
  data <- ols_stepwise_plot_data(x, pred, x$others$criteria)
  info <- ols_metric_info(x, x$others$direction, x$others$criteria)
  ols_stepaic_plot_build(data$d, data$d2, data$xmin, data$xmax, data$ymin, data$ymax, info, x$others$direction, x$others$criteria)

}

ols_stepwise_plot_data <- function(x, pred, metric = "r2") {

  step <- x$metrics$step
  aic  <- x$metrics[[metric]]
  xmin <- 0
  xmax <- max(step) + 1
  if (min(aic) < 0) {
    ymin <- min(aic) + (min(aic) * 0.05)
    ymax <- max(aic) - (max(aic) * 0.05)
  } else {
    ymin <- min(aic) - (min(aic) * 0.05)
    ymax <- max(aic) + (max(aic) * 0.05)
  }
  d2   <- data.frame(x = step, y = aic, tx = pred)
  d    <- data.frame(a = step, b = aic)

  return(list(d = d, d2 = d2, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax))
}


ols_step_plot_text <- function(x, direction = c("forward", "backward", "both"), details = TRUE, metric = "r2", digits = 3) {

  method <- match.arg(direction)

  if (method == "forward" || method == "backward") {
    if (details) {
      pred <- paste0("[", x$metrics$variable, ", ", format(round(x$metrics[[metric]], digits), nsmall = digits), "]")
    } else {
      pred <- x$metrics$variable
    }
  } else {
    if (details) {
      pred <- ifelse(x$metrics$method == "addition",
                             paste0("[+", x$metrics$variable, ", ", format(round(x$metrics[[metric]], digits), nsmall = digits), "]"),
                             paste0("[-", x$metrics$variable, ", ", format(round(x$metrics[[metric]], digits), nsmall = digits), "]"))
    } else {
      pred <- ifelse(x$metrics$method == "addition",
                               paste0("+", x$metrics$variable),
                               paste0("-", x$metrics$variable))
    }
  }

  return(pred)
}


ols_metric_info <- function(x, direction = c("forward", "backward", "both"), metric = "r2") {

  method <- match.arg(direction)

  if (metric == "r2") {
    met <- "rsq"
  } else if (metric == "adj_r2") {
    met <- "adjr"
  } else {
    met <- metric
  }

  if (method == "forward" || method == "both") {
    the_model <- x$others$base_model
    the_info  <- "Base Model  : "
  } else {
    the_model <- x$others$full_model
    the_info  <- " Full Model  : "
  }

  np <- coeff_names(the_model)
  if (is.null(np)) {
    mi <- null_model_metrics(the_model, x$others$full_model)
  } else {
    mi <- ols_regress(the_model)
  }

  base_model_met  <- round(mi[[met]], 3)
  final_model_met <- round(ols_regress(x$model)[[met]], 3)
  metric_info <- paste0(the_info, format(base_model_met, nsmall = 3), "\n",
                        "Final Model : ", format(final_model_met, nsmall = 3))

  return(metric_info)

}

ols_stepaic_plot_build <- function(d, d2, xmin, xmax, ymin, ymax, metric_info, direction = c("forward", "backward", "both"), criteria = "aic") {

  method <- match.arg(direction)

  mat  <- switch(criteria,
    aic    = "AIC",
    sbc    = "SBC",
    sbic   = "SBIC",
    r2     = "R2",
    adj_r2 = "Adj. R2")

  if (method == "forward") {
    title <- paste0("Stepwise ", mat , " Forward Selection")
    nudge <- 0.1
  } else if (method == "backward") {
    title <- paste0("Stepwise ", mat , " Backward Elimination")
    nudge <- 0.1
  } else {
    title <- paste0("Stepwise ", mat, " Both Direction Selection")
    nudge <- 0.5
  }

  y_lab  <- mat
  v_just <- "bottom"
  h_just <- 1.2
  ann_x  <- Inf

  ols_step_ggplot(d, d2, xmin, xmax, ymin, ymax, y_lab, title, v_just, h_just, nudge, ann_x, metric_info)

}


ols_plot_stepwise <- function(x, metric = "r2", y_lab = "R-Square", details = TRUE,
  direction = c("forward", "backward", "both")) {

  type <- match.arg(direction)
  pred <- ols_step_plot_text(x, type, details, metric)
  data <- ols_stepwise_plot_data(x, pred, metric)
  info <- ols_metric_info(x, type, metric)

  v_just <- ifelse(metric %in% c("aic", "rmse"), "bottom", "top")
  h_just <- ifelse(metric %in% c("aic", "rmse"), 1.2, 0)
  ann_x  <- ifelse(metric %in% c("aic", "rmse"), Inf, 0)

  if (metric == "r2") {
    title <- "R-Square"
  } else if (metric == "adj_r2") {
    title <- "Adjusted R-Square"
  } else if (metric == "aic") {
    title <- "Akaike Information Criteria"
  } else {
    title <- "Root Mean Squared Error"
  }

  ols_step_ggplot(data$d, data$d2, data$xmin, data$xmax, data$ymin, data$ymax,
    y_lab, title, v_just, h_just, nudge = 0.05, ann_x, info)

}

#' @importFrom ggplot2 aes_string
ols_step_ggplot <- function(d, d2, xmin, xmax, ymin, ymax, y_lab, title, v_just, h_just, nudge = 0.05, ann_x, metric_info) {

  ggplot(d, aes(x = a, y = b)) +
    geom_line(color = "blue") +
    geom_point(color = "blue", shape = 1, size = 2) +
    geom_text(data = d2, aes(x = x, y = y, label = tx), size = 3,
              hjust = "left", vjust = v_just, nudge_x = nudge) +
    annotate("text", x = ann_x, y = Inf, hjust = h_just, vjust = 2,
             family = "serif", fontface = "bold", size = 3,
             label = metric_info) +
    xlab("Step") +
    ylab(y_lab) +
    ggtitle(title) +
    xlim(c(xmin, xmax)) +
    ylim(c(ymin, ymax))
}

ols_print_final_model <- function(data) {
  cat("\n\n")
  cat("Final Model Output", "\n")
  cat(rep("-", 18), sep = "", "\n\n")
  print(ols_regress(data$model))
  cat("\n")
}


print_step_output <- function(data, direction = c("forward", "backward", "both")) {

  method <- match.arg(direction)

  print_step_zero(data, method)

  mi         <- print_step_mi(data, method)
  metrics    <- print_step_metrics(data, mi)
  predictors <- print_step_predictors(data, method)

  ols_print_output(metrics, predictors)
  ols_print_final_model(data)

}

print_step_zero <- function(data, method) {

  if (length(data$metrics$step) < 1) {
    if (method == "forward") {
      stop("No variables have been added to the model.")
    } else if (method == "backward") {
      stop("No variables have been removed from the model.")
    } else {
      stop("No variables have been added to or removed from the model.")
    }
  }

}

print_step_mi <- function(data, method) {

  if (method == "forward" || method == "both") {
    np <- coeff_names(data$others$base_model)
    mi <- null_model_metrics(data$others$base_model, data$others$full_model)
  } else {
    mi <- ols_regress(data$others$full_model)
  }

  return(mi)
}

print_step_metrics <- function(data, mi) {

  aic   <- c(mi$aic, data$metrics$aic)
  sbc   <- c(mi$sbc, data$metrics$sbc)
  sbic  <- c(mi$sbic, data$metrics$sbic)
  r2    <- c(mi$rsq, data$metrics$r2)
  adjr  <- c(mi$adjr, data$metrics$adj_r2)
  step  <- c(0, data$metrics$step)
  ln    <- length(aic)

  return(list(aic = aic, sbc = sbc, sbic = sbic, r2 = r2, adjr = adjr, step = step, ln = ln))

}

print_step_predictors <- function(data, method) {

  if (method == "both") {
    data$metrics$sign <- ifelse(data$metrics$method == "addition", "+", "-")
    variable <- paste0(data$metrics$variable, " (", data$metrics$sign, ")")
  }

  if (method == "forward") {
    predictors <- c("Base Model", data$metrics$variable)
  } else if (method == "backward") {
    predictors <- c("Full Model", data$metrics$variable)
  } else {
    predictors <- c("Base Model", variable)
  }

  return(predictors)
}

ols_print_output <- function(metrics, predictors) {

  w1 <- nchar("Step")
  w2 <- max(nchar(predictors))
  w3 <- max_nchar("AIC", metrics$aic)
  w4 <- max_nchar("SBC", metrics$sbc)
  w5 <- max_nchar("SBIC", metrics$sbic)
  w6 <- max_nchar("R2", metrics$r2, 5, 5)
  w7 <- max_nchar("Adj. R2", metrics$adjr, 5, 5)
  w  <- sum(w1, w2, w3, w4, w5, w6, w7, 24)

  cat("\n\n", format("Stepwise Summary", width = w, justify = "centre"), "\n")
  cat(rep("-", w), sep = "", "\n")

  cat(format("Step", width = w1, justify = "centre"), fs(),
    fl("Variable", w2), fs(),
    fc("AIC", w3), fs(),
    fc("SBC", w4), fs(),
    fc("SBIC", w5), fs(),
    fc("R2", w6), fs(),
    fc("Adj. R2", w7), "\n")

  cat(rep("-", w), sep = "", "\n")

  for (i in seq_len(metrics$ln)) {
    cat(format(as.character(metrics$step[i]), width = w1, justify = "centre"), fs(),
      fl(predictors[i], w2), fs(),
      fg(format(round(metrics$aic[i], 3), nsmall = 3), w3), fs(),
      fg(format(round(metrics$sbc[i], 3), nsmall = 3), w4), fs(),
      fg(format(round(metrics$sbic[i], 3), nsmall = 3), w5), fs(),
      fg(format(round(metrics$r2[i], 5), nsmall = 5), w6), fs(),
      fg(format(round(metrics$adjr[i], 5), nsmall = 5), w7), "\n")
  }

  cat(rep("-", w), sep = "")
}


ols_rsquared_init <- function(include, metric, response, rsq_base) {

  cat("\n")
    if (is.null(include)) {
      if (metric == "r2") {
        cat("Step   => 0", "\n")
        cat("Model  =>", paste(response, "~", 1, "\n"))
        cat("R2     =>", round(rsq_base, 3), "\n\n")
      } else {
        cat("Step     => 0", "\n")
        cat("Model    =>", paste(response, "~", 1, "\n"))
        cat("Adj. R2  =>", round(rsq_base, 3), "\n\n")
      }
    } else {
      if (metric == "r2") {
        cat("Step   => 0", "\n")
        cat("Model  =>", paste(response, "~", paste(include, collapse = " + "), "\n"))
        cat("R2     =>", round(rsq_base, 3), "\n\n")
      } else {
        cat("Step     => 0", "\n")
        cat("Model    =>", paste(response, "~", paste(include, collapse = " + "), "\n"))
        cat("Adj. R2  =>", round(rsq_base, 3), "\n\n")
      }
    }
    cat("Initiating stepwise selection...", "\n\n")
}

ols_rsquared_selected <- function(metric, step, preds, response, rsq1) {

  if (metric == "r2") {
        cat("Step      =>", step, "\n")
        cat("Selected  =>", tail(preds, n = 1), "\n")
        cat("Model     =>", paste(response, "~", paste(preds, collapse = " + "), "\n"))
        cat("R2        =>", round(rsq1, 3), "\n\n")
      } else {
        cat("Step      =>", step, "\n")
        cat("Selected  =>", tail(preds, n = 1), "\n")
        cat("Model     =>", paste(response, "~", paste(preds, collapse = " + "), "\n"))
        cat("Adj. R2   =>", round(rsq1, 3), "\n\n")
      }
}


ols_rsquared_removed <- function(metric, step, rpred, preds, response, aic_f) {
  if (metric == "r2") {
          cat("Step     =>", step, "\n")
          cat("Removed  =>", tail(rpred, n = 1), "\n")
          cat("Model    =>", paste(response, "~", paste(preds, collapse = " + "), "\n"))
          cat("R2       =>", aic_f, "\n\n")
        } else {
          cat("Step     =>", step, "\n")
          cat("Removed  =>", tail(rpred, n = 1), "\n")
          cat("Model    =>", paste(response, "~", paste(preds, collapse = " + "), "\n"))
          cat("Adj. R2  =>", aic_f, "\n\n")
        }
}

ols_base_criteria <- function(model, criteria) {
  switch (criteria,
    aic    = ols_aic(model),
    sbc    = ols_sbc(model),
    sbic   = ols_sbic(model, model),
    rsq    = summary(model)$r.squared,
    adjrsq = summary(model)$adj.r.squared
  )
}

ols_sort_metrics <- function(data, criteria) {

  mat  <- switch(criteria,
    aic    = "aics",
    sbc    = "bics",
    sbic   = "sbics",
    rsq    = "rsq",
    adjrsq = "arsq")

  if (criteria == "aic" || criteria == "sbc" || criteria == "sbic") {
    data[order(data[[mat]]), ]
  } else {
    data[order(-data[[mat]]), ]
  }
}

ols_threshold <- function(data, criteria) {
  if (criteria == "aic" || criteria == "sbc" || criteria == "sbic") {
    which(data == min(data))
  } else {
    which(data == max(data))
  }
}

ols_f_criteria <- function(criteria, mat, minc, bmetric) {
  if (criteria == "aic" || criteria == "sbc" || criteria == "sbic") {
    mat[minc] < bmetric
  } else {
    mat[minc] > bmetric
  }
}

ols_next_criteria <- function(criteria, mat, minaic, laic, lpreds) {
  if (criteria == "aic" || criteria == "sbc" || criteria == "sbic") {
    mat[minaic] < laic[lpreds]
  } else {
    mat[minaic] > laic[lpreds]
  }
}

coeff_names <- function(model) {
  colnames(attr(model$terms, which = "factor"))
}

model_colnames <- function(model) {
  names(model.frame(model))
}

coeff_length <- function(predicts, gap) {
  sum(nchar(predicts)) + gap
}

#' Residual vs regressor plot
#'
#' @description
#' Graph to determine whether we should add a new predictor to the model already
#' containing other predictors. The residuals from the model is regressed on the
#' new predictor and if the plot shows non random pattern, you should consider
#' adding the new predictor to the model.
#'
#' @param model An object of class \code{lm}.
#' @param variable New predictor to be added to the \code{model}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt, data = mtcars)
#' ols_plot_resid_regressor(model, 'drat')
#'
#' @seealso [ols_plot_added_variable()], [ols_plot_comp_plus_resid()]
#'
#' @export
#'
ols_plot_resid_regressor <- function(model, variable, print_plot = TRUE) {

  check_model(model)

  d     <- ols_prep_rvsrplot_data(model)
  inter <- eval(model$call$data)[variable]
  x     <- inter[[1]]
  y     <- residuals(model)
  v     <- names(inter)
  k     <- data.frame(x = x, y = y)

  p <-
    ggplot(k, aes(x = x, y = y)) +
    geom_point(shape = 1, colour = "blue") +
    geom_hline(yintercept = 0, colour = "red") +
    xlab(paste(v)) +
    ylab("Residual") +
    ggtitle(paste("Residual vs", v))

  if (print_plot) {
    print(p)
  }

  return(p)

}

#' Residual plus component plot
#'
#' @description
#' The residual plus component plot indicates whether any non-linearity is
#' present in the relationship between response and predictor variables and can
#' suggest possible transformations for linearizing the data.
#'
#' @param model An object of class \code{lm}.
#' @param print_plot logical; if \code{TRUE}, prints the plot else returns a plot object.
#'
#' @references
#' Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
#'
#' Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
#' Chicago, IL., McGraw Hill/Irwin.
#'
#' @examples
#' model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
#' ols_plot_comp_plus_resid(model)
#'
#' @seealso [ols_plot_added_variable()], [ols_plot_resid_regressor()]
#'
#' @export
#'
ols_plot_comp_plus_resid <- function(model, print_plot = TRUE) {

  check_model(model)

  pl      <- cpout(model)
  myplots <- list()

  for (i in seq_len(pl$lmc)) {
    k <- cpdata(pl$data, pl$mc, pl$e, i)
    p <-
      eval(
        substitute(
          ggplot(k, aes(x = x, y = y)) +
          geom_point(colour = "blue", size = 2) +
          stat_smooth(method = "lm", se = FALSE) +
          xlab(pl$nam[i]) +
          ylab(paste0("Residual + Component (", pl$indvar, ")")),
          list(i = i)
        )
      )

    myplots[[i]] <- p
  }

  if (print_plot) {
    marrangeGrob(myplots, nrow = 2, ncol = 2, top = "Residual Plus Component Plot")
  } else {
    return(myplots)
  }

}


cpdata <- function(data, mc, e, i) {

  x <- data[[i]]
  y <- ((mc[i] * data[i]) + e)[[1]]
  data.frame(x = x, y = y)

}

cpout <- function(model) {

  e      <- residuals(model)
  mc     <- coefficients(model)[-1]
  data   <- as.data.frame(model.matrix(model))[, -1]
  lmc    <- length(mc)
  nam    <- names(data)
  indvar <- names(model.frame(model))[1]

  list(e = e, mc = mc, data = data, lmc = lmc, nam = nam, indvar = indvar)

}

#' \code{olsrr} package
#'
#' Tools for teaching and learning OLS regression
#'
#' See the README on
#' \href{https://github.com/rsquaredacademy/olsrr}{GitHub}
#'
#' @docType package
#' @keywords internal
#' @name olsrr
#' @aliases olsrr-package
"_PACKAGE"

## quiets concerns of R CMD check re: the .'s that appear in pipelines
if (getRversion() >= "2.15.1") {
  utils::globalVariables(c(".", "owner", "repo", "tag_name", "result",
    "a", "b", "tx", "mindex", "n", "x", "y", "k", "size", "shape", "rsquare",
    "cp", "adjr", "cps", "aic", "sbic", "sbc", "index", "betas", "rsq",
    "lpreds", "terms", "pvdata", "values", "d", "v", "r.squared", "obs",
    "txt", "cd", "fct_color", "ckd", "dbetas", "color", "pred", "ds", "dsr",
    "fstatistic", "hadi", "Df", "res", "pot", "lfit", "rerror", "ybar",
    "yhat", "predicted", "resid", "lev_thrsh", "leverage", "levrstud",
    "sdres", "x1", "x2", "y1", "y2"))
}

test_that("output from stepAIC both direction regression is as expected when adding and removing variables", {
  model <- lm(y ~ ., data = stepdata)
  expect_snapshot(ols_step_both_aic(model, progress = TRUE))
  expect_snapshot(ols_step_both_aic(model, details = TRUE))
})

test_that("output from stepSBC both direction regression is as expected when adding and removing variables", {
  model <- lm(y ~ ., data = stepdata)
  expect_snapshot(ols_step_both_sbc(model, progress = TRUE))
  expect_snapshot(ols_step_both_sbc(model, details = TRUE))
})

test_that("output from stepSBIC both direction regression is as expected when adding and removing variables", {
  model <- lm(y ~ ., data = stepdata)
  expect_snapshot(ols_step_both_sbic(model, progress = TRUE))
  expect_snapshot(ols_step_both_sbic(model, details = TRUE))
})

test_that("stepwise selection output matches the expected result", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_both_aic(model)
  expect_equal(k$metrics$step, 1:3)
  expect_equal(k$metrics$method, c("addition", "addition", "addition"), ignore_attr = TRUE)
  expect_equal(k$metrics$variable, c("x4", "x1", "x2"), ignore_attr = TRUE)
})

test_that("stepwise selection output matches when steps are specified", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_both_aic(model, steps = 2)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$method, c("addition", "addition"), ignore_attr = TRUE)
  expect_equal(k$metrics$variable, c("x4", "x1"), ignore_attr = TRUE)
})

test_that("output from stepaic_both matches the expected output when variables are locked in", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_aic(model, include = c("x6"))
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("x1", "x3", "x2", "x4"), ignore_attr = TRUE)

  h <- ols_step_both_aic(model, include = c(6))
  expect_equal(h$metrics$step, 1:4)
  expect_equal(h$metrics$variable, c("x1", "x3", "x2", "x4"), ignore_attr = TRUE)
})

test_that("output from stepaic_both matches the expected output when variables are locked out", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_aic(model, exclude = c("x2"))
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("x6", "x1", "x3", "x4"), ignore_attr = TRUE)

  h <- ols_step_both_aic(model, exclude = c(2))
  expect_equal(h$metrics$step, 1:4)
  expect_equal(h$metrics$variable, c("x6", "x1", "x3", "x4"), ignore_attr = TRUE)
})

test_that("output from stepaic_both matches the expected output when variables are locked in and out", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_aic(model, include = c("x6"), exclude = c("x2"))
  expect_equal(k$metrics$step, 1:3)
  expect_equal(k$metrics$variable, c("x1", "x3", "x4"), ignore_attr = TRUE)

  h <- ols_step_both_aic(model, include = c(6), exclude = c(2))
  expect_equal(h$metrics$step, 1:3)
  expect_equal(h$metrics$variable, c("x1", "x3", "x4"), ignore_attr = TRUE)
})

test_that("stepaic_both returns the appropriate error", {

  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  expect_error(ols_step_both_aic(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_both_aic(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_both_aic(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_both_aic(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

test_that("stepwise sbc selection output matches the expected result", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_both_sbc(model)
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$method, c("addition", "addition", "addition", "removal"), ignore_attr = TRUE)
  expect_equal(k$metrics$variable, c("x4", "x1", "x2", "x4"), ignore_attr = TRUE)
})

test_that("output from stepsbc_both matches the expected output when variables are locked in", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_sbc(model, include = c("x6"))
  expect_equal(k$metrics$step, 1:3)
  expect_equal(k$metrics$variable, c("x1", "x3", "x2"), ignore_attr = TRUE)

  h <- ols_step_both_sbc(model, include = c(6))
  expect_equal(h$metrics$step, 1:3)
  expect_equal(h$metrics$variable, c("x1", "x3", "x2"), ignore_attr = TRUE)
})

test_that("output from stepsbc_both matches the expected output when variables are locked out", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_sbc(model, exclude = c("x2"))
  expect_equal(k$metrics$step, 1:3)
  expect_equal(k$metrics$variable, c("x6", "x1", "x3"), ignore_attr = TRUE)

  h <- ols_step_both_sbc(model, exclude = c(2))
  expect_equal(h$metrics$step, 1:3)
  expect_equal(h$metrics$variable, c("x6", "x1", "x3"), ignore_attr = TRUE)
})

test_that("output from stepsbc_both matches the expected output when variables are locked in and out", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_sbc(model, include = c("x6"), exclude = c("x2"))
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("x1", "x3"), ignore_attr = TRUE)

  h <- ols_step_both_sbc(model, include = c(6), exclude = c(2))
  expect_equal(h$metrics$step, 1:2)
  expect_equal(h$metrics$variable, c("x1", "x3"), ignore_attr = TRUE)
})

test_that("stepsbc_both returns the appropriate error", {

  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  expect_error(ols_step_both_sbc(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_both_sbc(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_both_sbc(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_both_sbc(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

test_that("stepwise sbic selection output matches the expected result", {
  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  k <- ols_step_both_sbic(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$method, c("addition", "addition"), ignore_attr = TRUE)
  expect_equal(k$metrics$variable, c("x4", "x1"), ignore_attr = TRUE)
})

test_that("output from stepsbic_both matches the expected output when variables are locked in", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_sbic(model, include = c("x6"))
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("x1", "x3", "x2", "x4"), ignore_attr = TRUE)

  h <- ols_step_both_sbic(model, include = c(6))
  expect_equal(h$metrics$step, 1:4)
  expect_equal(h$metrics$variable, c("x1", "x3", "x2", "x4"), ignore_attr = TRUE)
})

test_that("output from stepsbic_both matches the expected output when variables are locked out", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_sbic(model, exclude = c("x2"))
  expect_equal(k$metrics$step, 1:4)
  expect_equal(k$metrics$variable, c("x6", "x1", "x3", "x4"), ignore_attr = TRUE)

  h <- ols_step_both_sbic(model, exclude = c(2))
  expect_equal(h$metrics$step, 1:4)
  expect_equal(h$metrics$variable, c("x6", "x1", "x3", "x4"), ignore_attr = TRUE)
})

test_that("output from stepsbic_both matches the expected output when variables are locked in and out", {

  model <- lm(y ~ ., data = stepdata)
  k <- ols_step_both_sbic(model, include = c("x6"), exclude = c("x2"))
  expect_equal(k$metrics$step, 1:3)
  expect_equal(k$metrics$variable, c("x1", "x3", "x4"), ignore_attr = TRUE)

  h <- ols_step_both_sbic(model, include = c(6), exclude = c(2))
  expect_equal(h$metrics$step, 1:3)
  expect_equal(h$metrics$variable, c("x1", "x3", "x4"), ignore_attr = TRUE)
})

test_that("stepsbic_both returns the appropriate error", {

  model <- lm(y ~ x1 + x2 + x3 + x4, data = cement)
  expect_error(ols_step_both_sbic(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_both_sbic(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_both_sbic(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_both_sbic(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

model <- lm(y ~ ., data = surgical)

test_that("output from stepAIC backward regression is as expected", {
  expect_snapshot(ols_step_backward_aic(model))
  expect_snapshot(ols_step_backward_aic(model, progress = TRUE))
  expect_snapshot(ols_step_backward_aic(model, details = TRUE))
})

test_that("output from stepSBC backward regression is as expected", {
  expect_snapshot(ols_step_backward_sbc(model))
  expect_snapshot(ols_step_backward_sbc(model, progress = TRUE))
  expect_snapshot(ols_step_backward_sbc(model, details = TRUE))
})

test_that("output from stepSBIC backward regression is as expected", {
  expect_snapshot(ols_step_backward_sbic(model))
  expect_snapshot(ols_step_backward_sbic(model, progress = TRUE))
  expect_snapshot(ols_step_backward_sbic(model, details = TRUE))
})

model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)

test_that("output from stepaic_backward matches the expected output", {
  k <- ols_step_backward_aic(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("disp", "drat"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 3), c(156.731, 156.652), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.837, 0.827), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.819, 0.815), ignore_attr = TRUE)

  k <- ols_step_backward_aic(model, include = c("drat"))
  expect_equal(k$metrics$variable, c("disp"), ignore_attr = TRUE)

  k <- ols_step_backward_aic(model, exclude = c("drat"))
  expect_equal(k$metrics$variable, c("disp"), ignore_attr = TRUE)

})

test_that("output from stepaic_backward matches when steps are specified", {
  k <- ols_step_backward_aic(model, steps = 1)
  expect_equal(k$metrics$step, 1)
  expect_equal(k$metrics$variable, c("disp"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 3), c(156.731), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.837), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.819), ignore_attr = TRUE)
})

test_that("stepaic_backward returns the appropriate error", {
  expect_error(ols_step_backward_aic(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_backward_aic(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_backward_aic(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_backward_aic(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

test_that("output from stepsbc_backward matches the expected output", {
  k <- ols_step_backward_sbc(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("disp", "drat"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 3), c(156.731, 156.652), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.837, 0.827), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.819, 0.815), ignore_attr = TRUE)

  k <- ols_step_backward_sbc(model, include = c("drat"))
  expect_equal(k$metrics$variable, c("disp"), ignore_attr = TRUE)

  k <- ols_step_backward_sbc(model, exclude = c("drat"))
  expect_equal(k$metrics$variable, c("disp"), ignore_attr = TRUE)

})

test_that("stepsbc_backward returns the appropriate error", {
  expect_error(ols_step_backward_sbc(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_backward_sbc(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_backward_sbc(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_backward_sbc(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

test_that("output from stepsbic_backward matches the expected output", {
  k <- ols_step_backward_sbic(model)
  expect_equal(k$metrics$step, 1:2)
  expect_equal(k$metrics$variable, c("disp", "drat"), ignore_attr = TRUE)
  expect_equal(round(k$metrics$aic, 3), c(156.731, 156.652), ignore_attr = TRUE)
  expect_equal(round(k$metrics$r2, 3), c(0.837, 0.827), ignore_attr = TRUE)
  expect_equal(round(k$metrics$adj_r2, 3), c(0.819, 0.815), ignore_attr = TRUE)

  k <- ols_step_backward_sbic(model, include = c("drat"))
  expect_equal(k$metrics$variable, c("disp"), ignore_attr = TRUE)

  k <- ols_step_backward_sbic(model, exclude = c("drat"))
  expect_equal(k$metrics$variable, c("disp"), ignore_attr = TRUE)

})

test_that("stepsbic_backward returns the appropriate error", {
  expect_error(ols_step_backward_sbic(model, include = c("dis")), "dis not part of the model and hence cannot be forcibly included. Please verify the variable names.")
  expect_error(ols_step_backward_sbic(model, exclude = c("hps")), "hps not part of the model and hence cannot be forcibly excluded. Please verify the variable names.")
  expect_error(ols_step_backward_sbic(model, include = c(5)), "Index of variable to be included should be between 1 and 4.")
  expect_error(ols_step_backward_sbic(model, exclude = c(5)), "Index of variable to be excluded should be between 1 and 4.")
})

---
title: "Introduction to olsrr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to olsrr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

The olsrr package provides following tools for teaching and learning OLS regression using R:

- comprehensive regression output
- residual diagnostics
- measures of influence
- heteroskedasticity tests
- collinearity diagnostics
- model fit assessment
- variable contribution assessment
- variable selection procedures

```{r, echo=FALSE, message=FALSE}
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)
```

This document is a quickstart guide to the tools offered by **olsrr**. Other vignettes provide more details on specific topics:

- Residual Diagnostics: Includes plots to examine residuals to validate OLS assumptions

- Variable selection: Differnt variable selection procedures such as all possible regression, best subset regression, stepwise regression, stepwise forward regression and stepwise backward regression

- Heteroskedasticity: Tests for heteroskedasticity include bartlett test, breusch pagan test,
score test and f test

- Measures of influence: Includes 10 different plots to detect and identify influential observations

- Collinearity diagnostics: VIF, Tolerance and condition indices to detect collinearity and plots for assessing mode fit and contributions of variables


## Regression

```{r regress}
ols_regress(mpg ~ disp + hp + wt + qsec, data = mtcars)
```

In the presence of interaction terms in the model, the predictors are
scaled and centered before computing the standardized betas. `ols_regress()` will detect interaction terms automatically but in case you have created a new variable instead of using the inline function `*`, you can indicate the presence of interaction terms by setting `iterm` to `TRUE`.

## Residual vs Fitted Values Plot

Plot to detect non-linearity, unequal error variances, and outliers.

```{r rvsfplot, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_fit(model)
```

## DFBETAs Panel

DFBETAs measure the difference in each parameter estimate with and without the influential observation.
`dfbetas_panel` creates plots to detect influential observations using DFBETAs.

```{r dfbpanel, fig.width=7, fig.height=7, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_dfbetas(model)
```

## Residual Fit Spread Plot

Plot to detect non-linearity, influential observations and outliers.

```{r rfsplot, fig.width=10, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_fit_spread(model)
```

## Breusch Pagan Test

Breusch Pagan test is used to test for herteroskedasticity (non-constant error variance). It tests whether the variance of the errors from a regression is dependent on the values of the independent variables. It is a $\chi^{2}$ test.

```{r bp1}
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model)
```

## Collinearity Diagnostics

```{r colldiag}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_coll_diag(model)
```

## Stepwise Regression

Build regression model from a set of candidate predictor variables by entering and removing predictors based on 
p values, in a stepwise manner until there is no variable left to enter or remove any more.

### Variable Selection

```{r stepwise1}
# stepwise regression
model <- lm(y ~ ., data = surgical)
ols_step_both_p(model)
```

### Plot

```{r stepwise2, fig.width=10, fig.height=15, fig.align='center'}
model <- lm(y ~ ., data = surgical)
k <- ols_step_both_p(model)
plot(k)
```


## Stepwise AIC Backward Regression

Build regression model from a set of candidate predictor variables by removing predictors based on 
Akaike Information Criteria, in a stepwise manner until there is no variable left to remove any more.

### Variable Selection

```{r stepaicb1}
# stepwise aic backward regression
model <- lm(y ~ ., data = surgical)
k <- ols_step_backward_aic(model)
k
```

### Plot

```{r stepaicb2, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(y ~ ., data = surgical)
k <- ols_step_backward_aic(model)
plot(k)
```

---
title: "Media"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Media}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Shiny App Demo

- [YouTube Playlist](https://www.youtube.com/watch?v=v-qSC6bVstE&list=PLDfCKV3ey8-fPVNyOxd6dJfptZLXKTpRl)

---
title: "Variable Selection Methods"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Variable Selection Methods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE, message=FALSE}
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)
```

<style type="text/css">
#best-subset-regression pre { /* Code block */
  font-size: 10px
}
</style>

## Introduction

Variable selection refers to the process of choosing the most relevant variables to include in a 
regression model. They help to improve model performance and avoid over fitting. 

Before we explore stepwise selection methods, let us take a quick look at all/best subset regression.
As they evaluate every possible variable combination, these methods are computationally intensive and may 
crash your system if used with a large set of variables. We have included them in the package purely for 
educational purpose.

### All Possible Regression

All subset regression tests all possible subsets of the set of potential independent variables. If there are K potential independent variables (besides the constant), then there are $2^{k}$ distinct subsets of them to be tested. For example, if you have 10 candidate independent variables, the number of subsets to be tested is $2^{10}$, which is 1024, and if you have 20 candidate variables, the number is $2^{20}$, which is more than one million.

```{r allsub}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_step_all_possible(model)
```

### Best Subset Regression

Select the subset of predictors that do the best at meeting some well-defined objective criterion, 
such as having the largest R2 value or the smallest MSE, Mallow's Cp or AIC.

```{r bestsub, size='tiny'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_step_best_subset(model)
```

## Stepwise Selection

Stepwise regression is a method of fitting regression models that involves the 
iterative selection of independent variables to use in a model. It can be 
achieved through forward selection, backward elimination, or a combination of 
both methods. The forward selection approach starts with no variables and adds 
each new variable incrementally, testing for statistical significance, while 
the backward elimination method begins with a full model and then removes the 
least statistically significant variables one at a time.

### Model

We will use the below model throughout this article except in the case of hierarchical selection. 
You can learn more about the data [here](https://olsrr.rsquaredacademy.com/reference/surgical).

```{r model}
model <- lm(y ~ ., data = surgical)
summary(model)
```

### Model specification 

Irrespective of the stepwise method being used, we have to specify the full model i.e. all the variabels/predictors 
under consideration as `olsrr` extracts the candidate variables for selection/elimination from the model specified.

##### Forward selection

```{r stepf1}
# stepwise forward regression
ols_step_forward_p(model)
```

##### Backward elimination

```{r stepb}
# stepwise backward regression
ols_step_backward_p(model)
```

### Criteria 

The criteria for selecting variables may be one of the following:

- p value
- akaike information criterion (aic)
- schwarz bayesian criterion (sbc)
- sawa bayesian criterion (sbic)
- r-square
- adjusted r-square
  
### Include/exclude variables 

We can force variables to be included or excluded from the model at all stages of variable selection. The
variables may be specified either by name or position in the model specified.

##### By name

```{r include_name}
ols_step_forward_p(model, include = c("age", "alc_mod"))
```

##### By index

```{r include_index}
ols_step_forward_p(model, include = c(5, 7))
```

### Standardized output

All stepwise selection methods display standard output which includes:

- selection summary
- model summary
- ANOVA
- parameter estimates

```{r output}
# adjusted r-square 
ols_step_forward_adj_r2(model)
```

### Visualization

Use the `plot()` method to visualize variable selection. It will display how the variable selection criteria 
changes at each step of the selection process along with the variable selected.

```{r visualize}
# adjusted r-square 
k <- ols_step_forward_adj_r2(model)
plot(k)
```

### Verbose output

To view the detailed regression output at each stage of variable selection/elimination, set `details` to `TRUE`. It will
display the following information at each step:

- step number
- variable selected/eliminated
- model 
- value of the criteria at that stage

```{r details}
# adjusted r-square 
ols_step_forward_adj_r2(model, details = TRUE)
```

### Progress

To view the progress in the variable selection procedure, set `progress` to `TRUE`. It will display the variable 
being selected/eliminated at each step until there are no more candidate variables left.

```{r progress}
# adjusted r-square 
ols_step_forward_adj_r2(model, progress = TRUE)
```

### Hierarchical selection 

When using `p` values as the criterion for selecting/eliminating variables, we can enable hierarchical
selection. In this method, the search for the most significant variable is restricted to the next available
variable. In the below example, as `liver_test` does not meet the threshold for selection, none of the 
variables after `liver_test` are considered for further selection i.e. the stepwise selection ends as soon
as it comes across a variable that does not meet the selection threshold. You can learn more about hierachichal
selection [here](https://www.stata.com/manuals/rstepwise.pdf).

```{r hierarchical}
# hierarchical selection
m <- lm(y ~ bcs + alc_heavy + pindex + enzyme_test + liver_test + age + gender + alc_mod, data = surgical)
ols_step_forward_p(m, 0.1, hierarchical = TRUE)
```

---
title: "Residual Diagnostics"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Residual Diagnostics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE, message=FALSE}
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)
```

# Introduction

olsrr offers tools for detecting violation of standard regression assumptions. Here we take a look 
at residual diagnostics. The standard regression assumptions include the following about residuals/errors:

- The error has a normal distribution (normality assumption).
- The errors have mean zero.
- The errors have same but unknown variance (homoscedasticity assumption).
- The error are independent of each other (independent errors assumption). 

## Residual QQ Plot

Graph for detecting violation of normality assumption.

```{r qqresid, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_qq(model)
```

## Residual Normality Test

Test for detecting violation of normality assumption.

```{r normtest}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_normality(model)
```

Correlation between observed residuals and expected residuals under normality.

```{r corrtest}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_correlation(model)
```

## Residual vs Fitted Values Plot

It is a scatter plot of residuals on the y axis and fitted values on the x axis to detect non-linearity, unequal error variances, and outliers. 

**Characteristics of a well behaved residual vs fitted plot:**

- The residuals spread randomly around the 0 line indicating that the relationship is linear.
- The residuals form an approximate horizontal band around the 0 line indicating homogeneity of error variance.
- No one residual is visibly away from the random pattern of the residuals indicating that there are no outliers.

```{r rvsfplot, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_fit(model)
```


## Residual Histogram

Histogram of residuals for detecting violation of normality assumption.

```{r residhist, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_hist(model)
```

---
title: "Heteroscedasticity"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Heteroscedasticity}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE, message=FALSE}
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)
```

# Introduction

One of the assumptions made about residuals/errors in OLS regression is that the errors have the
same but unknown variance. This is known as constant variance or homoscedasticity. When this 
assumption is violated, the problem is known as heteroscedasticity.

##### Consequences of Heteroscedasticity

- The OLS estimators and regression predictions based on them remains unbiased and consistent.
- The OLS estimators are no longer the BLUE (Best Linear Unbiased Estimators) because they are no longer efficient, so the regression predictions will be inefficient too.
- Because of the inconsistency of the covariance matrix of the estimated regression coefficients, the tests of hypotheses, (t-test, F-test) are no longer valid.

**olsrr** provides the following 4 tests for detecting heteroscedasticity:

- Bartlett Test
- Breusch Pagan Test
- Score Test
- F Test

## Bartlett Test

Bartlett's test is used to test if variances across samples is equal. It is sensitive to 
departures from normality. The Levene test is an alternative test that is less sensitive 
to departures from normality.

You can perform the test using 2 continuous variables, one continuous and one grouping variable,
a formula or a linear model. 

#### Use grouping variable
```{r bartlett1}
ols_test_bartlett(hsb, 'read', group_var = 'female')
```

#### Using variables
```{r bartlett2}
ols_test_bartlett(hsb, 'read', 'write')
```


## Breusch Pagan Test

Breusch Pagan Test was introduced by Trevor Breusch and Adrian Pagan in 1979. It is used to
test for heteroskedasticity in a linear regression model and assumes that the error terms are 
normally distributed. It tests whether the variance of the errors from a regression is dependent 
on the values of the independent variables. It is a $\chi^{2}$ test.

You can perform the test using the fitted values of the model, the predictors in the model and a
subset of the independent variables. It includes options to perform multiple tests and p value adjustments. 
The options for p value adjustments include Bonferroni, Sidak and Holm's method.

#### Use fitted values of the model

```{r bp1}
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model)
```

#### Use independent variables of the model

```{r bp2}
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model, rhs = TRUE)
```

#### Use independent variables of the model and perform multiple tests

```{r bp3}
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE)
```

#### Bonferroni p value Adjustment

```{r bp4}
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'bonferroni')
```

#### Sidak p value Adjustment

```{r bp5}
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'sidak')
```

#### Holm's p value Adjustment

```{r bp6}
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'holm')
```

## Score Test

Test for heteroskedasticity under the assumption that the errors are independent and identically distributed (i.i.d.). You can perform the test using the fitted values of the model, the predictors in the model and a
subset of the independent variables.

#### Use fitted values of the model

```{r score1}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_score(model)
```

#### Use independent variables of the model

```{r score2}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_score(model, rhs = TRUE)
```

#### Specify variables

```{r score3}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_score(model, vars = c('disp', 'hp'))
```

## F Test

F Test for heteroskedasticity under the assumption that the errors are independent and identically distributed (i.i.d.). You can perform the test using the fitted values of the model, the predictors in the model and a
subset of the independent variables.

#### Use fitted values of the model

```{r ftest1}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_f(model)
```

#### Use independent variables of the model

```{r ftest2}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_f(model, rhs = TRUE)
```

#### Specify variables

```{r ftest3}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_f(model, vars = c('disp', 'hp'))
```

---
title: "Measures of Influence"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Measures of Influence}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE, message=FALSE}
library(olsrr)
library(ggplot2)
library(gridExtra)
library(nortest)
library(goftest)
```

# Introduction

It is possible for a single observation to have a great influence on the results of a regression analysis. It is therefore important to detect influential observations and to take them into consideration when interpreting the results.

**olsrr** offers the following tools to detect influential observations:

- Cook's D Bar Plot
- Cook's D Chart
- DFBETAs Panel
- DFFITs Plot
- Studentized Residual Plot
- Standardized Residual Chart
- Studentized Residuals vs Leverage Plot
- Deleted Studentized Residual vs Fitted Values Plot
- Hadi Plot
- Potential Residual Plot

## Cook's D Bar Plot

Bar Plot of Cook's distance to detect observations that strongly influence fitted values of the model. Cook's distance was introduced by American statistician R Dennis Cook in 1977. It is used to identify influential data points. It depends on both the residual and leverage i.e it takes it account both the **x** value and **y** value of the observation.

**Steps to compute Cook's distance:**

- delete observations one at a time.
- refit the regression model on remaining $(n - 1)$ observations
- examine how much all of the fitted values change when the ith observation is deleted.

A data point having a large cook's d indicates that the data point strongly influences the fitted values.

```{r ckdbp, fig.width=7, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_cooksd_bar(model)
```

## Cook's D Chart

Chart of Cook's distance to detect observations that strongly influence fitted values of the model.

```{r ckchart, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_cooksd_chart(model)
```

## DFBETAs Panel

DFBETA measures the difference in each parameter estimate with and without the influential point. There is a DFBETA for each data point i.e if there are n observations and k variables, there will be $n * k$ DFBETAs. In general, large values of DFBETAS indicate observations that are influential in estimating a given parameter. Belsley, Kuh, and Welsch recommend 2 as a general cutoff value to indicate influential observations and 
$\frac{2}{\sqrt{n}}$ as a size-adjusted cutoff.

```{r dfbpanel, fig.width=7, fig.height=7, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_dfbetas(model)
```

## DFFITS Plot

Proposed by Welsch and Kuh (1977). It is the scaled difference between the $i^{th}$ fitted value obtained from the full data and the $i^{th}$ fitted value obtained by deleting the $i^{th}$ observation. DFFIT - difference in fits, is used to identify influential data points. It quantifies the number of standard deviations that the fitted value changes when the ith data point is omitted.
 
**Steps to compute DFFITs:**

- delete observations one at a time.
- refit the regression model on remaining \eqn{n - 1} observations
- examine how much all of the fitted values change when the ith observation is deleted.

An observation is deemed influential if the absolute value of its DFFITS value is greater than: 

$${2}*\sqrt{\frac{(p + 1)}{(n - p - 1)}}$$

A size-adjusted cutoff recommended by Belsley, Kuh, and Welsch is

$${2}*\sqrt{\frac{p}{n}}$$

where n is the number of observations and p is the number of predictors including intercept.

olsrr uses the size adjusted cutoff by default.

```{r dfitsplot, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_dffits(model)
```

## Studentized Residual Plot

Plot for detecting outliers. Studentized deleted residuals (or externally studentized residuals) is the deleted residual divided by its estimated standard deviation. Studentized residuals are going to be more effective for  detecting outlying Y observations than standardized residuals. If an observation has an externally studentized residual that is larger than 3 (in absolute value) we can call it an outlier.

```{r srplot, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_stud(model)
```

## Standardized Residual Chart

Chart for detecting outliers. Standardized residual (internally studentized) is the residual divided by estimated standard deviation.

```{r srchart, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_stand(model)
```

## Studentized Residuals vs Leverage Plot

Graph for detecting influential observations.

```{r studlev, fig.width=7, fig.height=5, fig.align='center'}
model <- lm(read ~ write + math + science, data = hsb)
ols_plot_resid_lev(model)
```

## Deleted Studentized Residual vs Fitted Values Plot

Graph for detecting outliers.

```{r dsrvsp, fig.width=7, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_stud_fit(model)
```

## Hadi Plot

Hadi's measure of influence based on the fact that influential observations can be present in either
the response variable or in the predictors or both. The plot is used to detect influential observations
based on Hadi's measure.

```{r hadiplot, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_hadi(model)
```

## Potential Residual Plot

Plot to aid in classifying unusual observations as high-leverage points, outliers, or a combination of both.

```{r potres, fig.width=5, fig.height=5, fig.align='center'}
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_pot(model)
```

# output from ftest is as expected

    Code
      ols_test_f(model)
    Output
      
       F Test for Heteroskedasticity
       -----------------------------
       Ho: Variance is homogenous
       Ha: Variance is not homogenous
      
       Variables: fitted values of mpg 
      
            Test Summary        
       -------------------------
       Num DF     =    1 
       Den DF     =    30 
       F          =    0.4920617 
       Prob > F   =    0.4884154 

---

    Code
      ols_test_f(model, rhs = TRUE)
    Output
      
       F Test for Heteroskedasticity
       -----------------------------
       Ho: Variance is homogenous
       Ha: Variance is not homogenous
      
       Variables: disp hp wt qsec 
      
            Test Summary        
       -------------------------
       Num DF     =    4 
       Den DF     =    27 
       F          =    0.4594694 
       Prob > F   =    0.7647271 

---

    Code
      ols_test_f(model, vars = c("disp", "hp"))
    Output
      
       F Test for Heteroskedasticity
       -----------------------------
       Ho: Variance is homogenous
       Ha: Variance is not homogenous
      
       Variables: disp hp 
      
            Test Summary        
       -------------------------
       Num DF     =    2 
       Den DF     =    29 
       F          =    0.4669306 
       Prob > F   =    0.631555 

# output from bartlett test is as expected when using variables

    Code
      ols_test_bartlett(hsb, "read", "write")
    Output
      
          Bartlett's Test of Homogenity of Variances    
      ------------------------------------------------
      Ho: Variances are equal across groups
      Ha: Variances are unequal for atleast two groups
      
              Data          
       ---------------------
       Variables: read write 
      
              Test Summary         
       ----------------------------
       DF            =    1 
       Chi2          =    1.222871 
       Prob > Chi2   =    0.2687979 

---

    Code
      ols_test_bartlett(descriptr::mtcarz, "mpg", group_var = "vs")
    Output
      
          Bartlett's Test of Homogenity of Variances    
      ------------------------------------------------
      Ho: Variances are equal across groups
      Ha: Variances are unequal for atleast two groups
      
              Test Summary         
       ----------------------------
       DF            =    1 
       Chi2          =    1.584876 
       Prob > Chi2   =    0.2080597 

# when fitted.values == TRUE, fitted vals from the reg are used

    Code
      ols_test_score(model)
    Output
      
       Score Test for Heteroskedasticity
       ---------------------------------
       Ho: Variance is homogenous
       Ha: Variance is not homogenous
      
       Variables: fitted values of mpg 
      
              Test Summary         
       ----------------------------
       DF            =    1 
       Chi2          =    1.268973 
       Prob > Chi2   =    0.2599594 

---

    Code
      ols_test_score(model, rhs = TRUE)
    Output
      
       Score Test for Heteroskedasticity
       ---------------------------------
       Ho: Variance is homogenous
       Ha: Variance is not homogenous
      
       Variables: disp hp wt drat qsec 
      
             Test Summary         
       ---------------------------
       DF            =    5 
       Chi2          =    2.515705 
       Prob > Chi2   =    0.774128 

---

    Code
      ols_test_score(model, vars = c("disp", "hp"))
    Output
      
       Score Test for Heteroskedasticity
       ---------------------------------
       Ho: Variance is homogenous
       Ha: Variance is not homogenous
      
       Variables: disp hp 
      
              Test Summary         
       ----------------------------
       DF            =    2 
       Chi2          =    0.9690651 
       Prob > Chi2   =    0.6159851 

---

    Code
      ols_test_score(model, rhs = TRUE, vars = c("disp", "hp"))
    Output
      
       Score Test for Heteroskedasticity
       ---------------------------------
       Ho: Variance is homogenous
       Ha: Variance is not homogenous
      
       Variables: disp hp wt drat qsec 
      
             Test Summary         
       ---------------------------
       DF            =    5 
       Chi2          =    2.515705 
       Prob > Chi2   =    0.774128 

# output from model performance is as expected

    Code
      ols_model_performance(model)
    Output
                    Model Performance               
      ---------------------------------------------
      R                 0.909       AIC     158.643 
      R-Squared         0.827       SBC     165.972 
      Adj. R-Squared    0.808       SBIC     68.933 
      Pred R-Squared    0.768       RMSE      2.468 
      ---------------------------------------------
        AIC: Akaike Information Criteria 
        SBC: Schwarz Bayesian Criteria 
       SBIC: Sawa's Bayesian Criteria 
       RMSE: Root Mean Square Error 
      

# output from all subsets regression is as expected

    Code
      ols_step_all_possible(model)
    Output
        Index N Predictors  R-Square Adj. R-Square Mallow's Cp
      1     1 1       disp 0.7183433     0.7089548    4.443792
      2     2 1         hp 0.6024373     0.5891853   17.794906
      3     3 2    disp hp 0.7482402     0.7308774    3.000000

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-model-info.R
\name{ols_get_df}
\alias{ols_get_df}
\title{Degrees of freedom}
\usage{
ols_get_df(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Residual degrees of freedom.
}
\description{
Returns residual degrees of freedom.
}
\examples{
model <- lm(mpg ~ wt + cyl * hp * disp + gear * drat, data = mtcars)
ols_get_df(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-residual-qqplot.R
\name{ols_plot_resid_qq}
\alias{ols_plot_resid_qq}
\title{Residual QQ plot}
\usage{
ols_plot_resid_qq(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Graph for detecting violation of normality assumption.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_resid_qq(model)

}
\seealso{
Other residual diagnostics: 
\code{\link{ols_plot_resid_box}()},
\code{\link{ols_plot_resid_fit}()},
\code{\link{ols_plot_resid_hist}()},
\code{\link{ols_test_correlation}()},
\code{\link{ols_test_normality}()}
}
\concept{residual diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-all-possible-regression.R
\name{ols_step_all_possible_betas}
\alias{ols_step_all_possible_betas}
\title{All possible regression variable coefficients}
\usage{
ols_step_all_possible_betas(object, ...)
}
\arguments{
\item{object}{An object of class \code{lm}.}

\item{...}{Other arguments.}
}
\value{
\code{ols_step_all_possible_betas} returns a \code{data.frame} containing:

\item{model_index}{model number}
\item{predictor}{predictor}
\item{beta_coef}{coefficient for the predictor}
}
\description{
Returns the coefficients for each variable from each model.
}
\examples{
\dontrun{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_step_all_possible_betas(model)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-correlations.R
\name{ols_correlations}
\alias{ols_correlations}
\title{Part and partial correlations}
\usage{
ols_correlations(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
\code{ols_correlations} returns an object of class \code{"ols_correlations"}.
An object of class \code{"ols_correlations"} is a data frame containing the
following components:

\item{Zero-order}{zero order correlations}
\item{Partial}{partial correlations}
\item{Part}{part correlations}
}
\description{
Zero-order, part and partial correlations.
}
\details{
\code{ols_correlations()} returns the relative importance of independent
variables in determining response variable. How much each variable uniquely
contributes to rsquare over and above that which can be accounted for by the
other predictors? Zero order correlation is the Pearson correlation
coefficient between the dependent variable and the independent variables.
Part correlations indicates how much rsquare will decrease if that variable
is removed from the model and partial correlations indicates amount of
variance in response variable, which is not estimated by the other
independent variables in the model, but is estimated by the specific
variable.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_correlations(model)

}
\references{
Morrison, D. F. 1976. Multivariate statistical methods. New York: McGraw-Hill.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-data-auto.R
\docType{data}
\name{auto}
\alias{auto}
\title{Test Data Set}
\format{
An object of class \code{tbl_df} (inherits from \code{tbl}, \code{data.frame}) with 74 rows and 11 columns.
}
\usage{
auto
}
\description{
Test Data Set
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-data-surgical.R
\docType{data}
\name{surgical}
\alias{surgical}
\title{Surgical Unit Data Set}
\format{
A data frame with 54 rows and 9 variables:
\describe{
\item{bcs}{blood clotting score}
\item{pindex}{prognostic index}
\item{enzyme_test}{enzyme function test score}
\item{liver_test}{liver function test score}
\item{age}{age, in years}
\item{gender}{indicator variable for gender (0 = male, 1 = female)}
\item{alc_mod}{indicator variable for history of alcohol use (0 = None, 1 = Moderate)}
\item{alc_heavy}{indicator variable for history of alcohol use (0 = None, 1 = Heavy)}
\item{y}{Survival Time}
}
}
\source{
Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
Chicago, IL., McGraw Hill/Irwin.
}
\usage{
surgical
}
\description{
A dataset containing data about survival of patients undergoing liver operation.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-dffits-plot.R
\name{ols_plot_dffits}
\alias{ols_plot_dffits}
\title{DFFITS plot}
\usage{
ols_plot_dffits(model, size_adj_threshold = TRUE, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{size_adj_threshold}{logical; if \code{TRUE} (the default), size
adjusted threshold is used to determine influential observations.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a
plot object.}
}
\value{
\code{ols_plot_dffits} returns  a list containing the
following components:

\item{outliers}{a \code{data.frame} with observation number and \code{DFFITs} that exceed \code{threshold}}
\item{threshold}{\code{threshold} for classifying an observation as an outlier}
}
\description{
Plot for detecting influential observations using DFFITs.
}
\details{
DFFIT - difference in fits, is used to identify influential data points. It
quantifies the number of standard deviations that the fitted value changes
when the ith data point is omitted.

Steps to compute DFFITs:

\itemize{
\item Delete observations one at a time.
\item Refit the regression model on remaining \eqn{n - 1} observations
\item examine how much all of the fitted values change when the ith observation is deleted.
}

An observation is deemed influential if the absolute value of its DFFITS value is greater than:
\deqn{2\sqrt((p + 1) / (n - p -1))}

A size-adjusted cutoff recommended by Belsley, Kuh, and Welsch is
\deqn{2\sqrt(p / n)} and is used by default in \strong{olsrr}.

where \code{n} is the number of observations and \code{p} is the number of predictors including intercept.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_dffits(model)
ols_plot_dffits(model, size_adj_threshold = FALSE)

}
\references{
Belsley, David A.; Kuh, Edwin; Welsh, Roy E. (1980). Regression
Diagnostics: Identifying Influential Data and Sources of Collinearity.

Wiley Series in Probability and Mathematical Statistics.
New York: John Wiley & Sons. ISBN 0-471-05856-4.
}
\seealso{
\code{\link[=ols_plot_dfbetas]{ols_plot_dfbetas()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-standardized-residual-chart.R
\name{ols_plot_resid_stand}
\alias{ols_plot_resid_stand}
\title{Standardized residual chart}
\usage{
ols_plot_resid_stand(model, threshold = NULL, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{threshold}{Threshold for detecting outliers. Default is 2.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\value{
\code{ols_plot_resid_stand} returns  a list containing the
following components:

\item{outliers}{a \code{data.frame} with observation number and \code{standardized resiudals} that
exceed \code{threshold}} for classifying an observation as an outlier
\item{threshold}{\code{threshold} for classifying an observation as an outlier}
}
\description{
Chart for identifying outliers.
}
\details{
Standardized residual (internally studentized) is the residual divided by
estimated standard deviation.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_resid_stand(model)
ols_plot_resid_stand(model, threshold = 3)

}
\seealso{
\code{\link[=ols_plot_resid_stud]{ols_plot_resid_stud()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-stepaic-both-regression.R
\name{ols_step_both_sbc}
\alias{ols_step_both_sbc}
\alias{ols_step_both_sbc.default}
\alias{plot.ols_step_both_sbc}
\title{Stepwise SBC regression}
\usage{
ols_step_both_sbc(model, ...)

\method{ols_step_both_sbc}{default}(
  model,
  include = NULL,
  exclude = NULL,
  progress = FALSE,
  details = FALSE,
  steps = NULL,
  ...
)

\method{plot}{ols_step_both_sbc}(x, print_plot = TRUE, details = TRUE, digits = 3, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{...}{Other arguments.}

\item{include}{Character or numeric vector; variables to be included in selection process.}

\item{exclude}{Character or numeric vector; variables to be excluded from selection process.}

\item{progress}{Logical; if \code{TRUE}, will display variable selection progress.}

\item{details}{Logical; if \code{TRUE}, details of variable selection will
be printed on screen.}

\item{steps}{Number of steps after which the stepwise procedures should stop.}

\item{x}{An object of class \code{ols_step_both_*}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}

\item{digits}{Number of decimal places to display.}
}
\value{
List containing the following components:

\item{model}{final model; an object of class \code{lm}}
\item{metrics}{selection metrics}
\item{others}{list; info used for plotting and printing}
}
\description{
Build regression model from a set of candidate predictor variables by
entering and removing predictors based on schwarz bayesian criterion, in a
stepwise manner until there is no variable left to enter or remove any more.
}
\examples{
\dontrun{
# stepwise regression
model <- lm(y ~ ., data = stepdata)
ols_step_both_sbc(model)

# stepwise regression plot
model <- lm(y ~ ., data = stepdata)
k <- ols_step_both_sbc(model)
plot(k)

# selection metrics
k$metrics

# final model
k$model

# include or exclude variables
# force variable to be included in selection process
model <- lm(y ~ ., data = stepdata)

ols_step_both_sbc(model, include = c("x6"))

# use index of variable instead of name
ols_step_both_sbc(model, include = c(6))

# force variable to be excluded from selection process
ols_step_both_sbc(model, exclude = c("x2"))

# use index of variable instead of name
ols_step_both_sbc(model, exclude = c(2))

# include & exclude variables in the selection process
ols_step_both_sbc(model, include = c("x6"), exclude = c("x2"))

# use index of variable instead of name
ols_step_both_sbc(model, include = c(6), exclude = c(2))

# steps
ols_step_both_sbc(model, steps = 3)
}

}
\references{
Venables, W. N. and Ripley, B. D. (2002) Modern Applied Statistics with S. Fourth edition. Springer.
}
\seealso{
Other both direction selection procedures: 
\code{\link{ols_step_both_adj_r2}()},
\code{\link{ols_step_both_aic}()},
\code{\link{ols_step_both_r2}()},
\code{\link{ols_step_both_sbic}()}
}
\concept{both direction selection procedures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-added-variable-plot.R
\name{ols_plot_added_variable}
\alias{ols_plot_added_variable}
\title{Added variable plots}
\usage{
ols_plot_added_variable(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Added variable plot provides information about the marginal importance of a
predictor variable, given the other predictor variables already in
the model. It shows the marginal importance of the variable in reducing the
residual variability.
}
\details{
The added variable plot was introduced by Mosteller and Tukey
(1977). It enables us to visualize the regression coefficient of a new
variable being considered to be included in a model. The plot can be
constructed for each predictor variable.

Let us assume we want to test the effect of adding/removing variable \emph{X} from a
model. Let the response variable of the model be \emph{Y}

Steps to construct an added variable plot:

\itemize{
\item Regress \emph{Y} on all variables other than \emph{X} and store the residuals (\emph{Y} residuals).
\item Regress \emph{X} on all the other variables included in the model (\emph{X} residuals).
\item Construct a scatter plot of \emph{Y} residuals and \emph{X} residuals.
}

What do the \emph{Y} and \emph{X} residuals represent? The \emph{Y} residuals represent the part
of \strong{Y} not explained by all the variables other than X. The \emph{X} residuals
represent the part of \strong{X} not explained by other variables. The slope of the line
fitted to the points in the added variable plot is equal to the regression
coefficient when \strong{Y} is regressed on all variables including \strong{X}.

A strong linear relationship in the added variable plot indicates the increased
importance of the contribution of \strong{X} to the model already containing the
other predictors.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_added_variable(model)

}
\references{
Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.

Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
Chicago, IL., McGraw Hill/Irwin.
}
\seealso{
\code{\link[=ols_plot_resid_regressor]{ols_plot_resid_regressor()}}, \code{\link[=ols_plot_comp_plus_resid]{ols_plot_comp_plus_resid()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-residual-plus-component-plot.R
\name{ols_plot_comp_plus_resid}
\alias{ols_plot_comp_plus_resid}
\title{Residual plus component plot}
\usage{
ols_plot_comp_plus_resid(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
The residual plus component plot indicates whether any non-linearity is
present in the relationship between response and predictor variables and can
suggest possible transformations for linearizing the data.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_comp_plus_resid(model)

}
\references{
Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.

Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
Chicago, IL., McGraw Hill/Irwin.
}
\seealso{
\code{\link[=ols_plot_added_variable]{ols_plot_added_variable()}}, \code{\link[=ols_plot_resid_regressor]{ols_plot_resid_regressor()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-stepaic-backward-regression.R
\name{ols_step_backward_sbic}
\alias{ols_step_backward_sbic}
\alias{ols_step_backward_sbic.default}
\alias{plot.ols_step_backward_sbic}
\title{Stepwise SBIC backward regression}
\usage{
ols_step_backward_sbic(model, ...)

\method{ols_step_backward_sbic}{default}(
  model,
  include = NULL,
  exclude = NULL,
  progress = FALSE,
  details = FALSE,
  steps = NULL,
  ...
)

\method{plot}{ols_step_backward_sbic}(x, print_plot = TRUE, details = TRUE, digits = 3, ...)
}
\arguments{
\item{model}{An object of class \code{lm}; the model should include all
candidate predictor variables.}

\item{...}{Other arguments.}

\item{include}{Character or numeric vector; variables to be included in selection process.}

\item{exclude}{Character or numeric vector; variables to be excluded from selection process.}

\item{progress}{Logical; if \code{TRUE}, will display variable selection progress.}

\item{details}{Logical; if \code{TRUE}, will print the regression result at
each step.}

\item{steps}{Number of steps after which the stepwise procedures should stop.}

\item{x}{An object of class \code{ols_step_backward_*}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}

\item{digits}{Number of decimal places to display.}
}
\value{
List containing the following components:

\item{model}{final model; an object of class \code{lm}}
\item{metrics}{selection metrics}
\item{others}{list; info used for plotting and printing}
}
\description{
Build regression model from a set of candidate predictor variables by
removing predictors based on sawa bayesian criterion, in a stepwise
manner until there is no variable left to remove any more.
}
\examples{
# stepwise backward regression
model <- lm(y ~ ., data = surgical)
ols_step_backward_sbic(model)

# stepwise backward regression plot
model <- lm(y ~ ., data = surgical)
k <- ols_step_backward_sbic(model)
plot(k)

# selection metrics
k$metrics

# final model
k$model

# include or exclude variable
# force variables to be included in the selection process
ols_step_backward_sbic(model, include = c("alc_mod", "gender"))

# use index of variable instead of name
ols_step_backward_sbic(model, include = c(7, 6))

# force variable to be excluded from selection process
ols_step_backward_sbic(model, exclude = c("alc_heavy", "bcs"))

# use index of variable instead of name
ols_step_backward_sbic(model, exclude = c(8, 1))

# steps
ols_step_backward_sbic(model, steps = 2)

}
\references{
Venables, W. N. and Ripley, B. D. (2002) Modern Applied Statistics with S. Fourth edition. Springer.
}
\seealso{
Other backward selection procedures: 
\code{\link{ols_step_backward_adj_r2}()},
\code{\link{ols_step_backward_aic}()},
\code{\link{ols_step_backward_p}()},
\code{\link{ols_step_backward_r2}()},
\code{\link{ols_step_backward_sbc}()}
}
\concept{backward selection procedures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-plots-data.R
\name{ols_prep_outlier_obs}
\alias{ols_prep_outlier_obs}
\title{Cooks' D outlier observations}
\usage{
ols_prep_outlier_obs(k)
}
\arguments{
\item{k}{Cooks' d bar plot data.}
}
\description{
Identify outliers in cook's d plot.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
k <- ols_prep_cdplot_data(model)
ols_prep_outlier_obs(k)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-influence-measures.R
\name{ols_leverage}
\alias{ols_leverage}
\title{Leverage}
\usage{
ols_leverage(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Leverage of the model.
}
\description{
The leverage of an observation is based on how much the observation's value
on the predictor variable differs from the mean of the predictor variable.
The greater an observation's leverage, the more potential it has to be an
influential observation.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_leverage(model)

}
\references{
Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
Chicago, IL., McGraw Hill/Irwin.
}
\seealso{
Other influence measures: 
\code{\link{ols_hadi}()},
\code{\link{ols_pred_rsq}()},
\code{\link{ols_press}()}
}
\concept{influence measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-model-info.R
\name{ols_count_parameters}
\alias{ols_count_parameters}
\title{Parameters count}
\usage{
ols_count_parameters(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Number of parameters.
}
\description{
Returns the number of parameters from a model.
}
\examples{
model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
ols_count_parameters(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-stepwise-backward-regression.R
\name{ols_step_backward_p}
\alias{ols_step_backward_p}
\alias{ols_step_backward_p.default}
\alias{plot.ols_step_backward_p}
\title{Stepwise backward regression}
\usage{
ols_step_backward_p(model, ...)

\method{ols_step_backward_p}{default}(
  model,
  p_val = 0.3,
  include = NULL,
  exclude = NULL,
  hierarchical = FALSE,
  progress = FALSE,
  details = FALSE,
  steps = NULL,
  ...
)

\method{plot}{ols_step_backward_p}(x, model = NA, print_plot = TRUE, details = TRUE, ...)
}
\arguments{
\item{model}{An object of class \code{lm}; the model should include all
candidate predictor variables.}

\item{...}{Other inputs.}

\item{p_val}{p value; variables with p more than \code{p_val} will be removed
from the model.}

\item{include}{Character or numeric vector; variables to be included in selection process.}

\item{exclude}{Character or numeric vector; variables to be excluded from selection process.}

\item{hierarchical}{Logical; if \code{TRUE}, performs hierarchical selection.}

\item{progress}{Logical; if \code{TRUE}, will display variable selection progress.}

\item{details}{Logical; if \code{TRUE}, will print the regression result at
each step.}

\item{steps}{Number of steps after which the stepwise procedures should stop.}

\item{x}{An object of class \code{ols_step_backward_p}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\value{
\code{ols_step_backward_p} returns an object of class \code{"ols_step_backward_p"}.
An object of class \code{"ols_step_backward_p"} is a list containing the
following components:

\item{model}{final model; an object of class \code{lm}}
\item{metrics}{selection metrics}
}
\description{
Build regression model from a set of candidate predictor variables by
removing predictors based on p values, in a stepwise manner until there is
no variable left to remove any more.
}
\examples{
# stepwise backward regression
model <- lm(y ~ ., data = surgical)
ols_step_backward_p(model)

# stepwise backward regression plot
model <- lm(y ~ ., data = surgical)
k <- ols_step_backward_p(model)
plot(k)

# selection metrics
k$metrics

# final model
k$model

# include or exclude variables
# force variable to be included in selection process
ols_step_backward_p(model, include = c("age", "alc_mod"))

# use index of variable instead of name
ols_step_backward_p(model, include = c(5, 7))

# force variable to be excluded from selection process
ols_step_backward_p(model, exclude = c("pindex"))

# use index of variable instead of name
ols_step_backward_p(model, exclude = c(2))

# hierarchical selection
model <- lm(y ~ bcs + alc_heavy + pindex + age + alc_mod, data = surgical)
ols_step_backward_p(model, 0.1, hierarchical = TRUE)

# plot
k <- ols_step_backward_p(model, 0.1, hierarchical = TRUE)
plot(k)

# steps
ols_step_backward_p(model, steps = 2)


}
\references{
Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
}
\seealso{
Other backward selection procedures: 
\code{\link{ols_step_backward_adj_r2}()},
\code{\link{ols_step_backward_aic}()},
\code{\link{ols_step_backward_r2}()},
\code{\link{ols_step_backward_sbc}()},
\code{\link{ols_step_backward_sbic}()}
}
\concept{backward selection procedures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-plots-data.R
\name{ols_prep_rfsplot_fmdata}
\alias{ols_prep_rfsplot_fmdata}
\alias{ols_prep_rfsplot_rsdata}
\title{Residual fit spread plot data}
\usage{
ols_prep_rfsplot_fmdata(model)

ols_prep_rfsplot_rsdata(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\description{
Data for generating residual fit spread plot.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_prep_rfsplot_fmdata(model)
ols_prep_rfsplot_rsdata(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-score-test.R
\name{ols_test_score}
\alias{ols_test_score}
\title{Score test}
\usage{
ols_test_score(model, fitted_values = TRUE, rhs = FALSE, vars = NULL)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{fitted_values}{Logical; if TRUE, use fitted values of regression model.}

\item{rhs}{Logical; if TRUE, specifies that tests for heteroskedasticity be
performed for the right-hand-side (explanatory) variables of the fitted
regression model.}

\item{vars}{Variables to be used for for heteroskedasticity test.}
}
\value{
\code{ols_test_score} returns an object of class \code{"ols_test_score"}.
An object of class \code{"ols_test_score"} is a list containing the
following components:

\item{score}{f statistic}
\item{p}{p value of \code{score}}
\item{df}{degrees of freedom}
\item{fv}{fitted values of the regression model}
\item{rhs}{names of explanatory variables of fitted regression model}
\item{resp}{response variable}
\item{preds}{predictors}
}
\description{
Test for heteroskedasticity under the assumption that the errors are
independent and identically distributed (i.i.d.).
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt, data = mtcars)

# using fitted values of the model
ols_test_score(model)

# using predictors from the model
ols_test_score(model, rhs = TRUE)

# specify predictors from the model
ols_test_score(model, vars = c('disp', 'wt'))

}
\references{
Breusch, T. S. and Pagan, A. R. (1979) A simple test for heteroscedasticity and random coefficient variation. Econometrica 47, 1287–1294.

Cook, R. D. and Weisberg, S. (1983) Diagnostics for heteroscedasticity in regression. Biometrika 70, 1–10.

Koenker, R. 1981. A note on studentizing a test for heteroskedasticity. Journal of Econometrics 17: 107–112.
}
\seealso{
Other heteroskedasticity tests: 
\code{\link{ols_test_bartlett}()},
\code{\link{ols_test_breusch_pagan}()},
\code{\link{ols_test_f}()}
}
\concept{heteroskedasticity tests}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-launch-app.R
\name{ols_launch_app}
\alias{ols_launch_app}
\title{Launch shiny app}
\usage{
ols_launch_app()
}
\description{
Launches shiny app for interactive model building.
}
\examples{
\dontrun{
ols_launch_app()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-regression.R
\name{ols_regress}
\alias{ols_regress}
\alias{ols_regress.lm}
\title{Ordinary least squares regression}
\usage{
ols_regress(object, ...)

\method{ols_regress}{lm}(object, ...)
}
\arguments{
\item{object}{An object of class "formula" (or one that can be coerced to
that class): a symbolic description of the model to be fitted or class
\code{lm}.}

\item{...}{Other inputs.}
}
\value{
\code{ols_regress} returns an object of class \code{"ols_regress"}.
An object of class \code{"ols_regress"} is a list containing the following
components:

\item{r}{square root of rsquare, correlation between observed and predicted values of dependent variable}
\item{rsq}{coefficient of determination or r-square}
\item{adjr}{adjusted rsquare}
\item{rmse}{root mean squared error}
\item{cv}{coefficient of variation}
\item{mse}{mean squared error}
\item{mae}{mean absolute error}
\item{aic}{akaike information criteria}
\item{sbc}{bayesian information criteria}
\item{sbic}{sawa bayesian information criteria}
\item{prsq}{predicted rsquare}
\item{error_df}{residual degrees of freedom}
\item{model_df}{regression degrees of freedom}
\item{total_df}{total degrees of freedom}
\item{ess}{error sum of squares}
\item{rss}{regression sum of squares}
\item{tss}{total sum of squares}
\item{rms}{regression mean square}
\item{ems}{error mean square}
\item{f}{f statistis}
\item{p}{p-value for \code{f}}
\item{n}{number of predictors including intercept}
\item{betas}{betas; estimated coefficients}
\item{sbetas}{standardized betas}
\item{std_errors}{standard errors}
\item{tvalues}{t values}
\item{pvalues}{p-value of \code{tvalues}}
\item{df}{degrees of freedom of \code{betas}}
\item{conf_lm}{confidence intervals for coefficients}
\item{title}{title for the model}
\item{dependent}{character vector; name of the dependent variable}
\item{predictors}{character vector; name of the predictor variables}
\item{mvars}{character vector; name of the predictor variables including intercept}
\item{model}{input model for \code{ols_regress}}
}
\description{
Ordinary least squares regression.
}
\section{Interaction Terms}{

If the model includes interaction terms, the standardized betas
are computed after scaling and centering the predictors.
}

\examples{
ols_regress(mpg ~ disp + hp + wt, data = mtcars)

# if model includes interaction terms set iterm to TRUE
ols_regress(mpg ~ disp * wt, data = mtcars, iterm = TRUE)

}
\references{
https://www.ssc.wisc.edu/~hemken/Stataworkshops/stdBeta/Getting\%20Standardized\%20Coefficients\%20Right.pdf
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-studentized-residual-plot.R
\name{ols_plot_resid_stud}
\alias{ols_plot_resid_stud}
\title{Studentized residual plot}
\usage{
ols_plot_resid_stud(model, threshold = NULL, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{threshold}{Threshold for detecting outliers. Default is 3.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\value{
\code{ols_plot_resid_stud} returns  a list containing the
following components:

\item{outliers}{a \code{data.frame} with observation number and \code{studentized residuals} that
exceed \code{threshold}} for classifying an observation as an outlier
\item{threshold}{\code{threshold} for classifying an observation as an outlier}
}
\description{
Graph for identifying outliers.
}
\details{
Studentized deleted residuals (or externally studentized residuals) is the
deleted residual divided by its estimated standard deviation. Studentized
residuals are going to be more effective for detecting outlying Y
observations than standardized residuals. If an observation has an externally
studentized residual that is larger than 3 (in absolute value) we can call
it an outlier.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_resid_stud(model)
ols_plot_resid_stud(model, threshold = 2)

}
\seealso{
\code{\link[=ols_plot_resid_stand]{ols_plot_resid_stand()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-model-info.R
\name{ols_get_predicted}
\alias{ols_get_predicted}
\title{Fitted values}
\usage{
ols_get_predicted(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Fitted values.
}
\description{
Returns the fitted values.
}
\examples{
model <- lm(mpg ~ wt + cyl + hp + disp + gear + drat, data = mtcars)
ols_get_predicted(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_sbic}
\alias{ols_sbic}
\title{Sawa's bayesian information criterion}
\usage{
ols_sbic(model, full_model)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{full_model}{An object of class \code{lm}.}
}
\value{
Sawa's Bayesian Information Criterion
}
\description{
Sawa's bayesian information criterion for model selection.
}
\details{
Sawa (1978) developed a model selection criterion that was derived from a
Bayesian modification of the AIC criterion. Sawa's Bayesian Information
Criterion (BIC) is a function of the number of observations n, the SSE, the
pure error variance fitting the full model, and the number of independent
variables including the intercept.

\deqn{SBIC = n * ln(SSE / n) + 2(p + 2)q - 2(q^2)}

where \eqn{q = n(\sigma^2)/SSE}, \emph{n} is the sample size, \emph{p} is the number of model parameters including intercept
\emph{SSE} is the residual sum of squares.
}
\examples{
full_model <- lm(mpg ~ ., data = mtcars)
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_sbic(model, full_model)

}
\references{
Sawa, T. (1978). “Information Criteria for Discriminating among Alternative Regression Models.” Econometrica
46:1273–1282.

Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
New York: John Wiley & Sons.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_aic}()},
\code{\link{ols_apc}()},
\code{\link{ols_fpe}()},
\code{\link{ols_hsp}()},
\code{\link{ols_mallows_cp}()},
\code{\link{ols_msep}()},
\code{\link{ols_sbc}()}
}
\concept{model selection criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_aic}
\alias{ols_aic}
\title{Akaike information criterion}
\usage{
ols_aic(model, method = c("R", "STATA", "SAS"), corrected = FALSE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{method}{A character vector; specify the method to compute AIC. Valid
options include R, STATA and SAS.}

\item{corrected}{Logical; if \code{TRUE}, returns corrected akaike information criterion for SAS method.}
}
\value{
Akaike information criterion of the model.
}
\description{
Akaike information criterion for model selection.
}
\details{
AIC provides a means for model selection. Given a collection of models for
the data, AIC estimates the quality of each model, relative to each of the
other models. R and STATA use loglikelihood to compute AIC. SAS uses residual
sum of squares. Below is the formula in each case:

\emph{R & STATA}
\deqn{AIC = -2(loglikelihood) + 2p}

\emph{SAS}
\deqn{AIC = n * ln(SSE / n) + 2p}

\emph{corrected}
\deqn{AIC = n * ln(SSE / n) + ((n * (n + p)) / (n - p - 2))}

where \emph{n} is the sample size and \emph{p} is the number of model parameters including intercept.
}
\examples{
# using R computation method
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_aic(model)

# using STATA computation method
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_aic(model, method = 'STATA')

# using SAS computation method
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_aic(model, method = 'SAS')

# corrected akaike information criterion
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_aic(model, method = 'SAS', corrected = TRUE)

}
\references{
Akaike, H. (1969). “Fitting Autoregressive Models for Prediction.” Annals of the Institute of Statistical
Mathematics 21:243–247.

Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
New York: John Wiley & Sons.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_apc}()},
\code{\link{ols_fpe}()},
\code{\link{ols_hsp}()},
\code{\link{ols_mallows_cp}()},
\code{\link{ols_msep}()},
\code{\link{ols_sbc}()},
\code{\link{ols_sbic}()}
}
\concept{model selection criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-outlier-test.R
\name{ols_test_outlier}
\alias{ols_test_outlier}
\title{Bonferroni Outlier Test}
\usage{
ols_test_outlier(model, cut_off = 0.05, n_max = 10, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{cut_off}{Bonferroni p-values cut off for reporting observations.}

\item{n_max}{Maximum number of observations to report, default is 10.}

\item{...}{Other arguments.}
}
\description{
Detect outliers using Bonferroni p values.
}
\examples{
# model
model <- lm(y ~ ., data = surgical)
ols_test_outlier(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-influence-measures.R
\name{ols_hadi}
\alias{ols_hadi}
\title{Hadi's influence measure}
\usage{
ols_hadi(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Hadi's measure of the model.
}
\description{
Measure of influence based on the fact that influential observations in
either the response variable or in the predictors or both.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_hadi(model)

}
\references{
Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
}
\seealso{
Other influence measures: 
\code{\link{ols_leverage}()},
\code{\link{ols_pred_rsq}()},
\code{\link{ols_press}()}
}
\concept{influence measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-data-cement.R
\docType{data}
\name{cement}
\alias{cement}
\title{Test Data Set}
\format{
An object of class \code{data.frame} with 13 rows and 6 columns.
}
\usage{
cement
}
\description{
Test Data Set
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_fpe}
\alias{ols_fpe}
\title{Final prediction error}
\usage{
ols_fpe(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Final prediction error of the model.
}
\description{
Estimated mean square error of prediction.
}
\details{
Computes the estimated mean square error of prediction for each model
selected assuming that the values of the regressors are fixed and that the
model is correct.

\deqn{MSE((n + p) / n)}

where \eqn{MSE = SSE / (n - p)}, n is the sample size and p is the number of predictors including the intercept
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_fpe(model)

}
\references{
Akaike, H. (1969). “Fitting Autoregressive Models for Prediction.” Annals of the Institute of Statistical
Mathematics 21:243–247.

Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
New York: John Wiley & Sons.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_aic}()},
\code{\link{ols_apc}()},
\code{\link{ols_hsp}()},
\code{\link{ols_mallows_cp}()},
\code{\link{ols_msep}()},
\code{\link{ols_sbc}()},
\code{\link{ols_sbic}()}
}
\concept{model selection criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-influence-measures.R
\name{ols_press}
\alias{ols_press}
\title{PRESS}
\usage{
ols_press(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Predicted sum of squares of the model.
}
\description{
PRESS (prediction sum of squares) tells you how well the model will predict
new data.
}
\details{
The prediction sum of squares (PRESS) is the sum of squares of the prediction
error. Each fitted to obtain the predicted value for the ith observation. Use
PRESS to assess your model's predictive ability. Usually, the smaller the
PRESS value, the better the model's predictive ability.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_press(model)

}
\references{
Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
Chicago, IL., McGraw Hill/Irwin.
}
\seealso{
Other influence measures: 
\code{\link{ols_hadi}()},
\code{\link{ols_leverage}()},
\code{\link{ols_pred_rsq}()}
}
\concept{influence measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-anova.R
\name{ols_anova}
\alias{ols_anova}
\title{ANOVA}
\usage{
ols_anova(model, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{...}{Other arguments.}
}
\value{
\code{ols_anova} returns an object of class \code{"ols_anova"}.
An object of class \code{"ols_anova"} is a list containing the following
components:

\item{error_df}{residual degrees of freedom}
\item{model_df}{regression degrees of freedom}
\item{total_df}{total degrees of freedom}
\item{ess}{error sum of squares}
\item{rss}{regression sum of squares}
\item{tss}{total sum of squares}
\item{rms}{regression mean square}
\item{ems}{error mean square}
\item{f}{f statistis}
\item{p}{p-value for \code{f}}
}
\description{
ANOVA for regression model.
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt, data = mtcars)

# anova
ols_anova(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-fitted-line-properties.R
\name{ols_plot_reg_line}
\alias{ols_plot_reg_line}
\title{Simple linear regression line}
\usage{
ols_plot_reg_line(response, predictor, print_plot = TRUE)
}
\arguments{
\item{response}{Response variable.}

\item{predictor}{Predictor variable.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Plot to demonstrate that the regression line always passes  through mean of
the response and predictor variables.
}
\examples{
ols_plot_reg_line(mtcars$mpg, mtcars$disp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-best-subsets-regression.R
\name{ols_step_best_subset}
\alias{ols_step_best_subset}
\alias{ols_step_best_subset.default}
\alias{plot.ols_step_best_subset}
\title{Best subsets regression}
\usage{
ols_step_best_subset(model, ...)

\method{ols_step_best_subset}{default}(
  model,
  max_order = NULL,
  include = NULL,
  exclude = NULL,
  metric = c("rsquare", "adjr", "predrsq", "cp", "aic", "sbic", "sbc", "msep", "fpe",
    "apc", "hsp"),
  ...
)

\method{plot}{ols_step_best_subset}(x, model = NA, print_plot = TRUE, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{...}{Other inputs.}

\item{max_order}{Maximum subset order.}

\item{include}{Character or numeric vector; variables to be included in selection process.}

\item{exclude}{Character or numeric vector; variables to be excluded from selection process.}

\item{metric}{Metric to select model.}

\item{x}{An object of class \code{ols_step_best_subset}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\value{
\code{ols_step_best_subset} returns an object of class \code{"ols_step_best_subset"}.
An object of class \code{"ols_step_best_subset"} is a list containing the following:

\item{metrics}{selection metrics}
}
\description{
Select the subset of predictors that do the best at meeting some
well-defined objective criterion, such as having the largest R2 value or the
smallest MSE, Mallow's Cp or AIC. The default metric used for selecting the
model is R2 but the user can choose any of the other available metrics.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_step_best_subset(model)
ols_step_best_subset(model, metric = "adjr")
ols_step_best_subset(model, metric = "cp")

# maximum subset
model <- lm(mpg ~ disp + hp + drat + wt + qsec, data = mtcars)
ols_step_best_subset(model, max_order = 3)

# plot
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
k <- ols_step_best_subset(model)
plot(k)

# return only models including `qsec`
ols_step_best_subset(model, include = c("qsec"))

# exclude `hp` from selection process
ols_step_best_subset(model, exclude = c("hp"))

}
\references{
Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
Chicago, IL., McGraw Hill/Irwin.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-data-step.R
\docType{data}
\name{stepdata}
\alias{stepdata}
\title{Test Data Set}
\format{
An object of class \code{data.frame} with 20000 rows and 7 columns.
}
\usage{
stepdata
}
\description{
Test Data Set
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-residual-fit-spread-plot.R
\name{ols_plot_resid_fit_spread}
\alias{ols_plot_resid_fit_spread}
\alias{ols_plot_fm}
\alias{ols_plot_resid_spread}
\title{Residual fit spread plot}
\usage{
ols_plot_resid_fit_spread(model, print_plot = TRUE)

ols_plot_fm(model, print_plot = TRUE)

ols_plot_resid_spread(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Plot to detect non-linearity, influential observations and outliers.
}
\details{
Consists of side-by-side quantile plots of the centered fit and the
residuals. It shows how much variation in the data is explained by the fit
and how much remains in the residuals. For inappropriate models, the spread
of the residuals in such a plot is often greater than the spread of the
centered fit.
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt, data = mtcars)

# residual fit spread plot
ols_plot_resid_fit_spread(model)

# fit mean plot
ols_plot_fm(model)

# residual spread plot
ols_plot_resid_spread(model)

}
\references{
Cleveland, W. S. (1993). Visualizing Data. Summit, NJ: Hobart Press.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_msep}
\alias{ols_msep}
\title{MSEP}
\usage{
ols_msep(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Estimated error of prediction of the model.
}
\description{
Estimated error of prediction, assuming multivariate normality.
}
\details{
Computes the estimated mean square error of prediction assuming that both
independent and dependent variables are multivariate normal.

\deqn{MSE(n + 1)(n - 2) / n(n - p - 1)}

where \eqn{MSE = SSE / (n - p)}, n is the sample size and p is the number of
predictors including the intercept
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_msep(model)

}
\references{
Stein, C. (1960). “Multiple Regression.” In Contributions to Probability and Statistics: Essays in Honor
of Harold Hotelling, edited by I. Olkin, S. G. Ghurye, W. Hoeffding, W. G. Madow, and H. B. Mann,
264–305. Stanford, CA: Stanford University Press.

Darlington, R. B. (1968). “Multiple Regression in Psychological Research and Practice.” Psychological
Bulletin 69:161–182.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_aic}()},
\code{\link{ols_apc}()},
\code{\link{ols_fpe}()},
\code{\link{ols_hsp}()},
\code{\link{ols_mallows_cp}()},
\code{\link{ols_sbc}()},
\code{\link{ols_sbic}()}
}
\concept{model selection criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_mallows_cp}
\alias{ols_mallows_cp}
\title{Mallow's Cp}
\usage{
ols_mallows_cp(model, fullmodel)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{fullmodel}{An object of class \code{lm}.}
}
\value{
Mallow's Cp of the model.
}
\description{
Mallow's Cp.
}
\details{
Mallows' Cp statistic estimates the size of the bias that is introduced into
the predicted responses by having an underspecified model. Use Mallows' Cp
to choose between multiple regression models. Look for models where
Mallows' Cp is small and close to the number of predictors in the model plus
the constant (p).
}
\examples{
full_model <- lm(mpg ~ ., data = mtcars)
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_mallows_cp(model, full_model)

}
\references{
Hocking, R. R. (1976). “The Analysis and Selection of Variables in a Linear Regression.” Biometrics
32:1–50.

Mallows, C. L. (1973). “Some Comments on Cp.” Technometrics 15:661–675.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_aic}()},
\code{\link{ols_apc}()},
\code{\link{ols_fpe}()},
\code{\link{ols_hsp}()},
\code{\link{ols_msep}()},
\code{\link{ols_sbc}()},
\code{\link{ols_sbic}()}
}
\concept{model selection criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-data-fitness.R
\docType{data}
\name{fitness}
\alias{fitness}
\title{Test Data Set}
\format{
An object of class \code{data.frame} with 31 rows and 7 columns.
}
\usage{
fitness
}
\description{
Test Data Set
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-model-performance.R
\name{ols_model_performance}
\alias{ols_model_performance}
\title{Model Performance}
\usage{
ols_model_performance(model, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{...}{Other arguments.}
}
\value{
\code{ols_model_performance} returns an object of class
\code{"ols_model_performance"}. An object of class \code{"ols_regress"} is a
list containing the following components:

\item{r}{square root of rsquare, correlation between observed and predicted values of dependent variable}
\item{rsq}{coefficient of determination or r-square}
\item{adjr}{adjusted rsquare}
\item{prsq}{predicted rsquare}
\item{aic}{akaike information criteria}
\item{sbc}{bayesian information criteria}
\item{sbic}{sawa bayesian information criteria}
\item{rmse}{root mean squared error}
}
\description{
Evaluate performance of regression models.
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt, data = mtcars)

# model performance
ols_model_performance(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-dfbetas-panel.R
\name{ols_plot_dfbetas}
\alias{ols_plot_dfbetas}
\title{DFBETAs panel}
\usage{
ols_plot_dfbetas(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\value{
list; \code{ols_plot_dfbetas} returns a list of \code{data.frame} (for intercept and each predictor)
with the observation number and DFBETA of observations that exceed the threshold for classifying
an observation as an outlier/influential observation.
}
\description{
Panel of plots to detect influential observations using DFBETAs.
}
\details{
DFBETA measures the difference in each parameter estimate with and without
the influential point. There is a DFBETA for each data point i.e if there are
n observations and k variables, there will be \eqn{n * k} DFBETAs. In
general, large values of DFBETAS indicate observations that are influential
in estimating a given parameter. Belsley, Kuh, and Welsch recommend 2 as a
general cutoff value to indicate influential observations and
\eqn{2/\sqrt(n)} as a size-adjusted cutoff.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_dfbetas(model)

}
\references{
Belsley, David A.; Kuh, Edwin; Welsh, Roy E. (1980). Regression
Diagnostics: Identifying Influential Data and Sources of Collinearity.

Wiley Series in Probability and Mathematical Statistics.
New York: John Wiley & Sons. pp. ISBN 0-471-05856-4.
}
\seealso{
\code{\link[=ols_plot_dffits]{ols_plot_dffits()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-resid-vs-regressor-plot-shiny.R
\name{rvsr_plot_shiny}
\alias{rvsr_plot_shiny}
\title{Residual vs regressors plot for shiny app}
\usage{
rvsr_plot_shiny(model, data, variable, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{data}{A \code{data.frame} or \code{tibble}.}

\item{variable}{Character; new predictor to be added to the \code{model}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Graph to determine whether we should add a new predictor to the
model already containing other predictors. The residuals from the model is
regressed on the new predictor and if the plot shows non random pattern,
you should consider adding the new predictor to the model.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
rvsr_plot_shiny(model, mtcars, 'drat')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-bartlett-test.R
\name{ols_test_bartlett}
\alias{ols_test_bartlett}
\alias{ols_test_bartlett.default}
\title{Bartlett test}
\usage{
ols_test_bartlett(data, ...)

\method{ols_test_bartlett}{default}(data, ..., group_var = NULL)
}
\arguments{
\item{data}{A \code{data.frame} or \code{tibble}.}

\item{...}{Columns in \code{data}.}

\item{group_var}{Grouping variable.}
}
\value{
\code{ols_test_bartlett} returns an object of class \code{"ols_test_bartlett"}.
An object of class \code{"ols_test_bartlett"} is a list containing the
following components:

\item{fstat}{f statistic}
\item{pval}{p-value of \code{fstat}}
\item{df}{degrees of freedom}
}
\description{
Test if k samples are from populations with equal variances.
}
\details{
Bartlett's test is used to test if variances across samples is equal.
It is sensitive to departures from normality. The Levene test
is an alternative test that is less sensitive to departures from normality.
}
\examples{
# using grouping variable
if (require("descriptr")) {
  library(descriptr)
  ols_test_bartlett(mtcarz, 'mpg', group_var = 'cyl')
}

# using variables
ols_test_bartlett(hsb, 'read', 'write')

}
\references{
Snedecor, George W. and Cochran, William G. (1989), Statistical Methods,
Eighth Edition, Iowa State University Press.
}
\seealso{
Other heteroskedasticity tests: 
\code{\link{ols_test_breusch_pagan}()},
\code{\link{ols_test_f}()},
\code{\link{ols_test_score}()}
}
\concept{heteroskedasticity tests}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-pure-error-anova.R
\name{ols_pure_error_anova}
\alias{ols_pure_error_anova}
\title{Lack of fit F test}
\usage{
ols_pure_error_anova(model, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{...}{Other parameters.}
}
\value{
\code{ols_pure_error_anova} returns an object of class
\code{"ols_pure_error_anova"}. An object of class \code{"ols_pure_error_anova"} is a
list containing the following components:

\item{lackoffit}{lack of fit sum of squares}
\item{pure_error}{pure error sum of squares}
\item{rss}{regression sum of squares}
\item{ess}{error sum of squares}
\item{total}{total sum of squares}
\item{rms}{regression mean square}
\item{ems}{error mean square}
\item{lms}{lack of fit mean square}
\item{pms}{pure error mean square}
\item{rf}{f statistic}
\item{lf}{lack of fit f statistic}
\item{pr}{p-value of f statistic}
\item{pl}{p-value pf lack of fit f statistic}
\item{mpred}{\code{data.frame} containing data for the response and predictor of the \code{model}}
\item{df_rss}{regression sum of squares degrees of freedom}
\item{df_ess}{error sum of squares degrees of freedom}
\item{df_lof}{lack of fit degrees of freedom}
\item{df_error}{pure error degrees of freedom}
\item{final}{data.frame; contains computed values used for the lack of fit f test}
\item{resp}{character vector; name of \code{response variable}}
\item{preds}{character vector; name of \code{predictor variable}}
}
\description{
Assess how much of the error in prediction is due to lack of model fit.
}
\details{
The residual sum of squares resulting from a regression can be decomposed
into 2 components:

\itemize{
\item Due to lack of fit
\item Due to random variation
}

If most of the error is due to lack of fit and not just random error, the
model should be discarded and a new model must be built.
}
\note{
The lack of fit F test works only with simple linear regression.
Moreover, it is important that the data contains repeat observations i.e.
replicates for at least one of the values of the predictor x. This
test generally only applies to datasets with plenty of replicates.
}
\examples{
model <- lm(mpg ~ disp, data = mtcars)
ols_pure_error_anova(model)

}
\references{
Kutner, MH, Nachtscheim CJ, Neter J and Li W., 2004, Applied Linear Statistical Models (5th edition).
Chicago, IL., McGraw Hill/Irwin.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-breusch-pagan-test.R
\name{ols_test_breusch_pagan}
\alias{ols_test_breusch_pagan}
\title{Breusch pagan test}
\usage{
ols_test_breusch_pagan(
  model,
  fitted.values = TRUE,
  rhs = FALSE,
  multiple = FALSE,
  p.adj = c("none", "bonferroni", "sidak", "holm"),
  vars = NA
)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{fitted.values}{Logical; if TRUE, use fitted values of regression model.}

\item{rhs}{Logical; if TRUE, specifies that tests for heteroskedasticity be
performed for the right-hand-side (explanatory) variables of the fitted
regression model.}

\item{multiple}{Logical; if TRUE, specifies that multiple testing be performed.}

\item{p.adj}{Adjustment for p value, the following options are available:
bonferroni, holm, sidak and none.}

\item{vars}{Variables to be used for heteroskedasticity test.}
}
\value{
\code{ols_test_breusch_pagan} returns an object of class \code{"ols_test_breusch_pagan"}.
An object of class \code{"ols_test_breusch_pagan"} is a list containing the
following components:

\item{bp}{breusch pagan statistic}
\item{p}{p-value of \code{bp}}
\item{fv}{fitted values of the regression model}
\item{rhs}{names of explanatory variables of fitted regression model}
\item{multiple}{logical value indicating if multiple tests should be performed}
\item{padj}{adjusted p values}
\item{vars}{variables to be used for heteroskedasticity test}
\item{resp}{response variable}
\item{preds}{predictors}
}
\description{
Test for constant variance. It assumes that the error terms are normally
distributed.
}
\details{
Breusch Pagan Test was introduced by Trevor Breusch and Adrian Pagan in 1979.
It is used to test for heteroskedasticity in a linear regression model.
It test whether variance of errors from a regression is dependent on the
values of a independent variable.

\itemize{
\item Null Hypothesis: Equal/constant variances
\item Alternative Hypothesis: Unequal/non-constant variances
}

Computation

\itemize{
\item Fit a regression model
\item Regress the squared residuals from the above model on the independent variables
\item Compute \eqn{nR^2}. It follows a chi square distribution with p -1 degrees of
freedom, where p is the number of independent variables, n is the sample size and
\eqn{R^2} is the coefficient of determination from the regression in step 2.
}
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)

# use fitted values of the model
ols_test_breusch_pagan(model)

# use independent variables of the model
ols_test_breusch_pagan(model, rhs = TRUE)

# use independent variables of the model and perform multiple tests
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE)

# bonferroni p value adjustment
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'bonferroni')

# sidak p value adjustment
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'sidak')

# holm's p value adjustment
ols_test_breusch_pagan(model, rhs = TRUE, multiple = TRUE, p.adj = 'holm')

}
\references{
T.S. Breusch & A.R. Pagan (1979), A Simple Test for Heteroscedasticity and
Random Coefficient Variation. Econometrica 47, 1287–1294

Cook, R. D.; Weisberg, S. (1983). "Diagnostics for Heteroskedasticity in Regression". Biometrika. 70 (1): 1–10.
}
\seealso{
Other heteroskedasticity tests: 
\code{\link{ols_test_bartlett}()},
\code{\link{ols_test_f}()},
\code{\link{ols_test_score}()}
}
\concept{heteroskedasticity tests}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-model-info.R
\name{ols_get_formula}
\alias{ols_get_formula}
\title{Model formula}
\usage{
ols_get_formula(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
An object of class \code{formula}.
}
\description{
Returns the model formula.
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_get_formula(model)

}
\concept{model info}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-all-possible-regression.R
\name{ols_step_all_possible}
\alias{ols_step_all_possible}
\alias{ols_step_all_possible.default}
\alias{plot.ols_step_all_possible}
\title{All possible regression}
\usage{
ols_step_all_possible(model, ...)

\method{ols_step_all_possible}{default}(model, max_order = NULL, ...)

\method{plot}{ols_step_all_possible}(x, model = NA, print_plot = TRUE, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{...}{Other arguments.}

\item{max_order}{Maximum subset order.}

\item{x}{An object of class \code{ols_step_all_possible}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\value{
\code{ols_step_all_possible} returns an object of class \code{"ols_step_all_possible"}.
An object of class \code{"ols_step_all_possible"} is a data frame containing the
following components:

\item{mindex}{model index}
\item{n}{number of predictors}
\item{predictors}{predictors in the model}
\item{rsquare}{rsquare of the model}
\item{adjr}{adjusted rsquare of the model}
\item{rmse}{root mean squared error of the model}
\item{predrsq}{predicted rsquare of the model}
\item{cp}{mallow's Cp}
\item{aic}{akaike information criteria}
\item{sbic}{sawa bayesian information criteria}
\item{sbc}{schwarz bayes information criteria}
\item{msep}{estimated MSE of prediction, assuming multivariate normality}
\item{fpe}{final prediction error}
\item{apc}{amemiya prediction criteria}
\item{hsp}{hocking's Sp}
}
\description{
Fits all regressions involving one regressor, two regressors, three
regressors, and so on. It tests all possible subsets of the set of potential
independent variables.
}
\examples{
model <- lm(mpg ~ disp + hp, data = mtcars)
k <- ols_step_all_possible(model)
k

# plot
plot(k)

# maximum subset
model <- lm(mpg ~ disp + hp + drat + wt + qsec, data = mtcars)
ols_step_all_possible(model, max_order = 3)

}
\references{
Mendenhall William and  Sinsich Terry, 2012, A Second Course in Statistics Regression Analysis (7th edition).
Prentice Hall
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-plots-data.R
\name{ols_prep_dfbeta_data}
\alias{ols_prep_dfbeta_data}
\title{DFBETAs plot data}
\usage{
ols_prep_dfbeta_data(d, threshold)
}
\arguments{
\item{d}{A \code{tibble} or \code{data.frame} with dfbetas.}

\item{threshold}{The threshold for outliers.}
}
\description{
Prepares the data for dfbetas plot.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
dfb <- dfbetas(model)
n <- nrow(dfb)
threshold <- 2 / sqrt(n)
dbetas  <- dfb[, 1]
df_data <- data.frame(obs = seq_len(n), dbetas = dbetas)
ols_prep_dfbeta_data(df_data, threshold)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-plots-data.R
\name{ols_prep_srchart_data}
\alias{ols_prep_srchart_data}
\title{Standardized residual chart data}
\usage{
ols_prep_srchart_data(model, threshold = NULL)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{threshold}{Threshold for detecting outliers. Default is 2.}
}
\description{
Generates data for standardized residual chart.
}
\examples{
model <- lm(read ~ write + math + science, data = hsb)
ols_prep_srchart_data(model)
ols_prep_srchart_data(model, threshold = 3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-hadi-plot.R
\name{ols_plot_hadi}
\alias{ols_plot_hadi}
\title{Hadi plot}
\usage{
ols_plot_hadi(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Hadi's measure of influence based on the fact that influential observations
can be present in either the response variable or in the predictors or both.
The plot is used to detect influential observations based on Hadi's measure.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_hadi(model)

}
\references{
Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
}
\seealso{
\code{\link[=ols_plot_resid_pot]{ols_plot_resid_pot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-influence-measures.R
\name{ols_pred_rsq}
\alias{ols_pred_rsq}
\title{Predicted rsquare}
\usage{
ols_pred_rsq(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Predicted rsquare of the model.
}
\description{
Use predicted rsquared to determine how well the model predicts responses for
new observations. Larger values of predicted R2 indicate models of greater
predictive ability.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_pred_rsq(model)

}
\seealso{
Other influence measures: 
\code{\link{ols_hadi}()},
\code{\link{ols_leverage}()},
\code{\link{ols_press}()}
}
\concept{influence measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-plots-data.R
\name{ols_prep_regress_y}
\alias{ols_prep_regress_y}
\title{Regress y on other predictors}
\usage{
ols_prep_regress_y(data, i)
}
\arguments{
\item{data}{A \code{data.frame}.}

\item{i}{A numeric vector (indicates the predictor in the model).}
}
\description{
Regress y on all the predictors except the ith predictor.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
data <- ols_prep_avplot_data(model)
ols_prep_regress_y(data, 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-dsresid-vs-pred-plot.R
\name{ols_plot_resid_stud_fit}
\alias{ols_plot_resid_stud_fit}
\title{Deleted studentized residual vs fitted values plot}
\usage{
ols_plot_resid_stud_fit(model, threshold = NULL, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{threshold}{Threshold for detecting outliers. Default is 2.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\value{
\code{ols_plot_resid_stud_fit} returns  a list containing the
following components:

\item{outliers}{a \code{data.frame} with observation number, fitted values and deleted studentized
residuals that exceed the \code{threshold} for classifying observations as
outliers/influential observations}
\item{threshold}{\code{threshold} for classifying an observation as an outlier/influential observation}
}
\description{
Plot for detecting violation of assumptions about residuals such as
non-linearity, constant variances and outliers. It can also be used to
examine model fit.
}
\details{
Studentized deleted residuals (or externally studentized residuals) is the
deleted residual divided by its estimated standard deviation. Studentized
residuals are going to be more effective for detecting outlying Y
observations than standardized residuals. If an observation has an externally
studentized residual that is larger than 2 (in absolute value) we can call
it an outlier.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_resid_stud_fit(model)
ols_plot_resid_stud_fit(model, threshold = 3)

}
\seealso{
\code{\link[=ols_plot_resid_lev]{ols_plot_resid_lev()}}, \code{\link[=ols_plot_resid_stand]{ols_plot_resid_stand()}},
\code{\link[=ols_plot_resid_stud]{ols_plot_resid_stud()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-normality-tests.R
\name{ols_test_correlation}
\alias{ols_test_correlation}
\title{Correlation test for normality}
\usage{
ols_test_correlation(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Correlation between fitted regression model residuals and expected
values of residuals.
}
\description{
Correlation between observed residuals and expected residuals under normality.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_correlation(model)

}
\seealso{
Other residual diagnostics: 
\code{\link{ols_plot_resid_box}()},
\code{\link{ols_plot_resid_fit}()},
\code{\link{ols_plot_resid_hist}()},
\code{\link{ols_plot_resid_qq}()},
\code{\link{ols_test_normality}()}
}
\concept{residual diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-residual-vs-predicted-plot.R
\name{ols_plot_resid_fit}
\alias{ols_plot_resid_fit}
\title{Residual vs fitted plot}
\usage{
ols_plot_resid_fit(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Scatter plot of residuals on the y axis and fitted values on the
x axis to detect non-linearity, unequal error variances, and outliers.
}
\details{
Characteristics of a well behaved residual vs fitted plot:

\itemize{
\item The residuals spread randomly around the 0 line indicating that the relationship is linear.
\item The residuals form an approximate horizontal band around the 0 line indicating homogeneity of error variance.
\item No one residual is visibly away from the random pattern of the residuals indicating that there are no outliers.
}
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_resid_fit(model)

}
\seealso{
Other residual diagnostics: 
\code{\link{ols_plot_resid_box}()},
\code{\link{ols_plot_resid_hist}()},
\code{\link{ols_plot_resid_qq}()},
\code{\link{ols_test_correlation}()},
\code{\link{ols_test_normality}()}
}
\concept{residual diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-observed-vs-predicted-plot.R
\name{ols_plot_obs_fit}
\alias{ols_plot_obs_fit}
\title{Observed vs fitted values plot}
\usage{
ols_plot_obs_fit(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Plot of observed vs fitted values to assess the fit of the model.
}
\details{
Ideally, all your points should be close to a regressed diagonal line. Draw
such a diagonal line within your graph and check out where the points lie. If
your model had a high R Square, all the points would be close to this
diagonal line. The lower the R Square, the weaker the Goodness of fit of your
model, the more foggy or dispersed your points are from this diagonal line.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_obs_fit(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_apc}
\alias{ols_apc}
\title{Amemiya's prediction criterion}
\usage{
ols_apc(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Amemiya's prediction error of the model.
}
\description{
Amemiya's prediction error.
}
\details{
Amemiya's Prediction Criterion penalizes R-squared more heavily than does
adjusted R-squared for each addition degree of freedom used on the
right-hand-side of the equation.  The lower the better for this criterion.

\deqn{((n + p) / (n - p))(1 - (R^2))}

where \emph{n} is the sample size, \emph{p} is the number of predictors including the intercept and
\emph{R^2} is the coefficient of determination.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_apc(model)

}
\references{
Amemiya, T. (1976). Selection of Regressors. Technical Report 225, Stanford University, Stanford, CA.

Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
New York: John Wiley & Sons.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_aic}()},
\code{\link{ols_fpe}()},
\code{\link{ols_hsp}()},
\code{\link{ols_mallows_cp}()},
\code{\link{ols_msep}()},
\code{\link{ols_sbc}()},
\code{\link{ols_sbic}()}
}
\concept{model selection criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-cooks-d-barplot.R
\name{ols_plot_cooksd_bar}
\alias{ols_plot_cooksd_bar}
\title{Cooks' D bar plot}
\usage{
ols_plot_cooksd_bar(model, type = 1, threshold = NULL, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{type}{An integer between 1 and 5 selecting one of the 5 methods for
computing the threshold.}

\item{threshold}{Threshold for detecting outliers.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a
plot object.}
}
\value{
\code{ols_plot_cooksd_bar} returns  a list containing the
following components:

\item{outliers}{a \code{data.frame} with observation number and \code{cooks distance} that exceed \code{threshold}}
\item{threshold}{\code{threshold} for classifying an observation as an outlier}
}
\description{
Bar Plot of cook's distance to detect observations that strongly influence
fitted values of the model.
}
\details{
Cook's distance was introduced by American statistician R Dennis Cook in
1977. It is used to identify influential data points. It depends on both the
residual and leverage i.e it takes it account both the \emph{x} value and
\emph{y} value of the observation.

Steps to compute Cook's distance:

\itemize{
\item Delete observations one at a time.
\item Refit the regression model on remaining \eqn{n - 1} observations
\item examine how much all of the fitted values change when the ith observation is deleted.
}

A data point having a large cook's d indicates that the data point strongly
influences the fitted values. There are several methods/formulas to compute
the threshold used for detecting or classifying observations as outliers and
we list them below.

\itemize{
\item \strong{Type 1} : 4 / n
\item \strong{Type 2} : 4 / (n - k - 1)
\item \strong{Type 3} : ~1
\item \strong{Type 4} : 1 / (n - k - 1)
\item \strong{Type 5} : 3 * mean(Vector of cook's distance values)
}

where \strong{n} and \strong{k} stand for

\itemize{
\item \strong{n}: Number of observations
\item \strong{k}: Number of predictors
}
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_cooksd_bar(model)
ols_plot_cooksd_bar(model, type = 4)
ols_plot_cooksd_bar(model, threshold = 0.2)

}
\seealso{
\code{\link[=ols_plot_cooksd_chart]{ols_plot_cooksd_chart()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-potential-residual-plot.R
\name{ols_plot_resid_pot}
\alias{ols_plot_resid_pot}
\title{Potential residual plot}
\usage{
ols_plot_resid_pot(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Plot to aid in classifying unusual observations as high-leverage points,
outliers, or a combination of both.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_resid_pot(model)

}
\references{
Chatterjee, Samprit and Hadi, Ali. Regression Analysis by Example. 5th ed. N.p.: John Wiley & Sons, 2012. Print.
}
\seealso{
\code{\link[=ols_plot_hadi]{ols_plot_hadi()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-f-test.R
\name{ols_test_f}
\alias{ols_test_f}
\title{F test}
\usage{
ols_test_f(model, fitted_values = TRUE, rhs = FALSE, vars = NULL, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{fitted_values}{Logical; if TRUE, use fitted values of regression model.}

\item{rhs}{Logical; if TRUE, specifies that tests for heteroskedasticity be
performed for the right-hand-side (explanatory) variables of the fitted
regression model.}

\item{vars}{Variables to be used for for heteroskedasticity test.}

\item{...}{Other arguments.}
}
\value{
\code{ols_test_f} returns an object of class \code{"ols_test_f"}.
An object of class \code{"ols_test_f"} is a list containing the
following components:

\item{f}{f statistic}
\item{p}{p-value of \code{f}}
\item{fv}{fitted values of the regression model}
\item{rhs}{names of explanatory variables of fitted regression model}
\item{numdf}{numerator degrees of freedom}
\item{dendf}{denominator degrees of freedom}
\item{vars}{variables to be used for heteroskedasticity test}
\item{resp}{response variable}
\item{preds}{predictors}
}
\description{
Test for heteroskedasticity under the assumption that the errors are
independent and identically distributed (i.i.d.).
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)

# using fitted values
ols_test_f(model)

# using all predictors of the model
ols_test_f(model, rhs = TRUE)

# using fitted values
ols_test_f(model, vars = c('disp', 'hp'))

}
\references{
Wooldridge, J. M. 2013. Introductory Econometrics: A Modern Approach. 5th ed. Mason, OH: South-Western.
}
\seealso{
Other heteroskedasticity tests: 
\code{\link{ols_test_bartlett}()},
\code{\link{ols_test_breusch_pagan}()},
\code{\link{ols_test_score}()}
}
\concept{heteroskedasticity tests}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-rstud-vs-lev-plot.R
\name{ols_plot_resid_lev}
\alias{ols_plot_resid_lev}
\title{Studentized residuals vs leverage plot}
\usage{
ols_plot_resid_lev(model, threshold = NULL, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{threshold}{Threshold for detecting outliers. Default is 2.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Graph for detecting outliers and/or observations with high leverage.
}
\examples{
model <- lm(read ~ write + math + science, data = hsb)
ols_plot_resid_lev(model)
ols_plot_resid_lev(model, threshold = 3)

}
\seealso{
\code{\link[=ols_plot_resid_stud_fit]{ols_plot_resid_stud_fit()}}, \code{\link[=ols_plot_resid_lev]{ols_plot_resid_lev()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-diagnostics-panel.R
\name{ols_plot_diagnostics}
\alias{ols_plot_diagnostics}
\title{Diagnostics panel}
\usage{
ols_plot_diagnostics(model, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Panel of plots for regression diagnostics.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_plot_diagnostics(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-stepaic-forward-regression.R
\name{ols_step_forward_adj_r2}
\alias{ols_step_forward_adj_r2}
\alias{ols_step_forward_adj_r2.default}
\alias{plot.ols_step_forward_adj_r2}
\title{Stepwise Adjusted R-Squared forward regression}
\usage{
ols_step_forward_adj_r2(model, ...)

\method{ols_step_forward_adj_r2}{default}(
  model,
  include = NULL,
  exclude = NULL,
  progress = FALSE,
  details = FALSE,
  steps = NULL,
  ...
)

\method{plot}{ols_step_forward_adj_r2}(x, print_plot = TRUE, details = TRUE, digits = 3, ...)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{...}{Other arguments.}

\item{include}{Character or numeric vector; variables to be included in selection process.}

\item{exclude}{Character or numeric vector; variables to be excluded from selection process.}

\item{progress}{Logical; if \code{TRUE}, will display variable selection progress.}

\item{details}{Logical; if \code{TRUE}, will print the regression result at
each step.}

\item{steps}{Number of steps after which the stepwise procedures should stop.}

\item{x}{An object of class \code{ols_step_forward_*}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}

\item{digits}{Number of decimal places to display.}
}
\value{
List containing the following components:

\item{model}{final model; an object of class \code{lm}}
\item{metrics}{selection metrics}
\item{others}{list; info used for plotting and printing}
}
\description{
Build regression model from a set of candidate predictor variables by
entering predictors based on adjusted r-squared, in a stepwise
manner until there is no variable left to enter any more.
}
\examples{
# stepwise forward regression
model <- lm(y ~ ., data = surgical)
ols_step_forward_adj_r2(model)

# stepwise forward regression plot
k <- ols_step_forward_adj_r2(model)
plot(k)

# selection metrics
k$metrics

# extract final model
k$model

# include or exclude variables
# force variable to be included in selection process
ols_step_forward_adj_r2(model, include = c("age"))

# use index of variable instead of name
ols_step_forward_adj_r2(model, include = c(5))

# force variable to be excluded from selection process
ols_step_forward_adj_r2(model, exclude = c("liver_test"))

# use index of variable instead of name
ols_step_forward_adj_r2(model, exclude = c(4))

# include & exclude variables in the selection process
ols_step_forward_adj_r2(model, include = c("age"), exclude = c("liver_test"))

# use index of variable instead of name
ols_step_forward_adj_r2(model, include = c(5), exclude = c(4))

# steps
ols_step_forward_adj_r2(model, steps = 2)

}
\references{
Venables, W. N. and Ripley, B. D. (2002) Modern Applied Statistics with S. Fourth edition. Springer.
}
\seealso{
Other forward selection procedures: 
\code{\link{ols_step_forward_aic}()},
\code{\link{ols_step_forward_p}()},
\code{\link{ols_step_forward_r2}()},
\code{\link{ols_step_forward_sbc}()},
\code{\link{ols_step_forward_sbic}()}
}
\concept{forward selection procedures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/olsrr.R
\docType{package}
\name{olsrr}
\alias{olsrr}
\alias{olsrr-package}
\title{\code{olsrr} package}
\description{
Tools for teaching and learning OLS regression
}
\details{
See the README on
\href{https://github.com/rsquaredacademy/olsrr}{GitHub}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://olsrr.rsquaredacademy.com/}
  \item \url{https://github.com/rsquaredacademy/olsrr}
  \item Report bugs at \url{https://github.com/rsquaredacademy/olsrr/issues}
}

}
\author{
\strong{Maintainer}: Aravind Hebbali \email{hebbali.aravind@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-plots-data.R
\name{ols_prep_cdplot_data}
\alias{ols_prep_cdplot_data}
\title{Cooks' D plot data}
\usage{
ols_prep_cdplot_data(model, type = 1)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{type}{An integer between 1 and 5 selecting one of the 6 methods for computing the threshold.}
}
\description{
Prepare data for cook's d bar plot.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_prep_cdplot_data(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-normality-tests.R
\name{ols_test_normality}
\alias{ols_test_normality}
\alias{ols_test_normality.lm}
\title{Test for normality}
\usage{
ols_test_normality(y, ...)

\method{ols_test_normality}{lm}(y, ...)
}
\arguments{
\item{y}{A numeric vector or an object of class \code{lm}.}

\item{...}{Other arguments.}
}
\value{
\code{ols_test_normality} returns an object of class \code{"ols_test_normality"}.
An object of class \code{"ols_test_normality"} is a list containing the
following components:

\item{kolmogorv}{kolmogorv smirnov statistic}
\item{shapiro}{shapiro wilk statistic}
\item{cramer}{cramer von mises statistic}
\item{anderson}{anderson darling statistic}
}
\description{
Test for detecting violation of normality assumption.
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_test_normality(model)

}
\seealso{
Other residual diagnostics: 
\code{\link{ols_plot_resid_box}()},
\code{\link{ols_plot_resid_fit}()},
\code{\link{ols_plot_resid_hist}()},
\code{\link{ols_plot_resid_qq}()},
\code{\link{ols_test_correlation}()}
}
\concept{residual diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-residual-vs-regressor-plot.R
\name{ols_plot_resid_regressor}
\alias{ols_plot_resid_regressor}
\title{Residual vs regressor plot}
\usage{
ols_plot_resid_regressor(model, variable, print_plot = TRUE)
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{variable}{New predictor to be added to the \code{model}.}

\item{print_plot}{logical; if \code{TRUE}, prints the plot else returns a plot object.}
}
\description{
Graph to determine whether we should add a new predictor to the model already
containing other predictors. The residuals from the model is regressed on the
new predictor and if the plot shows non random pattern, you should consider
adding the new predictor to the model.
}
\examples{
model <- lm(mpg ~ disp + hp + wt, data = mtcars)
ols_plot_resid_regressor(model, 'drat')

}
\seealso{
\code{\link[=ols_plot_added_variable]{ols_plot_added_variable()}}, \code{\link[=ols_plot_comp_plus_resid]{ols_plot_comp_plus_resid()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_hsp}
\alias{ols_hsp}
\title{Hocking's Sp}
\usage{
ols_hsp(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
Hocking's Sp of the model.
}
\description{
Average prediction mean squared error.
}
\details{
Hocking's Sp criterion is an adjustment of the residual sum of
Squares. Minimize this criterion.

\deqn{MSE / (n - p - 1)}

where \eqn{MSE = SSE / (n - p)}, n is the sample size and p is the number of predictors including the intercept
}
\examples{
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_hsp(model)

}
\references{
Hocking, R. R. (1976). “The Analysis and Selection of Variables in a Linear Regression.” Biometrics
32:1–50.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_aic}()},
\code{\link{ols_apc}()},
\code{\link{ols_fpe}()},
\code{\link{ols_mallows_cp}()},
\code{\link{ols_msep}()},
\code{\link{ols_sbc}()},
\code{\link{ols_sbic}()}
}
\concept{model selection criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-model-performance.R
\name{ols_compare_model_performance}
\alias{ols_compare_model_performance}
\title{Compare model performance}
\usage{
ols_compare_model_performance(...)
}
\arguments{
\item{...}{Multiple objects of class \code{lm}.}
}
\description{
Compare performance of regression models using multiple metrics.
}
\examples{
model_1 <- lm(mpg ~ disp + hp, data = mtcars)
model_2 <- lm(mpg ~ disp + hp + wt, data = mtcars)
model_3 <- lm(mpg ~ disp + hp + wt + qsec + + drat, data = mtcars)
ols_compare_model_performance(model_1, model_2, model_3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-model-info.R
\name{ols_get_variables}
\alias{ols_get_variables}
\title{Model variables}
\usage{
ols_get_variables(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
List with response and predictors.
}
\description{
Return model variables.
}
\examples{
model <- lm(mpg ~ wt + cyl * hp * disp + gear * drat, data = mtcars)
ols_get_variables(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-collinearity-diagnostics.R
\name{ols_coll_diag}
\alias{ols_coll_diag}
\alias{ols_vif_tol}
\alias{ols_eigen_cindex}
\title{Collinearity diagnostics}
\usage{
ols_coll_diag(model)

ols_vif_tol(model)

ols_eigen_cindex(model)
}
\arguments{
\item{model}{An object of class \code{lm}.}
}
\value{
\code{ols_coll_diag} returns an object of class \code{"ols_coll_diag"}.
An object of class \code{"ols_coll_diag"} is a list containing the
following components:

\item{vif_t}{tolerance and variance inflation factors}
\item{eig_cindex}{eigen values and condition index}
}
\description{
Variance inflation factor, tolerance, eigenvalues and condition indices.
}
\details{
Collinearity implies two variables are near perfect linear combinations of
one another. Multicollinearity involves more than two variables. In the
presence of multicollinearity, regression estimates are unstable and have
high standard errors.

\emph{Tolerance}

Percent of variance in the predictor that cannot be accounted for by other predictors.

Steps to calculate tolerance:

\itemize{
\item Regress the kth predictor on rest of the predictors in the model.
\item Compute \eqn{R^2} - the coefficient of determination from the regression in the above step.
\item \eqn{Tolerance = 1 - R^2}
}

\emph{Variance Inflation Factor}

Variance inflation factors measure the inflation in the variances of the parameter estimates due to
collinearities that exist among the predictors. It is a measure of how much the variance of the estimated
regression coefficient \eqn{\beta_k}  is inflated by the existence of correlation among the predictor variables
in the model. A VIF of 1 means that there is no correlation among the kth predictor and the remaining predictor
variables, and hence the variance of \eqn{\beta_k} is not inflated at all. The general rule of thumb is that VIFs
exceeding 4 warrant further investigation, while VIFs exceeding 10 are signs of serious multicollinearity
requiring correction.

Steps to calculate VIF:

\itemize{
\item Regress the kth predictor on rest of the predictors in the model.
\item Compute \eqn{R^2} - the coefficient of determination from the regression in the above step.
\item \eqn{Tolerance = 1 / 1 - R^2 = 1 / Tolerance}
}

\emph{Condition Index}

Most multivariate statistical approaches involve decomposing a correlation matrix into linear
combinations of variables. The linear combinations are chosen so that the first combination has
the largest possible variance (subject to some restrictions), the second combination
has the next largest variance, subject to being uncorrelated with the first, the third has the largest
possible variance, subject to being uncorrelated with the first and second, and so forth. The variance
of each of these linear combinations is called an eigenvalue. Collinearity is spotted by finding 2 or
more variables that have large proportions of variance (.50 or more) that correspond to large condition
indices. A rule of thumb is to label as large those condition indices in the range of 30 or larger.
}
\examples{
# model
model <- lm(mpg ~ disp + hp + wt + drat, data = mtcars)

# vif and tolerance
ols_vif_tol(model)

# eigenvalues and condition indices
ols_eigen_cindex(model)

# collinearity diagnostics
ols_coll_diag(model)

}
\references{
Belsley, D. A., Kuh, E., and Welsch, R. E. (1980). Regression Diagnostics: Identifying Influential Data and
Sources of Collinearity. New York: John Wiley & Sons.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols-information-criteria.R
\name{ols_sbc}
\alias{ols_sbc}
\title{Bayesian information criterion}
\usage{
ols_sbc(model, method = c("R", "STATA", "SAS"))
}
\arguments{
\item{model}{An object of class \code{lm}.}

\item{method}{A character vector; specify the method to compute BIC. Valid
options include R, STATA and SAS.}
}
\value{
The bayesian information criterion of the model.
}
\description{
Bayesian information criterion for model selection.
}
\details{
SBC provides a means for model selection. Given a collection of models for
the data, SBC estimates the quality of each model, relative to each of the
other models. R and STATA use loglikelihood to compute SBC. SAS uses residual
sum of squares. Below is the formula in each case:

\emph{R & STATA}
\deqn{AIC = -2(loglikelihood) + ln(n) * 2p}

\emph{SAS}
\deqn{AIC = n * ln(SSE / n) + p * ln(n)}

where \emph{n} is the sample size and \emph{p} is the number of model parameters including intercept.
}
\examples{
# using R computation method
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_sbc(model)

# using STATA computation method
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_sbc(model, method = 'STATA')

# using SAS computation method
model <- lm(mpg ~ disp + hp + wt + qsec, data = mtcars)
ols_sbc(model, method = 'SAS')

}
\references{
Schwarz, G. (1978). “Estimating the Dimension of a Model.” Annals of Statistics 6:461–464.

Judge, G. G., Griffiths, W. E., Hill, R. C., and Lee, T.-C. (1980). The Theory and Practice of Econometrics.
New York: John Wiley & Sons.
}
\seealso{
Other model selection criteria: 
\code{\link{ols_aic}()},
\code{\link{ols_apc}()},
\code{\link{ols_fpe}()},
\code{\link{ols_hsp}()},
\code{\link{ols_mallows_cp}()},
\code{\link{ols_msep}()},
\code{\link{ols_sbic}()}
}
\concept{model selection criteria}
